/******************************************************************************
   getframe: extract floating point image frame from time series image file

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 
  
   usage: getframe [optional parameters] < e3dFrameFile > imagefile

   parameters:
      frame   - frame number of image frame (starting at 1)
                header info will be printed if frame_number = 0
      sample  - sampling factor for image (default autosample = 0)

   NOTE: for same size images, sample should be set to sample factor in e3d

   example: getframe frame=5 sample=1 < image.Vx > image.1000.Vx

   compile: cc -g -c getframe.c imageframe.c -lm -lget

   author: shawn larsen
   date  : september 1990
   where : caltech
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "usage.h"

#define BIGINTEGER 2147483647

main(argc, argv)
int    argc;
char **argv;
{
   float *V, dt;
   int frame, sample, factor, Vx, Vy, Sx, Sy, t, fsize;
   unsigned long jump;

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* initialize */
   sample = 0;
   frame  = 0;

   /* get input parameters */
   setpar(argc, argv);
   getpar("sample", "d", &sample);
   getpar("frame",  "d", &frame);
   endpar();
   if (sample < 0) sample = 0;
   if (frame  < 0) frame  = 0;

   /* read header information */
   if (getheader(&Vx, &Vy, &t, &dt, stdin)) {
      fprintf(stderr, "Error: bad header\n");
      exit(-1);
      };

   /* print header info if frame = 0 */
   if (frame == 0) {
      fprintf(stderr, "x dimension (unsampled)  : %d\n", Vx);
      fprintf(stderr, "y dimension (unsampled)  : %d\n", Vy);
      fprintf(stderr, "image frames in file     : %d\n",  t);
      fprintf(stderr, "frame increment (seconds): %f\n", dt);
      exit(0);
      }

   /* error if frame request is larger than frames in file */
   if (frame > t) {
      fprintf(stderr, "Error: frame request larger than frames in file\n");
      exit(-1);
      }

   /* allocate memory for image frame */
   V = (float *) malloc(Vx*Vy*sizeof(float));

   /* read first image frame to get frame byte size */
   fsize = getimageframe(V, Vx, Vy, stdin);
   if (fsize < 0) {
      fprintf(stderr, "Error: error during image frame read\n");
      exit(-1);
      }

   /* jump to desired image frame and read */
   if (frame > 1) {
      jump = (long) ((frame - 2)*fsize);
      if (jump > (BIGINTEGER)) {
         fprintf(stderr, "Warning: seek may be too big\n");
         /*exit(-1);*/
         } 
      fseek(stdin, jump, (SEEK_CUR));
      if (getimageframe(V, Vx, Vy, stdin) != fsize) {
         fprintf(stderr, "Error: error during image frame read\n");
         exit(-1);
         }
      }

   /* sample the grid */
   factor = getsampler(V, Vx, Vy, sample);
   Sx     = (Vx - 1)/factor + 1;
   Sy     = (Vy - 1)/factor + 1;

   /* write desired image frame */
   putimage(V, Sx, Sy, stdout);

   /* free memory for image frame */
   free(V);

}



/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf0, statbuf1;
   int i, error, error0 = 1, error1 = 1;
 
   /* get input status */
   fstat(0, &statbuf0);
 
   /* normal input */
   if (statbuf0.st_mode & REGULAR_FILE) error0 = 0;
   if (statbuf0.st_mode & PIPE)         error0 = 0;

   /* get output status */
   fstat(1, &statbuf1);

   /* normal output */
   if (argc > 1) {
      if (statbuf1.st_mode & REGULAR_FILE) error1 = 0;
      if (statbuf1.st_mode & PIPE)         error1 = 0;
      }
   else {
      error1 = 0;
      }
   error1 = 0;

   /* abnormal input */
   if (argc > 3) error0 = 1;
 
   /* print usage and return exit */
   if (error0 || error1) {
      i = 0;
      while(getframeUsage[i]) fprintf(stderr, "%s", getframeUsage[i++]);
      }
   error = error0 + error1;
   return(error);
 
}

