/******************************************************************************
   getframes: read frames and output computed average as floating image 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   usage: getframes [optional parameters] < e3dFrameFile > imagefile

   parameters:
      mode      1 average (average at end); 
                2 integrate (average at end); 
                3 maximum (absolute) [default];
      sample  - sampling factor for image (default autosample = 0)
      time    - number of frames to average at end [default nframes/20]

   NOTE: for same size images, sample should be set to sample factor in e3d

   example: getframes mode=2 sample=1 time=10 < image.Vx > image.Dx

   compile: cc -g -c getframes.c imageframe.c -lm -lget

   author: shawn larsen
   date  : august 1990
   where : caltech
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "usage.h"

#define FRAMEFACTOR 20
#define BIGINTEGER 2147483647

main(argc, argv)
int    argc;
char **argv;
{
   float *V0, *V1, *V2, dt;
   int f, p, mode, sample, factor, Vx, Vy, Sx, Sy, t, fsize, time;
   unsigned long jump;

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* read header information */
   if (getheader(&Vx, &Vy, &t, &dt, stdin)) {
      fprintf(stderr, "Error: bad header\n");
      exit(-1);
      };
   if (t <= 0) {
      fprintf(stderr, "Error: no image frames\n");
      exit(-1);
      } 

   /* initialize */
   sample = 0;
   mode   = 3;
   time   = t/(FRAMEFACTOR);

   /* get input parameters */
   setpar(argc, argv);
   getpar("sample", "d", &sample);
   getpar("mode",   "d", &mode);
   getpar("time",   "d", &time);
   endpar();
   if (sample < 0) sample = 0;
   if (time   < 1) time   = 1;
   if (time   > t) time   = t;
   if (mode < 1 || mode > 3) {
      fprintf(stderr, "Error: bad mode\n");
      exit(-1);
      }

   /* allocate memory for image frames */
   V0 = (float *) malloc(Vx*Vy*sizeof(float));
   V1 = (float *) malloc(Vx*Vy*sizeof(float));
   V2 = (float *) malloc(Vx*Vy*sizeof(float));

   /* null image frames */
   for(p = 0; p < Vx*Vy; p++) {
      V0[p] = 0.;
      V1[p] = 0.;
      V2[p] = 0.;
      }

   /* read image frames and process */
   if (mode == 1) {
      getave(V0, V1, V2, Vx, Vy, time, t, dt, stdin);
      }
   if (mode == 2) {
      getint(V0, V1, V2, Vx, Vy, time, t, dt, stdin);
      }
   if (mode == 3) {
      getmax(V0, V1, V2, Vx, Vy, time, t, dt, stdin);
      }

   /* sample the grid to be output */
   factor = getsampler(V2, Vx, Vy, sample);
   Sx     = (Vx - 1)/factor + 1;
   Sy     = (Vy - 1)/factor + 1;

   /* write desired image frame */
   putimage(V2, Sx, Sy, stdout);

   /* free memory for image frame */
   free(V0);
   free(V1);
   free(V2);

}


getave(V0, V1, V2, Vx, Vy, time, t, dt, file)
float *V0, *V1, *V2, dt;
int Vx, Vy, time, t;
FILE *file;
{
   float factor;
   int f, f0, p, fsize;
   unsigned long jump;

   /* read first image to get size */
   fsize = getimageframe(V0, Vx, Vy, file);

   /* include if first image is part of average */
   if (time == t) for(p = 0; p < Vx*Vy; p++) V2[p] += V0[p];

   /* jump to desired image frame */
   if (time != t) {
      jump = (long) ((t - time - 1)*fsize);
      if (jump > (BIGINTEGER)) {
         fprintf(stderr, "Warning: seek may be too big\n");
         /*exit(-1);*/
         }
      fseek(file, jump, (SEEK_CUR));
      }

   /* read remaining image frames and store */
   f0 = ((t - time) == 0) ? (1) : (t - time);
   for(f = f0; f < t; f++) {
      if (getimageframe(V0, Vx, Vy, file) != fsize) {
         fprintf(stderr, "Error: error during image frame read\n");
         exit(-1);
         }
      for(p = 0; p < Vx*Vy; p++) V2[p] += V0[p];
      }

   /* compute average */
   factor = 1./((float) time);
   for(p = 0; p < Vx*Vy; p++) V2[p] *= factor;

}

getint(V0, V1, V2, Vx, Vy, time, t, dt, file)
float *V0, *V1, *V2, dt;
int Vx, Vy, time, t;
FILE *file;
{

   float factor;
   int f, p, fsize, fgood;

   /* loop over all frames and integrate (store during averaging interval) */
   for(f = 0; f < t; f++) {
      fsize = getimageframe(V0, Vx, Vy, file);
      if (f == 0) fgood = fsize;
      if (fsize != fgood) {
         fprintf(stderr, "Error: error during image frame read\n");
         exit(-1);
         }
      for(p = 0; p < Vx*Vy; p++) V1[p] += V0[p]*dt;
      if (f >= (t - time)) for(p = 0; p < Vx*Vy; p++) V2[p] += V1[p];
      }

   /* compute average */
   factor = 1./((float) time);
   for(p = 0; p < Vx*Vy; p++) V2[p] *= factor;

}

getmax(V0, V1, V2, Vx, Vy, time, t, dt, file)
float *V0, *V1, *V2, dt;
int Vx, Vy, time, t;
FILE *file;
{
   double fabs();
   float value;
   int f, p, fsize, fgood;

   /* loop over all frames and find maximum */
   for(f = 0; f < t; f++) {
      fsize = getimageframe(V0, Vx, Vy, file);
      if (f == 0) fgood = fsize;
      if (fsize != fgood) {
         fprintf(stderr, "Error: error during image frame read\n");
         exit(-1);
         }
      for(p = 0; p < Vx*Vy; p++) 
         if ((value = fabs(V0[p])) > V2[p]) V2[p] = value;
      }

}


/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf0, statbuf1;
   int i, error, error0 = 1, error1 = 1;
 
   /* get input status */
   fstat(0, &statbuf0);
 
   /* normal input */
   if (statbuf0.st_mode & REGULAR_FILE) error0 = 0;
   if (statbuf0.st_mode & PIPE)         error0 = 0;

   /* get output status */
   fstat(1, &statbuf1);

   /* normal output */
   if (argc > 1) {
      if (statbuf1.st_mode & REGULAR_FILE) error1 = 0;
      if (statbuf1.st_mode & PIPE)         error1 = 0;
      }
   else {
      error1 = 0;
      }
   error1 = 0;

   /* abnormal input */
   if (argc > 4) error0 = 1;
 
   /* print usage and return exit */
   if (error0 || error1) {
      i = 0;
      while(getframesUsage[i]) fprintf(stderr, "%s", getframesUsage[i++]);
      }
   error = error0 + error1;
   return(error);
 
}

