/******************************************************************************
   getimage: converts output fields from e3d to 24/32-bit images 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   usage: getimage [optional parameters] < outputfield > imagefile

   parameters:
      ascale - autoscale scale factor
      scale  - scale image by this amount (default 0: autoscale) 
      sample - sampling rate of image     (default 1)
      tfile  - topography file
      mode   - 32 bit mode (if = 32)
      file   - color table file (GMT format)

   compile: cc -g -c getimage.c -lm -lget

   note: EPSILON is 1.e-30

   author: shawn larsen
   date  : june 1990
   where : caltech
*****************************************************************************/

#include "../src/elas.h"
#include "../src/rasterfile.h"
#include "usage.h"

#define   INDEX   255
#define AFACTOR  0.70
#define ASCALE   100.
#define MAXCPT   1000

main(argc, argv)
int    argc;
char **argv;
{
   double fabs();
   float *V, value, scale, ascale, max, index;
   int i, j, k, nx, ny, sample, mode, imode, bsize;
   char cfile[82], tfile[82];
   unsigned char r, g, b, a, *image;

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* initialize */
   scale  = 0.;
   ascale = (ASCALE);
   sample = 1.;
   mode   = 24;
   index = (float) INDEX;
   g = (unsigned char) 0;
   tfile[0] = (char) '\0';
   cfile[0] = (char) '\0';

   /* get input parameters */
   setpar(argc, argv);
   getpar("scale" , "f", &scale);
   getpar("ascale", "f", &ascale);
   getpar("sample", "d", &sample);
   getpar("tfile",  "s",  tfile);
   getpar("file",   "s",  cfile);
   getpar("mode",   "d",  &mode);
   endpar();
   if (sample < 1) sample = 1;
   bsize = (mode == 32) ? (4) : (3);
   imode = (cfile[0]) ? (2) : (1);
   if (imode == 2 && scale == 0.) scale = 1.;

   /* get file size */
   fread(&nx, 4, 1, stdin);
   fread(&ny, 4, 1, stdin);

   /* allocate memory */
   V     = (float         *) malloc(    1*nx*ny*sizeof(float));
   image = (unsigned char *) malloc(bsize*nx*ny*sizeof(unsigned char));

   /* read floating point data */
   fread(V, 4, nx*ny, stdin);

   /* resample image */
   if (sample != 1) {
      k = 0;
      for(j = 0; j < ny; j += sample) for(i = 0; i < nx; i += sample) 
         V[k++] = V[i + nx*j];
      nx = (nx - 1)/sample + 1;
      ny = (ny - 1)/sample + 1;
      }

   /* determine scale for autoscale */
   if (scale == 0.) {
      max = 0.;
      for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
         value = fabs(V[i + nx*j]);
         if (value >  max) max = value;
         }
      if (max <  EPSILON) scale = 1.0;
      if (max >= EPSILON) scale = index/max*ASCALE;
      fprintf(stderr, "scale factor = %7.3e\n", scale);
      }

   /* rescale array image */
   if (scale != 1.0) for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) 
      V[i + nx*j] *= scale;

   /* convert output field to image */ 
   if (!cfile[0]) getImageScale1(image, V, nx, ny, mode       );
   if ( cfile[0]) getImageScale2(image, V, nx, ny, mode, cfile);

   /* white out above topography points */
   if (tfile[0]) getAboveTopo(image, nx, ny, sample, mode, tfile);

   /* write raster file */
   if (mode != 32) mode = 2;
   if (mode == 32) mode = 4;
   writeSunRaster(image, nx, ny, mode, 0, stdout);

   /* free memory */
   free(V);
   free(image);

}




/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf;
   int i, error = 1;
 
   /* get input status */
   fstat(0, &statbuf);
 
   /* normal input */
   if (statbuf.st_mode & REGULAR_FILE) error = 0;
   if (statbuf.st_mode & PIPE)         error = 0;

   /* abnormal input */
   if (argc > 5) error = 1;
 
   /* print usage and return exit */
   if (error) {
      i = 0;
      while(getimageUsage[i]) fprintf(stderr, "%s", getimageUsage[i++]);
      }
   return(error);
 
}


/******************************************************************************
   getImageScale1: get image using simple scaling

   usage: void = getImageScale1(image, V, nx, ny, mode)

   image:  byte image data 
   V:      floating point image data
   nx:     x dimension of image
   ny:     y dimension of image 
   mode:   image mode (!32 for 24-bit image; 32 for 32-bit image)
******************************************************************************/
getImageScale1(image, V, nx, ny, mode)
float *V;
int nx, ny, mode;
unsigned char *image;
{
   float value;
   int i, j;
   unsigned char r, g, b, a;

   for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
      r = (unsigned char) 0; 
      b = (unsigned char) 0;
      value = V[i + nx*j];
      if (value >  255.) value =  255.;
      if (value < -255.) value = -255.;
      r =                 (unsigned char) ((int) fabs(value        ));
      a =                 (unsigned char) ((int) fabs(value*AFACTOR));
      if (value < 0.) b = (unsigned char) ((int) fabs(value));
      /*
      if (value > 0.) r = (unsigned char) ((int)  value);
      if (value < 0.) b = (unsigned char) ((int) -value);
      */
      if (mode != 32) {
         image[3*(i + nx*j) + 0] = b;
         image[3*(i + nx*j) + 1] = g;
         image[3*(i + nx*j) + 2] = r;
         }
      if (mode == 32) {
         image[4*(i + nx*j) + 0] = a;
         image[4*(i + nx*j) + 1] = b;
         image[4*(i + nx*j) + 2] = g;
         image[4*(i + nx*j) + 3] = r;
         }
      } 

}

/******************************************************************************
   getImageScale2: get image using color map

   usage: void = getImageScale2(image, V, nx, ny, mode, cfile)

   image:  byte image data 
   V:      floating point image data
   nx:     x dimension of image
   ny:     y dimension of image 
   mode:   image mode (!32 for 24-bit image; 32 for 32-bit image)
   cfile:  color table file
******************************************************************************/
getImageScale2(image, V, nx, ny, mode, cfile)
float *V;
int nx, ny, mode;
char *cfile;
unsigned char *image;
{
   float value, V1[MAXCPT], V2[MAXCPT], Vmin, Vmax;
   float dV[MAXCPT];
   int i, j, k, n;
   int r1[MAXCPT], r2[MAXCPT], g1[MAXCPT], g2[MAXCPT], b1[MAXCPT], b2[MAXCPT]; 
   int F[3], B[3], N[3];
   char c, line[258], dummy[82];
   unsigned char r, g, b, a;
   FILE *file;

   /* initialize */
   F[0] =   0; F[1] =   0; F[2] =   0;
   N[0] =   0; N[1] =   0; N[2] =   0;
   B[0] = 255; B[1] = 255; B[2] = 255;
   Vmin = 0; 
   Vmax = 0;
   n    = 0;

   /* open ascii color table file */
   file = NULL;
   file = fopen(cfile, "r");
   if (file == NULL) {
      fprintf(stderr, "Error: can't open file \"%s\"\n", cfile);
      exit(1);
      }

   /* read color table file */
   while(fgets(line, 256, file)) {
      c = (char) valueLine(line, 256);
      if ((int) c ==  0 ) continue;
      if ((int) c == 'F') sscanf(line, "%s %d %d %d", dummy, F+0, F+1, F+2); 
      if ((int) c == 'B') sscanf(line, "%s %d %d %d", dummy, B+0, B+1, B+2); 
      if ((int) c == 'N') sscanf(line, "%s %d %d %d", dummy, N+0, N+1, N+2); 
      if ((int) c == 'F' || (int) c == 'B' || (int) c == 'N') continue;
      if (n >= MAXCPT) {
         n++;
         fprintf(stderr, "Error: too many values in color table file: %d\n", n);
         exit(1);
         } 
      sscanf(line, "%f %d %d %d %f %d %d %d", &V1[n], &r1[n], &g1[n], &b1[n],
                                              &V2[n], &r2[n], &g2[n], &b2[n]);
      n++;
      }

   /* determine differences */
   for(k = 0; k < n; k++) {
      dV[k] = V2[k] - V1[k];
      if (dV[k] == 0.) dV[k] = 1.0;
      }

   /* check color table */
   for(k = 0; k < n; k++) {
      if (V1[k] > V2[k]) {
         fprintf(stderr, "Error: bad color table values\n");
         exit(1);
         }
      if (r1[k] <   0) r1[k] =   0;
      if (g1[k] <   0) g1[k] =   0;
      if (b1[k] <   0) b1[k] =   0;
      if (r2[k] <   0) r2[k] =   0;
      if (g2[k] <   0) g2[k] =   0;
      if (b2[k] <   0) b2[k] =   0;
      if (r1[k] > 255) r1[k] = 255;
      if (g1[k] > 255) g1[k] = 255;
      if (b1[k] > 255) b1[k] = 255;
      if (r2[k] > 255) r2[k] = 255;
      if (g2[k] > 255) g2[k] = 255;
      if (b2[k] > 255) b2[k] = 255;
      }
   for(k = 0; k < 3; k++) {
      if (F[k] <   0) F[k] =   0;
      if (B[k] <   0) B[k] =   0;
      if (N[k] <   0) N[k] =   0;
      if (F[k] > 255) F[k] = 255;
      if (B[k] > 255) B[k] = 255;
      if (N[k] > 255) N[k] = 255;
      }

   /* determine minimum and maximum data */
   Vmin = V1[0]; Vmax = V2[0];
   for(k = 0; k < n; k++) {
      if (V1[k] < Vmin) Vmin = V1[k];
      if (V2[k] > Vmax) Vmax = V2[k];
      }

   /* loop through image and translate to color data */
   for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
      value = V[i + nx*j];
      r = (unsigned char) N[0]; 
      g = (unsigned char) N[1]; 
      b = (unsigned char) N[2];
      if (value < Vmin) {
         r = (unsigned char) F[0]; 
         g = (unsigned char) F[1]; 
         b = (unsigned char) F[2];
         }
      if (value > Vmax) {
         r = (unsigned char) B[0]; 
         g = (unsigned char) B[1]; 
         b = (unsigned char) B[2];
         }
      for(k = 0; k < n; k++) {
         if (value >= V1[k] && value <= V2[k]) {
            r = (unsigned char) r1[k] + ((value - V1[k])/dV[k])*(r2[k] - r1[k]);
            g = (unsigned char) g1[k] + ((value - V1[k])/dV[k])*(g2[k] - g1[k]);
            b = (unsigned char) b1[k] + ((value - V1[k])/dV[k])*(b2[k] - b1[k]);
            }
         }
      a = (unsigned char) ((int) fabs(255*(AFACTOR)));
      if (mode != 32) {
         image[3*(i + nx*j) + 0] = b;
         image[3*(i + nx*j) + 1] = g;
         image[3*(i + nx*j) + 2] = r;
         }
      if (mode == 32) {
         image[4*(i + nx*j) + 0] = a;
         image[4*(i + nx*j) + 1] = b;
         image[4*(i + nx*j) + 2] = g;
         image[4*(i + nx*j) + 3] = r;
         }
      }

   /* close file */
   fclose(file);

}

/* determine line type and return value */
valueLine(line, n)
int      n;
char *line;
{
   int i;
   char c;
   for(i = 0; i < n; i++) {
      c = line[i];
      if (c == '\n') return(0);
      if (c ==  ' ') continue;
      if (c == '\t') continue;
      if (c >= '0' && c <= '9') return((int) c);
      if (c == '.' || c == '-' || c == '+') return((int) c);
      if (c == 'F' || c == 'B' || c == 'N') return((int) c);
      return(0);
      }
   return(0);

}


/******************************************************************************
   writeSunRaster: write sun raster file

   usage: void = writeSunRaster(data, x, y, mode, color, file)

   data:  image data
   x:     x dimension of image in pixels
   y:     y dimension of image in pixels
   mode:  type of rasterfile (1 = 8 bit; 2 = 24 bit; 3 = 24 bit (8 in)
   color: color scheme (1 = gray scale; 2 = PS black; 3 = PS white)
   file:  output file
******************************************************************************/

writeSunRaster(data, x, y, mode, color, file)
int x, y, mode, color;
unsigned char *data;
FILE *file;
{
   Raster rFile[1];
   int i, j, k, m, n, p;
   int red, green, blue;
   unsigned char r[256], g[256], b[256], c;

   /* write raster file header */
   p = (mode == 1) ? 1 : 3;
   p = (mode == 4) ? 4 : p;
   m = (mode == 1) ? RMT_EQUAL_RGB : RMT_NONE;
   n = (mode == 1) ? 3*256 : 0;
   k = (mode == 1) ? RT_STANDARD : RT_STANDARD;
   rFile->ras_magic     =     RAS_MAGIC;
   rFile->ras_width     =             x;
   rFile->ras_height    =             y;
   rFile->ras_depth     =           p*8;
   rFile->ras_length    =         x*y*p;
   rFile->ras_type      =             k;
   rFile->ras_maptype   =             m;
   rFile->ras_maplength =             n;
   fwrite(rFile, sizeof(Raster), 1, file);

   /* get color table (gray scale only) */
   if (color == 1) 
      for(i = 0; i < 256; i++) r[i] = g[i] = b[i] = (unsigned char) i;
   if (color == 2) {
      for(i = 11; i < 245; i++) {
         r[i] = (!(i %   2)) ? (0) : (GETCOLORVALUE(i));
         g[i] = ( (i %   2)) ? (0) : (GETCOLORVALUE(i));
         b[i] = ( (i > 128)) ? (0) : (GETCOLORVALUE(i));
         }
      }
   if (color == 3) {
      for(i = 11; i < 245; i++) {
         r[i] = (!(i %   2)) ? (GETWHITEVALUE(i)) : (SATURATE);
         g[i] = ( (i %   2)) ? (GETWHITEVALUE(i)) : (SATURATE);
         b[i] = ( (i % 128)) ? (GETWHITEVALUE(i)) : (SATURATE);
         }
      }

   /* output color table */
   if (mode == 1) {
      fwrite(r, 256, 1, file);
      fwrite(g, 256, 1, file);
      fwrite(b, 256, 1, file);
      }

   /* null byte (since lines must be %16 bits) */
   c = (unsigned char) '\0';

   /* output file */
   switch(mode) {
      case 1:
         for(i = 0; i < y; i++) {
            fwrite(data + i*x,  1*x, 1, file);
            if (x%2) fwrite(&c,   1, 1, file);
            }
         break;
      case 2:
         for(i = 0; i < y; i++) {
            fwrite(data + 3*i*x,  3*x, 1, file);
            if (x%2) fwrite(&c,     1, 1, file);
            }
         break;
      case 3:
         for(i = 0; i < y; i++) {
            for(j = 0; j < x; j++) {
               fwrite(&r[(unsigned int) data[i*x + j]], 1, 1, file);
               fwrite(&g[(unsigned int) data[i*x + j]], 1, 1, file);
               fwrite(&b[(unsigned int) data[i*x + j]], 1, 1, file);
               }
            if (x%2) fwrite(&c, 1, 1, file);
            }
         break;
      case 4:
         for(i = 0; i < y; i++) {
            fwrite(data + 4*i*x,  4*x, 1, file);
            /*if (x%2) fwrite(&c,     1, 1, file);*/
            }
         break;
      default:
         fprintf(stderr, "ERROR: bad file mode\n");
         exit(-1);
         break;
      }

   return(0);

}



#define SACSIZE (158*4)
#define DDELTA       0
#define DNPTS       79

getAboveTopo(image, nx, ny, sample, mode, tfile)
int nx, ny, sample, mode;
char *tfile;
unsigned char *image;
{
   float *T, dt, z;
   int i, j, n, p;
   unsigned char header[SACSIZE];
   FILE *file;

   /* open topography file */
   file = fopen(tfile, "r");
   if (file == NULL) {
      fprintf(stderr, "Warning: can't open topography file\n");
      return(-1);
      }

   /* read topography file header */
   fread(header, SACSIZE, 1, file);
   dt = *((float *) (header + DDELTA*sizeof(float)));
   n  = *((int   *) (header + DNPTS *sizeof(int  )));
   if (((n-1)/sample + 1) != nx) {
      fprintf(stderr, "Warning: bad topography values\n");
      return(-1);
      }
   dt /= 1000.;
   fprintf(stderr, "Notice: topography dx is in meters\n");

   /* get memory */
   T = (float *) malloc(n*sizeof(float));

   /* read topography file data */
   fread(T, sizeof(float), n, file);

   /* resample topography file (in place) */
   n = (n-1)/sample + 1;
   for(i = 0; i < n; i++) T[i] = T[i*sample]; 

   /* close topography file */
   fclose(file);

   /* white out topography points */
/*
   for(i = 0; i < nx; i++) {
      p = (int) T[i]/dt;
      if (p*dt >= T[i]) p--;
      p--;
      if (p >= ny) p = ny;
      for(j = 0; j <= p; j++) {
         image[3*(i + nx*j) + 0] = 255;
         image[3*(i + nx*j) + 1] = 255;
         image[3*(i + nx*j) + 2] = 255;
         }
      }
*/
   if (mode != 32) {
      for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
         z = ((float) (j-1))*dt;
         if (z < T[i]) {
            image[3*(i + nx*j) + 0] = 255;
            image[3*(i + nx*j) + 1] = 255;
            image[3*(i + nx*j) + 2] = 255;
            }
         }
      }
   if (mode == 32) {
      for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
         z = ((float) (j-1))*dt;
         if (z < T[i]) {
            image[3*(i + nx*j) + 0] =   0;
            image[3*(i + nx*j) + 1] = 255;
            image[3*(i + nx*j) + 2] = 255;
            image[3*(i + nx*j) + 3] = 255;
            }
         }
      }

   /* free memory */
   free(T);

}
