/******************************************************************************
   getps: convert and combine P and S e3d output fields to 24/32-bit images 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   usage: getps [optional parameters] Pfile Sfile > imagefile

   parameters:
      scale  - scale   image by this amount (default 0: autoscale) 
      pratio - scale p image by this amount (default 1:     scale) 
      sample - sampling rate of image     (default 1)
      color  - color scheme (1 black bg; 2 white bg) (default 1)
      clip   - clipping color value (default 255)
      tfile  - topography file
      mode   - 32 bit mode (if = 32)

   compile: cc -g -c getps.c -lm -lget

   author: shawn larsen
   date  : june 1990
   where : caltech
*****************************************************************************/

#include "../src/elas.h"
#include "../src/rasterfile.h"
#include "usage.h"

#define INDEX    255
#define AFACTOR 0.75
#define ASCALE  100.

main(argc, argv)
int    argc;
char **argv;
{
   double fabs();
   float *P, *S, Pvalue, Svalue, Avalue, scale, pratio, max, index;
   int i, j, k, Pnx, Pny, Snx, Sny, nx, ny, sample, color, clip, mode, bsize;
   int red, green, blue, alpha;
   unsigned char r, g, b, a, *image;
   char Pname[82], Sname[82], tfile[82], name;
   FILE *Pfile, *Sfile;

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* initialize */
   scale  = 0.;
   pratio = 1.;
   sample = 1.;
   color  = 1 ;
   clip   = 255;
   mode   = 24;
   index  = (float) INDEX;
   tfile[0] = (char) '\0';

   /* get input parameters */
   setpar(argc, argv);
   getpar("scale" , "f", &scale);
   getpar("pratio", "f", &pratio);
   getpar("sample", "d", &sample);
   getpar("color" , "d", &color);
   getpar("clip"  , "d", &clip ); 
   getpar("tfile" , "s",  tfile);
   getpar("mode"  , "d", &mode );
   endpar();
   if (sample <   1) sample =   1;
   if (color !=   1) color  =   2;
   if (clip   <   0) clip   =   0;
   if (clip   > 255) clip   = 255;
   if (color ==   2) clip   = 255 - clip;
   bsize = (mode == 32) ? (4) : (3);

   /* open input files */
   strcpy(Pname, argv[argc-2]);
   strcpy(Sname, argv[argc-1]);
   Pfile = fopen(Pname, "r");
   if (Pfile == NULL) {
      fprintf(stderr, "Error: can't open file \"%s\"\n", Pname);
      exit(-1);
      }
   Sfile = fopen(Sname, "r");
   if (Sfile == NULL) {
      fprintf(stderr, "Error: can't open file \"%s\"\n", Sname);
      exit(-1);
      }
   
   /* get file size */
   fread(&Pnx, 4, 1, Pfile);
   fread(&Pny, 4, 1, Pfile);
   fread(&Snx, 4, 1, Sfile);
   fread(&Sny, 4, 1, Sfile);
   if (Pnx != Snx || Pny != Sny) {
      fprintf(stderr, "Error: file sizes are not identical\n");
      exit(-1);
      }
   nx = Pnx;
   ny = Sny;

   /* allocate memory */
   P     = (float         *) malloc(    1*Pnx*Pny*sizeof(float));
   S     = (float         *) malloc(    1*Snx*Sny*sizeof(float));
   image = (unsigned char *) malloc(bsize* nx* ny*sizeof(unsigned char));

   /* read floating point data */
   fread(P, 4, Pnx*Pny, Pfile);
   fread(S, 4, Snx*Sny, Sfile);

   /* resample image */
   if (sample != 1) {
      k = 0; 
      for(j = 0; j < ny; j += sample) for(i = 0; i < nx; i += sample) {
         P[k] = P[i + nx*j];
         S[k] = S[i + nx*j];
         k++;
         }
      nx = (nx - 1)/sample + 1;
      ny = (ny - 1)/sample + 1;
      }

   /* determine scale for autoscale */
   if (scale == 0.) {
      max = 0.;
      for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
         Pvalue = fabs(P[i + nx*j]);
         if (Pvalue >  max) max = Pvalue;
         Svalue = fabs(S[i + nx*j]);
         if (Svalue >  max) max = Svalue;
         }
      if (max == 0.) scale = 1.0;
      if (max != 0.) scale = index/max*(ASCALE);
      fprintf(stderr, "scale factor = %7.3e\n", scale);
      }

   /* rescale array image */
   if (scale != 1.0) for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
      P[i + nx*j] *= scale*pratio;
      S[i + nx*j] *= scale;
      }

   /* convert output field to 24/32-bit image */ 
   for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
      Pvalue = P[i + nx*j];
      Svalue = S[i + nx*j];
      if (Pvalue >  255.) Pvalue =  255.;
      if (Pvalue < -255.) Pvalue = -255.;
      if (Svalue >  255.) Svalue =  255.;
      if (Svalue < -255.) Svalue = -255.;
      Avalue = (AFACTOR)*(fabs(Pvalue) + fabs(Svalue));
      if (Avalue > 255.) Avalue = 255.;
      alpha = (int) Avalue;
      if (color == 1) {
         red   =                  (int) fabs(Pvalue);
         green =                  (int) fabs(Svalue);
         blue  =                                  0 ;
         if (Pvalue < 0.) blue += (int)     -Pvalue ;
         if (Svalue < 0.) blue += (int)     -Svalue ;
         if (blue > 255 ) blue  =               255 ;
         if (clip < 255 ) {
            if (red   > clip) red   = clip;
            if (green > clip) green = clip;
            if (blue  > clip) blue  = clip;
            }
         r = (unsigned char)   red;
         g = (unsigned char) green;
         b = (unsigned char)  blue;
         a = (unsigned char) alpha;
         }
      if (color == 2) {
         red   = 255 - ((int) fabs(Svalue)) + ((int) fabs(Pvalue));
         green = 255 - ((int) fabs(Pvalue)) + ((int) fabs(Svalue));
         blue  = 255;
         if (Pvalue > 0.) blue  -= (int) Pvalue;
         if (Svalue > 0.) blue  -= (int) Svalue;
         if (red   > 255) red    = 255;
         if (green > 255) green  = 255;
         if (blue  <   0) blue   =   0;
         if (clip < 255 ) {
            if (red   < clip) red   = clip;
            if (green < clip) green = clip;
            if (blue  < clip) blue  = clip;
            }
         r = (unsigned char)   red; 
         g = (unsigned char) green; 
         b = (unsigned char)  blue;
         a = (unsigned char) alpha;
         }
      if (mode != 32) {
         image[3*(i + nx*j) + 0] = b;
         image[3*(i + nx*j) + 1] = g;
         image[3*(i + nx*j) + 2] = r;
         }
      if (mode == 32) {
         image[4*(i + nx*j) + 0] = a;
         image[4*(i + nx*j) + 1] = b;
         image[4*(i + nx*j) + 2] = g;
         image[4*(i + nx*j) + 3] = r;
         }
      } 

   /* white out above topography points */
   if (tfile[0]) getAboveTopo(image, nx, ny, sample, mode, tfile);

   /* write raster file */
   if (mode != 32) mode = 2;
   if (mode == 32) mode = 4;
   writeSunRaster(image, nx, ny, mode, 0, stdout);

   /* free memory and close files */
   free(P);
   free(S);
   free(image);
   fclose(Pfile);
   fclose(Sfile);

}




/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf;
   int i, error = 0;
 
   /* get input status */
   fstat(0, &statbuf);
 
   /* normal input */
   /*
   if (statbuf.st_mode & REGULAR_FILE) error = 0;
   if (statbuf.st_mode & PIPE)         error = 0;
   */

   /* abnormal input */
   if (argc < 3) error = 1;
 
   /* print usage and return exit */
   if (error) {
      i = 0;
      while(getpsUsage[i]) fprintf(stderr, "%s", getpsUsage[i++]);
      }
   return(error);
 
}




/******************************************************************************
   writeSunRaster: write sun raster file

   usage: void = writeSunRaster(data, x, y, mode, color, file)

   data:  image data
   x:     x dimension of image in pixels
   y:     y dimension of image in pixels
   mode:  type of rasterfile (1=8 bit; 2=24 bit; 3=24 bit (8 in); 4=32 bit
   color: color scheme (1 = gray scale; 2 = PS black; 3 = PS white)
   file:  output file
******************************************************************************/

writeSunRaster(data, x, y, mode, color, file)
int x, y, mode, color;
unsigned char *data;
FILE *file;
{
   Raster rFile[1];
   int i, j, k, m, n, p;
   int red, green, blue;
   unsigned char r[256], g[256], b[256], c;

   /* write raster file header */
   p = (mode == 1) ? 1 : 3;
   p = (mode == 4) ? 4 : p;
   m = (mode == 1) ? RMT_EQUAL_RGB : RMT_NONE;
   n = (mode == 1) ? 3*256 : 0;
   k = (mode == 1) ? RT_STANDARD : RT_STANDARD;
   rFile->ras_magic     =     RAS_MAGIC;
   rFile->ras_width     =             x;
   rFile->ras_height    =             y;
   rFile->ras_depth     =           p*8;
   rFile->ras_length    =         x*y*p;
   rFile->ras_type      =             k;
   rFile->ras_maptype   =             m;
   rFile->ras_maplength =             n;
   fwrite(rFile, sizeof(Raster), 1, file);

   /* get color table (gray scale only) */
   if (color == 1) 
      for(i = 0; i < 256; i++) r[i] = g[i] = b[i] = (unsigned char) i;
   if (color == 2) {
      for(i = 11; i < 245; i++) {
         r[i] = (!(i %   2)) ? (0) : (GETCOLORVALUE(i));
         g[i] = ( (i %   2)) ? (0) : (GETCOLORVALUE(i));
         b[i] = ( (i > 128)) ? (0) : (GETCOLORVALUE(i));
         }
      }
   if (color == 3) {
      for(i = 11; i < 245; i++) {
         r[i] = (!(i %   2)) ? (GETWHITEVALUE(i)) : (SATURATE);
         g[i] = ( (i %   2)) ? (GETWHITEVALUE(i)) : (SATURATE);
         b[i] = ( (i % 128)) ? (GETWHITEVALUE(i)) : (SATURATE);
         }
      }

   /* output color table */
   if (mode == 1) {
      fwrite(r, 256, 1, file);
      fwrite(g, 256, 1, file);
      fwrite(b, 256, 1, file);
      }

   /* null byte (since lines must be %16 bits) */
   c = (unsigned char) '\0';

   /* output file */
   switch(mode) {
      case 1:
         for(i = 0; i < y; i++) {
            fwrite(data + i*x,  1*x, 1, file);
            if (x%2) fwrite(&c,   1, 1, file);
            }
         break;
      case 2:
         for(i = 0; i < y; i++) {
            fwrite(data + 3*i*x,  3*x, 1, file);
            if (x%2) fwrite(&c,     1, 1, file);
            }
         break;
      case 3:
         for(i = 0; i < y; i++) {
            for(j = 0; j < x; j++) {
               fwrite(&r[(unsigned int) data[i*x + j]], 1, 1, file);
               fwrite(&g[(unsigned int) data[i*x + j]], 1, 1, file);
               fwrite(&b[(unsigned int) data[i*x + j]], 1, 1, file);
               }
            if (x%2) fwrite(&c, 1, 1, file);
            }
         break;
      case 4:
         for(i = 0; i < y; i++) {
            fwrite(data + 4*i*x,  4*x, 1, file);
            /*if (x%2) fwrite(&c,     1, 1, file);*/
            }
         break;
      default:
         fprintf(stderr, "ERROR: bad file mode\n");
         exit(-1);
         break;
      }

   return(0);

}


#define SACSIZE (158*4)
#define DDELTA       0
#define DNPTS       79

getAboveTopo(image, nx, ny, sample, mode, tfile)
int nx, ny, sample, mode;
char *tfile;
unsigned char *image;
{
   float *T, dt, z;
   int i, j, n, p;
   unsigned char header[SACSIZE];
   FILE *file;

   /* open topography file */
   file = fopen(tfile, "r");
   if (file == NULL) {
      fprintf(stderr, "Warning: can't open topography file\n");
      return(-1);
      }

   /* read topography file header */
   fread(header, SACSIZE, 1, file);
   dt = *((float *) (header + DDELTA*sizeof(float)));
   n  = *((int   *) (header + DNPTS *sizeof(int  )));
   if (((n-1)/sample + 1) != nx) {
      fprintf(stderr, "Warning: bad topography values\n");
      return(-1);
      }
/*
   dt /= 1000.;
   fprintf(stderr, "Notice: topography dx is in meters\n");
*/

   /* get memory */
   T = (float *) malloc(n*sizeof(float));

   /* read topography file data */
   fread(T, sizeof(float), n, file);

   /* resample topography file (in place) */
   n = (n-1)/sample + 1;
   for(i = 0; i < n; i++) T[i] = T[i*sample]; 

   /* close topography file */
   fclose(file);

   /* white out topography points */
/*
   for(i = 0; i < nx; i++) {
      p = (int) T[i]/dt;
      if (p*dt >= T[i]) p--;
      p--;
      if (p >= ny) p = ny;
      for(j = 0; j <= p; j++) {
         image[3*(i + nx*j) + 0] = 255;
         image[3*(i + nx*j) + 1] = 255;
         image[3*(i + nx*j) + 2] = 255;
         }
      }
*/
   for(j = 0; j < ny; j++) for(i = 0; i < nx; i++) {
      z = ((float) (j-1))*dt;
      if (z < T[i]) {
         if (mode != 32) {
            image[3*(i + nx*j) + 0] = 255;
            image[3*(i + nx*j) + 1] = 255;
            image[3*(i + nx*j) + 2] = 255;
            }
         if (mode == 32) {
            image[4*(i + nx*j) + 0] = 255;
            image[4*(i + nx*j) + 1] = 255;
            image[4*(i + nx*j) + 2] = 255;
            image[4*(i + nx*j) + 3] = 255;
            }
         }
      }

   /* free memory */
   free(T);

}
