/******************************************************************************
   getsubfield: converts subfields from e3d to single field file 

   usage: getsubfield [optional parameters] < e3dfieldfile > fieldfile

   parameters:
      sample - sampling rate of field     (default autosample)

   compile: cc -g -c getsubfield.c -lm -lget

   NOTE: autosample will make result in an field size that is
         between SFACTOR and 4*SFACTOR (bytes). For an SFACTOR
         of 500000 this gives an field size of 0.5 to 2 MBytes.
         Specifying a sampling factor (even if 1) will turn off
         the autosampler.

   NOTE: format of subfield file is the following

      x   : x dimension of unsampled global grid    (4   bytes int)
      y   : y dimension of unsampled global grid    (4   bytes int)

   followed by n (n = x*y) tuples of the format:

      x   : x global starting coordinate            (4   bytes int)
      y   : y global starting coordinate            (4   bytes int)
      v   : field value of interest                 (4   byte float)

   author: shawn larsen
   date  : june 1992
   where : llnl
*****************************************************************************/

#include "../src/elas.h"
#include "usage.h"

#define SFACTOR  500000

main(argc, argv)
int    argc;
char **argv;
{
   double sqrt();
   float *V, *S;
   int sample, Vx, Vy, Sx, Sy;
   void sampleV();

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* initialize */
   sample = 0;

   /* get input parameters */
   setpar(argc, argv);
   getpar("sample", "d", &sample);
   endpar();

   /* read global header */
   fread(&Vx, 4, 1, stdin);
   fread(&Vy, 4, 1, stdin);

   /* allocate memory */
   V = (float *) malloc(Vx*Vy*sizeof(float));

   /* get auto sample factor for V */
   if (sample <= 0) {
      sample = (int) sqrt((double) ((Vx*Vy - 1)/SFACTOR)) + 1;
      if (sample < 1) sample = 1;
      }

   /* initialize and read in global field file */
   readSFile(V, Vx, Vy);

   /* get sampled V */
   Sx = (Vx - 1)/sample + 1;
   Sy = (Vy - 1)/sample + 1;
   if (sample > 1) sampleV(V, Vx, Vy, sample);

   /* write out field file */
   fwrite(&Sx, 4,     1, stdout); 
   fwrite(&Sy, 4,     1, stdout);
   fwrite(V  , 4, Sx*Sy, stdout);

   /* free memory */
   free(V);

}


readSFile(V, Vx, Vy)
float *V;
int Vx, Vy;
{
   float s;
   int x, y, p;
   for(p = 0; p < Vx*Vy; p++) V[p] = 0.;
   for(p = 0; p < Vx*Vy; p++) {
      fread(&x, sizeof(int)  , 1, stdin);
      fread(&y, sizeof(int)  , 1, stdin);
      fread(&s, sizeof(float), 1, stdin);
      V[y*Vx + x] = s;
      }
}



void
sampleV(V, Vx, Vy, sample)
float *V;
int Vx, Vy, sample;
{
   int i, j, index = 0;

   /* return if no sampling */
   if (sample <= 1) return;

   /* sample grid in place */
   for(j = 0; j < Vy; j += sample) for(i = 0; i < Vx; i += sample) 
      V[index++] = V[i + Vx*j];

}



/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf;
   int i, error = 1;
 
   /* get input status */
   fstat(0, &statbuf);
 
   /* normal input */
   if (statbuf.st_mode & REGULAR_FILE) error = 0;
   if (statbuf.st_mode & PIPE)         error = 0;

   /* abnormal input */
   if (argc > 3) error = 1;
 
   /* print usage and return exit */
   if (error) {
      i = 0;
      while(getsubimageUsage[i]) fprintf(stderr, "%s", getsubimageUsage[i++]);
      }
   return(error);
 
}

