/******************************************************************************
   getsubimage: converts subimages from e3d to single image file 

   usage: getsubimage [optional parameters] < e3dimagefile > imagefile

   parameters:
      sample - sampling rate of image     (default autosample)

   compile: cc -g -c getsubimage.c -lm -lget

   NOTE: autosample will make result in an image size that is
         between SFACTOR and 4*SFACTOR (bytes). For an SFACTOR
         of 500000 this gives and image size of 0.5 to 2 MBytes.
         Specifying a sampling factor (even if 1) will turn off
         the autosampler.

   NOTE: format of subimage file is the following

      x   : x dimension of unsampled global grid    (4   bytes int)
      y   : y dimension of unsampled global grid    (4   bytes int)

   followed by n "subgrids" of format

      x   : x dimension of unsampled subgrid        (4   bytes int)
      y   : y dimension of unsampled subgrid        (4   bytes int)
      x0  : x global starting coordinate of subgrid (4   bytes int)
      y0  : y global starting coordinate of subgrid (4   bytes int)
      f   : sampling factor                         (4   bytes int)
      data: field data                              (4*n bytes float)
            where n = ((x-1)/f + 1)*((y-1)/f + 1)

   author: shawn larsen
   date  : june 1992
   where : llnl
*****************************************************************************/

#include "../src/elas.h"
#include "usage.h"

#define SFACTOR  500000

main(argc, argv)
int    argc;
char **argv;
{
   double sqrt();
   float *V, *S, *getSubGrid();
   int sample, Vx, Vy, Sx, Sy;
   void sampleV();

   /* check usage */
   if (checkUsage(argc, argv)) exit(0);

   /* initialize */
   sample = 0;

   /* get input parameters */
   setpar(argc, argv);
   getpar("sample", "d", &sample);
   endpar();

   /* read global header */
   fread(&Vx, 4, 1, stdin);
   fread(&Vy, 4, 1, stdin);

   /* allocate memory */
   V = (float *) malloc(Vx*Vy*sizeof(float));

   /* loop over all subgrids */
   while(S = getSubGrid(V, Vx, Vy)) free(S);

   /* get auto sample factor for V */
   if (sample <= 0) {
      sample = (int) sqrt((double) ((Vx*Vy - 1)/SFACTOR)) + 1;
      if (sample < 1) sample = 1;
      }

   /* get sampled V */
   Sx = (Vx - 1)/sample + 1;
   Sy = (Vy - 1)/sample + 1;
   if (sample > 1) sampleV(V, Vx, Vy, sample);

   /* write out image file */
   fwrite(&Sx, 4,     1, stdout); 
   fwrite(&Sy, 4,     1, stdout);
   fwrite(V  , 4, Sx*Sy, stdout);

   /* free memory */
   free(V);

}


float *
getSubGrid(V, Vx, Vy)
float *V;
int Vx, Vy;
{
   float *S;
   int i, j, Sx, Sy, Gx, Gy, x0, y0, f;

   /* get subgrid header and test for EOF */
   if (!fread(&Gx, 4, 1, stdin)) return(NULL);
   if (!fread(&Gy, 4, 1, stdin)) return(NULL);
   if (!fread(&x0, 4, 1, stdin)) return(NULL);
   if (!fread(&y0, 4, 1, stdin)) return(NULL);
   if (!fread(&f , 4, 1, stdin)) return(NULL);

   /* determine if subgrid is within global grid */
   if (x0 < 0 || y0 < 0 || (x0 + Gx) > Vx || (y0 + Gy) > Vy) {
      fprintf(stderr, "Error: bad subgrid size\n");
      exit(-1);
      }

   /* determine dimensions of sampled subgrid */
   Sx = (Gx - 1)/f + 1;
   Sy = (Gy - 1)/f + 1;

   /* allocate memory for sampled subgrid */
   S = (float *) malloc(Sx*Sy*sizeof(float));

   /* read subgrid */
   if (fread(S, 4, Sx*Sy, stdin) == EOF) {
      free(S);
      return(NULL);
      }

   /* copy sampled subgrid to unsampled global grid and fill */
   for(j = 0; j < Gy; j++) for(i = 0; i < Gx; i++)
      V[(x0 + i) + Vx*(y0 + j)] = S[(i/f) + Sx*(j/f)];

   /* return sampled subgrid */
   return(S);

}



void
sampleV(V, Vx, Vy, sample)
float *V;
int Vx, Vy, sample;
{
   int i, j, index = 0;

   /* return if no sampling */
   if (sample <= 1) return;

   /* sample grid in place */
   for(j = 0; j < Vy; j += sample) for(i = 0; i < Vx; i += sample) 
      V[index++] = V[i + Vx*j];

}



/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv)

   return 0: correct usage - continue execution
          1: print   usage - stop     execution
******************************************************************************/
checkUsage(argc, argv)
int    argc;
char **argv;
{
   struct stat statbuf;
   int i, error = 1;
 
   /* get input status */
   fstat(0, &statbuf);
 
   /* normal input */
   if (statbuf.st_mode & REGULAR_FILE) error = 0;
   if (statbuf.st_mode & PIPE)         error = 0;

   /* abnormal input */
   if (argc > 3) error = 1;
 
   /* print usage and return exit */
   if (error) {
      i = 0;
      while(getsubimageUsage[i]) fprintf(stderr, "%s", getsubimageUsage[i++]);
      }
   return(error);
 
}

