/******************************************************************************
   imageframe: routines for reading sub grids and image frames from e3d output

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   getimageframe: read image frame from e3d output
   getsubgrid:    read subgrid from e3d output
   getsampler:    sample image frame (autosample option)
   putimage:      write floating point image file (2 int header)
   getheader:     read global header

   usage:
      framesize    = getimageframe(V, Vx, Vy,         file)
      subgridsize  = getsubgrid   (V, Vx, Vy,         file)
      samplefactor = getsampler   (V, Vx, Vy, sample      )
      int          = putimage     (V, Vx, Vy,         file)
      int          = getheader    (&Vx, &Vy, &t, &dt, file)

   parameters:
      V      - array containing image V[p]              (*float)
      Vx     - global image horizontal dimension        (   int)
      Vy     - global image vertical   dimension        (   int)
      file   - input file (or stdin)
      sample - sampling rate of image (autosample <= 0) (   int)
      t      - number of frames in image frame file     (   int)
      dt     - time increment of frames (seconds)       ( float)
      

   return:
      framesize    - size (bytes) of image frame (including headers) (error < 0)
      subgridsize  - size (bytes) of sub grid    (including header ) (error < 0)
      samplefactor - sampling factor used (note: Sn = (Vn-1)/sample+1)

   compile: cc -g -c imageframe.c 

   NOTE: autosample will make result in an image size that is
         between SFACTOR and 4*SFACTOR (bytes). For an SFACTOR
         of 500000 this gives and image size of 0.5 to 2 MBytes.
         Specifying a sampling factor (even if 1) will turn off
         the autosampler.

   NOTE: format of image frame file is the following

      x   : x dimension of unsampled global grid    (4   bytes int  )
      y   : y dimension of unsampled global grid    (4   bytes int  )
      t   : number of time steps (image frames)     (4   bytes int  )
      dt  : time step increment (seconds)           (4   bytes float)

      not used
      s   : global sampling factor                  (4   bytes int  )

   followed by t*(n "subgrids") of format

      x   : x dimension of unsampled subgrid        (4   bytes int)
      y   : y dimension of unsampled subgrid        (4   bytes int)
      x0  : x global starting coordinate of subgrid (4   bytes int)
      y0  : y global starting coordinate of subgrid (4   bytes int)
      f   : sampling factor of subgrid (withglobal) (4   bytes int)
      data: field data                              (4*n bytes float)
            where n = ((x-1)/f + 1)*((y-1)/f + 1)

   author: shawn larsen
   date  : october 1990
   where : caltech
*****************************************************************************/

#include <stdio.h>
#include <math.h>

#define GHEADER  4       /* number of  global header values        */
#define SHEADER  5       /* number of subgrid header values        */
#define SFACTOR  500000  /* minimum autosample image size          */

struct rbytes {
      int        real;   /* actual   bytes read (including header) */
      int    apparent;   /* apparent bytes read (including header) */
      };
typedef struct rbytes Rbytes;

int
getimageframe(V, Vx, Vy, file)
float       *V;
int     Vx, Vy;
FILE     *file;
{
   Rbytes rsize, sumsize, getsubgrid();
   int g;

   /* initialize */
   sumsize.real     = 0;
   sumsize.apparent = 0;
   g                = 0;
   
   /* loop over subgrids until size limit reached */
   do {
      rsize = getsubgrid(V, Vx, Vy, file);
      if (rsize.real < 0 || rsize.apparent < 0) return(-1);
      sumsize.real     += rsize.real;
      sumsize.apparent += rsize.apparent;
      g++;
      }
   while((sumsize.apparent - g*(SHEADER)*sizeof(int)) < (Vx*Vy*sizeof(float)));

   /* return size */
   if ((sumsize.apparent - g*(SHEADER)*sizeof(int)) != (Vx*Vy*sizeof(float))) 
      return(-1);
   return(sumsize.real);

}


Rbytes
getsubgrid(V, Vx, Vy, file)
float *V;
int Vx, Vy;
FILE *file;
{
   Rbytes rsize;
   float *S;
   int i, j, Sx, Sy, Gx, Gy, x0, y0, f;
   int size;

   /* initialize rsize to bad values */
   rsize.real     = -1;
   rsize.apparent = -1;

   /* get subgrid header and test for EOF */
   size = sizeof(int);
   if (!fread(&Gx, size, 1, file)) return(rsize);
   if (!fread(&Gy, size, 1, file)) return(rsize);
   if (!fread(&x0, size, 1, file)) return(rsize);
   if (!fread(&y0, size, 1, file)) return(rsize);
   if (!fread(&f , size, 1, file)) return(rsize);

   /* determine if subgrid is within global grid */
   if (x0 < 0 || y0 < 0 || (x0 + Gx) > Vx || (y0 + Gy) > Vy) {
      fprintf(stderr, "Error: bad subgrid size\n");
      exit(-1);
      }

   /* determine dimensions of sampled subgrid */
   Sx = (Gx - 1)/f + 1;
   Sy = (Gy - 1)/f + 1;

   /* allocate memory for sampled subgrid */
   S = (float *) malloc(Sx*Sy*sizeof(float));

   /* read subgrid */
   size = sizeof(float);
   if (fread(S, size, Sx*Sy, file) == EOF) {
      free(S);
      return(rsize);
      }

   /* copy sampled subgrid to unsampled global grid and fill */
   for(j = 0; j < Gy; j++) for(i = 0; i < Gx; i++)
      V[(x0 + i) + Vx*(y0 + j)] = S[(i/f) + Sx*(j/f)];

   /* free memory for sampled subgrid */
   free(S);

   /* return number of data and header bytes read */
   rsize.real      = (SHEADER)*sizeof(int);
   rsize.real     += Sx*Sy*sizeof(float);
   rsize.apparent  = (SHEADER)*sizeof(int);
   rsize.apparent += Gx*Gy*sizeof(float);
   return(rsize);

}


int
getsampler(V, Vx, Vy, sample)
float       *V;
int     Vx, Vy;
{
   double sqrt();
   int Sx, Sy;
   void sampleV();

   /* get auto sample factor for V */
   if (sample <= 0) {
      sample = (int) sqrt((double) ((Vx*Vy - 1)/(SFACTOR))) + 1;
      if (sample < 1) sample = 1;
      }

   /* get sampled V */
   Sx = (Vx - 1)/sample + 1;
   Sy = (Vy - 1)/sample + 1;
   if (sample > 1) sampleV(V, Vx, Vy, sample);

   /* return sampling factor */
   return(sample);

}


void
sampleV(V, Vx, Vy, sample)
float *V;
int Vx, Vy, sample;
{
   int i, j, index = 0;

   /* return if no sampling */
   if (sample <= 1) return;

   /* sample grid in place */
   for(j = 0; j < Vy; j += sample) for(i = 0; i < Vx; i += sample) 
      V[index++] = V[i + Vx*j];

}


int
putimage(V, Vx, Vy, file)
float       *V;
int     Vx, Vy;
FILE     *file;
{
   fwrite(&Vx, sizeof(int  ),     1, file); 
   fwrite(&Vy, sizeof(int  ),     1, file);
   fwrite(V  , sizeof(float), Vx*Vy, file);
}


int
getheader(Vx, Vy, t, dt, file)
int  *Vx, *Vy, *t, *dt;
FILE             *file;
{
   if (!fread(Vx, sizeof(int  ), 1, file)) return(-1);
   if (!fread(Vy, sizeof(int  ), 1, file)) return(-1);
   if (!fread(t , sizeof(int  ), 1, file)) return(-1);
   if (!fread(dt, sizeof(float), 1, file)) return(-1);
   return(0);
}

