/******************************************************************************
   active: handles various propagating grid scenarios

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   compile: cc -g -c active.c

   note: input units are km, s, km/s; g/cm3; dyne-cm;
   note: used  units are mks (m, s, m/s, kg/m3; N-m)

   author: shawn larsen
   author: craig schultz (getActiveGrid3: one-way propagation)
   date  : june 1991
   where : caltech
*****************************************************************************/

#include "elas.h"

getActiveGrid(par, source, grid, data, t)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
int t;
{
   int g;
   switch(par->active) {
      case ACTIVEDEFAULT :
         for(g = 0; g < par->g; g++) 
         getActiveGrid1(par, source, grid + g, data + g, t);
         break;
      case ACTIVESURFACE :
         for(g = 0; g < par->g; g++) 
         getActiveGrid2(par, source, grid + g, data + g, t);
         break;
      case ACTIVEREDUCING:
         for(g = 0; g < par->g; g++)
         getActiveGrid3(par, source, grid + g, data + g, t);
         break;
      case ACTIVEBOX:
         for(g = 0; g < par->g; g++)
         getActiveGrid4(par, source, grid + g, data + g, t);
         break;
      default:
         fprintf(stderr, "ERROR: undefined active grid type\n");
         exit(-1);
      }
}


getActiveGrid1(par, source, grid, data, t)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
int                t;
{
   float x, y, z, h, s;
   int i, lmin, lmax, mmin, mmax, nmin, nmax, smin, smax;

   /* initialize */
   if (par->f > 0) {
      data->lmin = grid->l; data->lmax = 0;
      data->mmin = grid->m; data->mmax = 0;
      data->nmin = grid->n; data->nmax = 0;
      }

   /* get maximum travel distance */
   s = ((float) t)*par->dt*par->maxP; 

   /* get limits */
   for(i = 0; i < par->f; i++) {
      x = source[i].x; y = source[i].y; z = source[i].z;
      if (source[i].type == BSOURCE) {
         x += source[i].s0*sin(source[i].strike/(RAD));
         y += source[i].s0*cos(source[i].strike/(RAD));
         h  = source[i].d0*cos(source[i].dip   /(RAD));
         if (source[i].dip < 0.) h = -h;
         x += h*cos(source[i].strike/(RAD));
         y -= h*sin(source[i].strike/(RAD));
         z  = source[i].depth + fabs(source[i].d0*sin(source[i].dip/(RAD)));
         }
      smin = (int) (((x - grid->x0 - s)/grid->dx) - 5.);
      smax = (int) (((x - grid->x0 + s)/grid->dx) + 5.);
      nmin = (smin <       0) ? (      0) : (smin);
      nmax = (smax > grid->n) ? (grid->n) : (smax);
      smin = (int) (((y - grid->y0 - s)/grid->dy) - 5.);
      smax = (int) (((y - grid->y0 + s)/grid->dy) + 5.);
      lmin = (smin <       0) ? (      0) : (smin);
      lmax = (smax > grid->l) ? (grid->l) : (smax);
      smin = (int) (((z - grid->z0 - s)/grid->dz) - 5.);
      smax = (int) (((z - grid->z0 + s)/grid->dz) + 5.);
      mmin = (smin <       0) ? (      0) : (smin);
      mmax = (smax > grid->m) ? (grid->m) : (smax);
      data->lmin = (data->lmin < lmin) ? (data->lmin) : (lmin);
      data->lmax = (data->lmax > lmax) ? (data->lmax) : (lmax);
      data->mmin = (data->mmin < mmin) ? (data->mmin) : (mmin);
      data->mmax = (data->mmax > mmax) ? (data->mmax) : (mmax);
      data->nmin = (data->nmin < nmin) ? (data->nmin) : (nmin);
      data->nmax = (data->nmax > nmax) ? (data->nmax) : (nmax);
      }

   /* add for join-sub CAS: change propagating grid to remove  */
   /* vertical motions - note that xjoin starts 2 grid points  */
   /* after the input x-location - this is shifted back to the */
   /* absorbing boundary once the driving function lasting for */
   /* par->tjoin time steps has passed                         */
   if(par->join == 1) {
      data->lmin = 0; data->lmax = grid->l;
      data->mmin = 0; data->mmax = grid->m;
      data->nmin = par->xjoin + 2;
      if(t > par->tjoin) data->nmin = 0;
      }
   /* in this case the x-location is at the usual point used */
   /* to incorporate the absorbing boundary */
   if(par->join == 2){
      data->lmin = 0; data->lmax = grid->l;
      data->mmin = 0; data->mmax = grid->m;
      }       

   /* take into account surface condition */
   mmin = 0;
   if ((grid->bz1 == SURFACE) || (grid->bz1 == TOPOGRAPHY)) mmin = 1;
   data->mmin = (data->mmin > mmin) ? (data->mmin) : (mmin);
   
}


getActiveGrid2(par, source, grid, data, t)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
int                t;
{
   float x, y, z, h, s;
   int i, lmin, lmax, mmin, mmax, nmin, nmax, smin, smax;

   /* initialize */
   if (par->f > 0) {
      data->lmin = grid->l; data->lmax = 0;
      data->mmin = grid->m; data->mmax = 0;
      data->nmin = grid->n; data->nmax = 0;
      }

   /* get maximum travel distance */
   s = ((float) t)*par->dt*par->maxP; 

   /* get limits */
   for(i = 0; i < par->f; i++) {
      x = source[i].x; y = source[i].y; z = source[i].z;
      if (source[i].type == BSOURCE) {
         x += source[i].s0*sin(source[i].strike/(RAD));
         y += source[i].s0*cos(source[i].strike/(RAD));
         h  = source[i].d0*cos(source[i].dip   /(RAD));
         if (source[i].dip < 0.) h = -h;
         x += h*cos(source[i].strike/(RAD));
         y -= h*sin(source[i].strike/(RAD));
         z  = source[i].depth + fabs(source[i].d0*sin(source[i].dip/(RAD)));
         }
      smin = (int) (((x - grid->x0 - s)/grid->dx) - 5.);
      smax = (int) (((x - grid->x0 + s)/grid->dx) + 5.);
      nmin = (smin <       0) ? (      0) : (smin);
      nmax = (smax > grid->n) ? (grid->n) : (smax);
      smin = (int) (((y - grid->y0 - s)/grid->dy) - 5.);
      smax = (int) (((y - grid->y0 + s)/grid->dy) + 5.);
      lmin = (smin <       0) ? (      0) : (smin);
      lmax = (smax > grid->l) ? (grid->l) : (smax);
      smin = (int) (((z - grid->z0 - s)/grid->dz) - 5.);
      smax = (int) (((z - grid->z0 + s)/grid->dz) + 5.);
      mmin = (smin <       0) ? (      0) : (smin);
      mmax = (smax > grid->m) ? (grid->m) : (smax);
      data->lmin = (data->lmin < lmin) ? (data->lmin) : (lmin);
      data->lmax = (data->lmax > lmax) ? (data->lmax) : (lmax);
      data->mmin = (data->mmin < mmin) ? (data->mmin) : (mmin);
      data->mmax = (data->mmax > mmax) ? (data->mmax) : (mmax);
      data->nmin = (data->nmin < nmin) ? (data->nmin) : (nmin);
      data->nmax = (data->nmax > nmax) ? (data->nmax) : (nmax);
      }

   /* get vertical grid limits for surface only output */
   s = ((float) (par->t - t))*par->dt*par->maxP + 5*par->dz;
   mmax = data->mmax;
   if (s < grid->z0) mmax = 0;
   if (s < grid->z0 + grid->z) mmax = (int) ((s - grid->z0)/grid->dz);
   data->mmax = (data->mmax < mmax) ? (data->mmax) : (mmax);


   /* take into account surface condition */
   mmin = 0;
   if ((grid->bz1 == SURFACE) || (grid->bz1 == TOPOGRAPHY)) mmin = 1;
   data->mmin = (data->mmin > mmin) ? (data->mmin) : (mmin);
   
}



getActiveGrid3(par, source, grid, data, t)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
int                t;
{
   float x, y, z, h, s, ds, vred;
   int i, lmin, lmax, mmin, mmax, nmin, nmax, smin, smax;

   /* initialize */
   vred = par->vred;
   if (vred <= 0.) vred = par->maxP;
   if (vred <  par->maxP) {
      fprintf(stderr, "Error: reducing velocity less than maximum velocity: ");
      fprintf(stderr, "vred = %f, maxP = %f\n", vred, grid->maxP);
      exit(-1);
      }
   if (par->f > 0) {
      data->lmin = grid->l; data->lmax = 0;
      data->mmin = grid->m; data->mmax = 0;
      data->nmin = grid->n; data->nmax = 0;
      }

   /* get maximum travel distance */
   s = ((float) t)*par->dt*par->maxP; 

   /* get minimum distance for reducing velocity */
   ds = (((float) t)*par->dt - par->tlen)*vred;

   /* get limits */
   for(i = 0; i < par->f; i++) {
      x = source[i].x; y = source[i].y; z = source[i].z;
      if (source[i].type == BSOURCE) {
         x += source[i].s0*sin(source[i].strike/(RAD));
         y += source[i].s0*cos(source[i].strike/(RAD));
         h  = source[i].d0*cos(source[i].dip   /(RAD));
         if (source[i].dip < 0.) h = -h;
         x += h*cos(source[i].strike/(RAD));
         y -= h*sin(source[i].strike/(RAD));
         z  = source[i].depth + fabs(source[i].d0*sin(source[i].dip/(RAD)));
         }
      smin = 0;   /* won't work if dx varies laterally */
      if (x + ds > 0) smin=(int) (((x + ds)/grid->dx) - 5.);
      smin -= grid->n0;
      if (smin <       0) smin = 0;
      if (smin > grid->n) smin = grid->n;
      nmin = (smin > grid->n) ? (grid->n ) : (  smin);
      nmin = (smin <       0) ? (      0 ) : (  nmin);
      /* integer formulation
      nmin = (n    >   -dmin) ? (n + dmin) : (     0);
      nmin = (nmin > grid->n) ? (grid->n ) : (  nmin);
      */
      smax = (int) (((x - grid->x0 + s)/grid->dx) + 5.);
      nmax = (smax > grid->n) ? (grid->n) : (smax);
      smin = (int) (((y - grid->y0 - s)/grid->dy) - 5.);
      smax = (int) (((y - grid->y0 + s)/grid->dy) + 5.);
      lmin = (smin <       0) ? (      0) : (smin);
      lmax = (smax > grid->l) ? (grid->l) : (smax);
      smin = (int) (((z - grid->z0 - s)/grid->dz) - 5.);
      smax = (int) (((z - grid->z0 + s)/grid->dz) + 5.);
      mmin = (smin <       0) ? (      0) : (smin);
      mmax = (smax > grid->m) ? (grid->m) : (smax);
      data->lmin = (data->lmin < lmin) ? (data->lmin) : (lmin);
      data->lmax = (data->lmax > lmax) ? (data->lmax) : (lmax);
      data->mmin = (data->mmin < mmin) ? (data->mmin) : (mmin);
      data->mmax = (data->mmax > mmax) ? (data->mmax) : (mmax);
      data->nmin = (data->nmin < nmin) ? (data->nmin) : (nmin);
      data->nmax = (data->nmax > nmax) ? (data->nmax) : (nmax);
      }

   /* get vertical grid limits for end of calculation */
   s = ((float) (par->t - t))*par->dt*par->maxP + 5*par->dz;
   mmax = data->mmax;
   if (s < grid->z0) mmax = 0;
   if (s < grid->z0 + grid->z) mmax = (int) ((s - grid->z0)/grid->dz);
   data->mmax = (data->mmax < mmax) ? (data->mmax) : (mmax);

   /* take into account surface condition */
   mmin = 0;
   if ((grid->bz1 == SURFACE) || (grid->bz1 == TOPOGRAPHY)) mmin = 1;
   data->mmin = (data->mmin > mmin) ? (data->mmin) : (mmin);
   
}


int
getActive(par)
Parameter *par;
{
   int i, time, lmin, lmax, mmin, mmax, nmin, nmax, *flag, total;
   char line[82];
   FILE *file;

   /* return if not active box */
   if (par->active != ACTIVEBOX) return(0);

   /* open active grid file */
   file = fopen(par->afile, "r");
   if (file == NULL) {
      fprintf(stderr, "Error: Can't open active grid file %s\n", par->afile);
      exit(1);
      }

   /* get flag array and initialize values to null */
   flag = (int *) malloc(par->t*sizeof(int));
   for(i = 0; i < par->t; i++) flag[i] = 0;
   total = 0;

   /* read file and input */
   while(fgets(line, 82, file)) {
      sscanf(line, "%d %d %d %d %d %d %d", 
             &time, &lmin, &lmax, &mmin, &mmax, &nmin, &nmax);
      /* go from inclusive input to exclusive output */
      lmax++; mmax++; nmax++; 
      /*
      if (time  < 1     ) time =      1;
      if (time  > par->t) time = par->t;
      */
      if (time  < 1     ) continue;
      if (time  > par->t) continue;
      time--;
      if (lmin  < 0     ) lmin =      0;
      if (mmin  < 0     ) mmin =      0;
      if (nmin  < 0     ) mmin =      0;
      if (lmax  > par->l) lmax = par->l; 
      if (mmax  > par->m) mmax = par->m; 
      if (nmax  > par->n) nmax = par->n; 
      if (par->d == 2) {lmin = 0; lmax = 1;}
      if (lmin >= lmax || mmin >= mmax || nmin >= nmax) {
         fprintf(stderr, "Error: bad active input values\n");
         exit(1);
         }
      par->almin[time] = lmin;
      par->almax[time] = lmax;
      par->ammin[time] = mmin;
      par->ammax[time] = mmax;
      par->anmin[time] = nmin;
      par->anmax[time] = nmax;
      flag[time] = 1;
      total++;
      }
   if (total == 0) {
      fprintf(stderr, "Error: bad active box input file\n");
      exit(1);
      }

   /* make sure first timestep is filled */
   for(i = 0; i < par->t; i++) {
      if (flag[i]) {
         if (i != 0) {
            par->almin[0] = par->almin[i];
            par->almax[0] = par->almax[i];
            par->ammin[0] = par->ammin[i];
            par->ammax[0] = par->ammax[i];
            par->anmin[0] = par->anmin[i];
            par->anmax[0] = par->anmax[i];
            }
         break;
         }
      }
   
   /* process remaining points */
   for(i = 1; i < par->t; i++) {
      if (!flag[i]) {
         par->almin[i] = par->almin[i-1];
         par->almax[i] = par->almax[i-1];
         par->ammin[i] = par->ammin[i-1];
         par->ammax[i] = par->ammax[i-1];
         par->anmin[i] = par->anmin[i-1];
         par->anmax[i] = par->anmax[i-1];
         }
      }

   /* close */
   free(flag);
   fclose(file);
   
}




/* doesn't work with variable density grid */
getActiveGrid4(par, source, grid, data, t)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
int                t;
{
   int lmin, lmax, mmin, mmax, nmin, nmax; 

   /* get local grid values */
   lmin = par->almin[t] - grid->l0;
   lmax = par->almax[t] - grid->l0;
   mmin = par->ammin[t] - grid->m0;
   mmax = par->ammax[t] - grid->m0;
   nmin = par->anmin[t] - grid->n0;
   nmax = par->anmax[t] - grid->n0;

   /* adjust values */
   if (lmin <       0) lmin =       0;
   if (lmax > grid->l) lmax = grid->l;
   if (mmin <       0) mmin =       0;
   if (mmax > grid->m) mmax = grid->m;
   if (nmin <       0) nmin =       0;
   if (nmax > grid->n) nmax = grid->n;

   /* set values */
   data->lmin = lmin;
   data->lmax = lmax;
   data->mmin = mmin;
   data->mmax = mmax;
   data->nmin = nmin;
   data->nmax = nmax;

   /* take into account surface condition */
   mmin = 0;
   if ((grid->bz1 == SURFACE) || (grid->bz1 == TOPOGRAPHY)) mmin = 1;
   data->mmin = (data->mmin > mmin) ? (data->mmin) : (mmin);
   
}

