/******************************************************************************
   attenuate: routines for viscoelastic attenuation finite-difference 

    ================================================================= 
   | (c) Copyright 1994 The Regents of the University of California. |
   |     All Rights Reserved.                                        |
   |                                                                 |
   |     This work was produced under the sponsorship of the         |
   |     United States Department of Energy.                         |
   |     The Government retains certain rights therein.              |
    ================================================================= 

   attenuate2D: 2D viscoelastic attenuation

   par  - input parameters
   grid - input parameters for grid
   data - input data
   Vx   - velocity in x direction
   Vy   - velocity in y direction
   Vz   - velocity in z direction
   Txx  - stress (normal x)
   Tyy  - stress (normal y)
   Tzz  - stress (normal z)
   Txy  - stress (shear   )
   Txz  - stress (shear   )
   Tyz  - stress (shear   )
   l    - grid points in y direction (plane )  
   m    - grid points in z direction (row   ) (positive down)
   n    - grid points in x direction (column)

   compile: cc -g -c attenuate.c

   note: used  units are mks (m, s, m/s, kg/m3)

   author: shawn larsen
   date  : august 1994
   where : llnl
*****************************************************************************/

#include "elas.h"

attenuate2D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float dt, *dv1, *dv2, c1, c2, ***Vx, ***Vz; 
   float *Txx, *Tzz, *Txz, *lamda, *mu;
   float *F0, *F1, *F2, *F3, *q1, *q2, *q3;
   float s1, s2, s3, D1, D2; 
   int d, l, m, n, mmin;

   /* initialize and allocate memory for derivative arrays */
   Vx = data->Vx; 
   Vz = data->Vz; 
   dt = par->dt;
   c1 = C1/grid->dx;
   c2 = C2/grid->dx;
   dv1 = (float *) malloc(grid->n*sizeof(float));
   dv2 = (float *) malloc(grid->n*sizeof(float));
   d = data->nmax - data->nmin;
   l = 0;
   mmin = data->mmin;
   /*if (mmin < 1) mmin = 1;*/
   if (grid->bz1 == SURFACE && mmin < 1) mmin = 1;

   /* calculate dVx/dx and dVz/dz derivatives; update q1,q2,q3; stress */
   for(m = mmin; m < data->mmax; m++) {
      n = data->nmin;
      Txx = data->Txx[l][m]; Tzz   = data->Tzz  [l][m];
      F0  = grid->F0 [l][m]; F1    = grid->F1   [l][m]; 
      F2  = grid->F2 [l][m]; F3    = grid->F3   [l][m];
      q1  = data->q1 [l][m]; q2    = data->q2   [l][m];
      mu  = grid->mu [l][m]; lamda = grid->lamda[l][m];
      /* use c derivatives 
      getDV(dv1+n,Vx[l][m]+n-1,Vx[l][m]+n,Vx[l][m]+n+1,Vx[l][m]+n+2,c1,c2,d);
      getDV(dv2+n,Vz[l][m-2]+n,Vz[l][m-1]+n,Vz[l][m]+n,Vz[l][m+1]+n,c1,c2,d);
      for(n = data->nmin; n < data->nmax; n++) {
         s1 = q1[n];
         s2 = q2[n];
         q1[n]=F0[n]*s1 + F1[n]*(dv1[n] + dv2[n]) + F2[n]*dv2[n];
         q2[n]=F0[n]*s2 + F1[n]*(dv1[n] + dv2[n]) + F2[n]*dv1[n];
         s1 = (s1 + q1[n])*0.5;
         s2 = (s2 + q2[n])*0.5;
         D1 = (lamda[n] + 2.*mu[n]);
         D2 = (lamda[n]);
         Txx[n] += dt*(D1*dv1[n] + D2*dv2[n] + s1);
         Tzz[n] += dt*(D2*dv1[n] + D1*dv2[n] + s2);
         }
      */
      dvf_(dv1+n,Vx[l][m]+n-1,Vx[l][m]+n,Vx[l][m]+n+1,Vx[l][m]+n+2,&c1,&c2,&d);
      dvf_(dv2+n,Vz[l][m-2]+n,Vz[l][m-1]+n,Vz[l][m]+n,Vz[l][m+1]+n,&c1,&c2,&d);
      da1f_(q1+n,q2+n,Txx+n,Tzz+n,lamda+n,mu+n,F0+n,F1+n,F2+n,dv1+n,dv2+n,&dt,&d);
      }

   /* calculate dVx/dz and dVz/dx derivatives; update e3; stress */
   for(m = mmin; m < data->mmax; m++) {
      n = data->nmin;
      F0  = grid->F0 [l][m]; F3 = grid->F3[l][m];
      Txz = data->Txz[l][m]; 
      q3  = data->q3 [l][m]; 
      mu  = grid->mu [l][m];
      /* use c derivatives and formulation
      getDV(dv2+n,Vx[l][m-1]+n,Vx[l][m]+n,Vx[l][m+1]+n,Vx[l][m+2]+n,c1,c2,d);
      getDV(dv1+n,Vz[l][m]+n-2,Vz[l][m]+n-1,Vz[l][m]+n,Vz[l][m]+n+1,c1,c2,d);
      for(n = data->nmin; n < data->nmax; n++) {
         s3 = q3[n];
         q3[n] = F0[n]*s3 + F3[n]*(dv1[n] + dv2[n]);
         s3 = (s3 + q3[n])*0.5;
         Txz[n] += dt*mu[n]*(dv1[n] + dv2[n]) +dt*s3;
         }
      */
      dvf_(dv2+n,Vx[l][m-1]+n,Vx[l][m]+n,Vx[l][m+1]+n,Vx[l][m+2]+n,&c1,&c2,&d);
      dvf_(dv1+n,Vz[l][m]+n-2,Vz[l][m]+n-1,Vz[l][m]+n,Vz[l][m]+n+1,&c1,&c2,&d);
      da2f_(q3+n, Txz+n, mu+n, F0+n, F3+n, dv1+n, dv2+n, &dt, &d);
      }

   /* free memory */
   free(dv1);
   free(dv2);

}



/******************************************************************************
   getAttenuation: convert Q parameters into finite-difference constants (2D)

   usage: getAttenuation(par, grid)
      par  - input parameters
      grid - input parameters for grid

   note: rT  is stress relaxation time for P and SV
   note: rEp is strain relaxation time for P        
   note: rEs is strain relaxation time for       SV
******************************************************************************/
getAttenuation(par, grid)
Parameter *par;
FdGrid    *grid;
{
   float ***lamda, ***mu, Qp, Qs, Qw, rT, rEp, rEs, F0, F1, F2, F3, dt;
   float getF02D(), getF12D(), getF22D(), getF32D();
   int l, m, n;

   /* initialize */
   lamda = grid->lamda;
   mu    = grid->mu;
   dt    =  par->dt;

   /* 2D version */
   if (par->d == 2) {
      for(m = 0; m < grid->m; m++) {
         for(n = 0; n < grid->n; n++) {
            Qp = grid->Qp[0][m][n];
            Qs = grid->Qs[0][m][n];
            Qw = grid->Qw[0][m][n];
            /*
            Qp /= (PI);
            Qs /= (PI);
            */
            /* Shawn Q min method */
            rT  = (sqrt(Qp*Qp + 1) - 1)/(Qp*Qw);
            rEp = (sqrt(Qp*Qp + 1) + 1)/(Qp*Qw);
            rEs = (sqrt(Qs*Qs + 1) + 1)/(Qs*Qw);
            /**/
            /* Rice Q min method (one relaxation method only) */
            /*
            rT  = 1./Qw;
            rEp = (2. + Qp)/(Qw*Qp);
            rEs = (2. + Qs)/(Qw*Qs);
            */
            /**/
            F0 = getF02D(rT, dt);
            F1 = getF12D(lamda[0][m][n], mu[0][m][n], rT, rEp, dt);
            F2 = getF22D(mu[0][m][n], rT, rEp, rEs, dt);
            F3 = getF32D(mu[0][m][n], rT, rEs, dt);
            grid->F0[0][m][n] = F0;
            grid->F1[0][m][n] = F1;
            grid->F2[0][m][n] = F2;
            grid->F3[0][m][n] = F3;
            }
         }
      }

   /* 3D version */
   if (par->d == 3) {
      for(l = 0; l < grid->l; l++) {
         for(m = 0; m < grid->m; m++) {
            for(n = 0; n < grid->n; n++) {
               Qp = grid->Qp[l][m][n];
               Qs = grid->Qs[l][m][n];
               Qw = grid->Qw[l][m][n];
               /*
               Qp /= (PI);
               Qs /= (PI);
               */
               /* Shawn Q min method */
               rT  = (sqrt(Qp*Qp + 1) - 1)/(Qp*Qw);
               rEp = (sqrt(Qp*Qp + 1) + 1)/(Qp*Qw);
               rEs = (sqrt(Qs*Qs + 1) + 1)/(Qs*Qw);
               /**/
               /* Rice Q min method (one relaxation method only */
               /*
               rT  = 1./Qw;
               rEp = (2. + Qp)/(Qw*Qp);
               rEs = (2. + Qs)/(Qw*Qs);
               */
               /**/
               F0 = getF02D(rT, dt);
               F1 = getF12D(lamda[l][m][n], mu[l][m][n], rT, rEp, dt);
               F2 = getF22D(mu[l][m][n], rT, rEp, rEs, dt);
               F3 = getF32D(mu[l][m][n], rT, rEs, dt);
               grid->F0[l][m][n] = F0;
               grid->F1[l][m][n] = F1;
               grid->F2[l][m][n] = F2;
               grid->F3[l][m][n] = F3;
               }
            }
         }
      }

}



float
getF02D(rT, dt)
float rT, dt;
{
   float F, F0;
   F = dt/(2.*rT);
   F0 = (1. - F)/(1. + F);
   return(F0);
}

float
getF12D(lamda, mu, rT, rEp, dt)
float lamda, mu, rT, rEp, dt;
{
   float F, D1, F1;
   F  = (1. + dt/(2.*rT));
   D1 = (lamda + 2.*mu)/rEp*(1. - rEp/rT);
   F1 = D1*dt/F;
   return(F1);
}

float
getF22D(mu, rT, rEp, rEs, dt)
float mu, rT, rEp, rEs, dt;
{
   float F, D2, F2;
   F  = (1. + dt/(2.*rT));
   /*
   D2 = -2.*mu/rEs*(1. - rEp/rT);
   */
   D2 = -2.*mu/rEs*(1. - rEs/rT);
   F2 = D2*dt/F;
   return(F2);
}

float
getF32D(mu, rT, rEs, dt)
float mu, rT, rEs, dt;
{
   float D3, F3;
   D3 = mu/rEs*(1. - rEs/rT);
   F3 = D3*dt/(1. + dt/(2.*rT));
   return(F3);
}
