/******************************************************************************
   check: check grid and parameters and get run-time statistics

    ================================================================= 
   | (c) Copyright 1992 The Regents of the University of California. |
   |     All Rights Reserved.                                        |
   |                                                                 |
   |     This work was produced under the sponsorship of the         |
   |     United States Department of Energy.                         |
   |     The Government retains certain rights therein.              |
    ================================================================= 

   compile: cc -g -c check.c -lm -lget

   author: shawn larsen
   date  : january 1993
   where : llnl
*****************************************************************************/

#include "elas.h"


/******************************************************************************
   checkGrid: check grid and parameters and get run-time statistics

   usage: checkGrid(par, grid, sac, image)

   par   - parameters
   grid  - finite-difference grid
   sac   - seismograms and paramters
   image - image planes
******************************************************************************/
checkGrid(par, grid, sac, image)
Parameter    *par;
FdGrid      *grid;
Seismogram   *sac;
Image      *image;
{
   float t, dtP, dtS, dh, maxP, maxS, factor;
   int g, error = 0;

   /* check certain parameters */
   if (grid->dx != grid->dz || grid->dx != grid->dy || grid->dy != grid->dz) {
      fprintf(stderr, "Error: dx must equal dy must equal dz\n");
      error = 1;
      }
   if ( par->dt <= 0.0) {
      fprintf(stderr, "Error: dt must be greater than 0.0\n");
      error = 1;
      }
   for(g = 0; g < par->g; g++) if (grid[g].l <= 0) {
      fprintf(stderr, "Error: l must be greater than 0\n"); 
      error = 1;
      }
   for(g = 0; g < par->g; g++) if (grid[g].m <= 0) {
      fprintf(stderr, "Error: m must be greater than 0\n"); 
      error = 1;
      }
   for(g = 0; g < par->g; g++) if (grid[g].n <= 0) {
      fprintf(stderr, "Error: n must be greater than 0\n");
      error = 1;
      }
   if (error) fprintf(stderr, "\n");

   /* print velocity and density limits */
   for(g = 0; g < par->g; g++) printGridLimits(par, grid + g);

   /* check grid for 0 velocity */
   for(g = 0; g < par->g; g++) if (grid[g].minP == 0.) {
      fprintf(stderr, "Error: alpha=0 (P-wave) velocity found in grid\n\n");
      error = 1;
      }

   /* check grid for non-zero shear velocity for acoustic model */
   if (par->model == 2) {
      for(g = 0; g < par->g; g++) if (grid[g].minS!=0. || grid[g].maxS!=0.){
         fprintf(stderr, "Error: non-zero shear velocity for acoustic model\n");
         fprintf(stderr, "\n");
         error = 1;
         }
      }

   /* check courant condition */
   dtP = INFINITY; dtS = INFINITY;
   factor = (par->d == 2) ? (0.606) : (0.494);
   for(g = 0; g < par->g; g++) {
      dh = grid[g].dx;
      maxP = grid[g].maxP;
      maxS = grid[g].maxS;
      dtP = MIN(dtP, ((maxP > 0) ? (factor*dh/maxP) : (INFINITY)));
      dtS = MIN(dtS, ((maxS > 0) ? (factor*dh/maxS) : (INFINITY)));
      }
   printCourant(par->dt, MIN(dtP,dtS));
   if (par->dt > MIN(dtP,dtS)) {
      fprintf(stderr, "Error: Courant condition failed - decrease dt\n\n");
      error = 1;
      }

   /* determine and output runtime statistics */
   /*
   t = (float) (grid->m*grid->n*grid->l)*
       (STOH)*(FLOPPERNODE)/(1.e6)*((float) par->t);
   printRuntime(t);
   */

   /* determine and print hard storage statistics */
   /*
   printDiskUsage(par, grid, sac, image);
   */

   /* exit if error condition */
   if (error) exit(-1);

}




printGridLimits(par, grid)
Parameter *par;
FdGrid    *grid;
{
   fprintf(stderr, "Grid Limits\n");
   fprintf(stderr, "-----------\n");
   fprintf(stderr, "min Vp  = %6.2f km/s\n",     grid->minP/KMTOM);
   fprintf(stderr, "max Vp  = %6.2f km/s\n",     grid->maxP/KMTOM);
   fprintf(stderr, "min Vs  = %6.2f km/s\n",     grid->minS/KMTOM);
   fprintf(stderr, "max Vs  = %6.2f km/s\n",     grid->maxS/KMTOM);
   fprintf(stderr, "min rho = %6.2f gm/cm**3\n", grid->minR/GM3TOKGM3);
   fprintf(stderr, "max rho = %6.2f gm/cm**3\n", grid->maxR/GM3TOKGM3);
   if (par->q) {
      fprintf(stderr, "min Qp = %8.1f\n", grid->minQp);
      fprintf(stderr, "max Qp = %8.1f\n", grid->maxQp);
      fprintf(stderr, "min Qs = %8.1f\n", grid->minQs);
      fprintf(stderr, "max Qs = %8.1f\n", grid->maxQs);
      fprintf(stderr, "min Qf = %8.1f\n", grid->minQw/(2.*PI));
      fprintf(stderr, "max Qf = %8.1f\n", grid->maxQw/(2.*PI));
      }
   fprintf(stderr, "\n");
}




printCourant(dt, ct)
float dt, ct;
{
   fprintf(stderr, "Courant condition requires dt < ct\n");
   fprintf(stderr, "----------------------------------\n");
   fprintf(stderr, "  dt    = %9.6f\n", dt);
   fprintf(stderr, "  ct    = %9.6f\n", ct);
   fprintf(stderr, "  ct/dt = %5.2f\n", ct/dt); 
   fprintf(stderr, "\n");
}



/*
printRuntime(t)
float t;
{
   fprintf(stderr, "Estimated Run Time (Optimized)\n");
   fprintf(stderr, "------------------------------\n\n");
   fprintf(stderr, "Sparc1: %7.3f hours\n", t/SPARC1O);
   fprintf(stderr, "Sparc2: %7.3f hours\n", t/SPARC2O);
   fprintf(stderr, "\n");
}
*/




/*
printDiskUsage(par, grid, sac, image)
Parameter    *par;
FdGrid       *grid;
Seismogram   *sac;
Image        *image;
{
   int g, n, disk = 0;
   g = (grid->m + grid->top + grid->bot)*(grid->n + grid->lhs + grid->rhs);
   n = par->p;
   disk += n*(2*g*sizeof(float) + sizeof(int));
   disk += par->s*(4*sac[0].t*sizeof(float) + 2*sizeof(int));
   fprintf(stderr, "Estimated Disk Usage\n");
   fprintf(stderr, "--------- ---- -----\n");
   fprintf(stderr, "Hard Storage = %d bytes\n", disk);
   fprintf(stderr, "\n");
}
*/
