/******************************************************************************
   damping: damping boundary conditions for 2D/3D elastic finite-difference

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   damp     : amplitude reduction boundary condition

   compile: cc -g -c damping.c 

   note: input units are km, s, km/s, g/cm3
   note: used  units are mks (m, s, m/s, kg/m3, N-m)

   author: shawn larsen
   date  : december 1989
   where : caltech
*****************************************************************************/

#include "elas.h"

damp2D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float **Vx, **Vz;
   float adamp, a1, a2, *cdamp;
   int i, j, m, n, nx, nz, left, right, top, bottom, ndamp, xdamp, zdamp;

   /* initialize */
   ndamp = par->damp;
   adamp = par->adamp;
   left = right = top = bottom = 0;
   if (grid->bx1 == ABSORBING) left   = 1;
   if (grid->bx2 == ABSORBING) right  = 1;
   if (grid->bz1 == ABSORBING) top    = 1;
   if (grid->bz2 == ABSORBING) bottom = 1;
   Vx = data->Vx[0];
   Vz = data->Vz[0];
   nx = grid->n;
   nz = grid->m;
   xdamp = (ndamp < nx) ? (ndamp) : (nx);
   zdamp = (ndamp < nz) ? (ndamp) : (nz);

   /* allocate damping memory */
   cdamp = (float *) malloc(ndamp*sizeof(float));

   /* get damping array */
   a1 = adamp;
   a2 = sqrt(-log(a1)/((ndamp-1)*(ndamp-1)));
   for(i = 0; i < ndamp; i++) cdamp[i] = exp(-(a2*a2*(ndamp-i-1)*(ndamp-i-1)));

   /* left boundary */
   if (left) {
      for(m = 0; m < nz; m++) {
         for(n = 0; n < xdamp; n++) {
            Vx[m][n] *= cdamp[n];
            Vz[m][n] *= cdamp[n];
            }
         }
      }

   /* right boundary */
   if (right) {
      for(m = 0; m < nz; m++) {
         for(n = nx - xdamp; n < nx; n++) {
            Vx[m][n] *= cdamp[nx-n-1];
            Vz[m][n] *= cdamp[nx-n-1];
            }
         }
      }

   /* top boundary */
   if (top) {
      for(m = 0; m < zdamp; m++) {
         for(n = 0; n < nx; n++) {
            Vx[m][n] *= cdamp[m];
            Vz[m][n] *= cdamp[m];
            }
         }
      }

   /* bottom boundary */
   if (bottom) {
      for(m = nz - zdamp; m < nz; m++) {
         for(n = 0; n < nx; n++) {
            Vx[m][n] *= cdamp[nz-m-1];
            Vz[m][n] *= cdamp[nz-m-1];
            }
         }
      }

   /* free memory */
   free(cdamp);

}

damp3D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz;
   float adamp, a1, a2, *cdamp, value;
   int i, j, l, m, n, nx, ny, nz, left, right, top, bottom, front, back; 
   int ndamp, xdamp, ydamp, zdamp;

   /* initialize */
   ndamp = par->damp;
   adamp = par->adamp;
   left = right = top = bottom = front = back = 0;
   if (grid->bx1 == ABSORBING) left   = 1;
   if (grid->bx2 == ABSORBING) right  = 1;
   if (grid->by1 == ABSORBING) front  = 1;
   if (grid->by2 == ABSORBING) back   = 1;
   if (grid->bz1 == ABSORBING) top    = 1;
   if (grid->bz2 == ABSORBING) bottom = 1;
   Vx = data->Vx;
   Vy = data->Vy;
   Vz = data->Vz;
   nx = grid->n;
   ny = grid->l;
   nz = grid->m;
   xdamp = (ndamp < nx) ? (ndamp) : (nx);
   ydamp = (ndamp < ny) ? (ndamp) : (ny);
   zdamp = (ndamp < nz) ? (ndamp) : (nz);

   /* allocate damping memory */
   cdamp = (float *) malloc(ndamp*sizeof(float));

   /* get damping array */
   a1 = adamp;
   a2 = sqrt(-log(a1)/((ndamp-1)*(ndamp-1)));
   for(i = 0; i < ndamp; i++) cdamp[i] = exp(-(a2*a2*(ndamp-i-1)*(ndamp-i-1)));

   /* left boundary */
   if (left) {
      for(l = 0; l < ny; l++) {
         for(m = 0; m < nz; m++) {
              for(n = 0; n < xdamp; n++) {
                 Vx[l][m][n] *= cdamp[n];
                 Vy[l][m][n] *= cdamp[n];
                 Vz[l][m][n] *= cdamp[n];
                 }
            }
           }
      }

   /* right boundary */
   if (right) {
      for(l = 0; l < ny; l++) {
         for(m = 0; m < nz; m++) {
            for(n = nx - xdamp; n < nx; n++) {
               Vx[l][m][n] *= cdamp[nx-n-1];
               Vy[l][m][n] *= cdamp[nx-n-1];
               Vz[l][m][n] *= cdamp[nx-n-1];
               }
            }
         }
      }

   /* top boundary */
   if (top) {
      for(l = 0; l < ny; l++) {
         for(m = 0; m < zdamp; m++) {
            value = cdamp[m];
            getqDV(Vx[l][m], value, nx);
            getqDV(Vy[l][m], value, nx);
            getqDV(Vz[l][m], value, nx);
            /*
            for(n = 0; n < nx; n++) {
               Vx[l][m][n] *= cdamp[m];
               Vy[l][m][n] *= cdamp[m];
               Vz[l][m][n] *= cdamp[m];
               }
            */
            }
         }
      }

   /* bottom boundary */
   if (bottom) {
      for(l = 0; l < ny; l++) {
         for(m = nz - zdamp; m < nz; m++) {
            value = cdamp[nz-m-1];
            getqDV(Vx[l][m], value, nx);
            getqDV(Vy[l][m], value, nx);
            getqDV(Vz[l][m], value, nx);
            /*
            for(n = 0; n < nx; n++) {
               Vx[l][m][n] *= cdamp[nz-m-1];
               Vy[l][m][n] *= cdamp[nz-m-1];
               Vz[l][m][n] *= cdamp[nz-m-1];
               }
            */
            }
         }
      }

   /* front boundary */
   if (front) {
      for(l = 0; l < ydamp; l++) {
         for(m = 0; m < nz; m++) {
            value = cdamp[l];
            getqDV(Vx[l][m], value, nx);
            getqDV(Vy[l][m], value, nx);
            getqDV(Vz[l][m], value, nx);
            /*
            for(n = 0; n < nx; n++) {
               Vx[l][m][n] *= cdamp[l];
               Vy[l][m][n] *= cdamp[l];
               Vz[l][m][n] *= cdamp[l];
               }
            */
            }
         }
      }

   /* back boundary */
   if (back) {
      for(l = ny - ydamp; l < ny; l++) {
         for(m = 0; m < nz; m++) {
            value = cdamp[ny-l-1];
            getqDV(Vx[l][m], value, nx);
            getqDV(Vy[l][m], value, nx);
            getqDV(Vz[l][m], value, nx);
            /*
            for(n = 0; n < nx; n++) {
               Vx[l][m][n] *= cdamp[ny-l-1];
               Vy[l][m][n] *= cdamp[ny-l-1];
               Vz[l][m][n] *= cdamp[ny-l-1];
               }
            */
            }
         }
      }

   /* free memory */
   free(cdamp);

}

damp2Dold(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float factor;
   int mm, nn;
   int m, n;

   /* damp left hand side */
   if (grid->bx1 == ABSORBING) {
      for(n = 0; n < par->damp; n++) {
         nn = par->damp - n;
         factor = 1. - ((float) nn / (float) par->damp)*(1. - par->adamp);
         for(m = 0; m < grid->m; m++) {
            data->Vx[0][m][n] *= factor;
            data->Vz[0][m][n] *= factor;
            }
         }
      }

   /* damp right hand side */
   if (grid->bx2 == ABSORBING) {
      for(n = grid->n - par->damp; n < grid->n; n++) {
         nn = n - (grid->n - par->damp) + 1;
         factor = 1. - ((float) nn / (float) par->damp)*(1. - par->adamp);
         for(m = 0; m < grid->m; m++) {
            data->Vx[0][m][n] *= factor;
            data->Vz[0][m][n] *= factor;
            }
         }
      }

   /* damp top */
   if (grid->bz1 == ABSORBING) {
      /*
      for(m = grid->m - par->damp; m < grid->m; m++) {
      */
      for(m = 0; m < par->damp; m++) {
         mm = par->damp - m;
         factor = 1. - ((float) mm / (float) par->damp)*(1. - par->adamp);
         for(n = 0; n < grid->n; n++) {
            data->Vx[0][m][n] *= factor;
            data->Vz[0][m][n] *= factor;
            }
         }
      }

   /* damp bottom */
   if (grid->bz2 == ABSORBING) {
      for(m = grid->m - par->damp; m < grid->m; m++) {
         mm = m - (grid->m - par->damp) + 1;
         factor = 1. - ((float) mm / (float) par->damp)*(1. - par->adamp);
         for(n = 0; n < grid->n; n++) {
            data->Vx[0][m][n] *= factor;
            data->Vz[0][m][n] *= factor;
            }
         }
      }

}



damp3Dold(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, **Vx2, **Vy2, **Vz2, *Vx1, *Vy1, *Vz1;
   float factor;
   int ll, mm, nn;
   int l, m, n;

   Vx = data->Vx;
   Vy = data->Vy;
   Vz = data->Vz;

   /* damp left hand side */
   if (grid->bx1 == ABSORBING) {
      for(n = 0; n < par->damp; n++) {
         nn = par->damp - n;
         factor = 1. - ((float) nn / (float) par->damp)*(1. - par->adamp);
         for(l = 0; l < grid->l; l++) {
            Vx2 = Vx[l];
            Vy2 = Vy[l];
            Vz2 = Vz[l];
            for(m = 0; m < grid->m; m++) {
               Vx2[m][n] *= factor;
               Vy2[m][n] *= factor;
               Vz2[m][n] *= factor;
               }
            }
         }
      }

   /* damp right hand side */
   if (grid->bx2 == ABSORBING) {
      for(n = grid->n - par->damp; n < grid->n; n++) {
         nn = n - (grid->n - par->damp) + 1;
         factor = 1. - ((float) nn / (float) par->damp)*(1. - par->adamp);
         for(l = 0; l < grid->l; l++) {
            Vx2 = Vx[l];
            Vy2 = Vy[l];
            Vz2 = Vz[l];
            for(m = 0; m < grid->m; m++) {
               Vx2[m][n] *= factor;
               Vy2[m][n] *= factor;
               Vz2[m][n] *= factor;
               }
            }
         }
      }

   /* damp top */
   if (grid->bz1 == ABSORBING) {
      /*
      for(m = grid->m - par->damp; m < grid->m; m++) {
      */
      for(m = 0; m < par->damp; m++) {
         mm = par->damp - m;
         factor = 1. - ((float) mm / (float) par->damp)*(1. - par->adamp);
         for(l = 0; l < grid->l; l++) {
            Vx1 = Vx[l][m];
            Vy1 = Vy[l][m];
            Vz1 = Vz[l][m];
            for(n = 0; n < grid->n; n++) {
               Vx1[n] *= factor;
               Vy1[n] *= factor;
               Vz1[n] *= factor;
               }
            }
         }
      }

   /* damp bottom */
   if (grid->bz2 == ABSORBING) {
      for(m = grid->m - par->damp; m < grid->m; m++) {
         mm = m - (grid->m - par->damp) + 1;
         factor = 1. - ((float) mm / (float) par->damp)*(1. - par->adamp);
         for(l = 0; l < grid->l; l++) {
            Vx1 = Vx[l][m];
            Vy1 = Vy[l][m];
            Vz1 = Vz[l][m];
            for(n = 0; n < grid->n; n++) {
               Vx1[n] *= factor;
               Vy1[n] *= factor;
               Vz1[n] *= factor;
               }
            }
         }
      }

   /* damp front side */
   if (grid->by1 == ABSORBING) {
      /*
      for(l = grid->l - par->damp; l < grid->l; l++) {
      */
      for(l = 0; l < par->damp; l++) {
         ll = par->damp - l;
         factor = 1. - ((float) ll / (float) par->damp)*(1. - par->adamp);
         for(m = 0; m < grid->m; m++) {
            Vx1 = Vx[l][m];
            Vy1 = Vy[l][m];
            Vz1 = Vz[l][m];
            for(n = 0; n < grid->n; n++) {
               Vx1[n] *= factor;
               Vy1[n] *= factor;
               Vz1[n] *= factor;
               }
            }
         }
      }

   /* damp back  side */
   if (grid->by2 == ABSORBING) {
      for(l = grid->l - par->damp; l < grid->l; l++) {
         ll = l - (grid->l - par->damp) + 1;
         factor = 1. - ((float) ll / (float) par->damp)*(1. - par->adamp);
         for(m = 0; m < grid->m; m++) {
            Vx1 = Vx[l][m];
            Vy1 = Vy[l][m];
            Vz1 = Vz[l][m];
            for(n = 0; n < grid->n; n++) {
               Vx1[n] *= factor;
               Vy1[n] *= factor;
               Vz1[n] *= factor;
            }
            }
         }
      }

}

/* quick fix for corner dampers */
cdamp2D(par, grid, data, mode)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int             mode;
{
   float **Vx, **Vz;
   float adamp, a1, a2, cdamp[30];
   int i, j, m, n, nx, nz, ctl, ctr, cbl, cbr, ndamp, xdamp, zdamp; 

   /* initialize */
   ndamp = 20;
   adamp = 0.08;
   ctl = ctr = cbl = cbr = 0.;
   if (grid->bx1 == ABSORBING  || grid->bx1 == SURFACE) {ctl = 1; cbl = 1;}
   if (grid->bx1 == TOPOGRAPHY || grid->bx1 == WATER  ) {ctl = 1; cbl = 1;}
   if (grid->bx2 == ABSORBING  || grid->bx2 == SURFACE) {ctr = 1; cbr = 1;}
   if (grid->bx2 == TOPOGRAPHY || grid->bx2 == WATER  ) {ctr = 1; cbr = 1;}
   if (grid->bz1 == ABSORBING  || grid->bz1 == SURFACE) {ctl = 1; ctr = 1;}
   if (grid->bz1 == TOPOGRAPHY || grid->bz1 == WATER  ) {ctl = 1; ctr = 1;}
   if (grid->bz2 == ABSORBING  || grid->bz2 == SURFACE) {cbl = 1; cbr = 1;}
   if (grid->bz2 == TOPOGRAPHY || grid->bz2 == WATER  ) {cbl = 1; cbr = 1;}
   Vx = data->Vx[0];
   Vz = data->Vz[0];
   nx = grid->n;
   nz = grid->m;
   xdamp = (ndamp < nx) ? (ndamp) : (nx);
   zdamp = (ndamp < nz) ? (ndamp) : (nz);

   /* get damping array */
   a1 = 1. - adamp;
   a2 = sqrt(-log(a1)/((ndamp-1)*(ndamp-1)));
   for(i = 0; i < ndamp; i++) cdamp[i] = exp(-(a2*a2*(ndamp-i-1)*(ndamp-i-1)));

   /* top left */
   if (ctl) {
      for(m = 0; m < zdamp; m++) {
         for(n = 0; n < xdamp; n++) {
            Vx[m][n] *= cdamp[n];
            Vz[m][n] *= cdamp[n];
            }
         }
      }

   /* bottom left */
   if (cbl) {
      for(m = nz - zdamp; m < nz; m++) {
         for(n = 0; n < xdamp; n++) {
            /*
            Vx[m][n] *= cdamp[n];
            Vz[m][n] *= cdamp[n];
            */
            Vx[m][n] *= cdamp[nz-m-1];
            Vz[m][n] *= cdamp[nz-m-1];
            }
         }
      }

   /* top right */
   if (ctr) {
      for(m = 0; m < zdamp; m++) {
         for(n = nx - xdamp; n < nx; n++) {
            Vx[m][n] *= cdamp[nx-n-1];
            Vz[m][n] *= cdamp[nx-n-1];
            }
         }
      }

   /* bottom right */
   if (cbr) {
      for(m = nz - zdamp; m < nz; m++) {
         for(n = nx - xdamp; n < nx; n++) {
            /*
            Vx[m][n] *= cdamp[nx-n-1];
            Vz[m][n] *= cdamp[nx-n-1];
            */
            Vx[m][n] *= cdamp[nz-m-1];
            Vz[m][n] *= cdamp[nz-m-1];
            }
         }
      }

}

getqDV(d, c, l)
float *d, c;
int l;
{
   int n;
   for(n = 0; n < l; n++) *(d++) *= c;
}

