/*******************************************************************************
   e3d: 2D/3D elastic finite-difference synthetics on staggered grid (Levandar)

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   usage: e3d [options] < inputFile

   compile: cc -g -o elas elas.c -lm -lget

   note: input units are km, s, km/s, g/cm3, dyne-cm
   note: used  units are mks (m, s, m/s, kg/m3, N-m)

   author: shawn larsen
   date  : may 1990
   where : caltech
*****************************************************************************/

#include "elas.h"
#include "usage.h"

main(argc, argv)
int argc;
char **argv;
{
   Parameter                   par[1];
   Input                     input[1];
   Image                    visual[1];
   FdGrid               grid[MAXGRID];
   VelocityStress       data[MAXGRID];
/*
   FdSource         source[MAXSOURCE];
   Image             image[MAXIMAGE ];
   Seismogram          sac[MAXSAC   ];
   Medium           medium[MAXMEDIUM];
   Traces           traces[MAXTRACES];
*/
   FdSource                   *source;
   Image                       *image;
   Seismogram                    *sac;
   Medium                     *medium;
   Traces                     *traces;

   FILE *getFile(), *file;
   int t, g, node;


   /* initialize parallelization */
   node = mpInit(&argc, &argv);


   /* check for no input (print usage message) */
   if (checkUsage(argc, argv, node)) exit(0);


   /* read input file and parameters */
   file = getFile(argc, argv);
   readInput(input, argc, argv, file);
   expandInput(input);
   if (file != stdin) fclose(file);


   /* determine and get allocation for some structures */
   getSize(input, par, &source, &image, &sac, &medium, &traces);


   /* get input */
   getInput(input, par, source, grid, image, visual, sac, traces, medium);


   /* get memory */
   getMemory(par, source, grid, data, image, sac, traces);


   /* get grid medium parameters */
   getGrid(par, grid, medium);


   /* check initial grid and run-scheme (output statistics) */
   if (node == 0) checkGrid(par, grid, image, sac);


   /* get initial solution (source term) */
   initGrid(par, grid, data, image, sac);


   /* get source time function(s) and distributed slip */
   getSourceF(par,       source);
   getSourceD(par, grid, source);


   /* get topography */
   getTopography(par, grid);


   /* get trace output information */
   getTraces(par, grid, traces);


   /* get active grid (box) values */
   getActive(par);


   /* calculate solution at each time step and output results */
/*
if (par->node == 0) fprintf(stderr, "***using test source***\n");
if (par->node == 0) fprintf(stderr, "***WATER BOUNDARY CONDITIONS***\n");
if (par->node == 0) fprintf(stderr, "***USING SEG VFILE FILLING***\n");
*/
   getClock(par, grid, 0, 0);
   getClock(par, grid, 1, 0);
   for(t = par->restart; t < par->t; t++) {
      addSourceT(par, grid, data, source, t);
      timeStep(par, source, grid, data, t);
      writeOutput(par, grid, data, image, visual, sac, traces, source, t);
      for(g = 0; g < par->g; g++) (data + g)->t++;
      }
   getClock(par, grid, 1, 1);
   getClock(par, grid, 0, 2);
   

   /* free memory */
   freeMemory(par, source, grid, data, image, sac, traces);
   freeSize(input, par, source, image, sac, medium, traces);


   /* close parallelization */
   mpClose();

}





/******************************************************************************
   checkUsage: check usage and print usage message if no input

   usage: checkUsage(argc, argv, node)
 
   return 0: correct usage - continue execution
          1: print   usage - stop     execution

   note: use include file "usage.h"
******************************************************************************/
checkUsage(argc, argv, node)
int    argc, node;
char **argv;
{
   struct stat statbuf;
   char string[256];
   int n = 0;

   /* go into vi editor for -v argument */
   if (argc == 2) if (!strncmp(argv[1], "-v", 2)) {
      if (node == 0) {
         string[0] = (char) '\0';
         strcat(string, "cp "); 
         strcat(string, DOCFILE); 
         strcat(string, " /tmp/e3d.doc");
         system(string);
         system("vi /tmp/e3d.doc");
         system("rm /tmp/e3d.doc");
         }
      return(1);
      }
 
   /* get input status */
   fstat(0, &statbuf);
 
   /* return if normal input */
   if (statbuf.st_mode & REGULAR_FILE) return(0);
   if (statbuf.st_mode & PIPE)         return(0);
   if (argc == 2)                      return(0);
 
   /* print usage and return exit */
   if (node == 0) while(elasUsage[n]) fprintf(stderr, "%s", elasUsage[n++]);
   return(1);

}

