/******************************************************************************
   elas.h: include file for elastic seismic finite differencing and utilities 

   author: shawn larsen
   date:   may 1990
   where:  caltech
******************************************************************************/


/************
INCLUDE FILES
************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <memory.h>
#include <stdlib.h>


/******************
COMPILER DIRECTIVES
******************/


/************
DOCUMENTATION
************/

#define DOCFILE "/us/shawn/src/e3d/doc/e3d.doc"


/*****************
GENERAL PARAMETERS
*****************/

#define PI         3.14159265358979
#define RAD        (180./PI)
#define INFINITY   1.e30
#define EPSILON    1.e-30
#define SQRT2      1.4142136
#define SQRT3      1.7320508



/***************
UNIT CONVERSIONS
***************/

#define KMTOM             1000.0 /* km to meters                       */
#define KMD            111.31953 /* km per degree                      */
#define GM3TOKGM3         1000.0 /* grams/cm**3 to kilograms/meter**3  */
#define DYNECMTONM       (1.e-7) /* dyne-cm to newton-meter (moments)  */
#define STOH          (1./3600.) /* seconds to hours                   */
#define FLOPPERNODE  (3.30*80.0) /* convert computations to MFLOPS     */
#define SPARC1N              1.4 /* speed for nonoptimized sparc1 code *//**/
#define SPARC1O              1.8 /* speed for    optimized sparc1 code *//**/
#define SPARC2N              4.2 /* speed for nonoptimized sparc2 code *//**/
#define SPARC2O              7.0 /* speed for    optimized sparc2 code *//**/



/***********************
FINITE DIFFERENCE LIMITS
***********************/

#define MAXTIMESTEP 100000   /* maximum number of time steps            *//**/  
#define MAXGRID          3   /* maximum number of grids                 */
#define MAXSAC         100   /* maximum number of seismograms on output *//**/
#define MAXTRACES      100   /* maximum number of trace files           *//**/
#define MAXSOURCE      100   /* maximum number of sources               *//**/
#define MAXLINES       100   /* maximum number of input lines (default) */
#define MAXIMAGE       100   /* maximum number of images to output      *//**/
#define MAXMEDIUM      100   /* maximum number of grid medium units     *//**/
#define MAXPOLY        100   /* maximum number of grid polygon points   *//**/
#define MAXNODE        256   /* maximum nodes in any one dimension(file)*//**/



/**********************
GRID MEDIUM CONVERSIONS
**********************/

#define GETMU(alpha, beta, rho)     (rho*beta*beta)
#define GETLAMDA(alpha, beta, rho) ((rho*alpha*alpha)-2.*GETMU(alpha,beta,rho)) 
#define GETALPHA(lamda, mu, rho)   (sqrt(((lamda) + 2.*(mu))/(rho)))
#define GETBETA(lamda, mu, rho)   (sqrt((mu)/(rho)))
#define GETRHOL(alpha)            (252.0+0.3788*(alpha))
#define MIN(a, b)                  ((a < b) ? a : b)
#define MAX(a, b)                  ((a > b) ? a : b)
#define MIN3(a, b, c)         ((a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c))



/*************************
DEFAULT GRID MEDIUM VAULES
*************************/

#define SPRATIO                    0.577350269     /* S/P ratio 1/(sqrt 3)   */
#define ALPHA                      6.0             /* P wave velocity (km/s) */
#define BETA                       (ALPHA*SPRATIO) /* S wave velocity (km/s) */
#define RHO                        3.0             /* density (gm/cm**3)     */
#define RT                         0.0             /* attenuation parameter  */
#define REP                        0.0             /* attenuation parameter  */
#define RES                        0.0             /* attenuation parameter  */
#define QP                     10000.0             /* P attenuation          */
#define QS                     10000.0             /* S attenuation          */
#define QW                         1.0             /* attenuation frequency  */
#define QF                         1.0             /* factor for frequence   */



/**********************************
4TH ORDER DIFFERENCING COEFFICIENTS
**********************************/

#define C1  1.125000000 /* ( 9./8. ) 4th order finite difference coefficient */
#define C2 -0.041666667 /* (-1./24.) 4th order finite difference coefficient */



/****************
OUTPUT FILE NAMES
****************/

#define NAMESAC       "sac"     /* default file root name for sac files    */
#define NAMEIMAGE   "image"     /* default file root name for images       */
#define NAMEVOLUME  "volume"    /* default file root name for volumes      */
#define NAMETRACES "traces"     /* default file root name for ouput traces */



/***************
LINE INPUT NAMES
***************/

#define LINE_COMMENT -2 /* getLine: hard comment (# 1st non-white)           */
#define LINE_NULL    -3 /* getLine: null line ('\0' as first character)      */
#define LINE_BLANK   -4 /* getLine: blank line (all ' ' or '\t' before '\n') */
#define LINE_VALUE   -5 /* getLine: numeric 1st non-white (0-9, +, -, ., \)  */
#define LINEFILE     15 /* read in additional files     */
#define LINEGRID     11 /*   grid parameters            */
#define LINETIME     12 /*   time parameters            */
#define LINESOURCE   13 /* source parameters            */
#define LINEFAULT    14 /* fault source parameters      */
#define LINESAC      17 /* output sac files             */
#define LINETRACES   18 /* output seismic traces        */
#define LINEVISUAL   21 /* output runtime visualization */
#define LINEIMAGE    22 /* output image files           */
#define LINEMEDIUM   23 /* grid medium lines            */
#define LINETOPO     24 /* topography parameters        */
#define LINEPARALLEL 25 /* parallelization parameters   */
#define LINEHYBRID   26 /* hybrid joining technique     */
#define LINEBLOCK    27 /* grid medium lines (block)    */
#define LINEVFILE    28 /* grid medium lines (file)     */
#define LINEPOLY2D   29 /* grid medium lines (2/3d poly)*/
#define LINEPOLY3D   30 /* grid medium lines (2/3d poly)*/
#define LINEVOLUME   31 /* output data volumes          */
#define LINERESTART  32 /* restart paramters            */



/*****************
GRID ELEMENT TYPES
*****************/

#define EMPTY      0       /* nothing                    */
#define BLOCK      1       /* rectangular block          */
#define SLOPE      2       /* sloping layer              */
#define POLYN      3       /* polygon of points          */
#define PLANE      4       /* 3D plane parameterization  */
#define BFILL      5       /* fill rectangular block     */
#define POLY3      6       /* fill 2/3D polygons (lomax) */
#define VFILE     10       /* velocity file              */
#define PFILE     11       /* velocity file (P velocity) */
#define SFILE     12       /* velocity file (S velocity) */
#define RFILE     13       /* velocity file (density)    */
#define QFILE     14       /* velocity file (Qp and Qs)  */
#define YFILE     15       /* velocity file (Qp)         */
#define ZFILE     16       /* velocity file (Qs)         */
#define WFILE     17       /* velocity file (Qw)         */



/************************
VELOCITY FILE INPUT TYPES
************************/

#define ASCIIP          1    /* ascii  input P only     */
#define BINARYP         2    /* binary input P only     */
#define ASCIIALL        3    /* ascii  input all        */
#define BINARYALL       4    /* binary input all        */
#define VMODE      ASCIIP    /* default file input type */



/***********
SOURCE TYPES      
***********/

#define PSOURCE         1    /* compressional source               */ 
#define SSOURCE         2    /* shear source                       */
#define TSOURCE         3    /* compressional (add to stress)      */
#define MSOURCE         4    /* source defined by moment tensor    */
#define FSOURCE         5    /* source defined by forces           */
#define ASOURCE         6    /* source defined by fault parameters */
#define BSOURCE         7    /* fault  defined by fault parameters */
#define CSOURCE         8    /* distributed source from file       */
#define ESOURCE         9    /* distributed source with rise times */ 
#define RSOURCE        10    /* velocity source for reciprocity    */
#define DSOURCE   PSOURCE    /* default source                     */
#define DFAMPM0   (1.e20)    /* default source amplitude (dyne-cm) */



/***********
OUTPUT TYPES
***********/

#define SACS            1    /* print regular seismograms          */
#define SACM            2    /* print moment  seismograms          */
#define TRACEVX         1    /* output Vx traces                   */
#define TRACEVY         2    /* output Vy traces                   */
#define TRACEVZ         4    /* output Vz traces                   */
#define TRACEP          8    /* output P  traces                   */
#define TRACES         16    /* output S  traces                   */
#define WAVEPS          1    /* floating point image array (p, s)  */
#define WAVEXZ          2    /* floating point image array (Vx,Vz) */
#define WAVEIM          3    /* raster image                       */
#define WAVEMV          4    /* maximum velocity array             */
#define WAVEEN          5    /* energy array                       */
#define WAVEVP          6    /* p medium velocity                  */
#define WAVEVS          7    /* s medium velocity                  */
#define WAVEVR          8    /* r medium density                   */
#define WAVEVX         11    /* floating point data array (Vx )    */
#define WAVEVY         12    /* floating point data array (Vy )    */
#define WAVEVZ         13    /* floating point data array (Vz )    */
#define WAVETXX        14    /* floating point data array (Txx)    */
#define WAVETYY        15    /* floating point data array (Tyy)    */
#define WAVETZZ        16    /* floating point data array (Tzz)    */
#define WAVETXY        17    /* floating point data array (Txy)    */
#define WAVETXZ        18    /* floating point data array (Txz)    */
#define WAVETYZ        19    /* floating point data array (Tyz)    */
#define WAVEFP         20    /* floating point data array (P  )    */
#define WAVEFS         21    /* floating point data array (S  )    */
#define VOLUMEP        22    /* floating point volume array (P)(1) */
#define VOLUMES        23    /* floating point volume array (S)(2) */
#define D2PLANE         0    /* plane in xz dimension (2D)         */
#define XZPLANE         1    /* plane in xz dimension (3D)         */
#define XYPLANE         2    /* plane in xy dimension (3D)         */
#define YZPLANE         3    /* plane in yz dimension (3D)         */
#define S2PLANE         4    /* surface plane (line)  (2D)         */
#define S3PLANE         5    /* surface plane         (3D)         */



/*************
BOUNDARY TYPES
*************/

#define RHS             1    /* right hand side boundary */
#define LHS             2    /* left  hand side boundary */
#define BOT             4    /* bottom     side boundary */
#define TOP             8    /* top        side boundary */
#define FNT            16    /* front      side boundary */
#define BCK            32    /* back       side boundary */
#define ALL            63    /* all sides                */
#define REFLECTING      1    /* no boundary condition    */
#define ABSORBING       2    /* clayton and enquist w/o  surface */
#define SURFACE         3    /* clayton and enquist with surface */
#define WATER           4    /* modified SURFACE (for 3-D acstc) */
#define TOPOGRAPHY      5    /* clayton and enquist with topography */
#define MULTIGRID       6    /* variable density grid               */
#define MESSAGEPASS     7    /* message passing                     */

/****************
ACTIVE GRID TYPES
****************/

#define ACTIVEDEFAULT    1    /* default (expands with wavefield)        */
#define ACTIVESURFACE    2    /* default and limits to surface at end    */
#define ACTIVEREDUCING   3    /* oneway prop. at specified reducing vel  */
#define ACTIVEBOX        4    /* moves with predefined active box        */
#define FTLEN         0.75    /* default factor for time length (reduce) */



/*********************************
PARAMETERS FOR FINITE DIFFERENCING
*********************************/

struct elas_parameters {    
            float               dt; /* time step                            */
            int                  t; /* number of time steps                 */
            int                  d; /* dimensions of grid (d=2, d=3)        */
            int                  f; /* number of source terms               */
            int                  p; /* number of image planes               */
            int                  s; /* number of seismogram points          */
            int                  v; /* number of grid medium units          */
            int                  g; /* number of grids                      */
            int                  a; /* number of trace arrays (files)       */
            int              model; /* type of model (1 def; 2 ac; 3 slow)  */
            int              water; /* special "water" b.c. flag (def:0;1y) */
            int            restart; /* restart flag and parameter           */
            int               dump; /* restart dump interval                */
            int             nolast; /* flag to delete last restart file     */
            char         rfile[80]; /* restart file                         */
            int               node; /* parallel node number                 */
            int              nnode; /* number of parallel nodes             */
            int                  l; /* global grid nodes in y direction     */
            int                  m; /* global grid nodes in z direction     */
            int                  n; /* global grid nodes in x direction     */
            float                x; /* global grid dimensions in x (m)      */
            float                y; /* global grid dimensions in y (m)      */
            float                z; /* global grid dimensions in z (m)      */
            float               dx; /* global grid increment (m) (top grid) */
            float               dy; /* global grid increment (m) (top grid) */
            float               dz; /* global grid increment (m) (top grid) */
            float               x0; /* global starting x coordinate (m)     */
            float               y0; /* global starting y coordinate (m)     */
            float               z0; /* global starting z coordinate (m)     */
            int                 nx; /* number of parallel nodes in x        */
            int                 ny; /* number of parallel nodes in y        */
            int                 nz; /* number of parallel nodes in z        */
            char         pfile[80]; /* file defining mpp decomposition      */
            float            minP ; /* global minimum P-wave velocity       */
            float            maxP ; /* global maximum P-wave velocity       */
            float            minS ; /* global minimum S-wave velocity       */
            float            maxS ; /* global maximum S-wave velocity       */
            float            minR ; /* global minimum density               */
            float            maxR ; /* global maximum density               */
            float            minQp; /* global minimum P attenuation         */
            float            maxQp; /* global maximum P attenuation         */
            float            minQs; /* global minimum S attenuation         */
            float            maxQs; /* global maximum S attenuation         */
            float            minQw; /* global minimum attenuation frequency */
            float            maxQw; /* global maximum attenuation frequency */
            int             active; /* flag (mode) for active grid (1-3:d1) */
            float             vred; /* reducing velocity (for active=3)     */
            float             tlen; /* time-window       (for active=3)     */
            int             *almin; /* minimum global l value(for active=4) */
            int             *almax; /* maximum global l value(for active=4) */
            int             *ammin; /* minimum global m value(for active=4) */
            int             *ammax; /* maximum global m value(for active=4) */
            int             *anmin; /* minimum global n value(for active=4) */
            int             *anmax; /* maximum global n value(for active=4) */
            char         afile[82]; /* active grid file name (for active=4) */
            int           boundary; /* boundary condition mode              */
            int                  q; /* flag for attenuation                 */
            int               topo; /* flag for topography                  */
            int                ext; /* density extinguishing flag           */
            int                ave; /* medium averaging flag (default 0)    */
            float            depth; /* depth for flat interface             */
            char         tfile[80]; /* topography file                      */
            int               damp; /* damping flag and damping points      */
            float            adamp; /* damping amplitude                    */
            int              clock; /* do timings                           */
            float            time0; /* reference time                       */
            float            time1; /* start time                           */
            float            time2; /* end time                             */
            float            timec; /* computational time                   */
            float            timep; /* paraxial boundary condition time     */
            float            timem; /* mpp message massing time             */
            int               join; /* 0 = no, 1 = yes, 2 = test case     j */
            int              xjoin; /* n point for vertical joining       j */
            int              tjoin; /* n point for vertical joining       j */
            float              Vx0; /* velocity at the surface            j */
            float              Vz0; /* velocity dx/2 below surface        j */
            float             VxL0; /* velocity at neighboring surface pt j */
            float             VxR0; /* velocity at neighboring surface pt j */
            float             VzL0; /* velocity at neighboring surface pt j */
            float             VzR0; /* velocity at neighboring surface pt j */
            };



/******************************************
WAVE FIELD IN VELOCITY/STRESS STATE VECTORS
******************************************/

struct velocity_stress {
            float  *data;     /* big array to reference all data  */
            float  ***Vx;     /* velocity in x direction          */
            float  ***Vy;     /* velocity in y direction          */
            float  ***Vz;     /* velocity in z direction          */
            float ***Txx;     /* normal stress in x  direction    */
            float ***Tyy;     /* normal stress in y  direction    */
            float ***Tzz;     /* normal stress in z  direction    */
            float ***Txy;     /* shear  stress in xy direction    */
            float ***Txz;     /* shear  stress in xz direction    */
            float ***Tyz;     /* shear  stress in yz direction    */
            float ***T  ;     /* shear  stress for acoustic case  */
            float  ***q1;     /* memory variable (for Txx update) */
            float  ***q2;     /* memory variable (for Tzz update) */
            float  ***q3;     /* memory variable (for Txz update) */
            float  ***q4;     /* memory variable (for Tyy update) */
            float  ***q5;     /* memory variable (for Txy update) */
            float  ***q6;     /* memory variable (for Tyz update) */
            int        t;     /* current time step                */
            int     lmin;     /* minimum active plane (3D)        */
            int     lmax;     /* maximum active plane (3D)        */
            int     mmin;     /* minimum active row               */
            int     mmax;     /* maximum active row               */
            int     nmin;     /* minimum active column            */
            int     nmax;     /* maximum active column            */
            float  **Ux;      /* velocity in x direction          *//**/
            float  **Uz;      /* velocity in y direction          *//**/
            };



/*********************
FINITE-DIFFERENCE GRID
*********************/

struct fd_grid {
            int                   d;  /* dimensions of grid (d=2, d=3)       */
            int                   g;  /* multi grid number (start 0)         */
            int                   s;  /* multigrid sampling factor (power 2) */
            int                   n;  /* grid points x direction (no bound.) */
            int                   l;  /* grid points y direction (no bound.) */
            int                   m;  /* grid points z direction (no bound.) */
            float                 x;  /* local x length of grid              */
            float                 y;  /* local x length of grid              */
            float                 z;  /* local x length of grid              */
            int                   q;  /* flag for attenuation                */
            int                   b;  /* boundary condition type             */
            int                 ave;  /* medium averaging flag (default 0)   */
            int                 top;  /* top  boundary grid points (surface) */
            int                 bot;  /* bottom    boundary grid points      */
            int                 lhs;  /* lhs       boundary grid points      */
            int                 rhs;  /* rhs       boundary grid points      */
            int                 fnt;  /* front     boundary grid points      */
            int                 bck;  /* back      boundary grid points      */
            int                 bx1;  /* boundary condition type on left     */
            int                 bx2;  /* boundary condition type on right    */
            int                 by1;  /* boundary condition type on front    */
            int                 by2;  /* boundary condition type on back     */
            int                 bz1;  /* boundary condition type on top      */
            int                 bz2;  /* boundary condition type on bottom   */
            int                 px1;  /* parallel node on left   (-1 = none) */
            int                 px2;  /* parallel node on right  (-1 = none) */
            int                 py1;  /* parallel node on front  (-1 = none) */
            int                 py2;  /* parallel node on back   (-1 = none) */
            int                 pz1;  /* parallel node on top    (-1 = none) */
            int                 pz2;  /* parallel node on bottom (-1 = none) */
            float                x0;  /* upper-left in global grid           */
            float                y0;  /* upper-left in global grid           */
            float                z0;  /* upper-left in global grid           */
            int                  l0;  /* upper-left location in global grid  */
            int                  m0;  /* upper-left location in global grid  */
            int                  n0;  /* upper-left location in global grid  */
            float                dx;  /* grid spacing in x direction         */
            float                dy;  /* grid spacing in y direction         */
            float                dz;  /* grid spacing in z direction         */
            float          ***alpha;  /* P-wave velocity                     */
            float           ***beta;  /* S-wave velocity                     */
            float            ***rho;  /* density                             */
            float          ***lamda;  /* lame' coefficient                   */
            float           ***mu  ;  /* rigidity                            */
            float           ***muxy;  /* rigidity at xy stress subnode       */
            float           ***muxz;  /* rigidity at xz stress subnode       */
            float           ***muyz;  /* rigidity at yz stress subnode       */
            float          ***irho ;  /* inverse of density                  */
            float          ***irhox;  /* inverse of density at x vel subnode */
            float          ***irhoy;  /* inverse of density at y vel subnode */
            float          ***irhoz;  /* inverse of density at z vel subnode */
            float             ***Qp;  /* P wave attenuation                  */
            float             ***Qs;  /* S wave attenuation                  */
            float             ***Qw;  /* Frequency of attenuation            */
            float             ***F0;  /* FD parameter for attenuation        */
            float             ***F1;  /* FD parameter for attenuation        */
            float             ***F2;  /* FD parameter for attenuation        */
            float             ***F3;  /* FD parameter for attenuation        */
            float               *zi;  /* topography points 2D                */
            float             **ziM;  /* topography points 3D                */
            float              minP;  /* minimum P-wave velocity             */
            float              maxP;  /* maximum P-wave velocity             */
            float              minS;  /* minimum S-wave velocity             */
            float              maxS;  /* maximum S-wave velocity             */
            float              minR;  /* minimum density                     */
            float              maxR;  /* maximum density                     */
            float             minQp;  /* minimum P attenuation               */
            float             maxQp;  /* maximum P attenuation               */
            float             minQs;  /* minimum S attenuation               */
            float             maxQs;  /* maximum S attenuation               */
            float             minQw;  /* minimum attenuation frequency       */
            float             maxQw;  /* maximum attenuation frequency       */
            };



/******************
GRID MEDIUM ELEMENT
******************/

struct grid_element {
            float             alpha;  /* P wave velocity (m/s)          */
            float              beta;  /* S wave velocity (m/s)          */
            float               rho;  /* density (kg/m3)                */
            float          gradient;  /* vertical velocity gradient     */
            float           spratio;  /* S to P ratio                   */
            float                Qp;  /* P wave attenuation             */
            float                Qs;  /* S wave attenuation             */
            float                Qw;  /* frequency of attenuation       */
            float                x1;  /* starting x coordinate          */
            float                x2;  /* ending   x coordinate          */
            float                y1;  /* starting y coordinate          */
            float                y2;  /* ending   y coordinate          */
            float                z1;  /* starting z coordinate          */
            float                z2;  /* ending   z coordinate          */
            float               lon;  /* zero longitude for 3D poly     */
            float               lat;  /* zero latitude  for 3D poly     */
            float             angle;  /* x angle ccw from east 3D poly  */
            /*
            float        x[MAXPOLY];
            float        z[MAXPOLY];
            */
            float                *x;  /* x vertices of polygon          */
            float                *z;  /* z vertices of polygon          */
            int                   p;  /* # polygon vertices (inc close) */
            int                  n1;  /* lhs grid element               */
            int                  n2;  /* rhs grid element               */
            int                  l1;  /* fnt grid element               */
            int                  l2;  /* bck grid element               */
            int                  m1;  /* top grid element               */
            int                  m2;  /* bottom grid element            */
            int                   g;  /* grid element is in             */
            int                   c;  /* compress mode (0:no, 1:rle)    */
            int                type;  /* type of grid element           */
            int                mode;  /* mode of input                  */
            int                 seg;  /* SPECIAL SEG FLAG               */
            char           file[82];  /* name of velocity file          */
            };



/****************
SOURCE PARAMETERS
****************/

struct source_parameters {    
            int               type; /* source type (P=1;S=2;T=3;M=4;F=6;F=7)  */
            float              amp; /* source amplitude                       */
            float            alpha; /* gaussian parameter for source          */
            float               t0; /* time-shift factor for source           */
            float             freq; /* central frequency of source            */
            float                x; /* global source location (m)             */
            float                y; /* global source location (m)             */
            float                z; /* global source location (m)             */
            int                  l; /* global source y location               */
            int                  m; /* global source z location               */
            int                  n; /* global source x location               */
            float           length; /* length of fault (m)                    */
            float            width; /* width  of fault (m)                    */
            float            depth; /* depth to top of fault (m)              */
            float               s0; /* along strike rupture coord(m from x,y) */
            float               d0; /* along dip rupture coord (m from top)   */
            float                v; /* rupture velocity (m/s)                 */
            int                  g; /* grid of source                         */
            int                  t; /* number of time-steps                   */
            float               *f; /* time history of forcing function       */
            float           strike; /* strike of fault (cw from north (y))    */
            float              dip; /* dip    of fault (right along stike)    */
            float             rake; /* rake   of fault (ccw from strike  )    */
            float              M0 ; /* magnitude of moment tensor             */
            float              Mxx; /* xx component of moment tensor          */
            float              Myy; /* yy component of moemnt tensor          */
            float              Mzz; /* zz component of moemnt tensor          */
            float              Mxy; /* xy component of moment tensor          */
            float              Mxz; /* xz component of moemnt tensor          */
            float              Myz; /* yz component of moemnt tensor          */
            float               Fx; /* velocity force in x direction          */
            float               Fy; /* velocity force in y direction          */
            float               Fz; /* velocity force in z direction          */
            int                  p; /* number of points in distributed source */
            int               *dfl; /* array of source grid points (l)        */
            int               *dfm; /* array of source grid points (m)        */
            int               *dfn; /* array of source grid points (l)        */
            int               *dft; /* array of source initiation times       */
            float             *dfa; /* amplitude of source                    */
            float             *dfr; /* array of rise time functions           */
            char          file[82]; /* source time function file              */
            char         dfile[82]; /* distributed source time function file  */
            };



/*************************
SEISMOGRAMS (Vx, Vz, P, S)
*************************/

struct seismogram {
            float     *Vx;     /* x component of seismogram      */
            float     *Vy;     /* y component of seismogram      */
            float     *Vz;     /* z component of seismogram      */
            float      *P;     /* P potential                    */    
            float      *S;     /* S potential                    */    
            float    *Mxx;     /* xx moment component            */
            float    *Myy;     /* yy moment component            */
            float    *Mzz;     /* zz moment component            */
            float    *Mxy;     /* xy moment component            */
            float    *Mxz;     /* xz moment component            */
            float    *Myz;     /* yz moment component            */
            float      dt;     /* sampling time increment        */
            float   scale;     /* seismogram scale factor        */
            float       x;     /* x location (meters)            */
            float       y;     /* y location (meters)            */
            float       z;     /* z location (meters)            */
            int         l;     /* y grid component (no boundary) */
            int         m;     /* z grid component (no boundary) */
            int         n;     /* x grid component (no boundary) */ 
            int         g;     /* grid of seismogram             */
            int      node;     /* node of seismogram (parallel)  */
            int         t;     /* number of samples              */
            int    sample;     /* sampling rate on output        */
            int      mode;     /* type of output                 */
            char  file[82];    /* path name root for sac files   */
            };



/*********************
SEISMIC TRACE (OUTPUT) 
*********************/

struct trace_files {
            int     nglobal;    /* global number of traces                 */
            int      nlocal;    /* local  number of traces                 */
            int        mode;    /* output components (1=x;2=y;3=z;4=p;5=s) */
            int      *order;    /* order of traces relative to input file  */
            int          *l;    /* array of l (y)  index of traces         */
            int          *m;    /* array of m (z)  index of traces         */
            int          *n;    /* array of n (x)  index of traces         */
            float        *x;    /* array of x coordinate of traces         */
            float        *y;    /* array of y coordinate of traces         */
            float        *z;    /* array of z coordinate of traces         */
            float        x1;    /* starting x coordinate of trace line     */
            float        x2;    /* ending   x coordinate of trace line     */
            float        y1;    /* starting y coordinate of trace line     */
            float        y2;    /* ending   y coordinate of trace line     */
            float        z1;    /* starting z coordinate of trace line     */
            float        z2;    /* ending   z coordinate of trace line     */
            int           p;    /* number of traces                        */
            int      sample;    /* sampling factor                         */
            float     scale;    /* scale factor (not used)                 */
            char  tfile[82];    /* input file                              */
            char   file[82];    /* pathname root for file name             */
            };




/*********************
IMAGES FORMAT (OUTPUT) 
*********************/

struct image_files {
            int             t;  /* output time for image               */
            int        sample;  /* sampling factor for image           */
            int          mode;  /* mode of image                       */
            int         movie;  /* increment to print out images       */
            int         plane;  /* plane number of image               */
            int             d;  /* image dimension (xz=1; xy=2; yz=3)  */
            float           w;  /* floating point distance of plane    */
            float       scale;  /* scale factor for image              */
            float          *v;  /* floating point value                */
            int             p;  /* number of points for field variable */
            char     file[82];  /* path name root for image file       */
            };



/********************
TOPOGRAPHY PARAMETERS
********************/

struct topography_parameters {
            int           ext;  /* density extinquishing flag          */
            float         *zi;  /* height of interface at each point   */
            float       **ziM;  /* height of interface at each point   */
            float       depth;  /* depth of a flat interface           */
            char     file[82];  /* digitized topography in meters file */
            };



/************************
GRID/SUB GRID COORDINATES
************************/

struct plane_coordinates {
            int             x;   /* x coordinate or dimension  */
            int             y;   /* y coordinate or dimension  */ 
            int             z;   /* z coordinate or dimension  */
            int             l;   /* number l nodes             */
            int             m;   /* number m nodes             */ 
            int             n;   /* number n nodes             */
            int            l0;   /* node parameter y direction */
            int            m0;   /* node parameter z direction */
            int            n0;   /* node parameter x direction */
            int            l1;   /* intermed node y direction */
            int            m1;   /* intermed node z direction */
            int            n1;   /* intermed node x direction */
            };



/************************
INPUT FILE AND PARAMETERS
************************/

struct input_file_parameters {
            FILE         *file;    /* file to input from               */
            char   **fileLines;    /* lines in input file              */
            int         nLines;    /* number of lines in input file    */
            int           argc;    /* number of command line arguments */
            char        **argv;    /* command line arguments           */
            };




/***********************
PARAMETER REDECLARATIONS
***********************/

typedef struct elas_parameters          Parameter;
typedef struct source_parameters        FdSource;
typedef struct fd_grid                  FdGrid;
typedef struct velocity_stress          VelocityStress;
typedef struct image_files              Image;
typedef struct seismogram               Seismogram;
typedef struct grid_element             Medium;
typedef struct input_file_parameters    Input;
typedef struct topo                     Topography;
typedef struct state_2d                 State2D;
typedef struct topography_parameters    FdTopo;
typedef struct trace_files              Traces;

