/******************************************************************************
   getGrid: get finite difference grid properties

   usage: getGrid(par, grid, medium)
      par    - parameters
      grid   - grid parameters
      medium - grid medium parameters

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)
******************************************************************************/

#include "elas.h"

getGrid(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   int g, i;
   FILE *velocityFile = NULL;

   /* get equivalence arrays */
   gridEquivalence(par, grid, 0);

   /* initialize grid */
   for(g = 0; g < par->g; g++) initMGrid(par, grid + g);

   /* get grid velocities and parameters */
   for(g = 0; g < par->g; g++) for(i = 0; i < par->v; i++) 
      getVelocity(par, grid + g, medium + i);

   /**** initial call to getBoundary for if topography is above surface ****/
   for(g = 0; g < par->g; g++) getBoundary(par, grid + g);

   /* get topography parameters */
   g = par->g - 1;
   if (par->topo && par->d == 2 && grid[g].b == TOPOGRAPHY)
      getTopogX2d(par, grid + g); 
   if (par->topo && par->d == 3 && grid[g].b == TOPOGRAPHY)
      getTopogX3d(par, grid + g);

   /* get minimum and maximum grid parameters */
   for(g = 0; g < par->g; g++) getGridLimit(par, grid + g);
   getGlobalGridLimit(par, grid);
   getCheckGrid(par, grid);


   /* convert velocities to medium parameters */
   for(g = 0; g < par->g; g++) convertGrid(par, grid + g);


   /* get medium parameters around boundary */
   for(g = 0; g < par->g; g++) getBoundary(par, grid + g);


   /* dereference equivalence arrays */
   gridEquivalence(par, grid, 1);

}




/******************************************************************************
   gridEquivalence: set and dereference grid equivalence arrays

   usage: gridEquivalence(par, grid, mode)
      par    - parameters
      grid   - grid parameters
      mode   - grid memory reference mode (0 = reference; 1 = dereference)
******************************************************************************/
gridEquivalence(par, grid, mode)
Parameter   *par;
FdGrid     *grid;
int         mode;
{
   int g;

   if (mode == 0) for(g = 0; g < par->g; g++) {
      grid[g].alpha = grid[g].lamda;
      if (par->model != 2)
      grid[g].beta  = grid[g].mu;
      /*grid[g].rho   = grid[g].irho;*/
      grid[g].Qp    = grid[g].F1;
      grid[g].Qs    = grid[g].F2;
      grid[g].Qw    = grid[g].F3;
      }

   if (mode == 1) for(g = 0; g < par->g; g++) {
      grid[g].alpha = NULL;
      if (par->model != 2)
      grid[g].beta  = NULL;
      /*grid[g].rho   = NULL;*/
      grid[g].Qp    = NULL;
      grid[g].Qs    = NULL;
      grid[g].Qw    = NULL;
      }

}


initMGrid(par, grid)
Parameter   *par;
FdGrid      *grid;
{
   int l, m, n;
   for(l = -grid->fnt; l < grid->l + grid->bck; l++) {
      for(m = -grid->top; m < grid->m + grid->bot; m++) {
         for(n = -grid->lhs; n < grid->n + grid->rhs; n++) {
            grid->alpha[l][m][n] = -1.;
            if (par->model != 2)
            grid->beta [l][m][n] = -1.;
            grid->rho  [l][m][n] = -1.;
            if (par->q) {
               grid->Qp[l][m][n] = -1.;
               grid->Qs[l][m][n] = -1.;
               grid->Qw[l][m][n] = -1.;
               }
            }
         }
      }
}


/******************************************************************************
   getVelocity: get grid velocities from medium elements
******************************************************************************/
getVelocity(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{

   switch(medium->type) {
      case BLOCK:
         getBlock(par, grid, medium);
         break;
      case BFILL:
         blockFill(par, grid, medium);
         break;
      case PLANE:
         getPlane(par, grid, medium);
         break;
      case POLYN:
         if (par->d == 2) getPolyn2D(par, grid, medium);
         if (par->d == 3) {
            fprintf(stderr, "Error: polygon grid input not supported in 3D\n");
            exit(-1);
            }
         break;
      case POLY3:
         getPoly3D(par, grid, medium);
         break;
      case VFILE:
         getVFile(par, grid, medium);
         break;
      case SLOPE:
         fprintf(stderr, "Error: slope grid input no longer supported\n");
         exit(-1);
      case EMPTY:
         break;
      default:
         break;
      }
   
}



getBlock(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float alpha, beta, rho, dalpha, dbeta, drho;
   int l, m, n, l1, l2, m1, m2, n1, n2;

   /* get starting coordinates (was floor)), not rint()) */
   n1 = (int) rint ((medium->x1 - grid->x0)/grid->dx);
   n2 = (int) rint ((medium->x2 - grid->x0)/grid->dx);
   m1 = (int) rint ((medium->z1 - grid->z0)/grid->dz);
   m2 = (int) rint ((medium->z2 - grid->z0)/grid->dz);
   l1 = (int) rint ((medium->y1 - grid->y0)/grid->dy);
   l2 = (int) rint ((medium->y2 - grid->y0)/grid->dy);
   if (par->d == 2) {l1 = 0; l2 = 0;}
   if (n1 >= grid->n || n2 < 0) return(0);
   if (m1 >= grid->m || m2 < 0) return(0);
   if (l1 >= grid->l || l2 < 0) return(0);
   if (n1 <           0) n1 = 0;
   if (n2 > grid->n - 1) n2 = grid->n - 1;
   if (m1 <           0) m1 = 0;
   if (m2 > grid->m - 1) m2 = grid->m - 1;
   if (l1 <           0) l1 = 0;
   if (l2 > grid->l - 1) l2 = grid->l - 1;
   
   /* fill in grid */
   for(l = l1; l <= l2; l++) {
      alpha  = medium->alpha;
      beta   = medium->beta;
      rho    = medium->rho;
      dalpha = medium->gradient*grid->dz;
      dbeta  = medium->gradient*grid->dz*(medium->beta/medium->alpha);
      drho   = 0.;
      for(m = m1; m <= m2; m++) {
         for(n = n1; n <= n2; n++) {
            beta = (medium->beta == 0.) ? (alpha*medium->spratio) : (beta);
            rho  = (medium->rho  == 0.) ? (GETRHOL(alpha)) : (rho);
            grid->alpha[l][m][n] = alpha;
            grid->beta [l][m][n] = beta;
            grid->rho  [l][m][n] = rho; 
            if (par->q) {
               grid->Qp[l][m][n] = medium->Qp;
               grid->Qs[l][m][n] = medium->Qs;
               grid->Qw[l][m][n] = medium->Qw;
               }
            }
         alpha += dalpha;
         beta  += dbeta;
         rho   += drho;
         }
      }

}




/* this is the new method using block input (as opposed to medium input) */
blockFill(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float P, S, R, Qp, Qs, Qw, SP, dP, dS, dR;
   int l, m, n, l1, l2, m1, m2, n1, n2;
int dm;

   /* get starting coordinates (was floor(), not rint()) */
   n1 = (int) rint ((medium->x1 - grid->x0)/grid->dx);
   n2 = (int) rint ((medium->x2 - grid->x0)/grid->dx);
   m1 = (int) rint ((medium->z1 - grid->z0)/grid->dz);
   m2 = (int) rint ((medium->z2 - grid->z0)/grid->dz);
   l1 = (int) rint ((medium->y1 - grid->y0)/grid->dy);
   l2 = (int) rint ((medium->y2 - grid->y0)/grid->dy);
   if (par->d == 2) {l1 = 0; l2 = 0;}
   if (n1 >= grid->n || n2 < 0) return(0);
   if (m1 >= grid->m || m2 < 0) return(0);
   if (l1 >= grid->l || l2 < 0) return(0);
   if (n1 <           0) n1 = 0;
   if (n2 > grid->n - 1) n2 = grid->n - 1;
/* bill/allan */
dm = 0; 
if (m1 < 0) dm = -m1;
   if (m1 <           0) m1 = 0;
   if (m2 > grid->m - 1) m2 = grid->m - 1;
   if (l1 <           0) l1 = 0;
   if (l2 > grid->l - 1) l2 = grid->l - 1;

   /* get parameters */
   P   = medium->alpha;
   S   = medium->beta ;
   R   = medium->rho  ;
   Qp  = medium->Qp;
   Qs  = medium->Qs;
   Qw  = medium->Qw;
   SP  = medium->spratio;
   dP  = medium->gradient*grid->dz;
   dS  = medium->gradient*grid->dz; if (P > 0.          ) dS *= S/P;
   dR  = medium->gradient*grid->dz; if (P > 0. || S > 0.) dR = 0.;
/* bill/allan */
if (dm > 0) {
P += ((float) dm)*dP;
S += ((float) dm)*dS;
R += ((float) dm)*dR;
}

   /* fill grid */
   if (P  >= 0.) gfill(grid->alpha, P,  dP,         l1, l2, m1, m2, n1, n2); 
   if (par->model != 2)
   if (S  >= 0.) gfill(grid->beta,  S,  dS,         l1, l2, m1, m2, n1, n2); 
   if (par->model != 2)
   if (SP >= 0.) ffill(grid->beta, grid->alpha, SP, l1, l2, m1, m2, n1, n2); 
   if (R  >= 0.) gfill(grid->rho,   R,  dR,         l1, l2, m1, m2, n1, n2); 
   if (Qp >= 0.) gfill(grid->Qp,   Qp,  0.,         l1, l2, m1, m2, n1, n2); 
   if (Qs >= 0.) gfill(grid->Qs,   Qs,  0.,         l1, l2, m1, m2, n1, n2); 
   if (Qw >= 0.) gfill(grid->Qw,   Qw,  0.,         l1, l2, m1, m2, n1, n2); 
   
}


gfill(array, value, gradient, l1, l2, m1, m2, n1, n2)
float ***array, value, gradient;
int l1, l2, m1, m2, n1, n2;
{
   int l, m, n;
   for(m = m1; m <= m2; m++) {
      for(l = l1; l <= l2; l++) for(n = n1; n <= n2; n++) 
         array[l][m][n] = value;
      value += gradient;
      }
}


ffill(array1, array2, factor, l1, l2, m1, m2, n1, n2)
float ***array1, ***array2, factor;
int l1, l2, m1, m2, n1, n2;
{
   int l, m, n;
   for(m = m1; m <= m2; m++) {
      for(l = l1; l <= l2; l++) for(n = n1; n <= n2; n++) 
         array1[l][m][n] = factor*array2[l][m][n];
      }
}




getPoly3D(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float alpha, beta, rho, dalpha, dbeta, drho;
   float P, S, R, Qp, Qs, Qw, SP, dP, dS, dR;
   int l1, l2, m1, m2, n1, n2;

   /* get starting coordinates (was floor(), not rint()) */
   n1 = (int) rint ((medium->x1 - grid->x0)/grid->dx);
   n2 = (int) rint ((medium->x2 - grid->x0)/grid->dx);
   m1 = (int) rint ((medium->z1 - grid->z0)/grid->dz);
   m2 = (int) rint ((medium->z2 - grid->z0)/grid->dz);
   l1 = (int) rint ((medium->y1 - grid->y0)/grid->dy);
   l2 = (int) rint ((medium->y2 - grid->y0)/grid->dy);
   if (par->d == 2) {l1 = 0; l2 = 0;}
   if (n1 >= grid->n || n2 < 0) return(0);
   if (m1 >= grid->m || m2 < 0) return(0);
   if (l1 >= grid->l || l2 < 0) return(0);
   if (n1 <           0) n1 = 0;
   if (n2 > grid->n - 1) n2 = grid->n - 1;
   if (m1 <           0) m1 = 0;
   if (m2 > grid->m - 1) m2 = grid->m - 1;
   if (l1 <           0) l1 = 0;
   if (l2 > grid->l - 1) l2 = grid->l - 1;

   /* fill in grid */
   getppoint(par, grid, medium, l1, l2, m1, m2, n1, n2);

   /* get override parameters if explicitly specified (if parameter >= 0) */
   P   = medium->alpha;
   S   = medium->beta ;
   R   = medium->rho  ;
   Qp  = medium->Qp;
   Qs  = medium->Qs;
   Qw  = medium->Qw;
   SP  = medium->spratio;
   dP  = medium->gradient*grid->dz;
   dS  = medium->gradient*grid->dz; if (P > 0.) dS *= S/P;
   dR  = 0.;

   /* fill grid with override values */
   if (P  >= 0.) gfill(grid->alpha, P,  dP,         l1, l2, m1, m2, n1, n2);
   if (par->model != 2)
   if (S  >= 0.) gfill(grid->beta,  S,  dS,         l1, l2, m1, m2, n1, n2);
   if (par->model != 2)
   if (SP >= 0.) ffill(grid->beta, grid->alpha, SP, l1, l2, m1, m2, n1, n2);
   if (R  >= 0.) gfill(grid->rho,   R,  dR,         l1, l2, m1, m2, n1, n2);
   if (Qp >= 0.) gfill(grid->Qp,   Qp,  0.,         l1, l2, m1, m2, n1, n2);
   if (Qs >= 0.) gfill(grid->Qs,   Qs,  0.,         l1, l2, m1, m2, n1, n2);
   if (Qw >= 0.) gfill(grid->Qw,   Qw,  0.,         l1, l2, m1, m2, n1, n2);

}


getppoint(par, grid, medium, l1, l2, m1, m2, n1, n2)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
int       l1, l2, m1, m2, n1, n2;
{
   double angle, cs, sn, lon, lat, lon0, lat0;
   double x, y, z, dx, dy, dz, x0, y0, z0, xr, yr;
   float P, S, R, Qp, Qs, Qw;
   int l, m, n;
   FILE *file;

   /* initialize longitude and latitude */
   angle =  medium->angle;
   lon0  =  medium->lon;
   lat0  =  medium->lat;
   cs    =  cos(angle/(RAD));
   sn    =  sin(angle/(RAD));
   x0    =  ((double) grid->x0)/(KMTOM);
   y0    =  ((double) grid->y0)/(KMTOM);
   z0    =  ((double) grid->z0)/(KMTOM);
   dx    =  ((double) grid->dx)/(KMTOM);
   dy    =  ((double) grid->dy)/(KMTOM);
   dz    =  ((double) grid->dz)/(KMTOM);

   /* open 3D polygon file (lomex) */
   file = fopen(medium->file, "r");
   if (file == NULL) {
      fprintf(stderr,"Error: can't open 3D polygon file \"%s\"\n",medium->file);
      exit(-1);
      }

   /* fill in grid */
   for(l = l1; l <= l2; l++) {
      for(n = n1; n <= n2; n++) {
         x = x0 + ((double) n)*dx;
         y = y0 + ((double) l)*dy;
         xr = (x*cs - y*sn);
         yr = (x*sn + y*cs);
         lat = lat0 + (yr/(KMD));
         lon = lon0 + (xr/(KMD))/cos(((lat0 + lat)/2.)/(RAD));
         z = z0;
         for(m = m1; m <= m2; m++) {
            ppoint(&P, &S, &R, &Qp, &Qs, &Qw, lon, lat, z, file);
            if (P >= 0.) grid->alpha[l][m][n] = P*(KMTOM);
            if (par->model != 2)
            if (S >= 0.) grid->beta [l][m][n] = S*(KMTOM);
            if (R >= 0.) grid->rho  [l][m][n] = R*(GM3TOKGM3);
            if (par->q) {
               if (Qp >= 0.) grid->Qp[l][m][n] = Qp;
               if (Qs >= 0.) grid->Qs[l][m][n] = Qs;
               if (Qw >= 0.) grid->Qw[l][m][n] = Qw;
               }
            z += dz;
            }
         }
      }

   /* fill in grid */
/*
   dx = grid->dx; dy = grid->dy; dz = grid->dz;
   y = grid->y0 + ((float) l1)*dy;
   for(l = l1; l <= l2; l++) {
      z = grid->z0 + ((float) m1)*dz;
      for(m = m1; m <= m2; m++) {
         x = grid->x0 + ((float) n1)*dx;
         for(n = n1; n <= n2; n++) {
            ppoint(&P, &S, &R, &Qp, &Qs, &Qw, x, y, z, file);
            if (P >= 0.) grid->alpha[l][m][n] = P;
            if (S >= 0.) grid->beta [l][m][n] = S;
            if (R >= 0.) grid->rho  [l][m][n] = R;
            if (par->q) {
               if (Qp >= 0.) grid->Qp[l][m][n] = Qp;
               if (Qs >= 0.) grid->Qs[l][m][n] = Qs;
               if (Qw >= 0.) grid->Qw[l][m][n] = Qw;
               }
            x += dx;
            }
         z += dz;
         }
      y += dy;
      }
*/

   /* close 3D polygon file */
   fclose(file);

}


/******************************************************************************
   getPlane: get finite difference grid properties (dipping plane)

   usage: getPolyn2D(par, grid, medium)
      par    - parameters
      grid   - grid parameters
      medium - grid medium parameters

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)

   NOTE: THIS ROUTINE IS NOT COMPLETE
******************************************************************************/
getPlane(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float x, x1, x2, y, y1, y2, z, z1, z2, z3, z4, a, b, c, d, zmax, zmin;
   int l, l1, l2, m, m1, m2, n, n1, n2;

   /* get node values for grid */

   /* get plane parameterization */
/* comment to avoid compiler warnings on tera
   getPlaneABCD(x1, y1, z1, x2, y1, z2, x2, y2, z3, x1, y2, z4, &a, &b, &c, &d);
*/

   /* check plane */

   /* fill dipping plane */
/* comment to avoid compiler warnings on tera
   y = y1;
   for(l = l1; l <= l2; l++) {
      z = z1;
      for(m = m1; m <= m2; m++) {
         x = x1;
         for(n = n1; n <= n2; n++) {
            if (a*x + b*y + c*z + d >= 0.) {
               }
            x += grid->dx;
            }
         z += grid->dz;
         }
      y += grid->dy;
      }
*/

}




/******************************************************************************
   getPolyn2D: get finite difference grid properties (polygon)

   usage: getPolyn2D(par, grid, medium)
      par    - parameters
      grid   - grid parameters
      medium - grid medium parameters

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)

   SPECIAL NOTE: doesn't quite work right at top
******************************************************************************/
getPolyn2D(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float alpha, beta, rho, Qp, Qs, Qw, dalpha, dbeta, drho, *x, *z;
   int i, m, n, p, nx, nz, ix[MAXPOLY], iz[MAXPOLY];

   /* initialize */
   x  = medium->x;
   z  = medium->z;
   p  = medium->p;
   nx = grid->n;
   nz = grid->m;

   /* determine grid points */
   for(i = 0; i < p; i++) {
      ix[i] = (int) floor((x[i] - grid->x0)/grid->dx);
      iz[i] = (int) floor((z[i] - grid->z0)/grid->dz);
      }

   /* determine medium parameters */
   alpha = medium->alpha;
   beta  = medium->beta;
   rho   = medium->rho;
   Qp    = medium->Qp;
   Qs    = medium->Qs;
   Qw    = medium->Qw;
   beta  = (medium->beta == 0.) ? (alpha*medium->spratio) : (beta);
   rho   = (medium->rho  == 0.) ? (GETRHOL(alpha)) : (rho);

   /* determine gradients (change per grid point) */
   dalpha = medium->gradient*grid->dz;
   dbeta  = medium->gradient*grid->dz*(medium->beta/medium->alpha);
   drho   = 0.;

   /* get and fill polygon array */
   grademodel(grid->alpha[0], nx, nz, p, ix, iz, alpha, dalpha);
   grademodel(grid->beta [0], nx, nz, p, ix, iz, beta , dbeta);
   grademodel(grid->rho  [0], nx, nz, p, ix, iz, rho  , drho);
   if (par->q) {
      grademodel(grid->Qp[0], nx, nz, p, ix, iz, Qp, 0.);
      grademodel(grid->Qs[0], nx, nz, p, ix, iz, Qs, 0.);
      grademodel(grid->Qw[0], nx, nz, p, ix, iz, Qw, 0.);
      }

}



/******************************************************************************
   getVFile: get grid parameters from grid file

   usage: getVFile(par, grid, medium)
      par    - parameters
      grid   - grid parameters
      medium - grid medium parameters

   note: works for multigrid only for independent multigrids (parallel OK)
   note: format of input file is l, m, n where n varies fastests
   note: input is binary, 4 bytes per point
   note: only one medium parameter is read per file
   note: input (l,m,n)1,2 are start/end global grid coordinates for input file
         used  (l,m,n)1,2 are start/end local  grid coordinates for input file
   note: (l,m,n)0 are start coordinates for grid
   note: (l,m,n)3 are end   coordinates for grid
   note: all start and end coordinates are inclusive

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)
******************************************************************************/
getVFile(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float ***value, factor, v;
   int l, m, n, l0, m0, n0, l1, n1, m1, l2, m2, n2, l3, m3, n3, d, compress;
   long offset, loffset0, loffset1, moffset0, moffset1, noffset0, noffset1, p;
   FILE *file;

   /* check for acoustic model */
   if (medium->mode == SFILE && par->model == 2) {
      fprintf(stderr, "Warning: shear velocity file input\n");
      return;
      }

   /* initialize */
   l1 = medium->l1 - grid->l0; l2 = medium->l2 - grid->l0;
   m1 = medium->m1 - grid->m0; m2 = medium->m2 - grid->m0;
   n1 = medium->n1 - grid->n0; n2 = medium->n2 - grid->n0;
   l0 = (l1 < 0) ? (0) : (l1);
   m0 = (m1 < 0) ? (0) : (m1);
   n0 = (n1 < 0) ? (0) : (n1);
   l3 = (l2 >= grid->l) ? (grid->l-1) : (l2);
   m3 = (m2 >= grid->m) ? (grid->m-1) : (m2);
   n3 = (n2 >= grid->n) ? (grid->n-1) : (n2);
   d  = n3 - n0 + 1;
   compress = medium->c;

   /* begin parallel synchronization */
   mpNodeOrder(0);

   /* return if subgrid does not span velcity file (end parallel sync) */
   if ((l1 > l2) || (m1 > m2) || (n1 > n2)) {mpNodeOrder(1); return(-1);}
   if (l1 >= grid->l || l2 < 0)             {mpNodeOrder(1); return( 1);}
   if (m1 >= grid->m || m2 < 0)             {mpNodeOrder(1); return( 2);}
   if (n1 >= grid->n || n2 < 0)             {mpNodeOrder(1); return( 3);}

   /* get offsets */
   loffset0 = (long) (4*((l1 <  0) ? (  -l1):(0))*((m2-m1+1)*(n2-n1+1)));
   moffset0 = (long) (4*((m1 <  0) ? (  -m1):(0))*(          (n2-n1+1)));
   noffset0 = (long) (4*((n1 <  0) ? (  -n1):(0))                      );
   loffset1 = (long) (4*((l2 > l3) ? (l2-l3):(0))*((m2-m1+1)*(n2-n1+1)));
   moffset1 = (long) (4*((m2 > m3) ? (m2-m3):(0))*(          (n2-n1+1)));
   noffset1 = (long) (4*((n2 > n3) ? (n2-n3):(0))                      );

   /* determine value parameter */
   if (medium->mode == PFILE) value = grid->alpha;
   if (medium->mode == SFILE) value = grid->beta ;
   if (medium->mode == RFILE) value = grid->rho  ;
   if (medium->mode == QFILE) value = grid->Qp   ;
   if (medium->mode == YFILE) value = grid->Qp   ;
   if (medium->mode == ZFILE) value = grid->Qs   ;
   if (medium->mode == WFILE) value = grid->Qw   ;

   /* open file */
   file = fopen(medium->file, "r");
   if (file == NULL) {
      fprintf(stderr, "Error: can't open grid file \"%s\"\n", medium->file);
      mpNodeOrder(1);
      exit(-1);
      }

   /* check file size */
   if (compress == 0) {fseek(file, 0L, SEEK_END); offset = ftell(file);}
   if (compress == 1) {offset = rlefilesize(file);}
   rewind(file);
   if ((long) (4*(l2-l1+1)*(m2-m1+1)*(n2-n1+1)) != offset) {
      fprintf(stderr, "Error: incompatable size for file \"%s\"\n",
      medium->file);
      exit(-1);
      }

   /* read file uncompressed file */
   if (compress == 0) {
      fseek(file, loffset0, SEEK_SET);
      for(l = l0; l <= l3; l++) {
         fseek(file, moffset0, SEEK_CUR);
         for(m = m0; m <= m3; m++) {
            fseek(file, noffset0, SEEK_CUR);
            fread(value[l][m]+n0, 4, d, file);
            fseek(file, noffset1, SEEK_CUR);
            }
         fseek(file, moffset1, SEEK_CUR);
         }
      }

   /* read rle compressed file */
   if (compress == 1) {
      rleseek(file, loffset0, SEEK_SET, &p, &v);
      for(l = l0; l <= l3; l++) {
         rleseek(file, moffset0, SEEK_CUR, &p, &v);
         for(m = m0; m <= m3; m++) {
            rleseek(file, noffset0, SEEK_CUR, &p, &v);
            rleread(value[l][m]+n0, 4, d, file, &p, &v);
            rleseek(file, noffset1, SEEK_CUR, &p, &v);
            }
         rleseek(file, moffset1, SEEK_CUR, &p, &v);
         }
      }

   /* close file */
   fclose(file);

   /* end parallel synchronization */
   mpNodeOrder(1);

   /* change units */
   factor = 0.;
   if (medium->mode == PFILE) factor = KMTOM;
   if (medium->mode == SFILE) factor = KMTOM;
   if (medium->mode == RFILE) factor = GM3TOKGM3;
   if (medium->mode == WFILE) factor = 2.*(PI);
   if (factor > 0.) {
      for(l = l0; l <= l3; l++) 
         for(m = m0; m <= m3; m++) 
            for(n = n0; n <= n3; n++) 
               value[l][m][n] *= factor;
      }
   

   /* fill grid with other parameters */
   {
   float P, S, R, SP, Qp, Qs, QR, Qw, dP, dS, dR;
   P   = medium->alpha; if (medium->mode == PFILE) P  = -1.;
   S   = medium->beta ; if (medium->mode == SFILE) S  = -1.;
   R   = medium->rho  ; if (medium->mode == RFILE) R  = -1.;
   Qp  = medium->Qp   ; if (medium->mode == QFILE) Qp = -1.;
   Qp  = medium->Qp   ; if (medium->mode == YFILE) Qp = -1.;
   Qs  = medium->Qs   ; if (medium->mode == ZFILE) Qs = -1.;
   Qw  = medium->Qw   ; if (medium->mode == WFILE) Qw = -1.;
   SP  = medium->spratio;
   dP  = medium->gradient*grid->dz;
   dS  = medium->gradient*grid->dz; if (P > 0.) dS *= S/P;
   dR  =  0.;
   QR  = -1.;
   if (medium->mode == QFILE) {Qs = -1.; QR = 1.;}
   if (P  >= 0.) gfill(grid->alpha, P,  dP,         l0, l3, m0, m3, n0, n3);
   if (par->model != 2)
   if (S  >= 0.) gfill(grid->beta,  S,  dS,         l0, l3, m0, m3, n0, n3);
   if (par->model != 2)
   if (SP >= 0.) ffill(grid->beta, grid->alpha, SP, l0, l3, m0, m3, n0, n3);
   if (R  >= 0.) gfill(grid->rho,   R,  dR,         l0, l3, m0, m3, n0, n3);
   if (Qp >= 0.) gfill(grid->Qp,   Qp,  0.,         l0, l3, m0, m3, n0, n3);
   if (Qs >= 0.) gfill(grid->Qs,   Qs,  0.,         l0, l3, m0, m3, n0, n3);
   if (Qw >= 0.) gfill(grid->Qw,   Qw,  0.,         l0, l3, m0, m3, n0, n3);
   if (QR >= 0.) ffill(grid->Qs,   grid->Qp,    QR, l0, l3, m0, m3, n0, n3);
   }

   /* convert to seg model (special implementation) */
   if (medium->seg == 1) segfill(par, grid, l0, l3, m0, m3, n0, n3);

}

rlefilesize(file)
FILE *file;
{
   float v;
   long int n, nsum;
   nsum = 0;
   while(fread(&n, sizeof(int), 1, file)) {
      nsum += n;
      fread(&v, sizeof(float), 1, file);
      }
   nsum *= sizeof(float);
   return(nsum);
}

rleseek(file, offset, mode, p, v)
float *v;
long int offset, *p;
int mode;
FILE *file;
{
   int b;
   if (offset  < 0) {
      fprintf(stderr, "Error: rleseek can't handle negative offsets\n");
      exit(-1);
      }
   if (mode == SEEK_END) {
      fprintf(stderr, "Error: can't do SEEK_END rleseek mode\n");
      exit(-1);
      }
   if (mode == SEEK_SET) {
      rewind(file);
      *p = 0;
      }
   if (*p >= offset) {
      *p -= offset;
      offset = 0;
      }
   if (*p < offset) {
      offset -= *p;
      while(offset > 0) {
         b = fread(p, sizeof(int  ), 1, file);
         b = fread(v, sizeof(float), 1, file);
         *p *= sizeof(float);
         offset -= *p;
         if (b == 0) {
            fprintf(stderr, "Error: bad rle compression read 1\n");
            /*exit(-1);*/
            }
         }
      *p = -offset;
      }
}

rleread(array, rsize, d, file, p, v)
float *array, *v;
long int *p;
int rsize, d;
FILE *file;
{
   int b, n;
   if (*p < 0) {
      fprintf(stderr, "Error: error in rleread (compression)\n");
      exit(-1);
      }
   for(n = 0; n < d; n++) {
      if (*p == 0) {
         while(*p <= 0) {
            b = fread(p, sizeof(int  ), 1, file);
            b = fread(v, sizeof(float), 1, file);
            *p *= sizeof(float);
            if (b == 0) {
               fprintf(stderr, "Error: bad rle compression read 2\n");
               /*exit(-1);*/
               }
            }
         }
      array[n] = *v;
      *p -= rsize;
      }
}



getGridLimit(par, grid)
Parameter *par;
FdGrid    *grid;
{
   float alpha, beta, rho, Qp, Qs, Qw;
   int l, m, n, mtopo;
   Qp = 0.; Qs = 0.; Qw = 0.;
   for(l = 0; l < grid->l; l++) {
      for(n = 0; n < grid->n; n++) {
         mtopo = 0;
         if (grid->b == TOPOGRAPHY) {
            if (par->d == 2) mtopo  = ((int) (grid->zi    [n]/grid->dz));
            if (par->d == 3) mtopo  = ((int) (grid->ziM[l][n]/grid->dz));
            if (par->d == 2) if (mtopo*grid->dz != grid->zi    [n]) mtopo++;
            if (par->d == 3) if (mtopo*grid->dz != grid->ziM[l][n]) mtopo++;
            mtopo -= grid->m0;
            if (mtopo < 0) mtopo = 0;
            }
         for(m = mtopo; m < grid->m; m++) {
            alpha = grid->alpha[l][m][n];
            if (par->model != 2)
            beta  = grid->beta [l][m][n];
            if (par->model == 2) beta = 0.;
            rho   = grid->rho  [l][m][n];
            if (par->q) {
               Qp = grid->Qp   [l][m][n];
               Qs = grid->Qs   [l][m][n];
               Qw = grid->Qw   [l][m][n];
               }
            setGridLimit(grid, alpha, beta, rho, Qp, Qs, Qw);
            }
         }
      }
}



setGridLimit(grid, alpha, beta, rho, Qp, Qs, Qw)
FdGrid *grid;
float alpha, beta, rho, Qp, Qs, Qw;
{
   grid->minP  = MIN(grid->minP , alpha);
   grid->maxP  = MAX(grid->maxP , alpha);
   grid->minS  = MIN(grid->minS , beta );
   grid->maxS  = MAX(grid->maxS , beta );
   grid->minR  = MIN(grid->minR , rho  );
   grid->maxR  = MAX(grid->maxR , rho  );
   grid->minQp = MIN(grid->minQp, Qp   );
   grid->maxQp = MAX(grid->maxQp, Qp   );
   grid->minQs = MIN(grid->minQs, Qs   );
   grid->maxQs = MAX(grid->maxQs, Qs   );
   grid->minQw = MIN(grid->minQw, Qw   );
   grid->maxQw = MAX(grid->maxQw, Qw   );
}



getGlobalGridLimit(par, grid)
Parameter   *par;
FdGrid     *grid;
{
   int g;
   if (par->nnode > 1) for(g = 0; g < par->g; g++) mpGridLimit(par, grid + g);
   for(g = 0; g < par->g; g++) {
      par->minP  = MIN(grid[g].minP , par->minP );
      par->maxP  = MAX(grid[g].maxP , par->maxP );
      par->minS  = MIN(grid[g].minS , par->minS );
      par->maxS  = MAX(grid[g].maxS , par->maxS );
      par->minR  = MIN(grid[g].minR , par->minR );
      par->maxR  = MAX(grid[g].maxR , par->maxR );
      par->minQp = MIN(grid[g].minQp, par->minQp);
      par->maxQp = MAX(grid[g].maxQp, par->maxQp);
      par->minQs = MIN(grid[g].minQs, par->minQs);
      par->maxQs = MAX(grid[g].maxQs, par->maxQs);
      par->minQw = MIN(grid[g].minQw, par->minQw);
      par->maxQw = MAX(grid[g].maxQw, par->maxQw);
      }
}




/******************************************************************************
   getCheckGrid: determine if grid values reasonable

   usage: getCheckGrid(par, grid)
      par    - parameters
      grid   - grid parameters

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)
******************************************************************************/
getCheckGrid(par, grid)
Parameter    *par;
FdGrid      *grid;
{
   float minP, maxP, minS, maxS, minR, maxR;
   float minQp, maxQp, minQs, maxQs, minQw, maxQw;
   int error = 0;

   minP  = par->minP/KMTOM;
   maxP  = par->maxP/KMTOM;
   minS  = par->minS/KMTOM;
   maxS  = par->maxS/KMTOM;
   minR  = par->minR/GM3TOKGM3;
   maxR  = par->maxR/GM3TOKGM3;
   minQp = par->minQp;
   maxQp = par->maxQp;
   minQs = par->minQs;
   maxQs = par->maxQs;
   minQw = par->minQw;
   maxQw = par->maxQw;

   if (minP <= 0.) {
      fprintf(stderr, "Error: P velocity (%6.2f km/s)     <= 0.\n", minP);
      error = 1;
      }
   if (maxP > 50.) {
      fprintf(stderr, "Error: P velocity (%6.2f km/s)     > 50.\n", maxP);
      error = 1;
      }
   if (minS <  0.) {
      fprintf(stderr, "Error: S velocity (%6.2f km/s)     <  0.\n", minS);
      error = 1;
      }
   if (maxS > 50.) {
      fprintf(stderr, "Error: S velocity (%6.2f km/s)     > 50.\n", maxS);
      error = 1;
      }
   if (minR <= 0.) {
      fprintf(stderr, "Error: density    (%6.2f gm/cm**3) <= 0.\n", minR);
      error = 1;
      }
   if (maxR > 50.) {
      fprintf(stderr, "Error: density    (%6.2f gm/cm**3) > 50.\n", maxR);
      error = 1;
      }

   if (par->q) {
      if (minQp <= 0.) {
         fprintf(stderr, "Error: Qp (%6.2f) <= 0.\n", minQp);
         error = 1;
         }
      if (minQs <= 0.) {
         fprintf(stderr, "Error: Qs (%6.2f) <= 0.\n", minQs);
         error = 1;
         }
      if (minQw <= 0.) {
         fprintf(stderr, "Error: Qw (%6.2f) <= 0.\n", minQw);
         error = 1;
         }
      }

   if (error) exit(-1);

}


/******************************************************************************
   convertGrid: convert grid parameters

   usage: convertGrid(par, grid)
      par    - parameters
      grid   - grid parameters

   note: convert grid velocities to elastic parameters
   note: invert density (for numerical efficiency)
   note: get attenuation constants

   note: input units are km, s, km/s; g/cm3
   note: used  units are mks (m, s, m/s, kg/m3)
******************************************************************************/
convertGrid(par, grid)
Parameter    *par;
FdGrid      *grid;
{
   float alpha, beta, rho, Qp, Qs, Qw;
   int l, m, n;

   /* convert velocities to elastic parameters and invert density */
   for(l = 0; l < grid->l; l++) {
      for(m = 0; m < grid->m; m++) {
         for(n = 0; n < grid->n; n++) {
            alpha = grid->alpha[l][m][n];
            if (par->model != 2)
            beta  = grid->beta [l][m][n];
            if (par->model == 2) beta = 0.;
            rho   = grid->rho  [l][m][n];
            grid->lamda[l][m][n] = GETLAMDA(alpha, beta, rho);
            if (par->model != 2)
            grid->mu   [l][m][n] = GETMU   (alpha, beta, rho);
            grid->irho [l][m][n] = 1./rho;
            }
         }
      }

   /* convert attenuation parameters to FD attenuation parameters */
   if (par->q) getAttenuation(par, grid);

}




/******************************************************************************
   getBoundary: get grid medium parameters around boundary
   NOTE: multigrid case isn't implemented
******************************************************************************/
getBoundary(par, grid)
Parameter  *par;
FdGrid     *grid;
{
   float **lamda, **mu, **irho, **rho;
   int l, m, n, top, bot, lhs, rhs, fnt, bck;

   /* initialize parallel parameters */
   top = grid->top;
   bot = grid->bot;
   rhs = grid->rhs;
   lhs = grid->lhs;
   fnt = grid->fnt;
   bck = grid->bck;

   /* for surface boundary */
   for(l = 0; l < grid->l; l++) {
      lamda = grid->lamda[l];
      if (par->model != 2)
      mu    = grid->mu   [l];
      rho   = grid->rho  [l];
      irho  = grid->irho [l];
      for(m = -top; m < 0; m++) {
         for(n = 0; n < grid->n ; n++) {
            lamda[m][n] = lamda[0][n];
            if (par->model != 2)
            mu   [m][n] = mu   [0][n];
            rho  [m][n] = rho  [0][n];
            irho [m][n] = irho [0][n];
            }
         }
      }

   /* for bottom boundary */
   for(l = 0; l < grid->l; l++) {
      lamda = grid->lamda[l];
      if (par->model != 2)
      mu    = grid->mu   [l];
      rho   = grid->rho  [l];
      irho  = grid->irho [l];
      for(m = grid->m; m < grid->m + bot; m++) {
         for(n = 0; n < grid->n; n++) {
            lamda[m][n] = lamda[grid->m - 1][n];
            if (par->model != 2)
            mu   [m][n] = mu   [grid->m - 1][n];
            rho  [m][n] = rho  [grid->m - 1][n];
            irho [m][n] = irho [grid->m - 1][n];
            }
         }
      }

   /* for lhs boundary */
   for(l = 0; l < grid->l; l++) {
      lamda = grid->lamda[l];
      if (par->model != 2)
      mu    = grid->mu   [l];
      rho   = grid->rho  [l];
      irho  = grid->irho [l];
      for(n = -lhs; n < 0; n++) {
         for(m = -top; m < grid->m + bot; m++) {
            lamda[m][n] = lamda[m][0];
            if (par->model != 2)
            mu   [m][n] = mu   [m][0];
            rho  [m][n] = rho  [m][0];
            irho [m][n] = irho [m][0];
            }
         }
      }

   /* for rhs boundary */
   for(l = 0; l < grid->l; l++) {
      lamda = grid->lamda[l];
      if (par->model != 2)
      mu    = grid->mu   [l];
      rho   = grid->rho  [l];
      irho  = grid->irho [l];
      for(n = grid->n; n < grid->n + rhs; n++) {
         for(m = -top; m < grid->m + bot; m++) {
            lamda[m][n] = lamda[m][grid->n - 1];
            if (par->model != 2)
            mu   [m][n] = mu   [m][grid->n - 1];
            rho  [m][n] = rho  [m][grid->n - 1];
            irho [m][n] = irho [m][grid->n - 1];
            }
         }
      }

   /* for fnt boundary */
   for(l = -fnt; l < 0; l++) {
      for(n = -lhs; n < grid->n + rhs; n++) {
         for(m = -top; m < grid->m + bot; m++) {
            grid->lamda[l][m][n] = grid->lamda[0][m][n];
            if (par->model != 2)
            grid->mu   [l][m][n] = grid->mu   [0][m][n];
            grid->rho  [l][m][n] = grid->rho  [0][m][n];
            grid->irho [l][m][n] = grid->irho [0][m][n];
            }
         }
      }

   /* for bck boundary */
   for(l = grid->l; l < grid->l + bck; l++) {
      for(n = -lhs; n < grid->n + rhs; n++) {
         for(m = -top; m < grid->m + bot; m++) {
            grid->lamda[l][m][n] = grid->lamda[grid->l - 1][m][n];
            if (par->model != 2)
            grid->mu   [l][m][n] = grid->mu   [grid->l - 1][m][n];
            grid->rho  [l][m][n] = grid->rho  [grid->l - 1][m][n];
            grid->irho [l][m][n] = grid->irho [grid->l - 1][m][n];
            }
         }
      }

   /* for lower  left corner */

   /* for lower right corner */

   /* for upper  left corner */

   /* for upper right corner */

   if (par->nnode > 1 && par->d == 2) mpUpdate2D(par, grid, NULL, 3);
   if (par->nnode > 1 && par->d == 3) mpUpdate3D(par, grid, NULL, 3);

}








/* OLD STUFF */
/******************************************************************************
   getVelocity: get grid velocities from medium elements
******************************************************************************/
/*
getVelocity(par, grid, medium)
Parameter   *par;
FdGrid     *grid;
Medium   *medium;
{
   float dz, alpha, beta, rho, dalpha, dbeta, drho, slope;
   int l, m, n, m1, m2, n1, n2, mstart, mend, npoint, nstart, nend, mode;
   FILE *file;
*/

   /* get initial conditions */
/*
   alpha  = medium->alpha;
   beta   = medium->beta;
   rho    = medium->rho;
   dalpha = medium->gradient*grid->dz;
   dbeta  = medium->gradient*grid->dz*(medium->beta/medium->alpha);
   drho   = 0.;

   switch(medium->type) {
      case BLOCK:
         for(m = medium->m1; m <= medium->m2; m++) {
            for(n = medium->n1; n <= medium->n2; n++) {
               beta = (medium->beta == 0.) ? (alpha*medium->spratio) : (beta);
               rho  = (medium->rho  == 0.) ? (GETRHOL(alpha)) : (rho);
               grid->lamda[0][m][n] = GETLAMDA(alpha, beta, rho);
               grid->mu   [0][m][n] = GETMU(alpha, beta, rho);
               grid->rho  [0][m][n] = rho; 
               if (par->q) {
                  grid->rT    [0][m][n] = medium->rT    ;
                  grid->rEp   [0][m][n] = medium->rEp   ;
                  grid->rEs   [0][m][n] = medium->rEs   ;
                  }
               }
            alpha += dalpha;
            beta  += dbeta;
            rho   += drho;
            }
         break;
      case SLOPE:
         slope = (float) (medium->m2 - medium->m1)/
                 (float) (medium->n2 - medium->n1);
         mstart = (medium->m1 < medium->m2) ? medium->m1 : medium->m2;
         mend   = (medium->m1 > medium->m2) ? medium->m1 : medium->m2;
         if (mstart <              0) mstart = 0;
         if (mend   >= grid->m) mend   = grid->m - 1;
         for(m = mstart; m <= mend; m++) {
            npoint = (m - medium->m1)/slope + medium->n1;
            nstart = (medium->m1 > medium->m2) ? (npoint) : 
                     (medium->n2 > medium->n1) ? medium->n1 : medium->n2;
            nend   = (medium->m1 < medium->m2) ? (npoint) :
                     (medium->n2 > medium->n1) ? medium->n2 : medium->n1; 
            if (nstart <              0) nstart = 0;
            if (nend   >= grid->n) nend   = grid->n - 1;
            for(n = nstart; n <= nend; n++) {
               beta = (medium->beta == 0.) ? (alpha*medium->spratio) : (beta);
               rho = (medium->rho == 0.) ? (GETRHOL(alpha)) : (rho);
               grid->lamda[0][m][n] = GETLAMDA(alpha, beta, rho);
               grid->mu   [0][m][n] = GETMU(alpha, beta, rho);
               grid->rho  [0][m][n] = rho; 
               if (par->q) {
                  grid->rT    [0][m][n] = medium->rT    ;
                  grid->rEp   [0][m][n] = medium->rEp   ;
                  grid->rEs   [0][m][n] = medium->rEs   ;
                  }
               }
            alpha += dalpha;
            beta  += dbeta;
            rho   += drho;
            }
         break;
      case VFILE:
         file = fopen(medium->file, "r");
         if (file == NULL) {
            fprintf(stderr, "Error: can't open v file \"%s\"\n", medium->file);
            exit(-1);
            }
         m1 = medium->m1;
         m2 = medium->m2;
         n1 = medium->n1;
         n2 = medium->n2;
         if (medium->mode == ASCIIALL || medium->mode == BINARYALL) {
            mode = (medium->mode == ASCIIALL) ? (1) : (2);
            readMediumFile(grid->lamda[0], m1, m2, n1, n2, mode, file);
            readMediumFile(grid->mu   [0], m1, m2, n1, n2, mode, file);
            readMediumFile(grid->rho  [0], m1, m2, n1, n2, mode, file);
            for(m = m1; m <= m2; m++) for(n = n1; n <= n2; n++) {
               alpha = grid->lamda[0][m][n]*KMTOM;
               beta  = grid->mu   [0][m][n]*KMTOM;
               rho   = grid->rho  [0][m][n]*GM3TOKGM3;
               grid->lamda[0][m][n] = GETLAMDA(alpha, beta, rho);
               grid->mu   [0][m][n] = GETMU   (alpha, beta, rho);
               grid->rho  [0][m][n] = rho;
               }
            }
         else if (medium->mode == ASCIIP || medium->mode == BINARYP) {
            mode = (medium->mode == ASCIIP) ? (1) : (2);
            readMediumFile(grid->lamda[0], m1, m2, n1, n2, mode, file);
            for(m = m1; m <= m2; m++) for(n = n1; n <= n2; n++) {
               alpha = grid->lamda[0][m][n]*KMTOM;
               beta  = medium->beta;
               if (beta == 0.) beta = alpha*medium->spratio;
               rho   = medium->rho;
               if (rho == 0.) rho   = GETRHOL(alpha);
               grid->lamda[0][m][n] = GETLAMDA(alpha, beta, rho);
               grid->mu   [0][m][n] = GETMU   (alpha, beta, rho);
               grid->rho  [0][m][n] = rho;
               }
            }
         if (par->q) {
            for(m = medium->m1; m <= medium->m2; m++) {
               for(n = medium->n1; n <= medium->n2; n++) {
                  grid->rT    [0][m][n] = medium->rT    ;
                  grid->rEp   [0][m][n] = medium->rEp   ;
                  grid->rEs   [0][m][n] = medium->rEs   ;
                  }
               }
            }
         fclose(file);
         break;
      case EMPTY:
         break;
      default:
         break;
      }
*/


   /* copy plane to all 3D elements (including attenuation) */
/*
   for(l = 1; l < grid->l; l++) {
      for(m = 0; m < grid->m; m++) { 
         for(n = 0; n < grid->n; n++) {
            grid->lamda[l][m][n] = grid->lamda[0][m][n];
            grid->mu   [l][m][n] = grid->mu   [0][m][n];
            grid->rho  [l][m][n] = grid->rho  [0][m][n];
            if (par->q) {
               grid->rT    [l][m][n] = grid->rT    [0][m][n];
               grid->rEp   [l][m][n] = grid->rEp   [0][m][n];
               grid->rEs   [l][m][n] = grid->rEs   [0][m][n];
               }
            }
         }
      }

}
*/




readMediumFile(value, m1, m2, n1, n2, mode, file)
float **value;
int m1, m2, n1, n2, mode;
FILE *file;
{
   int m, n;
   if (mode == 1) {
      for(m = m1; m <= m2; m++) {
         for(n = n1; n <= n2; n++) {
            fscanf(file, "%f", &(value[m][n]));
            }
         }
      }
   if (mode == 2) 
      for(m = m1; m <= m2; m++) fread(value[m] + n1, (n2 - n1 + 1), 4, file);
}

