/*******************************************************************************
   input: input routines for 2D/3D elastic finite-difference code 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   note: input units are km, s, km/s; g/cm3, dyne-cm;
   note: used  units are mks (m, s, m/s, kg/m3, N-m)

   compile: cc -g -c input.c 

   author: shawn larsen
   date  : october 1989
   where : caltech
*****************************************************************************/

#include "elas.h"


FILE *
getFile(argc, argv) 
int    argc;
char **argv;
{
   FILE *file;
   if (argc == 1) return(stdin);
   file = fopen(argv[1], "r");
   if (file == NULL) {
      fprintf(stderr, "Error: Can't open input file \"%s\"\n", argv[1]);
      exit(-1);
      }
   return(file);
}


readInput(input, argc, argv, file)
Input *input;
int    argc;
char **argv;
FILE *file;
{
   int n, maxLines;
   unsigned size;
 
   /* initialize */
   maxLines = MAXLINES;
   n = 0;
 
   /* read input file and store */
   size = maxLines*sizeof(char *);
   input->fileLines = (char **) malloc(size);
   while((n += argvFile(input->fileLines + n, MAXLINES, file)) >= maxLines) {
      maxLines += MAXLINES;
      size = (unsigned) maxLines*sizeof(char *);
      input->fileLines = (char **) realloc((char *) input->fileLines, size);
      };
   input->nLines = n;
     
   /* null last pointer */
   input->fileLines[input->nLines] = NULL;

   /* get command line arguments */
   input->argc = argc;
   input->argv = argv;

}




expandInput(input)
Input *input;
{
   Input FInput[1];
   int i, n, maxLines;
   char **line, **newlines, name[82];
   unsigned size;
   FILE *file;
 
   /* set line types */
   setLine("file ", LINEFILE);

   /* initialize */
   maxLines = (input->nLines/MAXLINES + 1)*MAXLINES;
   newlines = (char **) malloc(maxLines*sizeof(char *));
   line = input->fileLines;

   /* loop over input file lines */
   n = 0;
   while((*line)) {
      newlines[n++] = *line;
      if (getLine(*line) == LINEFILE) {
         name[0] = (char) '\0';
         file    = NULL;
         setArg(*line);
         getArg("file", "s", name);
         endArg();
         if (name[0]) {
            n--;
            file = fopen(name, "r");
            if (file == NULL) {
               fprintf(stderr, "Error: input file \"%s\" not found\n", name);
               exit(-1);
               }
            readInput(FInput, 0, NULL, file);
            maxLines = ((input->nLines + FInput->nLines)/MAXLINES + 1)*MAXLINES;
            size = (unsigned) maxLines*sizeof(char *);
            newlines = (char **) realloc((char *) newlines, size);
            for(i = 0; i < FInput->nLines; i++) {
               newlines[n++] = FInput->fileLines[i];
               }
            input->nLines += FInput->nLines - 1;
            if (FInput->nLines) free(FInput->fileLines);
            }
         }
      line++;
      }

   /* reset input array */
   if (input->nLines) free(input->fileLines);
   input->fileLines = newlines; 

   /* null last pointer */
   input->fileLines[input->nLines] = NULL;
   
   /* end line types */
   endLine();

}




getSize(input, par, source, image, sac, medium, traces)
Input        *input;
Parameter    *par;
FdSource    **source;
Image       **image;
Seismogram  **sac;
Medium      **medium;
Traces      **traces;
{
   int p, v;
   char **line;

   /* initialize counters */
   par->a = 0;
   par->f = 0;
   par->p = 0;
   par->s = 0;
   par->v = 0;

   /* initialize line types */
   setLine("source ", LINESOURCE);
   setLine("sac "   , LINESAC   );
   setLine("image " , LINEIMAGE );
   setLine("volume ", LINEVOLUME);
   setLine("medium ", LINEMEDIUM);
   setLine("block " , LINEBLOCK );
   setLine("poly3d ", LINEPOLY3D);
   setLine("vfile " , LINEVFILE );
   setLine("traces ", LINETRACES);

   /* count lines */
   line = input->fileLines;
   while((*line)) {
      switch(getLine(*line)) {
         case LINESOURCE:
            setArg(*line);
            par->f++;
            endArg();
            break;
         case LINESAC:
            setArg(*line);
            par->s++;
            endArg();
            break;
         case LINEIMAGE:
         case LINEVOLUME:
            setArg(*line);
            par->p++;
            endArg();
            break;
         case LINEMEDIUM:
         case LINEBLOCK :
         case LINEPOLY3D:
         case LINEVFILE :
            setArg(*line);
            par->v++;
            endArg();
            break;
         case LINETRACES:
            setArg(*line);
            par->a++;
            endArg();
            break;
         default:
            break;
         }
      line++;
      }

   /* end line types */
   endLine();

   /* allocate memory for arrays */
   *source = (FdSource   *) malloc(par->f*sizeof(FdSource  ));
   *sac    = (Seismogram *) malloc(par->s*sizeof(Seismogram));
   *image  = (Image      *) malloc(par->p*sizeof(Image     ));
   *medium = (Medium     *) malloc(par->v*sizeof(Medium    ));
   *traces = (Traces     *) malloc(par->a*sizeof(Traces    ));

   /* get secondary parameters */
   v = 0;
   setLine("medium ", LINEMEDIUM);
   setLine("block " , LINEBLOCK );
   setLine("poly3d ", LINEPOLY3D);
   setLine("vfile " , LINEVFILE );
   line = input->fileLines;
   while((*line)) {
      switch(getLine(*line)) {
         case LINEMEDIUM:
            p = 0;
            setArg(*line);
            getArg("p", "d", &p);
            endArg();
            if (p != 0) {
               (*medium)[v].x = (float *) malloc((p+1)*sizeof(float));
               (*medium)[v].z = (float *) malloc((p+1)*sizeof(float));
               }
            v++;
            break;
         case LINEBLOCK: 
         case LINEPOLY3D:
         case LINEVFILE: 
            v++;
            break;
         default:
            break;
         }
      line++;
      }
   endLine();

}


freeSize(input, par, source, image, sac, medium, traces)
Input       *input;
Parameter   *par;
FdSource    *source;
Image       *image;
Seismogram  *sac;
Medium      *medium;
Traces      *traces;
{
   int v;
   for(v = 0; v < par->v; v++) if (medium[v].p != 0) {
      free(medium[v].x);
      free(medium[v].z);
      }
   free(source);
   free(sac   );
   free(image );
   free(medium);
   free(traces);
}


getInput(input, par, source, grid, image, visual, sac, traces, medium)
Input       *input;
Parameter   *par;
FdSource    *source;
FdGrid      *grid;
Image       *image;
Image       *visual;
Seismogram  *sac;
Traces      *traces;
Medium      *medium;
{
   float PSratio;
   int  i, n, g, t, a, mode, mg[MAXGRID];
   int maxsource, maximage, maxsac, maxmedium, maxtraces;
   char **line, name[20], type[20];

   /* get default parameters */
   maxtraces = par->a;
   maxsource = par->f;
   maximage  = par->p;
   maxsac    = par->s;
   maxmedium = par->v;
   par->a = 0;
   par->f = 0;
   par->p = 0;
   par->s = 0;
   par->v = 0;
   par->restart = 0;
   par->dump    = 0;
   getDefault(par, source, grid, visual, sac);

   /* initialize line types */
   setLine("grid "    , LINEGRID);
   setLine("time "    , LINETIME);
   setLine("source "  , LINESOURCE);
   setLine("visual "  , LINEVISUAL);
   setLine("sac "     , LINESAC);
   setLine("image "   , LINEIMAGE);
   setLine("volume "  , LINEVOLUME);
   setLine("medium "  , LINEMEDIUM);
   setLine("block "   , LINEBLOCK);
   setLine("poly3d "  , LINEPOLY3D);
   setLine("vfile "   , LINEVFILE);
   setLine("topo "    , LINETOPO);
   setLine("parallel ", LINEPARALLEL);
   setLine("traces "  , LINETRACES);
   setLine("trace "   , LINETRACES);
   setLine("restart " , LINERESTART);

   /* get line argument parameters */
   line = input->fileLines;
   while((*line)) {
      switch(getLine(*line)) {
         case LINEGRID:
            setArg(*line);
            par->x    = 0.; par->dx = 0.; par->n = 0; 
            par->y    = 0.; par->dy = 0.; par->l = 0;
            par->z    = 0.; par->dz = 0.; par->m = 0;
            par->vred = 0.;
            par->tlen = 0.;
            par->afile[0] = (char) '\0';
            getArg("l"     , "d", &par->l);
            getArg("m"     , "d", &par->m);
            getArg("n"     , "d", &par->n);
            getArg("x"     , "f", &par->x);
            getArg("y"     , "f", &par->y);
            getArg("z"     , "f", &par->z);
            getArg("dh"    , "f", &par->dx);
            getArg("dh"    , "f", &par->dy);
            getArg("dh"    , "f", &par->dz);
            getArg("dx"    , "f", &par->dx);
            getArg("dy"    , "f", &par->dy);
            getArg("dz"    , "f", &par->dz);
            getArg("b"     , "d", &par->boundary);
            getArg("q"     , "d", &par->q);
            getArg("damp"  , "d", &par->damp);
            getArg("adamp" , "f", &par->adamp);
            getArg("model" , "d", &par->model);
            getArg("active", "d", &par->active);
            getArg("afile" , "s",  par->afile);
            if (par->active == ACTIVEREDUCING) {
               getArg("vred","f", &par->vred);
               getArg("tlen","f", &par->tlen);
               }
            for(g = 0; g < MAXGRID; g++) mg[g] = 0;
            par->g = getArg("m", "vd", mg);
            if (par->g == 0) par->g = 1;
            endArg();
            break;
         case LINEPARALLEL:  /* temporary */
            setArg(*line);
            par->pfile[0] = (char) '\0';
            getArg("nx",       "d",  &par->nx);
            getArg("ny",       "d",  &par->ny);
            getArg("nz",       "d",  &par->nz);
            getArg("file",     "s",   par->pfile);
            endArg();
            break;
         default:
            break;
         }
      line++;
      }
   getGridValid     (par, grid, mg);
   getGridAdjust    (par, grid, mg);
   getGridDimensions(par, grid, mg);

/* temporary parallelization (NON VDENSITY) */
   /* get parallelization */
   par->node  = mpNode();
   par->nnode = mpN   ();
   if (grid->l <= 1) par->ny = 1;
   if (par->nx*par->ny*par->nz != mpN()) {
      fprintf(stderr, "Error: bad parallel nodes\n");
      exit(-1);
      }
   for(g = 0; g < par->g; g++) mpBoundary(par, grid + g);
   for(g = 0; g < par->g; g++) {
      if (grid[g].px1 >= 0) grid[g].bx1 = MESSAGEPASS;
      if (grid[g].px2 >= 0) grid[g].bx2 = MESSAGEPASS;
      if (grid[g].py1 >= 0) grid[g].by1 = MESSAGEPASS;
      if (grid[g].py2 >= 0) grid[g].by2 = MESSAGEPASS;
      if (grid[g].pz1 >= 0) grid[g].bz1 = MESSAGEPASS;
      if (grid[g].pz2 >= 0) grid[g].bz2 = MESSAGEPASS;
      }
   /* NON VDENSITY (par-nnode check) */
   for(g = 0; g < par->g; g++) if (par->nnode > 1) mpDecomposition(par, grid + g);


   line = input->fileLines;
   while((*line)) {
      switch(getLine(*line)) {
         case LINETIME:
            setArg(*line);
            mstArg("dt", "f", &par->dt);
            mstArg("t" , "d", &par->t);
            endArg();
            break;
         case LINESOURCE:
/*
            if (par->f >= MAXSOURCE) {
               fprintf(stderr, "Error: too many sources (%d)\n", MAXSOURCE+1);
               exit(-1);
               };
*/
            if (par->f >= maxsource) {
               fprintf(stderr, "Error: too many sources (%d)\n", maxsource+1);
               exit(-1);
               };
            i = par->f;
            source[i].type   =     DSOURCE;
            source[i].amp    =     DFAMPM0;
            source[i].freq   =          1.;
            source[i].t0     =         0.0;
            source[i].l      = (par->d == 2) ? (0) : (par->l/2.);
            source[i].m      = par->m/2.;
            source[i].n      = par->n/2.;
            source[i].p      = -1;
            source[i].length = -1.;
            source[i].width  = -1.;
            source[i].depth  =  0.;
            source[i].v      = -1.;
            source[i].s0     =  0.;
            source[i].d0     = -1.;
            source[i].M0     = 1.;
            source[i].strike = 0.;
            source[i].dip    = 90.;
            source[i].rake   = 90.;
            source[i].Mxx    = 0.0; source[i].Myy = 0.0; source[i].Mzz = 0.0;
            source[i].Mxy    = 0.0; source[i].Mxz = 0.0; source[i].Myz = 0.0;
            source[i].Fx     = 0.0; source[i].Fy  = 0.0; source[i].Fz  = 0.0;
            source[i].file [0] = (char) '\0';
            source[i].dfile[0] = (char) '\0';
            source[i].g       = 0;
            setArg(*line);
            mstArg("type"  ,  "d", &source[i].type);
            if (source[i].type == BSOURCE) source[i].m = 0;
            mstArg("amp",     "f", &source[i].amp);
            getArg("t0",      "f", &source[i].t0);
            getArg("freq",    "f", &source[i].freq);
            getArg("l",       "d", &source[i].l);
            getArg("m",       "d", &source[i].m);
            getArg("n",       "d", &source[i].n);
            source[i].x = (par->dx/KMTOM)*((float) source[i].n);
            source[i].y = (par->dy/KMTOM)*((float) source[i].l);
            source[i].z = (par->dz/KMTOM)*((float) source[i].m);
            getArg("x",       "f", &source[i].x);
            getArg("y",       "f", &source[i].y);
            getArg("z",       "f", &source[i].z);
            source[i].x *= KMTOM;
            source[i].y *= KMTOM;
            source[i].z *= KMTOM;
            source[i].depth = source[i].z/KMTOM;
            getArg("length",  "f", &source[i].length);
            getArg("width",   "f", &source[i].width);
            getArg("depth",   "f", &source[i].depth);
            source[i].length *= KMTOM;
            source[i].width  *= KMTOM;
            source[i].depth  *= KMTOM;
            getArg("v",       "f", &source[i].v);
            getArg("s0",      "f", &source[i].s0);
            getArg("d0",      "f", &source[i].d0);
            source[i].v  *= KMTOM;
            source[i].s0 *= KMTOM;
            source[i].d0 *= KMTOM;
            getArg("strike",  "f", &source[i].strike);
            getArg("dip",     "f", &source[i].dip);
            getArg("rake",    "f", &source[i].rake);
            getArg("Mxx,mxx", "f", &source[i].Mxx);
            getArg("Myy,myy", "f", &source[i].Myy);
            getArg("Mzz,mzz", "f", &source[i].Mzz);
            getArg("Mxy,mxy", "f", &source[i].Mxy);
            getArg("Mxz,mxz", "f", &source[i].Mxz);
            getArg("Myz,myz", "f", &source[i].Myz);
            getArg("Fx,fx",   "f", &source[i].Fx);
            getArg("Fy,fy",   "f", &source[i].Fy);
            getArg("Fz,fz",   "f", &source[i].Fz);
            getArg("file",    "s",  source[i].file);
            getArg("dfile",   "s",  source[i].dfile);
            getArg("p",       "d", &source[i].p);
            getArg("join",    "d", &par->join);
            getArg("xjoin",   "d", &par->xjoin);
            if (par->join) getArg("tjoin",   "d", &par->tjoin);
            endArg();
            if (source[i].type == ASOURCE) {
               getFaultMoment(par, source + i);
               source[i].type = MSOURCE;
               }
            if (source[i].type == BSOURCE) {
               getFaultMoment(par, source + i);
               }
            if (source[i].type == CSOURCE) {
               getFaultMoment(par, source + i);
               }
            if (source[i].type == ESOURCE) {
               getFaultMoment(par, source + i);
               }
            if (par->d == 2) {source[i].l = 0; source[i].y = 0.;}

            /**/
            /*
            g = source[i].g;
            */
            /*
            source[i].x = grid[g].x0 + ((float) source[i].n)*grid[g].dx;
            source[i].y = grid[g].y0 + ((float) source[i].l)*grid[g].dy;
            source[i].z = grid[g].z0 + ((float) source[i].m)*grid[g].dz;
            */
            /*
            source[i].x =              ((float) source[i].n)*grid[g].dx;
            source[i].y =              ((float) source[i].l)*grid[g].dy;
            source[i].z =              ((float) source[i].m)*grid[g].dz;
            */
            /*
            if (par->nnode == 1) source[i].y += grid[g].y0;
            */
            /**/

            par->f++;
            break;
         case LINESAC:
/*
            if (par->s >= MAXSAC) {
               fprintf(stderr, "Error: too many seismograms (%d)\n", MAXSAC+1);
               exit(-1);
               };
*/
            if (par->s >= maxsac) {
               fprintf(stderr, "Error: too many seismograms (%d)\n", maxsac+1);
               exit(-1);
               };
            i = par->s;
            sac[i].l      =   -1;
            sac[i].m      =    0;
            sac[i].n      =   -1;
            sac[i].sample =    1;
            sac[i].mode   = SACS;
            sac[i].scale  =  1.0;
            sac[i].g      =    0;
            strcpy(sac[i].file, NAMESAC);
            setArg(*line);
            getArg("l"     , "d",  &sac[i].l);
            getArg("m"     , "d",  &sac[i].m);
            getArg("n"     , "d",  &sac[i].n);
            sac[i].x = (par->dx/KMTOM)*((float) sac[i].n);
            sac[i].y = (par->dy/KMTOM)*((float) sac[i].l);
            sac[i].z = (par->dz/KMTOM)*((float) sac[i].m);
            getArg("x",       "f", &sac[i].x);
            getArg("y",       "f", &sac[i].y);
            getArg("z",       "f", &sac[i].z);
            sac[i].x *= KMTOM;
            sac[i].y *= KMTOM;
            sac[i].z *= KMTOM;
            getArg("scale" , "f" , &sac[i].scale);
            getArg("sample", "d" , &sac[i].sample);
            getArg("mode"  , "d" , &sac[i].mode);
            getArg("file"  , "s" ,  sac[i].file);
            endArg();
            par->s++;
            break;
         case LINETRACES:
/*
            if (par->a >= MAXTRACES) {
               fprintf(stderr, "Error: too many traces (%d)\n", MAXTRACES+1);
               exit(-1);
               };
*/
            if (par->a >= maxtraces) {
               fprintf(stderr, "Error: too many traces (%d)\n", maxtraces+1);
               exit(-1);
               };
            i = par->a;
            traces[i].x1     =   0.;
            traces[i].x2     =   0.;
            traces[i].y1     =   0.;
            traces[i].y2     =   0.;
            traces[i].z1     =   0.;
            traces[i].z2     =   0.;
            traces[i].p      =    0;
            traces[i].sample =    1;
            traces[i].scale  =  1.0;
            if (par->d == 2) traces[i].mode = TRACEVX + TRACEVZ +  
                                              TRACEP  + TRACES ;
            if (par->d == 3) traces[i].mode = TRACEVX + TRACEVY + TRACEVZ +
                                              TRACEP  + TRACES ;
            strcpy(traces[i].file, NAMETRACES);
            traces[i].tfile[0] = (char) '\0';
            setArg(*line);
            getArg("x1",     "f" , &traces[i].x1);
            getArg("x2",     "f" , &traces[i].x2);
            getArg("y1",     "f" , &traces[i].y1);
            getArg("y2",     "f" , &traces[i].y2);
            getArg("z1",     "f" , &traces[i].z1);
            getArg("z2",     "f" , &traces[i].z2);
            getArg("p",      "d" , &traces[i].p );
            traces[i].x1 *= KMTOM;
            traces[i].x2 *= KMTOM;
            traces[i].y1 *= KMTOM;
            traces[i].y2 *= KMTOM;
            traces[i].z1 *= KMTOM;
            traces[i].z2 *= KMTOM;
            if (par->d == 2) {traces[i].y1 = 0.; traces[i].y2 = 0.;}
            getArg("scale" , "f" , &traces[i].scale);
            getArg("sample", "d" , &traces[i].sample);
            getArg("mode"  , "d" , &traces[i].mode);
            getArg("file"  , "s" ,  traces[i].file);
            getArg("tfile" , "s" ,  traces[i].tfile);
            endArg();
            if (par->g > 1) {
               fprintf(stderr, "Error: trace output can't handle ");
               fprintf(stderr, "variable density grid\n");
               exit(1);
               }
            getTraceLimits(par, grid, traces + i);
            n = (int) strlen(traces[i].file);
            sprintf(traces[i].file + n, ".%d.\0", par->a);
            par->a++;
            break;
         case LINEIMAGE:
/*
            if (par->p >= MAXIMAGE) {
               fprintf(stderr, "Error: too many images (%d)\n", MAXIMAGE+1);
               exit(-1);
               };
*/
            if (par->p >= maximage) {
               fprintf(stderr, "Error: too many images (%d)\n", maximage+1);
               exit(-1);
               };
            i = par->p;
            image[i].scale   =               1.0;
            image[i].sample  =                 1;
            image[i].mode    =            WAVEPS;
            image[i].movie   =                 0;
            image[i].plane   =                -1;
            image[i].w       =    -par->dx/KMTOM;
            image[i].d       =           XZPLANE;
            image[i].t       =                 0;
            image[i].p       =                 0;
            strcpy(image[i].file, NAMEIMAGE);
            setArg(*line);
            getArg("t"     , "d",  &image[i].t);
            getArg("file"  , "s" ,  image[i].file);
            getArg("scale" , "f" , &image[i].scale);
            getArg("sample", "d" , &image[i].sample);
            getArg("movie" , "d" , &image[i].movie);
            getArg("mode"  , "d" , &image[i].mode); 
            if (getArg("m",  "d" , &image[i].plane)) image[i].d = XYPLANE;
            if (getArg("n",  "d" , &image[i].plane)) image[i].d = YZPLANE;
            if (getArg("l",  "d" , &image[i].plane)) image[i].d = XZPLANE;
            image[i].w = (par->dx/KMTOM)*((float) image[i].plane);
            if (getArg("z",  "f" , &image[i].w    )) image[i].d = XYPLANE;
            if (getArg("x",  "f" , &image[i].w    )) image[i].d = YZPLANE;
            if (getArg("y",  "f" , &image[i].w    )) image[i].d = XZPLANE;
            image[i].w *= KMTOM;
            image[i].plane = (int) (image[i].w/par->dx);
            endArg();
            par->p++;
            break;
         case LINEVOLUME:
/*
            if (par->p >= MAXIMAGE) {
               fprintf(stderr, "Error: too many images (%d)\n", MAXIMAGE+1);
               exit(-1);
               };
*/
            if (par->p >= maximage) {
               fprintf(stderr, "Error: too many images (%d)\n", maximage+1);
               exit(-1);
               };
            i = par->p;
            image[i].scale   =               1.0;
            image[i].sample  =                 1;
            image[i].mode    =           VOLUMEP;
            image[i].movie   =                 0;
            image[i].plane   =                -1;
            image[i].w       =    -par->dx/KMTOM;
            image[i].d       =           XZPLANE;
            image[i].t       =                 0;
            image[i].p       =                 0;
            strcpy(image[i].file, NAMEVOLUME);
            setArg(*line);
            getArg("t"     , "d",  &image[i].t);
            getArg("file"  , "s" ,  image[i].file);
            getArg("scale" , "f" , &image[i].scale);
            getArg("sample", "d" , &image[i].sample);
            getArg("movie" , "d" , &image[i].movie);
            getArg("mode"  , "d" , &image[i].mode); 
            endArg();
            if (image[i].mode == 1) image[i].mode = VOLUMEP;
            if (image[i].mode == 2) image[i].mode = VOLUMES;
            if (image[i].mode != VOLUMEP && image[i].mode != VOLUMES) {
               image[i].t     = 0;
               image[i].movie = 0;
               image[i].mode  = 0;
               }
            par->p++;
            break;
         case LINEVISUAL:
            visual->scale      =         1.0;
            visual->sample     =           1;
            visual->mode       =           1;
            visual->movie      =           0;
            visual->plane      =          -1;
            visual->w          =    -par->dx/KMTOM;
            visual->d          =     XZPLANE;
            setArg(*line);
            getArg("scale" ,   "f" , &visual->scale);
            getArg("sample",   "d" , &visual->sample);
            getArg("movie" ,   "d" , &visual->movie);
            getArg("mode"  ,   "d" , &visual->mode); 
            if (getArg("m",    "d" , &visual->plane)) visual->d = XYPLANE;
            if (getArg("n",    "d" , &visual->plane)) visual->d = YZPLANE;
            if (getArg("l",    "d" , &visual->plane)) visual->d = XZPLANE;
            visual->w = (par->dx/KMTOM)*((float) visual->plane);
            if (getArg("z",    "f" , &visual->w    )) visual->d = XYPLANE;
            if (getArg("x",    "f" , &visual->w    )) visual->d = YZPLANE;
            if (getArg("y",    "f" , &visual->w    )) visual->d = XZPLANE;
            visual->w *= KMTOM;
            visual->plane = (int) (visual->w/par->dx);
            endArg();
            break;
         case LINEMEDIUM:
/*
            if (par->v >= MAXMEDIUM) {
               fprintf(stderr, "Error: too many medium (%d)\n", MAXMEDIUM+1);
               exit(-1);
               };
*/
            if (par->v >= maxmedium) {
               fprintf(stderr, "Error: too many images (%d)\n", maxmedium+1);
               exit(-1);
               };
            i = par->v;
            medium[i].alpha    = ALPHA;
            medium[i].beta     = BETA;
            medium[i].beta     =   0.;
            medium[i].rho      = 0.; 
            medium[i].gradient = 0.;
            medium[i].spratio  = (1./SPRATIO);
            /*
            medium[i].rT       = RT ;
            medium[i].rEp      = REP;
            medium[i].rEs      = RES;
            */
            medium[i].Qp       = QP;
            medium[i].Qs       = QS;
            medium[i].Qw       = QW;
            medium[i].Qp       = -1.;
            medium[i].Qs       = -1.;
            medium[i].Qw       = -1.;
            medium[i].l1       = 0;
            medium[i].m1       = 0;
            medium[i].n1       = 0;
            medium[i].l2       = par->l - 1;
            medium[i].m2       = par->m - 1;
            medium[i].n2       = par->n - 1;
            medium[i].p        = 0;
            medium[i].type     = BLOCK;
            medium[i].mode     = VMODE;
            medium[i].file[0]  = (char) '\0';
            medium[i].g        = 0;
            medium[i].c        = 0;
            medium[i].seg      = 0;
            PSratio            = (1./SPRATIO);
            setArg(*line);
            getArg("alpha",         "f", &medium[i].alpha);
            getArg("PSratio,PS,ps", "f", &PSratio);
            medium[i].spratio = (PSratio <= 0.) ? (0.) : (1./PSratio);
            if (getArg("beta",      "f", &medium[i].beta))medium[i].spratio=0.;
            getArg("rho",           "f", &medium[i].rho);
            getArg("gradient",      "f", &medium[i].gradient);
            getArg("Qp",            "f", &medium[i].Qp);
            getArg("Qs",            "f", &medium[i].Qs);
            getArg("Q" ,            "f", &medium[i].Qp);
            getArg("Q" ,            "f", &medium[i].Qs);
            getArg("Qf",            "f", &medium[i].Qw);
            getArg("n1,n",          "d", &medium[i].n1);
            getArg("n2,n",          "d", &medium[i].n2);
            getArg("m1,m",          "d", &medium[i].m1);
            getArg("m2,m",          "d", &medium[i].m2);
            getArg("l1,l",          "d", &medium[i].l1);
            getArg("l2,l",          "d", &medium[i].l2);
            medium[i].x1 = (par->dx/KMTOM)*((float) medium[i].n1);
            medium[i].x2 = (par->dx/KMTOM)*((float) medium[i].n2);
            medium[i].y1 = (par->dy/KMTOM)*((float) medium[i].l1);
            medium[i].y2 = (par->dy/KMTOM)*((float) medium[i].l2);
            medium[i].z1 = (par->dz/KMTOM)*((float) medium[i].m1);
            medium[i].z2 = (par->dz/KMTOM)*((float) medium[i].m2);
            getArg("x1,x",          "f", &medium[i].x1);
            getArg("x2,x",          "f", &medium[i].x2);
            getArg("y1,y",          "f", &medium[i].y1);
            getArg("y2,y",          "f", &medium[i].y2);
            getArg("z1,z",          "f", &medium[i].z1);
            getArg("z2,z",          "f", &medium[i].z2);
            medium[i].x1 *= KMTOM;
            medium[i].x2 *= KMTOM;
            medium[i].y1 *= KMTOM;
            medium[i].y2 *= KMTOM;
            medium[i].z1 *= KMTOM;
            medium[i].z2 *= KMTOM;
            getArg("p",             "d", &medium[i].p);
            getArg("type",          "d", &medium[i].type);
            getArg("mode",          "d", &medium[i].mode);
            getArg("file",          "s",  medium[i].file);
            endArg();
            if (medium[i].p) getPolygonInput(medium + i, line);
            par->v++;
            break;
         case LINEBLOCK:
/*
            if (par->v >= MAXMEDIUM) {
               fprintf(stderr, "Error: too many mediums (%d)\n", MAXMEDIUM+1);
               exit(-1);
               };
*/
            if (par->v >= maxmedium) {
               fprintf(stderr, "Error: too many images (%d)\n", maxmedium+1);
               exit(-1);
               };
            i = par->v;
            medium[i].alpha    = -1.;
            medium[i].beta     = -1.;
            medium[i].rho      = -1.;
            medium[i].gradient =  0.;
            medium[i].Qp       = -1.;
            medium[i].Qs       = -1.;
            medium[i].Qw       = -1.;
            medium[i].spratio  = -1.;
            medium[i].p        = 0;
            medium[i].g        = 0;
            medium[i].c        = 0;
            medium[i].l1       = 0;
            medium[i].m1       = 0;
            medium[i].n1       = 0;
            medium[i].l2       = par->l - 1;
            medium[i].m2       = par->m - 1;
            medium[i].n2       = par->n - 1;
            medium[i].type     = BFILL;
            medium[i].mode     = EMPTY;
            medium[i].seg      = 0;
            medium[i].file[0]  = (char) '\0';
            setArg(*line);
            getArg("alpha,p,P",       "f",  &medium[i].alpha);
            getArg("beta,s,S",        "f",  &medium[i].beta); 
            getArg("PSratio,PS,ps",   "f",  &medium[i].spratio);
            getArg("rho,r,R",         "f",  &medium[i].rho);
            getArg("gradient",        "f",  &medium[i].gradient);
            getArg("Qp",              "f",  &medium[i].Qp);
            getArg("Qs",              "f",  &medium[i].Qs);
            getArg("Q" ,              "f",  &medium[i].Qp);
            getArg("Q" ,              "f",  &medium[i].Qs);
            getArg("Qf",              "f",  &medium[i].Qw);
            getArg("n1,n",            "d",  &medium[i].n1);
            getArg("n2,n",            "d",  &medium[i].n2);
            getArg("m1,m",            "d",  &medium[i].m1);
            getArg("m2,m",            "d",  &medium[i].m2);
            getArg("l1,l",            "d",  &medium[i].l1);
            getArg("l2,l",            "d",  &medium[i].l2);
            medium[i].x1 = (par->dx/KMTOM)*((float) medium[i].n1);
            medium[i].x2 = (par->dx/KMTOM)*((float) medium[i].n2);
            medium[i].y1 = (par->dy/KMTOM)*((float) medium[i].l1);
            medium[i].y2 = (par->dy/KMTOM)*((float) medium[i].l2);
            medium[i].z1 = (par->dz/KMTOM)*((float) medium[i].m1);
            medium[i].z2 = (par->dz/KMTOM)*((float) medium[i].m2);
            getArg("x1,x",            "f",  &medium[i].x1);
            getArg("x2,x",            "f",  &medium[i].x2);
            getArg("y1,y",            "f",  &medium[i].y1);
            getArg("y2,y",            "f",  &medium[i].y2);
            getArg("z1,z",            "f",  &medium[i].z1);
            getArg("z2,z",            "f",  &medium[i].z2);
            medium[i].x1 *= KMTOM;
            medium[i].x2 *= KMTOM;
            medium[i].y1 *= KMTOM;
            medium[i].y2 *= KMTOM;
            medium[i].z1 *= KMTOM;
            medium[i].z2 *= KMTOM;
            endArg();
            if (medium[i].beta    >= 0.) medium[i].spratio = -1.;
            if (medium[i].spratio>0.)medium[i].spratio = 1./medium[i].spratio;
            par->v++;
            break;
         case LINEPOLY3D:
/*
            if (par->v >= MAXMEDIUM) {
               fprintf(stderr, "Error: too many mediums (%d)\n", MAXMEDIUM+1);
               exit(-1);
               };
*/
            if (par->v >= maxmedium) {
               fprintf(stderr, "Error: too many images (%d)\n", maxmedium+1);
               exit(-1);
               };
            i = par->v;
            medium[i].alpha    = -1.;
            medium[i].beta     = -1.;
            medium[i].rho      = -1.;
            medium[i].gradient =  0.;
            medium[i].Qp       = -1.;
            medium[i].Qs       = -1.;
            medium[i].Qw       = -1.;
            medium[i].spratio  = -1.;
            medium[i].p        = 0;
            medium[i].g        = 0;
            medium[i].c        = 0;
            medium[i].l1       = 0;
            medium[i].m1       = 0;
            medium[i].n1       = 0;
            medium[i].l2       = par->l - 1;
            medium[i].m2       = par->m - 1;
            medium[i].n2       = par->n - 1;
            medium[i].lon      = 0.;
            medium[i].lat      = 0.;
            medium[i].angle    = 0.;
            medium[i].type     = POLY3;
            medium[i].mode     = EMPTY;
            medium[i].seg      = 0;
            medium[i].file[0]  = (char) '\0';
            setArg(*line);
            getArg("alpha,p,P",       "f",  &medium[i].alpha);
            getArg("beta,s,S",        "f",  &medium[i].beta); 
            getArg("PSratio,PS,ps",   "f",  &medium[i].spratio);
            getArg("rho,r,R",         "f",  &medium[i].rho);
            getArg("gradient",        "f",  &medium[i].gradient);
            getArg("Qp",              "f",  &medium[i].Qp);
            getArg("Qs",              "f",  &medium[i].Qs);
            getArg("Q" ,              "f",  &medium[i].Qp);
            getArg("Q" ,              "f",  &medium[i].Qs);
            getArg("Qf",              "f",  &medium[i].Qw);
            getArg("n1,n",            "d",  &medium[i].n1);
            getArg("n2,n",            "d",  &medium[i].n2);
            getArg("m1,m",            "d",  &medium[i].m1);
            getArg("m2,m",            "d",  &medium[i].m2);
            getArg("l1,l",            "d",  &medium[i].l1);
            getArg("l2,l",            "d",  &medium[i].l2);
            medium[i].x1 = (par->dx/KMTOM)*((float) medium[i].n1);
            medium[i].x2 = (par->dx/KMTOM)*((float) medium[i].n2);
            medium[i].y1 = (par->dy/KMTOM)*((float) medium[i].l1);
            medium[i].y2 = (par->dy/KMTOM)*((float) medium[i].l2);
            medium[i].z1 = (par->dz/KMTOM)*((float) medium[i].m1);
            medium[i].z2 = (par->dz/KMTOM)*((float) medium[i].m2);
            getArg("x1,x",            "f",  &medium[i].x1);
            getArg("x2,x",            "f",  &medium[i].x2);
            getArg("y1,y",            "f",  &medium[i].y1);
            getArg("y2,y",            "f",  &medium[i].y2);
            getArg("z1,z",            "f",  &medium[i].z1);
            getArg("z2,z",            "f",  &medium[i].z2);
            medium[i].x1 *= KMTOM;
            medium[i].x2 *= KMTOM;
            medium[i].y1 *= KMTOM;
            medium[i].y2 *= KMTOM;
            medium[i].z1 *= KMTOM;
            medium[i].z2 *= KMTOM;
            mstArg("lon,longitude",   "f",  &medium[i].lon);
            mstArg("lat,latitude",    "f",  &medium[i].lat);
            getArg("angle",           "f",  &medium[i].angle);
            mstArg("file",            "s",   medium[i].file);
            endArg();
            if (medium[i].beta    >= 0.) medium[i].spratio = -1.;
            if (medium[i].spratio>0.)medium[i].spratio = 1./medium[i].spratio;
            par->v++;
            break;
         case LINEVFILE:
/*
            if (par->v >= MAXMEDIUM) {
               fprintf(stderr, "Error: too many mediums (%d)\n", MAXMEDIUM+1);
               exit(-1);
               };
*/
            if (par->v >= maxmedium) {
               fprintf(stderr, "Error: too many images (%d)\n", maxmedium+1);
               exit(-1);
               };
            i = par->v;
            medium[i].alpha    = -1.;
            medium[i].beta     = -1.;
            medium[i].rho      = -1.;
            medium[i].gradient =  0.;
            medium[i].Qp       = -1.;
            medium[i].Qs       = -1.;
            medium[i].Qw       = -1.;
            medium[i].spratio  = -1.;
            medium[i].p        = 0;
            medium[i].g        = 0;
            medium[i].c        = 0;
            medium[i].type     = VFILE;
            medium[i].mode     = EMPTY;
            medium[i].seg      = 0;
            medium[i].file[0]  = (char) '\0';
            type[0]            = (char) '\0';
            setArg(*line);
            getArg("alpha,p,P",       "f",  &medium[i].alpha);
            getArg("beta,s,S",        "f",  &medium[i].beta); 
            getArg("PSratio,PS,ps",   "f",  &medium[i].spratio);
            getArg("rho,r,R",         "f",  &medium[i].rho);
            getArg("gradient",        "f",  &medium[i].gradient);
            getArg("Qp",              "f",  &medium[i].Qp);
            getArg("Qs",              "f",  &medium[i].Qs);
            getArg("Q" ,              "f",  &medium[i].Qp);
            getArg("Q" ,              "f",  &medium[i].Qs);
            getArg("Qf",              "f",  &medium[i].Qw);
            mstArg("n1,n",            "d",  &medium[i].n1);
            mstArg("n2,n",            "d",  &medium[i].n2);
            mstArg("m1,m",            "d",  &medium[i].m1);
            mstArg("m2,m",            "d",  &medium[i].m2);
            if (par->d == 3) {
            mstArg("l1,l",            "d",  &medium[i].l1);
            mstArg("l2,l",            "d",  &medium[i].l2);
            }
            getArg("c",               "d",  &medium[i].c   );
            getArg("type",            "s",             type);
            getArg("seg",             "d",  &medium[i].seg ); /*1 usesegfill*/
            mstArg("file",            "s",   medium[i].file);
            endArg();
            medium[i].mode = getGridFileType(medium[i].file, type);
            if (medium[i].beta    >= 0.) medium[i].spratio = -1.;
            if (medium[i].spratio>0.)medium[i].spratio = 1./medium[i].spratio;
            par->v++;
            break;
         case LINETOPO:
            par->topo   = 1;
            par->ext    = 1;
            par->depth  = 0.;
            par->ave    = 1;
            par->tfile[0] = (char) '\0';
            setArg(*line);
            getArg("file",     "s",   par->tfile);
            getArg("depth",    "f",  &par->depth);
            endArg();
            for(g = 0; g < MAXGRID; g++) grid[g].ave = 1;
            break;
         case LINEPARALLEL:
            setArg(*line);
            getArg("nx",       "d",  &par->nx);
            getArg("ny",       "d",  &par->ny);
            getArg("nz",       "d",  &par->nz);
            endArg();
            break;
         case LINERESTART:
            par->restart = 0;
            par->dump    = 0;
            par->nolast  = 0;
            par->rfile[0] = (char) '\0';
            setArg(*line);
            getArg("restart",  "d",  &par->restart);
            getArg("dump",     "d",  &par->dump   );
            getArg("nolast",   "d",  &par->nolast );
            getArg("file",     "s",   par->rfile);
            endArg();
            break;
         default:
            break;
         }
      line++;
      }

   /* end line types */
   endLine();


   /* check restart information */
   if (par->restart || par->dump) {
      if (par->restart < 0) par->restart = 0;
      if (par->dump    < 0) par->dump    = 0;
      if (par->restart >= par->t) {
         fprintf(stderr, "Warning: restart time > run time\n");
         par->restart = 0;
         }
      if (par->dump    >  par->t) {
         fprintf(stderr, "Warning: restart dump time > run time\n");
         par->dump = 0;
         }
      }
   if (par->restart || par->dump) {
      if (par->rfile[0] == (char) '\0') {
         fprintf(stderr, "Error: restart file not specified\n");
         exit(1);
         }
      if (par->d == 2) {
         fprintf(stderr, "Error: restart capabilities only work for 3-D\n");
         exit(1);
         }
      }

   /* check model type */
   if (par->model < 1 || par->model > 2) {
      fprintf(stderr, "Error: invalid model\n");
      exit(1);
      }
   if (par->model == 2 && par->d == 2) {
      fprintf(stderr, "Error: acoustic modeling option for 3D only\n");
      exit(1);
      }
   if (par->model == 2 && par->q >= 1) {
      fprintf(stderr, "Error: acoustic modeling option for non-attenuation\n");
      exit(1);
      }
   if (par->model == 2) {
      for(i = 0; i < par->s; i++) {
         if (sac[i].mode == 2) {
            fprintf(stderr, "Error: can't use sac mode 2 w acoustic model\n");
            exit(1);
            }
         }
      for(i = 0; i < par->p; i++) {
         if (image[i].mode == 6 || image[i].mode == 7 || image[i].mode == 8) {
            fprintf(stderr, "Error: can't write medium with acoustic model\n");
            exit(1);
            }
         if (image[i].mode >= 17 && image[i].mode <= 19) {
            fprintf(stderr, "Error: invalid image mode with acoustic model\n");
            exit(1);
            }
         }
      if (par->restart || par->dump || par->nolast) {
         fprintf(stderr, "Error: can't use restart options with acoustic model\n");
         exit(1);
         }
      if (par->join) { 
         fprintf(stderr, "Error: can't use join options with acoustic model\n");
         exit(1);
         }
      for(i = 0; i < par->f; i++) {
         if (source[i].type != 1 && source[i].type != 5) {
            fprintf(stderr, "Error: invalid source type with acoustic model\n");
            exit(1);
            }
         }
      for(i = 0; i < par->v; i++) {
         if (medium[i].type != BFILL && medium[i].type != VFILE && 
            medium[i].type != POLY3) {
            fprintf(stderr, "Error: invalid medium type with acoustic model\n");
            exit(1);
            }
         }
      }

   /* check for "water" boundary conditions */
   if (par->boundary == WATER) {
      par->boundary = SURFACE;
      par->water = 1;
      }
   if (par->water && par->d == 2) {
      fprintf(stderr, "Error: water boundary conditions valid only in 3-D\n");
      exit(1);
      }

   /* check multigrid information */
   if (par->g > MAXGRID) {
      fprintf(stderr, "Error: too many grids\n");
      exit(-1);
      }
   for(g = 0; g < MAXGRID; g++) {
      grid[g].b = SURFACE;
      if (par->topo == 1) grid[par->g - 1].b = TOPOGRAPHY;
      grid[g].g = g;
      }
   for(i = 0; i < par->v; i++) if (medium[i].g < 0 || medium[i].g >= par->g) {
      fprintf(stderr, "Error: incorrect medium grid\n");
      exit(-1);
      }
   for(g = 1; g < par->g; g++) {
      /*grid[g].n = 2*grid[g-1].n;*/
      if (par->d == 2) grid[g].l = 1; /* 2 D only */ 
      /*
      grid[g].dx = grid[g-1].dx/2.; 
      grid[g].dy = grid[g-1].dy/2.; 
      grid[g].dz = grid[g-1].dz/2.;
      */
      if (par->d == 2) grid[g].fnt = 0;
      if (par->d == 2) grid[g].bck = 0;
      }
   for(g = 0; g < par->g; g++) {
      grid[g].s = (1 << (par->g - g - 1));
      }

   /* check active grid */
   if (par->active < 0 || par->active > 4) par->active = ACTIVEDEFAULT;
   if (par->active == ACTIVEREDUCING) if (par->tlen <= 0.) 
      par->tlen = ((float) par->t)*par->dt*FTLEN;
   if (par->active == ACTIVEBOX && !par->afile[0]) {
      fprintf(stderr, "need active grid file for active=4\n");
      exit(1);
      }

   /* get boundary values */
   if (par->boundary < 1 || par->boundary > 3) par->boundary = SURFACE;
   for(g = 0; g < par->g; g++) {
      grid[g].by1 = REFLECTING;
      grid[g].by2 = REFLECTING;
      if (par->boundary == SURFACE) {
         grid[g].bz2 = (g ==          0) ? (ABSORBING) : (MULTIGRID);
         grid[g].bz1 = (g == par->g - 1) ? (  SURFACE) : (MULTIGRID);
         grid[g].bx1 = ABSORBING;
         grid[g].bx2 = ABSORBING;
         if (grid[g].d == 3) {
            grid[g].by1 = ABSORBING;
            grid[g].by2 = ABSORBING;
            }
         }
      if (par->boundary == ABSORBING) {
         grid[g].bz2 = (g ==          0) ? (ABSORBING) : (MULTIGRID);
         grid[g].bz1 = (g == par->g - 1) ? (ABSORBING) : (MULTIGRID);
         grid[g].bx1 = ABSORBING;
         grid[g].bx2 = ABSORBING;
         if (grid[g].d == 3) {
            grid[g].by1 = ABSORBING;
            grid[g].by2 = ABSORBING;
            }
         }
      if (par->boundary == REFLECTING) {
         grid[g].bz2 = (g ==          0) ? (REFLECTING) : (MULTIGRID);
         grid[g].bz1 = (g == par->g - 1) ? (REFLECTING) : (MULTIGRID);
         grid[g].bx1 = REFLECTING;
         grid[g].bx2 = REFLECTING;
         }
      }
   if (par->topo == 1) grid[par->g - 1].bz1 = ABSORBING;

   /* set no source parameters */
   if (par->f == 0) {
      source[0].l    = (par->d == 2) ? (0) : (par->l/2);
      source[0].y    = par->dy*((float) source[0].l);
      source[0].freq = 1.0;
      source[0].amp  = 0.0;
      }

   /* get 2D approximations */
   if (grid->l <= 1) {
      for(g = 0; g < par->g; g++) {
         (grid + g)->l   = 1;
         (grid + g)->fnt = 0;
         (grid + g)->bck = 0;
         (grid + g)->dy  = (grid + g)->dx;
         }
      visual->plane = 0;
      visual->w     = 0.;
      visual->d = XZPLANE; 
      for(i = 0; i < par->s; i++) { sac[i].l     = 0; sac[i].y  = 0.0;}
      /* i was not defined ??? (no for loop) */
      for(i = 0; i < par->p; i++) {
         if (image[i].mode == WAVEMV || image[i].mode == WAVEEN) {
            if (image[i].w < 0.) {
               image[i].plane = -1;
               image[i].w     = -par->dz;
               }
            image[i].d     = XZPLANE;
            if (image[i].movie <= 0) image[i].movie = 1;
            image[i].t = 0;
            }
         else {
            for(i = 0; i < par->p; i++) image[i].plane = 0;
            for(i = 0; i < par->p; i++) image[i].w     = 0.;
            for(i = 0; i < par->p; i++) image[i].d     = XZPLANE;
            }
         if (image[i].mode == VOLUMEP || image[i].mode == VOLUMES) {
            image[i].mode  = 0;
            image[i].t     = 0;
            image[i].movie = 0;
            }
         }
      for(i = 0; i < par->f; i++) {
         source[i].l =  0;
         source[i].y = 0.;
         source[i].length = 0.;
         source[i].s0     = 0.;
         }
      for(i = 0; i < par->v; i++) {medium[i].l1 = 0; medium[i].l2 = 0;}
      }

   /* get topography */
   if (par->topo) {
      }

   /* get default velocity grid if no velocities specified */
   if (par->v == 0) {
      medium[0].alpha    = ALPHA;
      medium[0].beta     = BETA ;
      medium[0].rho      = RHO  ;
      medium[0].gradient = 0.   ;
      medium[0].Qp       = QP;
      medium[0].Qs       = QS;
      medium[0].Qw       = QW*QF;
      if (par->f > 0)   medium[0].Qw = QF*source[0].freq;
      medium[0].x1 = 0.; medium[0].x2 = par->x;
      medium[0].y1 = 0.; medium[0].y2 = par->x;
      medium[0].z1 = 0.; medium[0].z2 = par->x;
      medium[i].type     = BLOCK;
      medium[i].file[0]  = (char) '\0';
      par->v++;
      fprintf(stderr, "Warning: no velocity file - default medium used\n");
      }

   /* check velocities and do unit conversions on medium parameters */
   for(i = 0; i < par->v; i++) {
/*
      g = medium[i].g;
*/
      t = medium[i].type;
      if (medium[i].alpha <= 0. && t != BFILL && t != VFILE && t != POLY3) {
         fprintf(stderr, "Error: P wave velocity <= 0.\n");
         exit(-1);
         }
      if (medium[i].beta  <  0. && t != BFILL && t != VFILE && t != POLY3) {
         fprintf(stderr, "Error: S wave velocity <  0.\n");
         exit(-1);
         }
      if (medium[i].rho   <  0. && t != BFILL && t != VFILE && t != POLY3) {
         fprintf(stderr, "Error: density <  0.\n");
         exit(-1);
         }
      medium[i].alpha *= KMTOM;
      medium[i].beta  *= KMTOM;
      medium[i].rho   *= GM3TOKGM3;
      if (medium[i].p  >= 3) medium[i].type = POLYN;
      /*
      if (medium[i].file[0]) medium[i].type = VFILE;
      */
      if (medium[i].type != BLOCK && medium[i].type != POLYN &&
          medium[i].type != VFILE && medium[i].type != BFILL &&
          medium[i].type != POLY3)  
            medium[i].type = EMPTY;
      if (medium[i].c != 0 && medium[i].c != 1) {
         fprintf(stderr, "Error: bad grid compression format\n");
         exit(-1);
         }
      }

   /* get attenuation parameters */
   if (par->q) {
      for(i = 0; i < par->v; i++) {
         t = medium[i].type;
         if (t != VFILE && t != BFILL && t != POLY3) {
            if (medium[i].Qs < 0.) medium[i].Qs = medium[i].Qp;
            if (medium[i].Qp < 0.) medium[i].Qp = QP;
            if (medium[i].Qs < 0.) medium[i].Qs = QS;
            if (medium[i].Qw < 0.) 
               medium[i].Qw = (par->f > 0) ? (QF*source[0].freq) : (QF*QW);
            /*   
            medium[i].Qw *= 2.*PI;
            */
            }
         /*
         */
         medium[i].Qw *= 2.*PI;
         }
      }
   if (!par->q) {
      for(i = 0; i < par->v; i++) {
         t = medium[i].type;
         if (t == VFILE || t == BFILL || t == POLY3) {
            medium[i].Qs = -1.;
            medium[i].Qp = -1.;
            medium[i].Qw = -1.;
            medium[i].Qs = -1.;
            }
         }
      }

   /* get runtime visual window (default source should be called first) */
   if (visual->movie <= 0 || visual->movie > par->t) 
       visual->movie  = par->t + 1;
   if (visual->d != XZPLANE && visual->d != XYPLANE && visual->d != YZPLANE)
       visual->d = XZPLANE; 
   if (visual->sample < 1) visual->sample = 1;
   if (par->d == 3 && visual->movie <= par->t) {
      if (visual->d == XZPLANE && (visual->w < 0. || visual->w > par->y)) {
         visual->w = par->y/2.;
         if (par->f == 1 && (source[0].y >= 0. && source[0].y < par->y))
            visual->w = source[0].y;
         visual->plane = (int) (visual->w/par->dy);
         }
      if (visual->d == YZPLANE && (visual->w < 0. || visual->w > par->x)) {
         visual->w = par->x/2.;
         if (par->f == 1 && (source[0].x >= 0. && source[0].x < par->x))
            visual->w = source[0].x;
         visual->plane = (int) (visual->w/par->dx);
         }
      if (visual->d == XYPLANE && (visual->w < 0. || visual->w > par->z)) {
         visual->w = par->z/2.;
         if (par->f == 1 && (source[0].z >= 0. && source[0].z < par->z))
            visual->w = source[0].z;
         visual->plane = (int) (visual->w/par->dz);
         }
      }


   /* check image parameters */
   for(i = 0; i < par->p; i++) {
      if (image[i].mode == WAVEVX  || image[i].mode == WAVEVY  || 
          image[i].mode == WAVEVZ  || image[i].mode == WAVETXX ||
          image[i].mode == WAVETYY || image[i].mode == WAVETZZ ||
          image[i].mode == WAVETXY || image[i].mode == WAVETXZ ||
          image[i].mode == WAVETYZ || image[i].mode == WAVEFP  ||
          image[i].mode == WAVEFS) {
         if (image[i].t > 0) image[i].t = 0;
         if (image[i].movie <= 0 || image[i].sample <= 0) {
            fprintf(stderr, "Error: need image movie and sample\n");
            exit(-1);
            }
         }
      if (image[i].t <= 0 || image[i].t > par->t) image[i].t = 0;
      if (image[i].t) image[i].movie = 0;
      if (image[i].sample < 1) image[i].sample = 1;
      if (image[i].d != XZPLANE && image[i].d != XYPLANE && 
          image[i].d != YZPLANE) image[i].d = XZPLANE;
      if (image[i].mode == WAVEMV || image[i].mode == WAVEEN) {
         if (par->d == 2) {
            g = par->g - 1;
            image[i].p = (grid[g].n - 1)/image[i].sample + 1;
            }
         if (par->d == 3) {
            g = par->g - 1;
            image[i].p  =  (grid[g].n - 1)/image[i].sample + 1 ;
            image[i].p *= ((grid[g].l - 1)/image[i].sample + 1);
            }
         }
      if (par->d == 3 && !image[i].p) {
         if (image[i].d==XZPLANE && (image[i].w < 0. || image[i].w > par->y)) {
            image[i].w = par->y/2.;
            if (par->f == 1 && (source[0].y >= 0. && source[0].y < par->y))
               image[i].w = source[0].y;
            image[i].plane = (int) (image[i].w/par->dy);
            }
         if (image[i].d==YZPLANE && (image[i].w < 0. || image[i].w > par->x)) {
            image[i].w = par->x/2.;
            if (par->f == 1 && (source[0].x >= 0. && source[0].x < par->x))
               image[i].w = source[0].x;
            image[i].plane = (int) (image[i].w/par->dx);
            }
         if (image[i].d==XYPLANE && (image[i].w < 0. || image[i].w > par->z)) {
            image[i].w = par->z/2.;
            if (par->f == 1 && (source[0].z >= 0. && source[0].z < par->z))
               image[i].w = source[0].z;
            image[i].plane = (int) (image[i].w/par->dz);
            }
         }
      if (par->d == 3 && image[i].p) {
         image[i].d == XZPLANE;
         if (image[i].movie <= 0) image[i].movie = 1;
         image[i].t = 0;
         if (image[i].w < 0.) {
            image[i].plane = -1;
            image[i].w     = -par->dz;
            }
         if (image[i].w > par->z) {
            fprintf(stderr, "Error: bad field image plane coordinate\n");
            exit(-1);
            }
         }
      }

   /* get and check seismogram parameters */
   for(i = 0; i < par->s; i++) {
      if (sac[i].x > par->x) sac[i].x = -par->dx;
      if (sac[i].y > par->y) sac[i].y = -par->dy;
      if (sac[i].z > par->z) sac[i].z = -par->dz;
      /*
      g = sac[i].g;
      if (sac[i].l >= grid[g].l) sac[i].l = -1;
      if (sac[i].m >= grid[g].m) sac[i].m = -1;
      if (sac[i].n >= grid[g].n) sac[i].n = -1;
      */
      if (sac[i].sample <=  0) sac[i].sample = 1;
      if (sac[i].mode   < 1 || sac[i].mode > 2) sac[i].mode = 1;
      sac[i].t  = (par->t + sac[i].sample - 1)/sac[i].sample;
      sac[i].dt = par->dt*sac[i].sample;
      }

   /* check damping (need to do for variable density grid) */
   if (par->damp < 0 || par->damp > (MIN(par->m, par->n))/2.) {
      fprintf(stderr, "ERROR: bad damping 1: %d %d %d\n", 
                       par->damp, par->m, par->n);
      exit(-1);
      }
   if (grid->l > 1) 
      if (par->damp > (MIN3(par->m, par->n, par->l))) {
         fprintf(stderr, "ERROR: bad damping 2: %d %d %d %d\n",
                          par->damp, par->l, par->m, par->n);
         exit(-1);
         }
   if (par->damp && par->adamp <= 0.) 
      par->adamp = 1. - 1./(2*((float) (par->damp)));
   if (par->adamp > 1.) par->damp = 0;

   /* get and check source parameters */
   for(i = 0; i < par->f; i++) source[i].t = par->t;
   for(i = 0; i < par->f; i++) {
      if (source[i].type == BSOURCE) {
         if (source[i].length<0.||source[i].width<0.||source[i].depth<0.) {
            fprintf(stderr, "Error: bad fault source parameters\n");
            exit(-1);
            }
         }
      if (source[i].type == CSOURCE || source[i].type == ESOURCE) {
         if (source[i].p < 0) {
            fprintf(stderr, "Error: need number of points in dsource file\n");
            exit(-1);
            }
         if (source[i].dfile[0] == (char) '\0') {
            fprintf(stderr, "Error: need to specify dsource file\n");
            exit(-1);
            }
         }
      }
   for(i = 0; i < par->f; i++) {
      /*
      if (source[i].type == BSOURCE && par->active > 0 && 
          source[i].v > par->maxP) {
         fprintf(stderr,"Error: fault rupture velocity too large for ");
         fprintf(stderr,"active grid\n");
         exit(-1);
         }
      */
      if (source[i].type == BSOURCE && par->active > 0 && par->node == 0) {
         fprintf(stderr, "Warning: active grid used with fault source\n");
         }
      if (source[i].type == CSOURCE && par->active > 0 && par->node == 0) {
         fprintf(stderr, "WARNING: active GRID used WITH distributed SOURCE\n");
         fprintf(stderr, "         make sure hypocenter specified (x,y,z)\n");
         }
      if (source[i].type == ESOURCE && par->active > 0 && par->node == 0) {
         fprintf(stderr, "WARNING: active GRID used WITH distributed SOURCE\n");
         fprintf(stderr, "         make sure hypocenter specified (x,y,z)\n");
         }
      if (source[i].type == BSOURCE) {
         source[i].z = 0.;
         if (source[i].d0 < 0.) source[i].d0 = 0.;
         if (source[i].d0 > source[i].width) source[i].d0 = source[i].width;
/*
         if (source[i].d0 < 0. || source[i].d0 > source[i].width) 
            source[i].d0 = source[i].width/2.;
*/
         if (source[i].s0 >  source[i].length/2.) 
            source[i].s0 = source[i].length/2.;
         if (source[i].s0 < -source[i].length/2.) 
            source[i].s0 = -source[i].length/2.;
         }
      }

   /* check traces parameters */
   for(i = 0; i < par->a; i++) {
      if (par->d == 2) mode = TRACEVX           + TRACEVZ + TRACEP + TRACES ;
      if (par->d == 3) mode = TRACEVX + TRACEVY + TRACEVZ + TRACEP + TRACES ;
      if (traces[i].sample <      1) traces[i].sample =      1;
      if (traces[i].sample > par->t) traces[i].sample = par->t;
      if (traces[i].mode   <      0) traces[i].mode   =      0;
      if (traces[i].mode   >   mode) traces[i].mode   =   mode;
      if (par->d == 2) if ((unsigned int) traces[i].mode & TRACEVY)
         traces[i].mode -= TRACEVY;
      if (traces[i].tfile[0] == (char) '\0' && traces[i].p < 2) {
         fprintf(stderr, "Error: need to define number of traces or tfile\n");
         exit(1);
         }
      if (traces[i].p == 1 || traces[i].p < 0) {
         fprintf(stderr, "Error: number of traces can\'t equal 1\n");
         exit(1);
         }
      }

   /* get attenuation parameters */
   for(i = 0; i < par->g; i++) grid[i].q = par->q;

   /* check join and active grids */
   if (par->join && par->active) {
      fprintf(stderr, "Error: active grid must be 0 with join\n");
      exit(-1);
      }
 
   /* change units */
   for(i = 0; i < par->f; i++) if (source[i].type != RSOURCE) 
      source[i].amp  *= DYNECMTONM;
   if (par->active == ACTIVEREDUCING) par->vred *= KMTOM;

   /* get parallelization (NON VDENSITY) */
   par->node  = mpNode();
   par->nnode = mpN   ();
   if (grid->l <= 1) par->ny = 1;
   if (par->nx*par->ny*par->nz != mpN()) {
      fprintf(stderr, "Error: bad parallel nodes\n");
      exit(-1);
      }
   for(g = 0; g < par->g; g++) mpBoundary(par, grid + g);
   for(g = 0; g < par->g; g++) {
      if (grid[g].px1 >= 0) grid[g].bx1 = MESSAGEPASS;
      if (grid[g].px2 >= 0) grid[g].bx2 = MESSAGEPASS;
      if (grid[g].py1 >= 0) grid[g].by1 = MESSAGEPASS;
      if (grid[g].py2 >= 0) grid[g].by2 = MESSAGEPASS;
      if (grid[g].pz1 >= 0) grid[g].bz1 = MESSAGEPASS;
      if (grid[g].pz2 >= 0) grid[g].bz2 = MESSAGEPASS;
      }
   /* NON VDENSITY (par-nnode check) */
/* done above
   for(g = 0; g < par->g; g++) if (par->nnode > 1) mpDecomposition(par, grid + g);
*/

}




getDefault(par, source, grid, visual, sac)
Parameter   *par;
FdSource    *source;
FdGrid      *grid;
Image       *visual;
Seismogram  *sac;
{
   int g;
   par->model         =           1;
   par->water         =           0;
   par->active        =           1;
   par->boundary      =     SURFACE;
   par->ave           =           0;
   par->q             =           0;
   par->topo          =           0;
   par->damp          =           0;
   par->adamp         =         0.0;
   par->nx            =           1;
   par->ny            =           1;
   par->nz            =           1;
   par->x0            =           0;
   par->y0            =           0;
   par->z0            =           0;
   par->join          =           0;
   par->xjoin         =           3;
   par->tjoin         =           0;
   par->nolast        =           0;
   visual->movie      =           0;
   for(g = 0; g < MAXGRID; g++) {
      grid[g].l   = 1;
      grid[g].top = 2;
      grid[g].bot = 3;
      grid[g].lhs = 3;
      grid[g].rhs = 2;
      grid[g].fnt = 3;
      grid[g].bck = 3;
      grid[g].ave = 0;
      grid[g].minP  =  INFINITY;
      grid[g].maxP  = -INFINITY;
      grid[g].minS  =  INFINITY;
      grid[g].maxS  = -INFINITY;
      grid[g].minR  =  INFINITY;
      grid[g].maxR  = -INFINITY;
      grid[g].minQp =  INFINITY;
      grid[g].maxQp = -INFINITY;
      grid[g].minQs =  INFINITY;
      grid[g].maxQs = -INFINITY;
      grid[g].minQw =  INFINITY;
      grid[g].maxQw = -INFINITY;
      }
   par->minP  =  INFINITY;
   par->maxP  = -INFINITY;
   par->minS  =  INFINITY;
   par->maxS  = -INFINITY;
   par->minR  =  INFINITY;
   par->maxR  = -INFINITY;
   par->minQp =  INFINITY;
   par->maxQp = -INFINITY;
   par->minQs =  INFINITY;
   par->maxQs = -INFINITY;
   par->minQw =  INFINITY;
   par->maxQw = -INFINITY;
}



getGridValid(par, grid, mg)
Parameter   *par;
FdGrid      *grid;
int         *mg;
{
   int g, error = 0, merror = 0;

   /* determine if valid x coordinates */
   if (par->x >  0. && (par->n <= 1 && par->dx <= 0.)) error += 1; 
   if (par->x <= 0. && (par->n <= 1 || par->dx <= 0.)) error += 1;

   /* determine if valid y coordinates */
   if (par->y >  0. && (par->l <= 1 && par->dy <= 0.)) error += 2; 
   if (par->y <= 0. && (par->l >  1 && par->dy <= 0.)) error += 2;

   /* determine if valid z coordinates */
   if (par->z >  0. && (par->m <= 1 && par->dz <= 0.)) error += 4; 
   if (par->z <= 0. && (par->m <= 1 || par->dz <= 0.)) error += 4;

   /* determine if multigrid is valid */
   if (mg[0] < 0 || mg[0] == 1) merror++;
   for(g = 1; g < par->g; g++) if (mg[g] <= 1) merror ++;
   if (merror) error += 8;

   /* error condition */
   if (error) {
      if (error & 1) fprintf(stderr, "ERROR: bad grid values (x)  \n");
      if (error & 2) fprintf(stderr, "ERROR: bad grid values (y)  \n");
      if (error & 4) fprintf(stderr, "ERROR: bad grid values (z)  \n");
      if (error & 8) fprintf(stderr, "ERROR: bad multi grid values\n");
      exit(-1);
      }

}



getGridAdjust(par, grid, mg)
Parameter   *par;
FdGrid      *grid;
int         *mg;
{
   int g, extra;

   /* reset y values for 2D case */
   if (par->y <= 0. && par->l <= 1) {
      par->y  = 0.;
      par->l  = 1 ;
      par->dy = par->dx; 
      }

   /* determine m for multi grid case (mg[0] is top of grid) */
   if (par->g > 1) for(g = 1; g < par->g; g++) par->m += (mg[g] << g); 

   /* determine x coordinates of full grid (precedence: x, dx, n) */
   if (par->x <= 0.) par->x = par->dx*((float) (par->n - 1));
   else {
      if (par->dx <= 0.) par->dx = par->x/((float) (par->n - 1)); 
      if (par->dx >  0.) {
         par->n  = par->x/par->dx;
         par->n += (par->n*par->dx == par->x) ? (1) : (2);
         par->x  = par->dx*((float) (par->n - 1));
         }
      }

   /* determine y coordinates of full grid (precedence: y, dy, l) */
   if (par->y <= 0.) par->y = par->dy*((float) (par->l - 1));
   else {
      if (par->dy <= 0.) par->dy = par->y/((float) (par->l - 1)); 
      if (par->dy >  0.) {
         par->l  = par->y/par->dy;
         par->l += (par->l*par->dy == par->y) ? (1) : (2);
         par->y  = par->dy*((float) (par->l - 1));
         }
      }

   /* determine z coordinates of full grid (precedence: z, dz, m) */
   if (par->z <= 0.) par->z = par->dz*((float) (par->m - 1));
   else {
      if (par->dz <= 0.) par->dz = par->z/((float) (par->m - 1)); 
      if (par->dz >  0.) {
         par->m  = par->z/par->dz;
         par->m += (par->m*par->dz == par->z) ? (1) : (2);
         par->z  = par->dz*((float) (par->m - 1));
         }
      }

   /* redetermine full grid coordinates if multigrid (mg[0] is top grid) */
   if (par->g > 1) {
      if ((extra = par->n%(1 << (par->g - 1)))) {
         par->n += (1 << (par->g - 1)) - extra;
         par->x  = par->dx*((float) (par->n - 1));
         }
      if ((extra = par->l%(1 << (par->g - 1))) && (par->l > 1)) {
         par->l += (1 << (par->g - 1)) - extra;
         par->y  = par->dy*((float) (par->l - 1));
         }
      }
/*
   if (par->g > 1) {
      if ((extra = par->n%(par->g + 1))) {
         par->n += (1 << par->g) - extra;
         par->x  = par->dx*((float) (par->n - 1));
         }
      if ((extra = par->l%(par->g + 1)) && (par->l > 1)) {
         par->l += (1 << par->g) - extra;
         par->y  = par->dy*((float) (par->l - 1));
         }
      }
*/

   /* reset dy value for 2D case */
   if (par->y <= 0. && par->l <= 1) par->dy = par->dx; 

   /* convert units */
   par->x  *= KMTOM;
   par->y  *= KMTOM;
   par->z  *= KMTOM;
   par->dx *= KMTOM;
   par->dy *= KMTOM;
   par->dz *= KMTOM;

}



getGridDimensions(par, grid, mg)
Parameter   *par;
FdGrid      *grid;
int         *mg;
{
   int g;

   /* initialize */
   for(g = 0; g < MAXGRID; g++) grid[g].m = 0;

   /* go from global grid to top subgrid (often one in the same) */
   g = par->g - 1;
   grid[g].n  = par->n;
   grid[g].l  = par->l;
   grid[g].m  = par->m;
   grid[g].dx = par->dx;
   grid[g].dy = par->dy;
   grid[g].dz = par->dz;

   /* determine multigrid dimensions */
   grid[par->g - 1].m = par->m;
   if (par->g > 1) {
      grid[par->g - 1].m = mg[0];
      fprintf(stderr, "Warning: correct for par->m in code\n");
      }
   for(g = 1; g < par->g; g++) grid[par->g - 1 - g].m = mg[g];

   for(g = par->g - 1 - 1; g >= 0; g--) {
      grid[g].n  = grid[g+1].n/2;
      grid[g].l  = grid[g+1].l/2;
      grid[g].dx = grid[g+1].dx*2.;
      grid[g].dy = grid[g+1].dy*2.;
      grid[g].dz = grid[g+1].dz*2.;
      }

   /* get dimensions of grid */
   par->d = (par->l > 1) ? (3) : (2);
   for(g = 0; g < par->g; g++) (grid + g)->d = par->d;
   if (par->d == 2) for(g = 0; g < par->g; g++) grid[g].l = 1;

   par->x0 = 0.;
   par->y0 = 0.;
   par->z0 = 0.;

   for(g = 0; g < par->g; g++) {
      (grid + g)->n0 = 0;
      (grid + g)->l0 = 0;
      (grid + g)->m0 = 0;
      }
   for(g = par->g - 2; g >= 0; g--) 
      (grid + g)->m0 = grid[g+1].m0 + grid[g+1].m*(1 << (par->g - g - 2));
/*
      (grid + g)->m0 = grid[g+1].m0 + grid[g+1].m*grid[g+1].s;
*/

   for(g = 0; g < par->g; g++) {
      (grid + g)->x0 = 0.;
      (grid + g)->y0 = 0.;
      (grid + g)->z0 = grid[g].m0*par->dz;
      }

   for(g = 0; g < par->g; g++) {
      grid[g].x = ((float) (grid[g].n - 1))*grid[g].dx;
      grid[g].y = ((float) (grid[g].l - 1))*grid[g].dy;
      grid[g].z = ((float) (grid[g].m - 1))*grid[g].dz;
      }

}



getPolygonInput(medium, line)
Medium  *medium;
char     **line;
{
   float x, z;
   int p;

   /* check limits */
   /*
   if (medium->p > MAXPOLY) {
      fprintf(stderr, "Error: too many polygon points (%d)\n", medium->p);
      exit(-1);
      }
   */

   /* initialize */
   line++;
   p = 0;

   /* read each line that contains legitimate values a polygon */
   while((*line) && (p < medium->p)) {
      switch(getLine(*line)) {
         case LINE_VALUE:
            sscanf((*line), "%f %f", &x, &z);
            medium->x[p] = x;
            medium->z[p] = z;
            p++;
            break;
         case LINE_COMMENT:
         case LINE_BLANK:
         case LINE_NULL:
            break;
         default:
            fprintf(stderr, "Error: bad polygon input: %s\n", *(line));
            exit(-1);
            break;
         }
      line++;
      }

   /* test for success */
   if (medium->p != p) {
      fprintf(stderr, "Error: bad polygon input\n");
      exit(-1);
      }

   /* test/correct polygon closure */
   if ((medium->x[0] != medium->x[p-1]) || (medium->z[0] != medium->z[p-1])) {
      /*
      if (p == MAXPOLY) {
         fprintf(stderr, "Error: too many polygon points (%d)\n", p);
         exit(-1);
         }
      */
      medium->x[p] = medium->x[0];
      medium->z[p] = medium->z[0];
      medium->p++;
      }

   /* convert units */
   for(p = 0; p < medium->p; p++) {
      medium->x[p] *= KMTOM;
      medium->z[p] *= KMTOM;
      }

}



getTraceLimits(par, grid, traces)
Parameter    *par;
FdGrid       *grid;
Traces       *traces;
{
   float x, y, z, dx, dy, dz; 
   int i, j, l, m, n, p, insideSubGrid, insideGrid;
   char line[82];
   FILE *file;

   /* initialize values */
   traces->nglobal = 0;
   traces->nlocal  = 0;
   p = traces->p;

   /* determine traces from trace coordinates */
   if (p > 1) {
      dx = (traces->x2 - traces->x1)/((float) (p - 1));
      dy = (traces->y2 - traces->y1)/((float) (p - 1));
      dz = (traces->z2 - traces->z1)/((float) (p - 1));
      for(j = 0; j < p; j++) {
         traces->nglobal++;
         x = traces->x1 + dx*((float) j);
         y = traces->y1 + dy*((float) j);
         z = traces->z1 + dz*((float) j);
         insideGrid = 0;
         for(i = 0; i < par->g; i++) {
            insideSubGrid = 1;
            l  = (int) floor((y - grid[i].y0)/grid[i].dy);
            m  = (int) floor((z - grid[i].z0)/grid[i].dz);
            n  = (int) floor((x - grid[i].x0)/grid[i].dx);
            if (l < 0 || l >= grid[i].l) insideSubGrid = 0;
            if (m < 0 || m >= grid[i].m) insideSubGrid = 0;
            if (n < 0 || n >= grid[i].n) insideSubGrid = 0;
            if (insideSubGrid) insideGrid = 1;
            }
         if (insideGrid) traces->nlocal++;
         }
      }

   /* open input file */
   if (traces->tfile[0]) {
      file = fopen(traces->tfile, "r");
      if (file == NULL) {
         fprintf(stderr, "Error: Can't open trace file \"%s\"\n", traces->tfile);
         exit(1);
         }
      }

   /* read input file and data lines */
   if (traces->tfile[0]) {
      while(fgets(line, 256, file) != NULL) {
         if (getLine(line) != LINE_VALUE) continue;
         traces->nglobal++;
         sscanf(line, "%f %f %f", &x, &y, &z);
         x *= (KMTOM);
         y *= (KMTOM);
         z *= (KMTOM);
         insideGrid = 0;
         for(i = 0; i < par->g; i++) {
            insideSubGrid = 1;
            l  = (int) floor((y - grid[i].y0)/grid[i].dy);
            m  = (int) floor((z - grid[i].z0)/grid[i].dz);
            n  = (int) floor((x - grid[i].x0)/grid[i].dx);
            if (l < 0 || l >= grid[i].l) insideSubGrid = 0;
            if (m < 0 || m >= grid[i].m) insideSubGrid = 0;
            if (n < 0 || n >= grid[i].n) insideSubGrid = 0;
            if (insideSubGrid) insideGrid = 1;
            }
         if (insideGrid) traces->nlocal++;
         }
      }

   /* close input file */
   if (traces->tfile[0]) fclose(file);

}



getFaultMoment(par, source)
Parameter   *par;
FdSource    *source;
{
   float strike, dip, rake;

   /* initialize */
   strike = source->strike;
   dip    = source->dip;
   rake   = source->rake;

   /* get 2D approximation */
   /*
   if (par->d == 2) {
      strike =   0.;
      rake   =  90.;
      }
   */

   getMoment3D(source);

}



int
getGridFileType(filename, type)
char *filename, *type;
{
   int l;

   /* determine specific types */
   if (!strcmp("p",  type)) return(PFILE);
   if (!strcmp("P",  type)) return(PFILE);
   if (!strcmp("s",  type)) return(SFILE);
   if (!strcmp("S",  type)) return(SFILE);
   if (!strcmp("r",  type)) return(RFILE);
   if (!strcmp("R",  type)) return(RFILE);
   if (!strcmp("Q",  type)) return(QFILE);
   if (!strcmp("Qp", type)) return(YFILE);
   if (!strcmp("Qs", type)) return(ZFILE);
   if (!strcmp("Qf", type)) return(WFILE);

   /* determine type from file extension */
   l = strlen(filename);
   if (l > 1) if (!strcmp(".p",  filename+l-2)) return(PFILE);
   if (l > 1) if (!strcmp(".P",  filename+l-2)) return(PFILE);
   if (l > 1) if (!strcmp(".s",  filename+l-2)) return(SFILE);
   if (l > 1) if (!strcmp(".S",  filename+l-2)) return(SFILE);
   if (l > 1) if (!strcmp(".r",  filename+l-2)) return(RFILE);
   if (l > 1) if (!strcmp(".R",  filename+l-2)) return(RFILE);
   if (l > 1) if (!strcmp(".Q",  filename+l-2)) return(QFILE);
   if (l > 2) if (!strcmp(".Qp", filename+l-3)) return(YFILE);
   if (l > 2) if (!strcmp(".Qs", filename+l-3)) return(ZFILE);
   if (l > 2) if (!strcmp(".Qf", filename+l-3)) return(WFILE);

   /* can't determine file type */
   fprintf(stderr, "Error: can't determine type of file \"%s\"\n", filename);
   exit(-1);
   
}
