/******************************************************************************
   interpolate: interpolate values on grid (node to real values)

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   l         : grid node point in y direction
   m         : grid node point in z direction
   n         : grid node point in x direction
   x         : fractional nodal distance in x direction
   y         : fractional nodal distance in y direction
   z         : fractional nodal distance in z direction

   compile: cc -g -c interpolate.c -lm

   NOTE: the P and S potentials are still computed at the (normal stress) node

   author: shawn larsen
   date  : september 1990
   where : caltech
*****************************************************************************/

#include "elas.h"


float
intVx2D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, V, V1, V2;
   Vx = data->Vx;
   V1 = (1. - fx)*Vx[0][m  ][n  ] + fx*Vx[0][m  ][n+1];
   V2 = (1. - fx)*Vx[0][m+1][n  ] + fx*Vx[0][m+1][n+1];
   V  = (1. - fz)*V1              + fz*V2             ;
   return(V);
}


float
intVz2D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vz, V, V1, V2;
   Vz = data->Vz;
   fx -= 0.5; fz -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   if (fz < 0.0) {fz += 1.0; m--;}
   V1 = (1. - fx)*Vz[0][m  ][n  ] + fx*Vz[0][m  ][n+1];
   V2 = (1. - fx)*Vz[0][m+1][n  ] + fx*Vz[0][m+1][n+1];
   V  = (1. - fz)*V1              + fz*V2             ;
   return(V);
}


float
intP_2D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, ***Vz, getP2D(), P, P1, P2, P3, P4, dx, dz;
   Vx = data->Vx;
   Vz = data->Vz;
   dx = grid->dx;
   dz = grid->dz;
   /*
   fx -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   P1 = (1. - fx)*getP2D(Vx[0], Vz[0], dx, dz, m  , n  );
   P2 = (     fx)*getP2D(Vx[0], Vz[0], dx, dz, m  , n+1);
   P3 = (1. - fx)*getP2D(Vx[0], Vz[0], dx, dz, m+1, n  );
   P4 = (     fx)*getP2D(Vx[0], Vz[0], dx, dz, m+1, n+1);
   P  = (1. - fz)*(P1 + P2) + fz*(P3 + P4);
   return(P);
   */
   /*HACK*/
   if (fx > 0.9998) n++;
   if (fz > 0.9998) m++;
   /*END HACK*/
   return(getP2D(Vx[0], Vz[0], dx, dz, m, n));
}


float
intS_2D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, ***Vz, getS2D(), S, S1, S2, S3, S4, dx, dz;
   Vx = data->Vx;
   Vz = data->Vz;
   dx = grid->dx;
   dz = grid->dz;
   /*
   fx -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   S1 = (1. - fx)*getS2D(Vx[0], Vz[0], dx, dz, m  , n  );
   S2 = (     fx)*getS2D(Vx[0], Vz[0], dx, dz, m  , n+1);
   S3 = (1. - fx)*getS2D(Vx[0], Vz[0], dx, dz, m+1, n  );
   S4 = (     fx)*getS2D(Vx[0], Vz[0], dx, dz, m+1, n+1);
   S  = (1. - fz)*(S1 + S2) + fz*(S3 + S4);
   return(S);
   */
   /*HACK*/
   if (fx > 0.9998) n++;
   if (fz > 0.9998) m++;
   /*END HACK*/
   return(getS2D(Vx[0], Vz[0], dx, dz, m, n));
}


float
intVx3D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, V, V1, V2, V3, V4, F1, F2;
   Vx = data->Vx;
   V1 = (1. - fx)*Vx[l  ][m  ][n  ] + fx*Vx[l  ][m  ][n+1];
   V2 = (1. - fx)*Vx[l  ][m+1][n  ] + fx*Vx[l  ][m+1][n+1];
   V3 = (1. - fx)*Vx[l+1][m  ][n  ] + fx*Vx[l+1][m  ][n+1];
   V4 = (1. - fx)*Vx[l+1][m+1][n  ] + fx*Vx[l+1][m+1][n+1];
   F1 = (1. - fz)*V1                + fz*V2               ;
   F2 = (1. - fz)*V3                + fz*V4               ;
   V  = (1. - fy)*F1                + fy*F2               ;
   return(V);
}


float
intVy3D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vy, V, V1, V2, V3, V4, F1, F2;
   Vy = data->Vy;
   fx -= 0.5; fy -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   if (fy < 0.0) {fy += 1.0; l--;}
   V1 = (1. - fx)*Vy[l  ][m  ][n  ] + fx*Vy[l  ][m  ][n+1];
   V2 = (1. - fx)*Vy[l  ][m+1][n  ] + fx*Vy[l  ][m+1][n+1];
   V3 = (1. - fx)*Vy[l+1][m  ][n  ] + fx*Vy[l+1][m  ][n+1];
   V4 = (1. - fx)*Vy[l+1][m+1][n  ] + fx*Vy[l+1][m+1][n+1];
   F1 = (1. - fz)*V1                + fz*V2               ;
   F2 = (1. - fz)*V3                + fz*V4               ;
   V  = (1. - fy)*F1                + fy*F2               ;
   return(V);
}


float
intVz3D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vz, V, V1, V2, V3, V4, F1, F2;
   Vz = data->Vz;
   fx -= 0.5; fz -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   if (fz < 0.0) {fz += 1.0; m--;}
   V1 = (1. - fx)*Vz[l  ][m  ][n  ] + fx*Vz[l  ][m  ][n+1];
   V2 = (1. - fx)*Vz[l  ][m+1][n  ] + fx*Vz[l  ][m+1][n+1];
   V3 = (1. - fx)*Vz[l+1][m  ][n  ] + fx*Vz[l+1][m  ][n+1];
   V4 = (1. - fx)*Vz[l+1][m+1][n  ] + fx*Vz[l+1][m+1][n+1];
   F1 = (1. - fz)*V1                + fz*V2               ;
   F2 = (1. - fz)*V3                + fz*V4               ;
   V  = (1. - fy)*F1                + fy*F2               ;
   return(V);
}


float
intP_3D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, ***Vy, ***Vz, getP3D(), dx, dy, dz;
   float P, P1, P2, P3, P4, P5, P6, P7, P8, F1, F2;
   Vx = data->Vx;
   Vy = data->Vy;
   Vz = data->Vz;
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;
   /*
   fx -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   P1 = (1. - fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l  , m  , n  );
   P2 = (     fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l  , m  , n+1);
   P3 = (1. - fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l  , m+1, n  );
   P4 = (     fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l  , m+1, n+1);
   P5 = (1. - fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l+1, m  , n  );
   P6 = (     fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l+1, m  , n+1);
   P7 = (1. - fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l+1, m+1, n  );
   P8 = (     fx)*getP3D(Vx, Vy, Vz, dx, dy, dz, l+1, m+1, n+1);
   F1 = (1. - fz)*(P1 + P2) + fz*(P3 + P4);
   F2 = (1. - fz)*(P5 + P6) + fz*(P7 + P8);
   P  = (1. - fy)*(F1     ) + fy*(F2     );
   return(P);
   */
   /*HACK*/
   if (fx > 0.9998) n++;
   if (fy > 0.9998) l++;
   if (fz > 0.9998) m++;
   /*END HACK*/
   return(getP3D(Vx, Vy, Vz, dx, dy, dz, l, m, n));
}


float
intS_3D(par, grid, data, l, m, n, fx, fy, fz)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
int l, m, n;
float fx, fy, fz;
{
   float ***Vx, ***Vy, ***Vz, getS3D(), dx, dy, dz;
   float S, S1, S2, S3, S4, S5, S6, S7, S8, F1, F2;
   Vx = data->Vx;
   Vy = data->Vy;
   Vz = data->Vz;
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;
   /*
   fx -= 0.5;
   if (fx < 0.0) {fx += 1.0; n--;}
   S1 = (1. - fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l  , m  , n  );
   S2 = (     fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l  , m  , n+1);
   S3 = (1. - fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l  , m+1, n  );
   S4 = (     fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l  , m+1, n+1);
   S5 = (1. - fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l+1, m  , n  );
   S6 = (     fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l+1, m  , n+1);
   S7 = (1. - fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l+1, m+1, n  );
   S8 = (     fx)*getS3D(Vx, Vy, Vz, dx, dy, dz, l+1, m+1, n+1);
   F1 = (1. - fz)*(S1 + S2) + fz*(S3 + S4);
   F2 = (1. - fz)*(S5 + S6) + fz*(S7 + S8);
   S  = (1. - fy)*(F1     ) + fy*(F2     );
   return(S);
   */
   /*HACK*/
   if (fx > 0.9998) n++;
   if (fy > 0.9998) l++;
   if (fz > 0.9998) m++;
   /*END HACK*/
   return(getS3D(Vx, Vy, Vz, dx, dy, dz, l, m, n));
}

