/******************************************************************************
   memory: routines for memory allocation for elastic 2D/3D finite-difference 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   getMemory  : driving function for memory allocation
   freeMemory : driving function for memory allocation
   getDataMem : allocate memory for data (velocity/stress arrays)
   getGridMem : allocate memory for finite difference grid
   getSacpMem : allocate memory for sac seismograms
   getSrctMem : allocate memory for source parameters
   getTopoMem : allocate memory for topography parameters
   getImagMem : allocate memory for field variable images
   getActvMem : allocate memory for active grid parameters
   freeDataMem: free memory for data (velocity/stress arrays)
   freeGridMem: free memory for finite difference grid
   freeSacpMem: free memory for sac seismograms
   freeSrctMem: free memory for source parameters
   freeTopoMem: free memory for topography parameters
   freeImagMem: free memory for field variable images
   freeActvMem: free memory for active grid parameters

   compile: cc -g -c memory.c

   note: used  units are mks (m, s, m/s, kg/m3)
   note: allocate big array and then self allocation of pointers
   note: need to change initGrid if memory allocation reverts to traditional

   author: shawn larsen
   date  : december 1990
   where : caltech
*****************************************************************************/

#include "elas.h"

int         *getArrayI1D();
float       *getArrayF1D();
float      **getArrayF2D();
float     ***getArrayF3D();
float    ****getArrayF4D();
double     **getArrayD2D();
void        freeArrayI1D();
void        freeArrayF1D();
void        freeArrayF2D();
void        freeArrayF3D();
void        freeArrayF4D();
void        freeArrayD2D();

getMemory(par, source, grid, data, image, sac, traces)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
Image          *image;
Seismogram     *sac;
Traces         *traces;
{
   int i, g;
   for(g = 0; g < par->g; g++) 
      getDataMem(par, grid + g, data + g);
   for(g = 0; g < par->g; g++) 
      getGridMem(par, grid + g);
   getSacpMem(par, sac);
   getSrctMem(par, source);
   if (par->topo) getTopoMem(par, grid + (par->g - 1));
   for(i = 0; i < par->p; i++) if (image[i].p)
      getImagMem(par, image + i);
   for(i = 0; i < par->a; i++) getTraceMem(traces + i);
   if (par->active == ACTIVEBOX) getActvMem(par);
}


freeMemory(par, source, grid, data, image, sac, traces)
Parameter      *par;
FdSource       *source;
FdGrid         *grid;
VelocityStress *data;
Image          *image;
Seismogram     *sac;
Traces         *traces;
{
   int i, g;
   for(g = 0; g < par->g; g++) freeDataMem(par, grid + g, data + g);
   for(g = 0; g < par->g; g++) freeGridMem(par, grid + g);
   freeSacpMem(par, sac);
   freeSrctMem(par, source);
   if (par->topo) freeTopoMem(par, grid + (par->g - 1));
   for(i = 0; i < par->p; i++) if (image[i].p)
      freeImageMem(par, image + i);
   for(i = 0; i < par->a; i++) freeTraceMem(traces + i);
   if (par->active == ACTIVEBOX) freeActvMem(par);
}



getDataMem(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int variables, l, m, n, lhs, rhs, top, bot, fnt, bck;

   /* initialize */
   lhs = grid->lhs; rhs = grid->rhs;
   top = grid->top; bot = grid->bot;
   fnt = grid->fnt; bck = grid->bck;
   l = grid->l;
   m = grid->m;
   n = grid->n;

   /* allocate memory for data arrays (2D problems will have 3D arrays) */
   data->Vx  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   data->Vz  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (par->model != 2) {
   data->Txx = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   data->Tzz = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   data->Txz = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   }
   if (par->model == 2)
   data->T   = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (grid->d == 3) {
      data->Vy  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      if (par->model != 2) {
      data->Tyy = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      data->Txy = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      data->Tyz = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      }
      }

   /* set temporary equivalence for acoustic case or NULL vectors */
   if (par->model == 2) {
      /*
      data->T = data->Txx;
      */
      data->Txx = NULL; data->Tyy = NULL; data->Tzz = NULL;
      data->Txy = NULL; data->Txz = NULL; data->Tyz = NULL;
      }

   /* allocate memory for attenuation reference variables (2D and 3D) */
   if (par->q) {
      data->q1  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      data->q2  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      data->q3  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      if (grid->d == 3) {
         data->q4  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
         data->q5  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
         data->q6  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
         }
      }

}




getGridMem(par, grid)
Parameter      *par;
FdGrid *grid;
{
   int m, n, l, lhs, rhs, top, bot, fnt, bck;

   /* initialize */
   lhs = grid->lhs; rhs = grid->rhs;
   top = grid->top; bot = grid->bot;
   fnt = grid->fnt; bck = grid->bck;
   l = grid->l;
   m = grid->m;
   n = grid->n;

   /* get memory for medium parameters */
   grid->lamda = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (par->model != 2)
   grid->mu    = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   grid->rho   = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
   grid->irho  = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);

   /* get memory for attenuation parameters */
   if (par->q) {
      grid->F0    = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      grid->F1    = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      grid->F2    = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      grid->F3    = getArrayF3D(l, fnt, bck, m, top, bot, n, lhs, rhs);
      }

   /* initialize equivalence arrays */
   grid->alpha = NULL;
   grid->beta  = NULL;
   /*
   grid->rho   = NULL;
   */
   grid->Qp    = NULL;
   grid->Qs    = NULL;
   grid->Qw    = NULL;

}




getSacpMem(par, sac)
Parameter  *par;
Seismogram *sac;
{
   int m, t;

   /* get seismogram memory */
   for(m = 0; m < par->s; m++) {
      if (sac[m].mode == SACS) {
         sac[m].Vx = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Vy = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Vz = getArrayF1D(sac[m].t, 0, 0);
         sac[m].P  = getArrayF1D(sac[m].t, 0, 0);
         sac[m].S  = getArrayF1D(sac[m].t, 0, 0);
         }
      if (sac[m].mode == SACM) {
         sac[m].Mxx = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Myy = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Mzz = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Mxy = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Mxz = getArrayF1D(sac[m].t, 0, 0);
         sac[m].Myz = getArrayF1D(sac[m].t, 0, 0);
         }
      }

}




getSrctMem(par, source)
Parameter *par;
FdSource  *source;
{
   int i, p;
   for(i = 0; i < par->f; i++) {
      source[i].f = getArrayF1D(source[i].t, 0, 0);
      if (source[i].type == CSOURCE || source[i].type == ESOURCE) {
         p = source[i].p;
         source[i].dfl = getArrayI1D(p, 0, 0);
         source[i].dfm = getArrayI1D(p, 0, 0);
         source[i].dfn = getArrayI1D(p, 0, 0);
         source[i].dft = getArrayI1D(p, 0, 0);
         source[i].dfa = getArrayF1D(p, 0, 0);
         if (source[i].type == ESOURCE) 
         source[i].dfr = getArrayF1D(p, 0, 0);
         }
      }
}




getTopoMem(par, grid)
Parameter      *par;
FdGrid *grid;
{
   int l, n;
   l = grid->l + 1;
   n = grid->n + 1;
   if (par->d == 2) grid->zi  = getArrayF1D(         n, 0, 0);
   if (par->d == 3) grid->ziM = getArrayF2D(l, 0, 0, n, 0, 0);
}




getImagMem(par, image)
Parameter      *par;
Image          *image;
{
   image->v = getArrayF1D(image->p, 0, 0);
}




getTraceMem(traces)
Traces *traces;
{
   int n;
   n = traces->nlocal;
   if (n > 0) {
      traces->x      = getArrayF1D(n, 0, 0);
      traces->y      = getArrayF1D(n, 0, 0);
      traces->z      = getArrayF1D(n, 0, 0);
      traces->order  = getArrayI1D(n, 0, 0);
      traces->l      = getArrayI1D(n, 0, 0);
      traces->m      = getArrayI1D(n, 0, 0);
      traces->n      = getArrayI1D(n, 0, 0);
      }
}




getActvMem(par)
Parameter *par;
{
   int t;
   t = par->t;
   par->almin = getArrayI1D(t, 0, 0);
   par->almax = getArrayI1D(t, 0, 0);
   par->ammin = getArrayI1D(t, 0, 0);
   par->ammax = getArrayI1D(t, 0, 0);
   par->anmin = getArrayI1D(t, 0, 0);
   par->anmax = getArrayI1D(t, 0, 0);
}




freeDataMem(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int l, m, n, lhs, rhs, top, bot, fnt, bck;

   /* initialize */
   lhs = grid->lhs; rhs = grid->rhs;
   top = grid->top; bot = grid->bot;
   fnt = grid->fnt; bck = grid->bck;
   l = grid->l;
   m = grid->m;
   n = grid->n;

   /* free memory for data arrays */
   freeArrayF3D(data->Vx , l, fnt, bck, m, top, bot, n, lhs, rhs);
   freeArrayF3D(data->Vz , l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (par->model != 2) {
   freeArrayF3D(data->Txx, l, fnt, bck, m, top, bot, n, lhs, rhs);
   freeArrayF3D(data->Tzz, l, fnt, bck, m, top, bot, n, lhs, rhs);
   freeArrayF3D(data->Txz, l, fnt, bck, m, top, bot, n, lhs, rhs);
   }
   if (par->model == 2) 
   freeArrayF3D(data->T  , l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (grid->d == 3) {
      freeArrayF3D(data->Vy , l, fnt, bck, m, top, bot, n, lhs, rhs);
      if (par->model != 2) {
      freeArrayF3D(data->Tyy, l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(data->Txy, l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(data->Tyz, l, fnt, bck, m, top, bot, n, lhs, rhs);
      }
      }

   /* free memory for attenuation arrays */
   if (par->q) {
      freeArrayF3D(data->q1 , l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(data->q2 , l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(data->q3 , l, fnt, bck, m, top, bot, n, lhs, rhs);
      if (par->d == 3) {
         freeArrayF3D(data->q4 , l, fnt, bck, m, top, bot, n, lhs, rhs);
         freeArrayF3D(data->q5 , l, fnt, bck, m, top, bot, n, lhs, rhs);
         freeArrayF3D(data->q6 , l, fnt, bck, m, top, bot, n, lhs, rhs);
         }
      }

}




freeGridMem(par, grid)
Parameter      *par;
FdGrid *grid;
{
   int l, m, n, lhs, rhs, top, bot, fnt, bck;

   /* initialize */
   lhs = grid->lhs; rhs = grid->rhs;
   top = grid->top; bot = grid->bot;
   fnt = grid->fnt; bck = grid->bck;
   l = grid->l;
   m = grid->m;
   n = grid->n;

   /* free memory for medium parameters */
   freeArrayF3D(grid->lamda, l, fnt, bck, m, top, bot, n, lhs, rhs);
   if (par->model != 2)
   freeArrayF3D(grid->mu   , l, fnt, bck, m, top, bot, n, lhs, rhs);
   freeArrayF3D(grid->rho  , l, fnt, bck, m, top, bot, n, lhs, rhs);
   freeArrayF3D(grid->irho , l, fnt, bck, m, top, bot, n, lhs, rhs);

   /* free memory for attenuation parameters */
   if (par->q) {
      freeArrayF3D(grid->F0, l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(grid->F1, l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(grid->F2, l, fnt, bck, m, top, bot, n, lhs, rhs);
      freeArrayF3D(grid->F3, l, fnt, bck, m, top, bot, n, lhs, rhs);
      }

}




freeSacpMem(par, sac)
Parameter  *par;
Seismogram *sac;
{
   int m, t;

   /* free seismogram memory */
   t = 0;
   for(m = 0; m < par->s; m++) {
      if (sac[m].mode == SACS) {
         freeArrayF1D(sac[m].Vx, t, 0, 0);
         freeArrayF1D(sac[m].Vy, t, 0, 0);
         freeArrayF1D(sac[m].Vz, t, 0, 0);
         freeArrayF1D(sac[m].P , t, 0, 0);
         freeArrayF1D(sac[m].S , t, 0, 0);
         }
      if (sac[m].mode == SACM) {
         freeArrayF1D(sac[m].Mxx, t, 0, 0);
         freeArrayF1D(sac[m].Myy, t, 0, 0);
         freeArrayF1D(sac[m].Mzz, t, 0, 0);
         freeArrayF1D(sac[m].Mxy, t, 0, 0);
         freeArrayF1D(sac[m].Mxz, t, 0, 0);
         freeArrayF1D(sac[m].Myz, t, 0, 0);
         }
      }

}




freeSrctMem(par, source)
Parameter *par;
FdSource  *source;
{
   int i, t;
   t = 0;
   for(i = 0; i < par->f; i++) {
      freeArrayF1D(source[i].f , t, 0, 0);
      if (source[i].type == CSOURCE || source[i].type == ESOURCE) {
         freeArrayI1D(source[i].dfl, t, 0, 0);
         freeArrayI1D(source[i].dfm, t, 0, 0);
         freeArrayI1D(source[i].dfn, t, 0, 0);
         freeArrayI1D(source[i].dft, t, 0, 0);
         freeArrayF1D(source[i].dfa, t, 0, 0);
         if (source[i].type == ESOURCE) freeArrayF1D(source[i].dfr, t, 0, 0);
         }
      }
}




freeTopoMem(par, grid)
Parameter *par;
FdGrid    *grid;
{
   int l, t;
   t = 0;
   l = grid->l + 1;
   if (par->d == 2) freeArrayF1D(grid->zi ,          t, 0, 0);
   if (par->d == 3) freeArrayF2D(grid->ziM, l, 0, 0, t, 0, 0);
}




freeImageMem(par, image)
Parameter *par;
Image     *image;
{
   int t = 0;
   freeArrayF1D(image->v, t, 0, 0);
}




freeTraceMem(traces)
Traces *traces;
{
   int t = 0;
   if (traces->nlocal > 0) {
      freeArrayI1D(traces->order, t, 0, 0);
      freeArrayI1D(traces->l    , t, 0, 0);
      freeArrayI1D(traces->m    , t, 0, 0);
      freeArrayI1D(traces->n    , t, 0, 0);
      freeArrayF1D(traces->x    , t, 0, 0);
      freeArrayF1D(traces->y    , t, 0, 0);
      freeArrayF1D(traces->z    , t, 0, 0);
      }
}




freeActvMem(par)
Parameter      *par;
{
   free(par->almin);
   free(par->almax);
   free(par->ammin);
   free(par->ammax);
   free(par->anmin);
   free(par->anmax);
}




/*****************************************************************************
   memoryv: memory allocation routines for elastic 2D/3D finate-difference

    getArrayI1D: allocate 1D int    array
    getArrayF1D: allocate 1D float  array
    getArrayF2D: allocate 2D float  array
    getArrayF3D: allocate 3D float  array
    getArrayF4D: allocate 4D float  array
    getArrayD2D: allocate 2D double array
   freeArrayI1D: free     1D int    array
   freeArrayF1D: free     1D float  array
   freeArrayF2D: free     2D float  array
   freeArrayF3D: free     3D float  array
   freeArrayF4D: free     4D float  array
   freeArrayD2D: free     2D double array

   n  fastest varying index
   n  length of array not counting boundaries
   n0 boundary on negitive index side (generally 0 or positive number)
   n1 boundary on positive index side (generally 0 or positive number)

   note: routines do their own memory management (to reduce memory cost)

   int       * =  getArrayI1D(n, n0, n1)
   float     * =  getArrayF1D(n, n0, n1)
   float    ** =  getArrayF2D(m, m0, m1, n, n0, n1)
   float   *** =  getArrayF3D(l, l0, l1, m, m0, m1, n, n0, n1)
   float  **** =  getArrayF4D(k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1)
   double   ** =  getArrayD2D(m, m0, m1, n, n0, n1)
   void        = freeArrayI1D(v, n, n0, n1)
   void        = freeArrayF1D(v, n, n0, n1)
   void        = freeArrayF2D(v, m, m0, m1, n, n0, n1)
   void        = freeArrayF3D(v, l, l0, l1, m, m0, m1, n, n0, n1)
   void        = freeArrayF4D(v, k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1)
   void        = freeArrayD2D(v, m, m0, m1, n, n0, n1)

*****************************************************************************/

int *
getArrayI1D(n, n0, n1)
int n, n0, n1;
{
   int *v;
   v = NULL;
   v = (int *) malloc((n + n0 + n1)*sizeof(int));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayI1D\n");
      exit(-1);
      }
   v += n0;
   return(v);
}

float *
getArrayF1D(n, n0, n1)
int n, n0, n1;
{
   float *v;
   v = NULL;
   v = (float *) malloc((n + n0 + n1)*sizeof(float));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF1D\n");
      exit(-1);
      }
   v += n0;
   return(v);
}

float **
getArrayF2D(m, m0, m1, n, n0, n1)
int m, m0, m1, n, n0, n1;
{
   float **v;
   int j;
   v = NULL;
   v = (float **) malloc((m + m0 + m1)*sizeof(float *));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF2D\n");
      exit(-1);
      }
   v += m0;
   for(j = -m0; j < m + m1; j++) {
      v[j] = NULL;
      v[j] = (float *) malloc((n + n0 + n1)*sizeof(float));
      if (v[j] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayF2D\n");
         exit(-1);
         }
      v[j] += n0;
      }
   return(v);
}

float ***
getArrayF3D(l, l0, l1, m, m0, m1, n, n0, n1)
int m, m0, m1, n, n0, n1;
{
   float ***v, *data;
   int j, k, p, pl, pm, pn, extra;
   extra = 1;
   pl = (l + l0 + l1);
   pm = (m + m0 + m1);
   pn = (n + n0 + n1);
   p  = pl*pm*pn; 
   data = NULL;
   data = (float *) malloc(p*sizeof(float));
   if (data == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF3D\n");
      exit(-1);
      }
   v = NULL;
   v = (float ***) malloc((l + l0 + l1 + extra)*sizeof(float **));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF3D\n");
      exit(-1);
      }
   v += l0;
   for(k = -l0; k < l + l1; k++) {
      v[k] = NULL;
      v[k] = (float **) malloc((m + m0 + m1)*sizeof(float *));
      if (v[k] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayF3D\n");
         exit(-1);
         }
      v[k] += m0;
      }
   /*
   for(k = -l0; k < l + l1; k++) for(j = -m0; j < m + m1; j++) {
      v[k][j] = NULL;
      v[k][j] = (float *) malloc((n + n0 + n1)*sizeof(float));
      if (v[k][j] == NULL) {
         fprintf(stderr, "Error: couldn't allocate memory in getArrayF3D\n");
         exit(-1);
         }
      v[k][j] += n0;
      }
   */
   v[l+l1] = NULL;
   v[l+l1] = (float **) malloc(1*sizeof(float *));
   if (v[l+l1] == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF3D\n");
      exit(-1);
      }
   v[l+l1][0] = data;
   for(k = -l0; k < l + l1; k++) for(j = -m0; j < m + m1; j++) {
      v[k][j] = data + (k+l0)*(pm)*(pn) + (j+m0)*(pn) + n0;
      }
   return(v);
}

float ****
getArrayF4D(k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1)
int k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1;
{
   float ****v;
   int h, i, j;
   v = NULL;
   v = (float ****) malloc((k + k0 + k1)*sizeof(float ***));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayF4D\n");
      exit(-1);
      }
   v += k0;
   for(h = -k0; h < k + k1; h++) {
      v[h] = NULL;
      v[h] = (float ***) malloc((l + l0 + l1)*sizeof(float **));
      if (v[h] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayF4D\n");
         exit(-1);
         }
      v[h] += l0;
      }
   for(h = -k0; h < k + k1; h++) for(i = -l0; i < l + l1; i++) { 
      v[h][i] = NULL;
      v[h][i] = (float **) malloc((m + m0 + m1)*sizeof(float *));
      if (v[h][i] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayF4D\n");
         exit(-1);
         }
      v[h][i] += m0;
      }
   for(h=-k0; h<k+k1; h++) for(i=-l0; i<l+l1; i++) for(j=-m0; j<m+m1; j++) { 
      v[h][i][j] = NULL;
      v[h][i][j] = (float *) malloc((n + n0 + n1)*sizeof(float));
      if (v[h][i][j] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayF4D\n");
         exit(-1);
         }
      v[h][i][j] += n0;
      }
   return(v);
}

double **
getArrayD2D(m, m0, m1, n, n0, n1)
int m, m0, m1, n, n0, n1;
{
   double **v;
   int j;
   v = NULL;
   v = (double **) malloc((m + m0 + m1)*sizeof(double *));
   if (v == NULL) {
      fprintf(stderr, "Error: can't allocate memory in getArrayD2D\n");
      exit(-1);
      }
   v += m0;
   for(j = -m0; j < m + m1; j++) {
      v[j] = NULL;
      v[j] = (double *) malloc((n + n0 + n1)*sizeof(double));
      if (v[j] == NULL) {
         fprintf(stderr, "Error: can't allocate memory in getArrayD2D\n");
         exit(-1);
         }
      v[j] += n0;
      }
   return(v);
}

void
freeArrayI1D(v, n, n0, n1)
int *v;
int n, n0, n1;
{
   free(v - n0);
}

void
freeArrayF1D(v, n, n0, n1)
float *v;
int n, n0, n1;
{
   free(v - n0);
}

void
freeArrayF2D(v, m, m0, m1, n, n0, n1)
float **v;
int m, m0, m1, n, n0, n1;
{
   int j;
   for(j = -m0; j < m + m1; j++) free(v[j] - n0);
   free(v - m0);
}

void
freeArrayF3D(v, l, l0, l1, m, m0, m1, n, n0, n1)
float ***v;
int l, l0, l1, m, m0, m1, n, n0, n1;
{
   float *data;
   int i, j, k;
   data = v[l+l1][0];
   /*
   for(k = -l0; k < l + l1; k++) 
      for(j = -m0; j < m + m1; j++) 
         free(v[k][j] - n0);
   */
   for(k = -l0; k < l + l1; k++) free(v[k] - m0);
   free(v - l0);
   free(data);
}

void
freeArrayF4D(v, k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1)
float ****v;
int k, k0, k1, l, l0, l1, m, m0, m1, n, n0, n1;
{
   int h, i, j;
   for(h = -k0; h < k + k1; h++) 
      for(i = -l0; i < l + l1; i++) 
         for(j = -m0; j < m + m1; j++) 
            free(v[h][i][j] - n0);
   for(h = -k0; h < k + k1; h++) 
      for(i = -l0; i < l + l1; i++) 
         free(v[h][i] - m0);
   for(h = -k0; h < k + k1; h++) 
      free(v[h] - l0);
   free(v - k0);
}

void
freeArrayD2D(v, m, m0, m1, n, n0, n1)
double **v;
int m, m0, m1, n, n0, n1;
{
   int j;
   for(j = -m0; j < m + m1; j++) free(v[j] - n0);
   free(v - m0);
}

