/******************************************************************************
   output: output routines for e3d (3D finite difference)

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   compile: cc -g -c output.c -lm -lget

   author: shawn larsen
   date  : january 1989
   where : caltech
*****************************************************************************/

#include "elas.h"
#include "sacutil.h"
#include "rasterfile.h"


/******************************************************************************
   writeOutput: write specified image type if at correct time

   usage: void = writeOutput(par, grid, data, image, t)

   par       : parameters
   grid      : grid  parameters
   data      : wavefield information
   image     : images to write
   t         : time of wavefield
******************************************************************************/
writeOutput(par, grid, data, image, visual, sac, traces, source, t)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
Image          *image;
Image          *visual;
Seismogram     *sac;
Traces         *traces;
FdSource       *source;
int             t;
{
   int g, p, s, a;
   if (par->p) writeImage(par, grid, data, image, t);
   if (t && !(t%visual->movie) && !par->topo) 
      writeVisual(par, grid, data, visual);
   if (t && !(t%visual->movie) &&  par->topo) 
      writeVisual(par, grid, data, visual); /* write topography */
   if (par->s) for(g = 0; g < par->g; g++) for(s = 0; s < par->s; s++) {
      /*
      g = sac[s].g;
      */
      putSeismogram(par, grid + g, data + g, sac + s, t);
      }
   if (par->node == 0) {
      if (t == 0) fprintf(stderr, "Time Step ");
      if (t != 0) fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
      fprintf(stderr, "%6d complete", t + 1);
      /*
      fprintf(stderr, "Time Step %6d complete\n", t + 1);
      */
      fflush(stderr);
      }
   if ((t + 1) == par->t) {
      if (par->node == 0) fprintf(stderr, "\n");
      if (par->s) writeSeismograms(par, grid, sac, source);
      }
   if ((t + 1) == par->t) for(p = 0; p < par->p; p++) if (image[p].p) 
      writeSField(par, grid + par->g - 1, image + p);
   if (t == 0) for(a = 0; a < par->a; a++) writeTracesH(par, traces + a);
   for(a = 0; a < par->a; a++) if (!((t+1)%traces[a].sample))
      writeTraces(par, grid, data, traces + a);
   if (par->dump) writeRestart(par, grid, data, t);
}




/******************************************************************************
   putSeismogram: save node values at seismograms 

   usage: void = putSeismogram(par, grid, data, sac, t)

   par       : parameters
   grid      : grid  parameters
   data      : wavefield information
   sac       : input parameters
   t         : current time-step

   NOTE: the P and S potentials are still computed at the (normal stress) node
   NOTE: the moment tensor output is still computed at the nodes
******************************************************************************/
putSeismogram(par, grid, data, sac, t)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
Seismogram     *sac; 
int t;
{
   float intVx2D(), intVz2D(), intVx3D(), intVy3D(), intVz3D();
   float intP_2D(), intS_2D(), intP_3D(), intS_3D();
   /*
   float getP2D(), getS2D(), getP3D(), getS3D(); 
   */
   float ***Vx, ***Vy, ***Vz, dx, dy, dz, fx, fy, fz;
   int l, m, n, time;

   /* initialize */
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;
   Vx = data->Vx;
   Vy = data->Vy;
   Vz = data->Vz;

   /* determine seismogram location */
   l  = (int) floor((sac->y - grid->y0)/dy); 
   m  = (int) floor((sac->z - grid->z0)/dz);
   n  = (int) floor((sac->x - grid->x0)/dx);
   fy = ((sac->y - grid->y0)/dy - (float) l); 
   fx = ((sac->x - grid->x0)/dx - (float) n);
   fz = ((sac->z - grid->z0)/dz - (float) m);
   if (par->d == 2) {l = 0; fy = 0.;}
   if (l < 0 || l >= grid->l) return(0);
   if (m < 0 || m >= grid->m) return(0);
   if (n < 0 || n >= grid->n) return(0);
   if (t%sac[0].sample)       return(0);

   /* put seismograms */
   time = t/sac[0].sample;
   if (sac[0].mode == SACS) {
      if (par->d == 2) {
         /*
         sac[0].Vx[time] = (Vx[l][m][n] + Vx[l][m][n+1])/2.;
         sac[0].Vz[time] = (Vz[l][m][n] + Vz[l][m-1][n])/2.;
         sac[0].P [time] = getP2D(Vx[l], Vz[l], dx, dz, m, n);
         sac[0].S [time] = getS2D(Vx[l], Vz[l], dx, dz, m, n);
         */
         sac[0].Vx[time] = intVx2D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].Vz[time] = intVz2D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].P [time] = intP_2D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].S [time] = intS_2D(par, grid, data, l, m, n, fx, fy, fz);
         }
      if (par->d == 3) {
         /*
         sac[0].Vx[time] = (Vx[l][m][n] + Vx[l][m][n+1])/2.;
         sac[0].Vy[time] = (Vy[l][m][n] + Vy[l-1][m][n])/2.;
         sac[0].Vz[time] = (Vz[l][m][n] + Vz[l][m-1][n])/2.;
         sac[0].P [time] = getP3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
         sac[0].S [time] = getS3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
         */
         sac[0].Vx[time] = intVx3D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].Vy[time] = intVy3D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].Vz[time] = intVz3D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].P [time] = intP_3D(par, grid, data, l, m, n, fx, fy, fz);
         sac[0].S [time] = intS_3D(par, grid, data, l, m, n, fx, fy, fz);
         }
      }
   else if (sac[0].mode == SACM) {
      if (par->d == 2) {
         sac[0].Mxx[time] = (Vx[l][m][n+1]   - Vx[l  ][m  ][n]);
         sac[0].Mzz[time] = (Vz[l][m][n  ]   - Vz[l  ][m-1][n]);
         sac[0].Mxy[time] = (Vx[l+1][m][n  ] + Vx[l+1][m][n+1] -
                             Vx[l-1][m][n  ] - Vx[l-1][m][n+1] -
                             Vy[l  ][m][n-1] - Vy[l-1][m][n-1] +
                             Vy[l  ][m][n+1] + Vy[l-1][m][n+1])/
                             4.;
         }
      if (par->d == 3) {
         sac[0].Mxx[time] = (Vx[l  ][m][n+1] - Vx[l  ][m  ][n]);
         sac[0].Myy[time] = (Vy[l  ][m][n  ] - Vy[l-1][m  ][n]);
         sac[0].Mzz[time] = (Vz[l  ][m][n  ] - Vz[l  ][m-1][n]);
sac[0].Mxx[time] = data->Txx[l][m][n];
sac[0].Myy[time] = data->Tyy[l][m][n];
sac[0].Mzz[time] = data->Tzz[l][m][n];
         sac[0].Mxy[time] = (Vx[l+1][m][n  ] + Vx[l+1][m][n+1] -
                             Vx[l-1][m][n  ] - Vx[l-1][m][n+1] -
                             Vy[l  ][m][n-1] - Vy[l-1][m][n-1] +
                             Vy[l  ][m][n+1] + Vy[l-1][m][n+1])/
                             4.;
         sac[0].Mxz[time] = (Vx[l][m+1][n  ] + Vx[l][m+1][n+1] -
                             Vx[l][m-1][n  ] - Vx[l][m-1][n+1] -
                             Vz[l][m-1][n-1] - Vz[l][m  ][n-1] +
                             Vz[l][m-1][n+1] + Vz[l][m  ][n+1])/
                             4.;
         sac[0].Myz[time] = (Vy[l-1][m+1][n] + Vy[l  ][m+1][n] -
                             Vy[l-1][m-1][n] - Vy[l  ][m-1][n] -
                             Vz[l-1][m-1][n] - Vz[l-1][m  ][n] +
                             Vz[l+1][m-1][n] + Vz[l+1][m  ][n])/
                             4.;
         }
      }

}




/******************************************************************************
   writeSeismograms: write seismograms

   usage: void = writeSeismograms(par, grid, sac, source)

   par       : parameters
   grid      : grid parameters
   sac       : input parameters
   source    : input source
******************************************************************************/
writeSeismograms(par, grid, sac, source)
Parameter    *par;
FdGrid       *grid;
Seismogram   *sac;
FdSource     *source;
{
   float dx, dy, dz;
   int g, s, l, m, n;

   /* initialize */
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;

   /* write seismogram */
   for(g = 0; g < par->g; g++) for(s = 0; s < par->s; s++) {
      l = (par->d == 2) ? (0) : ((int) floor((sac[s].y - grid[g].y0)/dy));
      m = (int) floor((sac[s].z - grid[g].z0)/dz);
      n = (int) floor((sac[s].x - grid[g].x0)/dx);
      if (l < 0 || l >= grid[g].l) continue;
      if (m < 0 || m >= grid[g].m) continue;
      if (n < 0 || n >= grid[g].n) continue;
      if (sac[s].mode == SACS) {
         if (par->d == 2) {
            writeSAC(sac + s, source, sac[s].Vx , s, "x");
            writeSAC(sac + s, source, sac[s].Vz , s, "z");
            writeSAC(sac + s, source, sac[s].P  , s, "p");
            writeSAC(sac + s, source, sac[s].S  , s, "s");
            }
         if (par->d == 3) {
            writeSAC(sac + s, source, sac[s].Vx , s, "x");
            writeSAC(sac + s, source, sac[s].Vy , s, "y");
            writeSAC(sac + s, source, sac[s].Vz , s, "z");
            writeSAC(sac + s, source, sac[s].P  , s, "p");
            writeSAC(sac + s, source, sac[s].S  , s, "s");
            }
         }
      else if (sac[s].mode == SACM) {
         if (par->d == 2) {
            writeSAC(sac + s, source, sac[s].Mxx, s,"xx");
            writeSAC(sac + s, source, sac[s].Mzz, s,"zz");
            writeSAC(sac + s, source, sac[s].Mxz, s,"xz");
            }
         if (par->d == 3) {
            writeSAC(sac + s, source, sac[s].Mxx, s,"xx");
            writeSAC(sac + s, source, sac[s].Myy, s,"yy");
            writeSAC(sac + s, source, sac[s].Mzz, s,"zz");
            writeSAC(sac + s, source, sac[s].Mxy, s,"xy");
            writeSAC(sac + s, source, sac[s].Mxz, s,"xz");
            writeSAC(sac + s, source, sac[s].Myz, s,"yz");
            }
         }
      }   

}




/******************************************************************************
   writeSAC: write specified seismogram in SAC format

   usage: void = writeSAC(sac, source, seismogram, s, type)

   sac       : input parameters
   source    : input source
   seismogram: array of time series
   s         : number of sac file
   type      : type of seismogram
******************************************************************************/
writeSAC(sac, source, seismogram, s, type)
Seismogram *sac;
FdSource   *source;
float      *seismogram;
int         s;
char       *type;
{
   SACParameter sacParameter[1]; 
   char fileName[82], ext[10], sacHeader[SACSIZE];
   FILE *fileOut;

   /* get SAC file name */
   strcpy(fileName, sac->file);
   sprintf(ext, ".%03d.%s\0", s, type);
   strcat(fileName, ext);

   /* open output SAC file */
   if ((fileOut = fopen(fileName, "w")) == NULL) {
      fprintf(stderr, "ERROR: can't open SAC file \"%s\"\n", fileName);
      exit(-1);
      }

   /* get sacFile header */
   getSACDefault(sacParameter);
   sacParameter->delta  = sac->dt;
   sacParameter->npts   = sac->t;
   sacParameter->b      = 0.;
   sacParameter->e      = (float) (sac->t - 1)*sac->dt;
   sacParameter->iftype = 1;
   sacParameter->leven  = 1;
   getSacParameter(sacParameter, sac, source, type);
   getSACHeader(sacHeader, sacParameter);

   /* write sac file */
   fwrite(sacHeader, SACSIZE, 1, fileOut);
   fwrite(seismogram, sizeof(float), sac->t, fileOut);

   /* close sac file */
   fclose(fileOut);

}

/******************************************************************************
   getSacParameter: routine to get special SAC parameters
******************************************************************************/
getSacParameter(sacParameter, sac, source, type)
SACParameter *sacParameter; 
Seismogram   *sac;
FdSource     *source;
char         *type;
{
   char dist[82];
   sacParameter->o      = 0.;
   sacParameter->stla   = 0.01;
   sacParameter->stlo   = ((sac->x - source->x)/1000.)/(KMD);
   sacParameter->stel   = sac->z/1000.;
   sacParameter->stdp   = sac->z/1000.;
   sacParameter->evla   = 0.01;
   sacParameter->evlo   = 0.;
   sacParameter->evel   = source->z/1000.;
   sacParameter->evdp   = source->z/1000.;
   sacParameter->cmpaz  = ((int) type[0] == 'y') ? (90.) : ( 0.);
   sacParameter->cmpinc = ((int) type[0] == 'z') ? ( 0.) : (90.);
   sacParameter->nzyear = 1970;
   sacParameter->nzjday = 1;
   sacParameter->nzhour = 0;
   sacParameter->nzmin  = 0;
   sacParameter->nzsec  = 0;
   sacParameter->nzmsec = 0;
   sprintf(dist, "%d\0", (int) ((sac->x - source->x)/1000.));
   strncpy(sacParameter->kstnm , dist, 8);
   strncpy(sacParameter->kcmpnm, type, 8);
}




/******************************************************************************
   writeImage: write specified image field if at correct time

   usage: void = writeImage(par, grid, data, image, t)

   par       : parameters
   grid      : grid  parameters
   data      : wavefield information
   image     : images to write
   t         : time of wavefield
******************************************************************************/
writeImage(par, grid, data, image, t)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
Image          *image;
int             t;
{
   int i, g;

   for(i = 0; i < par->p; i++) {
      if (!image[i].t && !image[i].movie) continue;
      if (image[i].movie) if ((t + 1)  % image[i].movie) continue;
      if (image[i].t    ) if ((t + 1) != image[i].t    ) continue;
      switch(image[i].mode) {
         case WAVEPS:   
            if (par->d == 2) writeField(par, grid, data, image + i, 1, t);
            if (par->d == 2) writeField(par, grid, data, image + i, 2, t);
            if (par->d == 3) writeField(par, grid, data, image + i, 1, t);
            if (par->d == 3) writeField(par, grid, data, image + i, 2, t);
            break;
         case WAVEXZ:
            if (par->d == 2) writeField(par, grid, data, image + i, 3, t);
            if (par->d == 2) writeField(par, grid, data, image + i, 5, t);
            if (par->d == 3) writeField(par, grid, data, image + i, 3, t);
            if (par->d == 3) writeField(par, grid, data, image + i, 4, t);
            if (par->d == 3) writeField(par, grid, data, image + i, 5, t);
            break;
         case WAVEVP:
            writeField(par, grid, data, image + i, WAVEVP, t);
            break;
         case WAVEVS:
            writeField(par, grid, data, image + i, WAVEVS, t);
            break;
         case WAVEVR:
            writeField(par, grid, data, image + i, WAVEVR, t);
            break;
         case WAVEMV:
            g = par->g - 1;
            putSField(par, grid + g, data + g, image + i, WAVEMV);
            break;
         case WAVEEN:
            g = par->g - 1;
            putSField(par, grid + g, data + g, image + i, WAVEEN);
            break;
         case WAVEVX:
         case WAVEVY:
         case WAVEVZ:
         case WAVETXX:
         case WAVETYY:
         case WAVETZZ:
         case WAVETXY:
         case WAVETXZ:
         case WAVETYZ:
         case WAVEFP:
         case WAVEFS:
            writeFieldT(par, grid, data, image + i, t);
            break;
         case VOLUMEP:
         case VOLUMES:
            writeVolume(par, grid, data, image + i, t);
            break;
         default:
            fprintf(stderr, "Warning: undefined image type\n");
            break;
         }
      }
}



/* only deals with 2D topography */
putSField(par, grid, data, image, mode)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
Image          *image;
int              mode;
{
   float *v, value, Vx, Vy, Vz, z;
   int l, n, m, p, sample;

   /* initialize */
   if (par->d == 3 && par->topo) {
      fprintf(stderr, "Error: can't have 3D topography in surface fields\n");
      exit(-1);
      }
   v = image->v;
   p = 0;
   sample = image->sample;
   if (sample > 1) {
      fprintf(stderr, "Error: bad surface field sample: %d\n", sample);
      exit(-1);
      }

   for(l = 0; l < grid->l; l++) for(n = 0; n < grid->n; n++) {
/*
      z = (par->topo) ? (grid->zi[n]) : (0.);
      m = (int) (z/grid->dz) - grid->m0 + 2;
*/
      z = (par->topo) ? (grid->zi[n]) : (image->w);
      m = (int) (z/grid->dz) - grid->m0;
      if (m >= 0 && m < grid->m) {
         if (mode == WAVEMV) {
            Vx = data->Vx[l][m][n];
            Vz = data->Vz[l][m][n];
            Vy = (par->d == 2) ? (0.) : (data->Vy[l][m][n]);
/*
            value = Vx*Vx + Vy*Vy + Vz*Vz;
*/
            value = Vx*Vx + Vy*Vy;
            if (value > v[p]) v[p] = value;
            }
         if (mode == WAVEEN) {
            Vx = data->Vx[l][m][n];
            Vz = data->Vz[l][m][n];
            Vy = (par->d == 2) ? (0.) : (data->Vy[l][m][n]);
            value  = Vx*Vx + Vy*Vy + Vz*Vz;
            v[p]  += value;
            }
         }
      p++;
      }

}


writeSField(par, grid, image)
Parameter     *par;
FdGrid       *grid;
Image       *image;
{
   double sqrt();
   float *v, value, z;
   int gl, gn, l, n, m, p, mode, type;
   FILE *file, *getFFile();

   /* initialize */
   if (par->d == 3 && par->topo) {
      fprintf(stderr, "Error: can't have 3D topography in surface fields\n");
      exit(-1);
      }
   v = image->v;
   p = 0;
   mode = image->mode;
   type = -1;
   if (mode == WAVEMV) type = 6;
   if (mode == WAVEEN) type = 7;

   /* begin parallel synchronization and open field file (write/append) */
   mpNodeOrder(0);
   file = getFFile(par, grid, image, type, par->t - 1);

   /* write global file header */
   if (par->node == 0) {
      n = par->n;
      l = par->l;
      fwrite(&n, sizeof(int), 1, file);
      fwrite(&l, sizeof(int), 1, file);
      }

   for(l = 0; l < grid->l; l++) for(n = 0; n < grid->n; n++) {
      z = (par->topo) ? (grid->zi[n]) : (0.);
      m = (int) (z/grid->dz) - grid->m0 + 2;
      if (par->nnode == 1) {
         value = image->v[p];
         if (mode == WAVEMV) {
            value = sqrt(value);
            fwrite(&value, sizeof(float), 1, file);
            }
         if (mode == WAVEEN) {
            if (m >= 0 && m < grid->m) value *= 0.5/grid->irho[l][m][n];
            fwrite(&value, sizeof(float), 1, file);
            }
         }
      if (m >= 0 && m < grid->m && par->nnode > 1) {
         value = image->v[p];
         gl = l + grid->l0;
         gn = n + grid->n0;
         if (mode == WAVEMV) {
            value = sqrt(value);
            fwrite(&gn,    sizeof(int  ), 1, file);
            fwrite(&gl,    sizeof(int  ), 1, file);
            fwrite(&value, sizeof(float), 1, file);
            }
         if (mode == WAVEEN) {
            value *= 0.5/grid->irho[l][m][n];
            fwrite(&gn,    sizeof(int  ), 1, file);
            fwrite(&gl,    sizeof(int  ), 1, file);
            fwrite(&value, sizeof(float), 1, file);
            }
         }
      p++;
      }

   /* close output image file and end parallel synchronization */
   fclose(file);
   mpNodeOrder(1);

}


/* does not do multigrid or parallelization */
writeField(par, grid, data, image, type, t)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
Image          *image;
int           type, t;
{
   int d, g, x, y, sample, ftype;
   FILE *file, *getFFile();

   /* initialize and return if no output */
   file   = NULL;
   d      = image->d;
   sample = image->sample;
   ftype  = type;
   if (image->sample < 1) return(0);
   if (type == WAVEVP) ftype =  8;
   if (type == WAVEVS) ftype =  9;
   if (type == WAVEVR) ftype = 10;

   /* begin parallel synchronization and open field file (write/append) */
   mpNodeOrder(0);
   file = getFFile(par, grid, image, ftype, t);

   /* write global header */
   if (par->g == 1 && par->nnode == 1) {
      x = (d == XYPLANE || d == XZPLANE) ? (par->n - 1) : (par->l - 1);
      y = (d == YZPLANE || d == XZPLANE) ? (par->m - 1) : (par->l - 1);
      x = x/sample + 1;
      y = y/sample + 1;
      fwrite(&x, 4, 1, file);
      fwrite(&y, 4, 1, file);
      }
   else if (par->node == 0) {
      x = (d == XYPLANE || d == XZPLANE) ? (par->n) : (par->l);
      y = (d == YZPLANE || d == XZPLANE) ? (par->m) : (par->l);
      fwrite(&x, 4, 1, file);
      fwrite(&y, 4, 1, file);
      }
   
   /* for each multigrid write out field data (including subgrid header) */
   for(g = 0; g < par->g; g++) {
      writeSubField(par, grid + g, data + g, image, type, file);
      }

   /* close output image file and end parallel synchronization */
   fclose(file);
   mpNodeOrder(1);

}


writeSubField(par, grid, data, image, type, file)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
Image          *image;
int              type;
FILE            *file;
{
   int x, y, x0, y0, l, m, n, l1, m1, n1, l2, m2, n2, f, sample, d;

   /* initialize and get sample */
   sample = image->sample;
   if (sample < 1) return(0);

   /* get grid limits */
   if (image->d == XYPLANE) {
      if ((m1 = image->plane - grid->m0) < 0) return(0);
      if ((m1 = m1/grid->s) >= grid->m) return(0);
      m2 = m1 + 1;
      l1 = 0; l2 = grid->l;
      n1 = 0; n2 = grid->n;
      x  = grid->n*grid->s;
      y  = grid->l*grid->s;
      x0 = grid->n0;
      y0 = grid->l0;
      }
   if (image->d == XZPLANE) {
      if ((l1 = image->plane - grid->l0) < 0) return(0);
      if ((l1 = l1/grid->s) >= grid->l) return(0);
      l2 = l1 + 1;
      m1 = 0; m2 = grid->m;
      n1 = 0; n2 = grid->n;
      x  = grid->n*grid->s;
      y  = grid->m*grid->s;
      x0 = grid->n0;
      y0 = grid->m0;
      }
   if (image->d == YZPLANE) {
      if ((n1 = image->plane - grid->n0) < 0) return(0);
      if ((n1 = n1/grid->s) >= grid->n) return(0);
      n2 = n1 + 1;
      l1 = 0; l2 = grid->l;
      m1 = 0; m2 = grid->m;
      x  = grid->l*grid->s;
      y  = grid->m*grid->s;
      x0 = grid->l0;
      y0 = grid->m0;
      }

   /* get apparent grid sampling factor */
   f = sample/grid->s; if (f == 0) f = 1;
   f = f*grid->s;

   /* write header */
   if (par->g > 1 || par->nnode > 1) {
      fwrite(&x,         4, 1, file);
      fwrite(&y,         4, 1, file);
      fwrite(&x0,        4, 1, file);
      fwrite(&y0,        4, 1, file);
      fwrite(&f,         4, 1, file);
      }

   /* determine actual grid sampling factor */
   f = sample/grid->s; if (f == 0) f = 1;

   /* write image file (all rows not including boundary) */
   d = image->d;
   if (type == 1) writeP(par, grid, data, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 2) writeS(par, grid, data, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 3) writeV(data->Vx, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 4) writeV(data->Vy, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 5) writeV(data->Vz, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 6) writeM(grid, 1, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 7) writeM(grid, 2, l1, l2, m1, m2, n1, n2, f, d, file);
   if (type == 8) writeM(grid, 3, l1, l2, m1, m2, n1, n2, f, d, file);
   
}



writeP(par, grid, data, l1, l2, m1, m2, n1, n2, sample, d, file)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
int l1, l2, m1, m2, n1, n2, sample, d;
FILE *file;
{
   float getP2D(), getP3D();
   float P, ***Vx, ***Vy, ***Vz, dx, dy, dz;
   int l, m, n;
   if (sample <= 0) return(0);
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   dx = grid->dx; dy = grid->dy; dz = grid->dz;
   if (par->d == 2) {
      for(m = m1; m < m2; m += sample) 
         for(n = n1; n < n2; n += sample) {
            P = getP2D(Vx[0], Vz[0], dx, dz, m, n);
            fwrite(&P, 1, 4, file);
            }
      }
   if (par->d == 3 && d != YZPLANE) {
      for(l = l1; l < l2; l += sample) 
         for(m = m1; m < m2; m += sample) 
            for(n = n1; n < n2; n += sample) {
               P = getP3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
               fwrite(&P, 1, 4, file);
               }
      }
   if (par->d == 3 && d == YZPLANE) {
      for(n = n1; n < n2; n += sample)
         for(m = m1; m < m2; m += sample) 
            for(l = l1; l < l2; l += sample) {
               P = getP3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
               fwrite(&P, 1, 4, file);
               }
      }

}




writeS(par, grid, data, l1, l2, m1, m2, n1, n2, sample, d, file)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
int l1, l2, m1, m2, n1, n2, sample, d;
FILE *file;
{
   float getS2D(), getS3D();
   float S, ***Vx, ***Vy, ***Vz, dx, dy, dz;
   int l, m, n;
   if (sample <= 0) return(0);
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   dx = grid->dx; dy = grid->dy; dz = grid->dz;
   if (par->d == 2) {
      for(m = m1; m < m2; m += sample) 
         for(n = n1; n < n2; n += sample) {
            S = getS2D(Vx[0], Vz[0], dx, dz, m, n);
            fwrite(&S, 1, 4, file);
            }
      }
   if (par->d == 3 && d != YZPLANE) {
      for(l = l1; l < l2; l += sample) 
         for(m = m1; m < m2; m += sample) 
            for(n = n1; n < n2; n += sample) {
               S = getS3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
               fwrite(&S, 1, 4, file);
               }
      }
   if (par->d == 3 && d == YZPLANE) {
      for(n = n1; n < n2; n += sample)
         for(m = m1; m < m2; m += sample) 
            for(l = l1; l < l2; l += sample) {
               S = getS3D(Vx, Vy, Vz, dx, dy, dz, l, m, n);
               fwrite(&S, 1, 4, file);
               }
      }

}




writeV(V, l1, l2, m1, m2, n1, n2, sample, d, file)
float ***V;
int l1, l2, m1, m2, n1, n2, sample, d;
FILE *file;
{
   int l, m, n;
   if (sample <= 0) return(0);
   if (d != YZPLANE) 
      for(l = l1; l < l2; l += sample) 
         for(m = m1; m < m2; m += sample) 
            for(n = n1; n < n2; n += sample)
               fwrite(&(V[l][m][n]), 1, 4, file);
   if (d == YZPLANE) 
      for(n = n1; n < n2; n += sample)
         for(m = m1; m < m2; m += sample) 
            for(l = l1; l < l2; l += sample) 
               fwrite(&(V[l][m][n]), 1, 4, file);
}


writeM(grid, mode, l1, l2, m1, m2, n1, n2, sample, d, file)
FdGrid  *grid;
int mode, l1, l2, m1, m2, n1, n2, sample, d;
FILE *file;
{
   float lamda, mu, rho, irho, value;
   int l, m, n;
   if (sample <= 0) return(0);
   value = -2.;
   if (d != YZPLANE) {
      for(l = l1; l < l2; l += sample) 
         for(m = m1; m < m2; m += sample) 
            for(n = n1; n < n2; n += sample) {
               lamda = grid->lamda[l][m][n];
               mu    = grid->mu   [l][m][n];
               irho  = grid->irho [l][m][n];
               if (irho != 0.) rho = 1./irho;
               if (mode == 1) value = GETALPHA(lamda, mu, rho);
               if (mode == 2) value = GETBETA (lamda, mu, rho);
               if (mode == 3) value = rho;
               fwrite(&value, 1, 4, file);
               }
      }
   if (d == YZPLANE) {
      for(n = n1; n < n2; n += sample) 
         for(m = m1; m < m2; m += sample) 
            for(l = l1; l < l2; l += sample) {
               lamda = grid->lamda[l][m][n];
               mu    = grid->mu   [l][m][n];
               irho  = grid->irho [l][m][n];
               if (irho != 0.) rho = 1./irho;
               if (mode == 1) value = GETALPHA(lamda, mu, rho);
               if (mode == 2) value = GETBETA (lamda, mu, rho);
               if (mode == 3) value = rho;
               fwrite(&value, 1, 4, file);
               }
      }

}


/* comment to avoid compiler warnings on tera
int
getIPlane(par, grid, image)
Parameter    *par;
FdGrid      *grid;
Image      *image;
{

}
*/



FILE *
getFFile(par, grid, image, type, t)
Parameter     *par;
FdGrid       *grid;
Image       *image;
int        type, t;
{
   int g;
   char fileName[82], ext[10], rwa[5];
   FILE *file;

   /* get wavefield file name */
   if (type ==  1) strcpy(ext, "p");
   if (type ==  2) strcpy(ext, "s");
   if (type ==  3) strcpy(ext, "x");
   if (type ==  4) strcpy(ext, "y");
   if (type ==  5) strcpy(ext, "z");
   if (type ==  6) strcpy(ext, "V");
   if (type ==  7) strcpy(ext, "E");
   if (type ==  8) strcpy(ext, "A");
   if (type ==  9) strcpy(ext, "B");
   if (type == 10) strcpy(ext, "R");
   if (type == 11) strcpy(ext, "VX");
   if (type == 12) strcpy(ext, "VY");
   if (type == 13) strcpy(ext, "VZ");
   if (type == 14) strcpy(ext, "XX");
   if (type == 15) strcpy(ext, "YY");
   if (type == 16) strcpy(ext, "ZZ");
   if (type == 17) strcpy(ext, "XY");
   if (type == 18) strcpy(ext, "XZ");
   if (type == 19) strcpy(ext, "YZ");
   if (type == 20) strcpy(ext, "FP");
   if (type == 21) strcpy(ext, "FS");

   /* open output image file */
   strcpy(rwa, "a"); 
   if (par->node == 0) strcpy(rwa, "w");
   getFileName(fileName, image->file, ext, t + 1, par->t);
   if ((file = fopen(fileName, rwa)) == NULL) {
      fprintf(stderr, "ERROR: can't open image file \"%s\"\n", fileName);
      exit(-1);
      }

   return(file);

}



getFileName(filename, head, tail, number, max)
char *filename, *head, *tail; 
int number, max;
{
   int digits;
   char ext[10], format[20];

   /* determine number of digits */
   digits = 1;
   if (max < 1) max = 1;
   while(max /= 10) digits++;

   /* determine format of number */
   sprintf(format, "%%0%dd\0", digits);
   sprintf(ext, format, number);

   /* get file name */
   strcpy(filename, head);
   strcat(filename,  ".");
   strcat(filename,  ext);
   strcat(filename,  ".");
   strcat(filename, tail);

}



/******************************************************************************
   getClock: write clock information

   usage: void = printClock(par, grid, type, mode)

   par       : parameters
   grid      : grid  parameters
   type      : type of determination (0=init; 1=global; 2=comp; 3=para; 4=mpp)
   mode      : mode of operation (0=init; 1=close; 2=print)

******************************************************************************/
getClock(par, grid, type, mode)
Parameter        *par;
FdGrid          *grid;
int        type, mode;
{
   float time, mflops, tflops;
   int nflops;

   if (type == 0 && mode != 2) {
      par->time0 = 0.;
      par->time1 = 0.;
      par->time2 = 0.;
      par->timec = 0.;
      par->timep = 0.;
      par->timem = 0.;
      return(0);
      }
   if (type == 1 && mode != 2) {
      if (mode == 0) par->time1  = ((float) mclock())/1000.;
      if (mode == 1) par->time2  = ((float) mclock())/1000.;
      return(0);
      }
   if (type == 2 && mode != 2) {
      if (mode == 0) par->time0  = ((float) mclock())/1000.;
      if (mode == 1) par->timec += ((float) mclock())/1000. -
                                   par->time0;
      return(0);
      }
   if (type == 3 && mode != 2) {
      if (mode == 0) par->time0  = ((float) mclock())/1000.;
      if (mode == 1) par->timep += ((float) mclock())/1000. -
                                   par->time0;
      return(0);
      }
   if (type == 4 && mode != 2) {
      if (mode == 0) par->time0  = ((float) mclock())/1000.;
      if (mode == 1) par->timem += ((float) mclock())/1000. -
                                   par->time0;
      return(0);
      }

   time = par->time2 - par->time1;

   nflops  = (grid->d == 2) ? (64) : (141);
   tflops  = (float) par->n*par->m*par->l;
   tflops *= (float) (par->t*nflops);
   mflops  = 0.;
   if (time > 0.) mflops = tflops/time/1000000.;

   if (par->node == 0) {
      fprintf(stderr, "TIME = %8.2f seconds; MFLOPS = %6.2f\n", time, mflops);
      }

   if (par->clock || !par->clock) {
      mpNodeOrder(0);
/*
      fprintf(stderr, "NODE %03d: cal = %6.2f; par = %6.2f; mpp = %6.2f\n",
              par->node, par->timec, par->timep, par->timem);
*/
      mpNodeOrder(1);
      }
   
}

/* doesn't work for multigrid or 2D */
writeRestart(par, grid, data, t)
Parameter        *par;
FdGrid          *grid;
VelocityStress  *data;
int                 t;
{
   int l, m, n, l0, m0, n0, l1, m1, n1;
   FILE *file, *getDFile();

   /* return if no restart dumps */
   if (par->dump  <= 0) return(0);
   if ((t+1)%par->dump) return(0);

   /* delete and do not create last restart file if requested */
   if(par->nolast && (t+1) == par->t) {
      delDFile(par, t);
      return(0);
      }

   /* initialize coordinates */
   l0 = -grid->fnt;
   m0 = -grid->top;
   n0 = -grid->lhs;
   l1 = grid->l + grid->bck; 
   m1 = grid->m + grid->bot; 
   n1 = grid->n + grid->rhs; 

   /* get dump file to write */
   file = getDFile(par, t);

   /* dump data */
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Vx[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Vy[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Vz[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Txx[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Tyy[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Tzz[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Txy[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Txz[l][m]+n0, sizeof(float), (n1-n0), file);
   for(l = l0; l < l1; l++) for(m = m0; m < m1; m++) 
      fwrite(data->Tyz[l][m]+n0, sizeof(float), (n1-n0), file);

   /* close dump file */
   fclose(file);

   /* delete previous restart file (or set to 0 length) */
   mpNodeOrder(0);
   mpNodeOrder(1);
   delDFile(par, t);

}

FILE *
getDFile(par, t)
Parameter *par;
int          t;
{
   char filename[82];
   FILE *file;
   sprintf(filename, "%s.%d.%d.E3D_RESTART\0", par->rfile, par->node, (t+1));
   if ((file = fopen(filename, "w")) == NULL) {
      fprintf(stderr, "ERROR: can't open dump file \"%s\"\n", filename);
      exit(1);
      }
   return(file);
}

delDFile(par, t)
Parameter *par;
int          t;
{
   int time;
   char filename[82], string[80];
   time = (t+1) - par->dump;
   if (time >= par->restart && time > 0) {
      strcpy(string, "rm ");
      sprintf(filename, "%s.%d.%d.E3D_RESTART\0", par->rfile, par->node, time);
      strcat(string, filename);
      system(string);
      }
}

