/******************************************************************************
   paraxial2d: paraxial boundary condition routines for 3D elastic code 

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   absorbLVR: Clayton-Engquist A1 absorbing boundary condition (right side)
   absorbLVL: Clayton-Engquist A1 absorbing boundary condition (left  side)
   absorbLVB: Clayton-Engquist A1 absorbing boundary condition (bottom)
   absorbLVT: Clayton-Engquist A1 absorbing boundary condition (top)
   cornerVTL: corner velocity boundary condition (top    - left)
   cornerVTR: corner velocity boundary condition (top    - right)
   cornerVBR: corner velocity boundary condition (bottom - right)
   cornerVBL: corner velocity boundary condition (bottom - left)
   boundTR  : stress boundary condition (right side)
   boundTL  : stress boundary condition (left  side)
   boundTB  : stress boundary condition (bottom)
   boundTT  : stress boundary condition (top)
   cornerTTL: corner stress boundary condition (top    - left)
   cornerTTR: corner stress boundary condition (top    - right)
   cornerTBR: corner stress boundary condition (bottom - right)
   cornerTBL: corner stress boundary condition (bottom - left)

   compile: cc -g -c boundary.c 

   note: input units are km, s, km/s, g/cm3, dyne-cm
   note: used  units are mks (m, s, m/s, kg/m3, N-m)

   author: shawn larsen
   date  : march 1989
   where : caltech
*****************************************************************************/

#include "elas.h"


/******************************************************************************
2D VELOCITY BOUNDARY CONDITIONS
******************************************************************************/
absorbLVR(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, factor;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   n = grid->n;
   for(m = 0; m < grid->m; m++) {
      VxLast = Vx[m][n];
      VzLast = Vz[m][n];
      update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
      update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
      alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
      beta  = sqrt((                 mu[m][n])*irho[m][n]);
      if (grid->ave) {
         beta  = sqrt((              mu[m][n])*2./(grid->rho[0][m  ][n  ] +
                                                   grid->rho[0][m+1][n  ]));
         }
      xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
      xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
      data->Vx[0][m][n+1] = VxLast + xP*(Vx[m][n+1] - Vx[m][n]);
      data->Vz[0][m][n+1] = VzLast + xS*(Vz[m][n+1] - Vz[m][n]);
      }
}




absorbLVLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   n = -1;
   for(m = 0; m < grid->m; m++) {
      update4Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
      VxLast = Vx[m][n  ];
      VzLast = Vz[m][n-1];
      update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n  );
      update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n-1);
      alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
      beta  = sqrt((                 mu[m][n])*irho[m][n]);
      if (grid->ave) {
         beta  = sqrt((              mu[m][n])*2./(grid->rho[0][m  ][n  ] +
                                                   grid->rho[0][m+1][n  ]));
         }
      xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
      xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
      data->Vx[0][m][n-1] = VxLast + xP*(Vx[m][n-1] - Vx[m][n]);
      data->Vz[0][m][n-2] = VzLast + xS*(Vz[m][n-2] - Vz[m][n-1]);
      }
}




absorbLVBa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = grid->m;
   for(n = 0; n < grid->n; n++) {
      update4Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
      VxLast = Vx[m+1][n];
      VzLast = Vz[m][n];
      update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m+1, n);
      update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
      alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
      beta  = sqrt((                 mu[m][n])*irho[m][n]);
      if (grid->ave) {
         beta  = sqrt((              mu[m][n])*2./(grid->rho[0][m  ][n  ] +
                                                   grid->rho[0][m  ][n-1]));
         }
      zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
      zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
      data->Vx[0][m+2][n] = VxLast + zS*(Vx[m+2][n] - Vx[m+1][n]);
      data->Vz[0][m+1][n] = VzLast + zP*(Vz[m+1][n] - Vz[m][n]);
      }
}
 



absorbLVT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = -1;
   for(n = 0; n < grid->n; n++)
      {
      VxLast = Vx[m][n];
      VzLast = Vz[m][n];
      update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
      update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
      alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
      beta  = sqrt((                 mu[m][n])*irho[m][n]);
      if (grid->ave) {
         beta  = sqrt((              mu[m][n])*2./(grid->rho[0][m  ][n  ] +
                                                   grid->rho[0][m  ][n-1]));
         }
      zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
      zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
      data->Vx[0][m-1][n] = VxLast + zS*(Vx[m-1][n] - Vx[m][n]);
      data->Vz[0][m-1][n] = VzLast + zP*(Vz[m-1][n] - Vz[m][n]);
      }
}




cornerVTLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = -1;
   n = -1;
   alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
   beta  = sqrt((                 mu[m][n])*irho[m][n]);
   xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
   xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
   zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
   zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
   VzLast = Vz[m][n];
   /*
   update2Vz2d(par, grid, data->Vz, data->Txz, data->Tzz, m, n);
   data->Vz[m-1][n] = VzLast + zP*(data->Vz[m-1][n] - data->Vz[m][n]);
   */
   aboveSurfaceVz2d(par, grid, Vx, Vz, m, n);
   data->Vz[0][m-1][n] = 0.;

   VxLast = Vx[m][n];
   VzLast = Vz[m][n-1];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n - 1);
   data->Vz[0][m  ][n-2] = VzLast + xS*(Vz[m  ][n-2] - Vz[m][n-1]);
   data->Vx[0][m  ][n-1] = VxLast + xP*(Vx[m  ][n-1] - Vx[m][n  ]);
   data->Vz[0][m-1][n-1] = VzLast + zP*(Vz[m-1][n-1] - Vz[m][n-1]);
   data->Vx[0][m-1][n  ] = VxLast + zS*(Vx[m-1][n  ] - Vx[m][n  ]);
}




cornerVTR(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m =      -1;
   n = grid->n;
   VxLast = Vx[m][n];
   VzLast = Vz[m][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
   alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
   beta  = sqrt((                 mu[m][n])*irho[m][n]);
   xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
   xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
   zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
   zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
   data->Vx[0][m-1][n] = VxLast + zS*(Vx[m-1][n] - Vx[m][n]);
   data->Vz[0][m-1][n] = VzLast + zP*(Vz[m-1][n] - Vz[m][n]);
   data->Vx[0][m][n+1] = VxLast + xP*(Vx[m][n+1] - Vx[m][n]);
   data->Vz[0][m][n+1] = VzLast + xS*(Vz[m][n+1] - Vz[m][n]);
}




cornerVBRa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = grid->m;
   n = grid->n;
   alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
   beta  = sqrt((                 mu[m][n])*irho[m][n]);
   xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
   xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
   zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
   zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
   VxLast = Vx[m][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
   data->Vx[0][m][n+1] = VxLast + xP*(Vx[m][n+1] - Vx[m][n]);
   VzLast = Vz[m  ][n];
   VxLast = Vx[m+1][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m    , n);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m + 1, n);
   data->Vz[0][m  ][n+1] = VzLast + xS*(Vz[m  ][n+1] - Vz[m  ][n]);
   data->Vx[0][m+1][n+1] = VxLast + xP*(Vx[m+1][n+1] - Vx[m+1][n]);
   data->Vz[0][m+1][n  ] = VzLast + zP*(Vz[m+1][n  ] - Vz[m  ][n]);
   data->Vx[0][m+2][n  ] = VxLast + zS*(Vx[m+2][n  ] - Vx[m+1][n]);

}




cornerVBL(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = grid->m;
   n = -1;
   VxLast = Vx[m][n];
   VzLast = Vz[m][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
   alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
   beta  = sqrt((                 mu[m][n])*irho[m][n]);
   xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
   xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
   zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
   zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
   data->Vx[0][m+1][n] = VxLast + zS*(Vx[m+1][n] - Vx[m][n]);
   data->Vz[0][m+1][n] = VzLast + zP*(Vz[m+1][n] - Vz[m][n]);
   data->Vx[0][m][n-1] = VxLast + xP*(Vx[m][n-1] - Vx[m][n]);
   data->Vz[0][m][n-1] = VzLast + xS*(Vz[m][n-1] - Vz[m][n]);
}
 
 


cornerVBLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float VxLast, VzLast, alpha, beta, xP, xS, zP, zS;
   float **Vx, **Vz, **lamda, **mu, **irho;
   int n, m;
   Vx = data->Vx[0]; Vz = data->Vz[0]; 
   lamda = grid->lamda[0]; mu = grid->mu[0]; irho = grid->irho[0];
   m = grid->m;
   n = -1;
   alpha = sqrt((lamda[m][n] + 2.*mu[m][n])*irho[m][n]);
   beta  = sqrt((                 mu[m][n])*irho[m][n]);
   xP = (grid->dx - alpha*par->dt)/(grid->dx + alpha*par->dt);
   xS = (grid->dx -  beta*par->dt)/(grid->dx +  beta*par->dt);
   zP = (grid->dz - alpha*par->dt)/(grid->dz + alpha*par->dt);
   zS = (grid->dz -  beta*par->dt)/(grid->dz +  beta*par->dt);
   VxLast = Vx[m][n];
   VzLast = Vz[m][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m, n);
   data->Vx[0][m][n-1] = VxLast + xP*(Vx[m][n-1] - Vx[m][n]);
   data->Vz[0][m+1][n] = VzLast + zP*(Vz[m+1][n] - Vz[m][n]);
   VxLast = Vx[m][n-1];
   VzLast = Vz[m+1][n];
   update2Vx2d(par, grid, Vx, data->Txx[0], data->Txz[0], m, n - 1);
   update2Vz2d(par, grid, Vz, data->Txz[0], data->Tzz[0], m + 1, n);
   data->Vx[0][m+2][n  ] = VxLast + zS*(Vx[m+2][n  ] - Vx[m+1][n  ]);
   data->Vz[0][m+1][n-1] = VzLast + zP*(Vz[m+1][n-1] - Vz[m  ][n-1]);
   data->Vx[0][m+1][n-1] = VxLast + xP*(Vx[m+1][n-1] - Vx[m+1][n  ]);
   data->Vz[0][m  ][n-2] = VzLast + xS*(Vz[m  ][n-2] - Vz[m  ][n-1]);
}




/******************************************************************************
2D STRESS BOUNDARY CONDITIONS
******************************************************************************/
boundTR(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   n = grid->n;
   for(m = 0; m < grid->m; m++) {
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n + 1);
      }
}




boundTLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   n = -1;
      /*
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], 0, n);
      data->Tzz[0][0][n] = 0.;
      */
      if (grid->bz1 == SURFACE) {
         update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], 0, n);
         data->Tzz[0][0][n] = 0.;
         }
      else {
         update4Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], 0, n);
         update4Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], 0, n);
         }
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], 0, n);
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], 0, n - 1);
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], 0, n - 1);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], 0, n - 1);
   for(m = 1; m < grid->m; m++) {
      update4Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
      update4Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n - 1);
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n - 1);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n - 1);
      }
/*
   for(m = 0; m < grid->m; m++) {
      update4Txxd(par, grid, data->Txx, data->Vx, data->Vz, m, n);
      update4Tzzd(par, grid, data->Tzz, data->Vx, data->Vz, m, n);
      update2Txzd(par, grid, data->Txz, data->Vx, data->Vz, m, n);
      update2Txxd(par, grid, data->Txx, data->Vx, data->Vz, m, n - 1);
      update2Tzzd(par, grid, data->Tzz, data->Vx, data->Vz, m, n - 1);
      update2Txzd(par, grid, data->Txz, data->Vx, data->Vz, m, n - 1);
      }
*/
}




boundTBa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = grid->m;
   for(n = 0; n < grid->n; n++) {
      update4Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
      update4Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m + 1, n);
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m + 1, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m + 1, n);
      }
}




boundTT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = -1;
   for(n = 0; n < grid->n; n++) {
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m - 1, n);
      }
}




cornerTTLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = -1;
   n = -1;
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m - 1, n);
}




cornerTTR(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m =      -1;
   n = grid->n;
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m - 1, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n + 1);
}




cornerTBRa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = grid->m;
   n = grid->n;
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n + 1);
}




cornerTBL(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = grid->m;
   n =      -1;
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txx2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Tzz2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m + 1, n);
}




cornerTBLa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = grid->m;
   n =      -1;
   update4Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
   update4Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n);
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m, n - 1);
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], m + 1, n);
   update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m + 1, n-1);
   update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], m + 1, n-1);
}
