/******************************************************************************
   readline: routines to read input file lines and return information

   argvFile  - store file in argv    -> lineNumbers   = argvFile(argv, n, file)
   readLine  - get line from file    -> lineCondition = readLine(line, n, file)
   setLine   - input line type list  -> setLine(string, lineType)
   getLine   - get line type         -> lineType = getLine(line)
   endLine   - clear line type list  -> endLine()
   setArg    - initialize attributes -> setArg(line)
   getArg    - get attribute         -> argCondition = getArg(name, type, value)
   mstArg    - must get attribute    -> argCondition = mstArg(name, type, value)
   endArg    - clear attribute list  -> endArg()
   getType   - get line type value   -> lineTypeValue = getType(lineType)

   Predefined Line Types and Values (readLine and getLine):

      LINE_FOUND    1   non-null line found with readLine
      FILE_END      0   EOF encountered with readLine
      LINE_END     -1   EOF encountered on continuation line (\) with readLine

      LINE_COMMENT -2   hard comment (# as first character) (readLine/getLine)

      LINE_NULL    -3   null line found in getLine ('\0' as first character)
      LINE_BLANK   -4   blank line  (all ' ' or '\t' before '\n')
      LINE_VALUE   -5   numeric value for first character (0-9, +, -, .)
      LINE_PAR     -6   parameter line found in getLine (parameter=value)
      LINE_UNKNOWN -9   unknown line type

   author: shawn larsen
   date:   march 1988
   where:  caltech
******************************************************************************/


#include <stdio.h>
#include <string.h>
#include "readline.h"

int nLineType = 0, *LineType;
char **LineString;

int nLineArg = 0;
char **ArgName, **ArgList, ArgLine[STRING_SIZE];



/******************************************************************************
   argvFile: read file and store lines in **char array

   usage: lineNumbers = argvFile(argv, maxline, file)

   lineNumbers: number of lines stored

   hard comment lines are not skipped (# as first non-white character)
   / as last character specifies continuation line - ignored for hard comment
   argv is a preallocated array of at least size argv[n]
******************************************************************************/
argvFile(argv, maxline, file)
int  maxline;
char  **argv;
FILE   *file;
{
   int n;
   char line[STRING_SIZE];
   n = 0;
   while(n < maxline) {
      if (!readLine(line, STRING_SIZE, file)) return(n);
      argv[n++] = (char *) strdup(line);
      }
   return(n);
}




/******************************************************************************
   readLine: read input line from file and return line condition

   usage: lineCondition = readLine(line, n, file)

   line - output line (memory already allocated)
   n    - maximum number of characters to read
   file - input file

   lineCondition: LINE_FOUND   - line read and stored
                  LINE_END     - EOF encountered on continuation line
                  LINE_COMMENT - hard comment line (#)
                  FILE_END     - EOF encountered on initial line

   hard comment lines are skipped (# as first non-white character)
   / as last character specifies continuation line - ignored for hard comment
******************************************************************************/
readLine(line, n, file)
char *line;
int      n;
FILE *file;
{
   int ptr = 0;
   char string[STRING_SIZE];

   /* read next line */
   if (fgets(line, n, file) == NULL) return(FILE_END);

   /* initial returns */
   if (line[0] == (char) '\0')  return(LINE_FOUND);
   string[0] = '\0';
   sscanf(line,"%s", string);
   if (string[0] == (char) '#') return(LINE_COMMENT);

   /* check for continuations, skipping comments */
   while((ptr = continueLine(line))) {
      line += ptr;
      if (fgets(line, n, file) == NULL) return(LINE_END);
      }
   return(LINE_FOUND);

}



int 
continueLine(line)
char *line;
{
   int ptr;
   ptr = strlen(line);
   if (ptr < 2)                    return(0);
   if (line[ptr-2] != (char) '\\') return(0); 
   if (line[ptr-1] != (char) '\n') return(0);
   return(ptr);
}



/******************************************************************************
   setLine: set input line parameters

   usage: setLine(string, lineType)
******************************************************************************/
setLine(string, lineType)
int lineType;
char *string;
{
   /* get memory if first pass */
   if (!nLineType)
      {
      LineType   = (int *)   malloc(MAX_LINE_TYPES*sizeof(int));
      LineString = (char **) malloc(MAX_LINE_TYPES*sizeof(char *));
      }

   /* error if maximum line types exceeding */
   if (nLineType == MAX_LINE_TYPES)
      {
      fprintf(stderr, "ERROR: setLine - maximum line types exceeded\n");
      exit(-1);
      }

   /* check bounds for line types */
   if (lineType <= 0) fprintf(stderr, "WARNING: declared line types <= 0\n");

   /* get and store line types */
   LineType[nLineType] = lineType;
   LineString[nLineType] = (char *) strdup(string);
   nLineType++; 

}



/******************************************************************************
   getLine: return line type

   usage: lineType = getLine(line)

   lineType: LINE_COMMENT - comment line (first non-white character #)
             LINE_NULL    - first character NULL 
             LINE_BLANK   - no strings in line
             LINE_UNKNOWN - unknown line type
             LINE_VALUE   - numeric for first non-white (or "+ - . ")
             LINE_PAR     - line is a parameter line (with "=")
             DEFINED      - user defined line types

   note: "-" and "+" must be followed by 0-9 or .
   note: "."         must be followed by 0-9
******************************************************************************/
getLine(line)
char *line;
{
   int n;
   char string[STRING_SIZE], c, c1, c2;

   /* determine if null or blank (no characters) line */
   if (line[0] == (char) '\0') return(LINE_NULL);
   if (line[0] == (char) '\n') return(LINE_BLANK);

   /* get first string */
   string[0] = '\0';
   string[1] = '\0';
   string[2] = '\0';
   n = sscanf(line, "%s", string); 
   if (n > 0) strcpy(string + strlen(string), " \0");
   c  = string[0];
   c1 = string[1];
   c2 = string[2];

   /* determine if line starts with numeric value (0-9) */
   if (c >= '0' && c <= '9') return(LINE_VALUE);

   /* determine if line starts with numeric character (+ , -,  .) */
   if (c == '+' || c == '-') {
      if (c1 >= '0' && c1 <= '9') return(LINE_VALUE);
      if (c1 == '.') if (c2 >= '0' && c2 <= '9') return(LINE_VALUE);
      }
   if (c == '.') if (c1 >= '0' && c1 <= '9') return(LINE_VALUE);

   /* determine if comment line */
   if (c == (char) '#') return(LINE_COMMENT);          

   /* determine if blank line */
   if (n <= 0) return(LINE_BLANK);

   /* determine if line is a parameter */
   for(n = 0; n < STRING_SIZE; n++) {
      if (string[n] == (char) '=') return(LINE_PAR);
      if (string[n] == (char) '\0') break;
      }

   /* determine defined line type */
   for(n = 0; n < nLineType; n++)
      if (!strncmp(string, LineString[n], strlen(LineString[n]))) 
         return(LineType[n]);

   /* line type not known */
   return(LINE_UNKNOWN);

}



/******************************************************************************
   endLine: clear defined line types

   usage: endLine()
******************************************************************************/
endLine()
{
   free(LineType);
   while(nLineType--) free(LineString[nLineType]);
   free(LineString);
   nLineType = 0;
}



/******************************************************************************
   setArg: initialize line argument list (clear previous)

   usage: setArg(line)
******************************************************************************/
setArg(line)
char *line;
{
   int n, ptr, cnt;
   char string[STRING_SIZE], c; 

   /* clear previous values */
   if (nLineArg) {
      while(nLineArg) {
         nLineArg--;
         free(ArgList[nLineArg]);
         free(ArgName[nLineArg]);
         }
      free(ArgList);
      free(ArgName);
      }

   /* get memory */
   ArgList = (char **) malloc(MAX_LINE_ARG*sizeof(char *));
   ArgName = (char **) malloc(MAX_LINE_ARG*sizeof(char *));

   /* read line and extract arguments */
   n = 0;
   ptr = 0;
   while((c = line[n]) && n < STRING_SIZE)
      switch(c)
         {
         case '\t':
         case '\n':
            ptr = 0;
            n++;
            break;
         case  ' ':
            if (ptr) {
               string[ptr] = '\0';
               ArgList[nLineArg] = (char *) strdup(string);
               ArgName[nLineArg] = (char *) strdup("\0");
               nLineArg++;
               }
            ptr = 0;
            n++;
            break;
         case  '=':
            c = line[++n];
            if (c != ' ' && c != '\t' && c != '\n' && c != '\0')
               {
               string[ptr] = '\0';
               ArgList[nLineArg] = (char *) strdup(string);
               n += getParString(line, string, n);
               ArgName[nLineArg] = (char *) strdup(string);
               nLineArg++;
               }
            ptr = 0;
            break;
         default  :
            string[ptr++] = c;
            n++;
            break;
         }

   /* copy line to line string */
   strcpy(ArgLine, line);

}



getParString(line, string, n)
int n;
char *line, *string;
{
   int cnt, ptr; 
   char c;

   /* initialize */
   string[0] = (char) '\0';
   line += n;
   cnt = 1;
   ptr = 0;

   /* get first character */
   c = line[0];

   /* case of string attribute */
   if (c == '"' || c == '\'') {
      c = line[++ptr];
      while(c && ++n < STRING_SIZE)
         switch(c) {
            case '\\': 
               cnt++;
               (line[++ptr] == '\n') ? (cnt++) : (string[ptr-cnt] = line[ptr]);
               c = line[++ptr];
               break;
            case  '"':
            case '\'':
               string[ptr - cnt] = '\0';
               return(++ptr);
            default  :
               string[ptr - cnt] = c;
               c = line[++ptr];
            }
      string[ptr - cnt] = (char) '\0';
      fprintf(stderr, "WARNING: string quotes not closed\n");
      return(ptr);
      }

   /* case of non-string attribute */
   while(c && c != ' ' && c != '\t' && c != '\n' && n++ < STRING_SIZE) {
      string[ptr] = c;
      c = line[++ptr];
      }
   string[ptr] = (char) '\0';
   return(ptr);

}



/******************************************************************************
   getArg: get line argument from list

   usage: argCondition = getArg(argname, argtype, value)

   argCondition: 1 - argument exists (return n for vector)
                 0 - argument not found

   argname: string name of argument
   argtype: argument type ("d" = int; "f" = float; "F" = double; "s" = string)
                           "v" = vector; "b" = boolean; "B" = negative boolean)
   value:   value of argument
******************************************************************************/
getArg(argname, argtype, value)
char *argname, *argtype;
int *value;
{
   double *dbl, atof();
   float  *flt;
   int i, n;
   char *str;

   n = nLineArg;
   while(n--) {
      if (getArgCmp(argname, ArgList[n])) continue;
      switch(argtype[0]) {
         case 'b':
            if (value) *value = (*ArgName[n]) ? (atoi(ArgName[n])) : (1);
            return(1);
         case 'B':
            if (value) *value = (*ArgName[n]) ? (atoi(ArgName[n])) : (0);
            return(1);
         case 'd':
            if (!(*ArgName[n])) return(0);
            *value = atoi(ArgName[n]);
            return(1);
         case 'f':
            if (!(*ArgName[n])) return(0);
             flt = (float *) value;
            *flt = atof(ArgName[n]);
            return(1);
         case 'F':
            if (!(*ArgName[n])) return(0);
             dbl = (double *) value;
            *dbl = atof(ArgName[n]);
            return(1);
         case 's':
            str = (char *) value;
            strcpy(str, ArgName[n]);
            return(1);
         case 'v':
            return(getArgVector(ArgName[n], argtype[1], value));
         default :
            fprintf(stderr, "WARNING: argument found; bad argument type\n");
            return(0);
         }
      }
   return(0);
}



getArgVector(name, type, value)
int  *value;
char *name, type;
{
   double *dbl, atof();
   float  *flt;
   int n = 0;
   switch((int) type) {
      case 'd':
         if (!(*name)) return(0);
         if (*name) value[n++] = atoi(name);
         while(*name) 
            if (*(name++) == (char) ',') 
               if (*name) value[n++] = atoi(name);
         break;
      case 'f':
         if (!(*name)) return(0);
         flt = (float *) value;
         if (*name) flt[n++] = atof(name);
         while(*name) 
            if (*(name++) == (char) ',') 
               if (*name) flt[n++] = atof(name);
         break;
      case 'F':
         if (!(*name)) return(0);
         dbl = (double *) value;
         if (*name) dbl[n++] = atof(name);
         while(*name) 
            if (*(name++) == (char) ',') 
               if (*name) dbl[n++] = atof(name);
         break;
      default :
         fprintf(stderr, "WARNING: argument found; bad argument type\n");
         break;
      }
   return(n);

}




getArgCmp(argname, arglist)
char *argname, *arglist;
{
   int n = 0;
   char name[STRING_SIZE];
   name[0] = (char) '\0';
   if (argname == NULL) return(1);
   if (arglist == NULL) return(1);
   while(*argname != '\0') {
      while(*argname != (char) ',' && *argname != (char) '\0') 
         name[n++] = *argname++;
      name[n] = (char) '\0';
      if (strcmp(name, arglist) == 0) return(0);
      if (*argname != (char) '\0') argname++;
      n = 0;
      }
   return(1);
}




/******************************************************************************
   mstArg: get required argument from list

   usage: argCondition = mstArg(argname, argtype, value)

   argCondition: 1 - argument exists
                 0 - argument not found (PROGRAM EXITS)

   argname: string name of argument
   argtype: argument type ("d" = int; "f" = float; "F" = double; "s" = string)
   value:   value of argument
******************************************************************************/
mstArg(argname, argtype, value)
char *argname, *argtype;
int *value;
{
   int n;
   n = getArg(argname, argtype, value);
   if (n == 0) {
      fprintf(stderr, "ERROR: required argument \"%s\" not found in \"%s\"\n", 
      argname, ArgLine);
      exit(-1);
      }
   return(n);
}




/******************************************************************************
   endArg: clear line argument list 

   usage: endArg()
******************************************************************************/
endArg()
{
   /* clear previous values */
   if (nLineArg) {
      while(nLineArg) {
         nLineArg--;
         free(ArgList[nLineArg]);
         free(ArgName[nLineArg]);
         }
      free(ArgList);
      free(ArgName);
      }
   ArgLine[0] = (char) '\0';
}



/******************************************************************************
   getType: return line type value

   usage: lineTypeValue = getType(lineType)

   lineTypeValue: LINE_COMMENT - hard comment line (#)
                  LINE_BLANK   - blank line
                  LINE_NULL    - null line
                  LINE_PAR     - parameter line (par=value)
                  LINE_VALUE   - line begins with numeric value
                  LINE_UNKNOWN - unknown line type

   lineType: line definition (string) ... "COMMENT", "BLANK", ...
******************************************************************************/
getType(type)
char *type;
{
   int value;
   if      (!strcmp("COMMENT", type)) return(LINE_COMMENT);
   else if (!strcmp("BLANK",   type)) return(LINE_BLANK);
   else if (!strcmp("NULL",    type)) return(LINE_NULL);
   else if (!strcmp("PAR",     type)) return(LINE_PAR);
   else if (!strcmp("VALUE",   type)) return(LINE_VALUE);
   else if (!strcmp("UNKNOWN", type)) return(LINE_UNKNOWN);
   else
      return(0);
}
