/******************************************************************************
   source: source terms for 3D finite difference

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   compile: cc -g -c source

   note: input units are km, s, km/s; g/cm3; dyne-cm;
   note: used  units are mks (m, s, m/s, kg/m3; N-m)

   author: shawn larsen
   date  : august 1989 
   where : caltech
*****************************************************************************/

#include "elas.h"
#include "sacutil.h"
#include <fcntl.h>

addSource(par, grid, data, source, t)
Parameter       *par;
FdGrid         *grid;
VelocityStress *data;
FdSource     *source;
int                t;
{
   int i, g;
   if (grid->d == 2) for(g = 0; g < par->g; g++) for(i = 0; i < par->f; i++) {
      addSource2D(par, grid + g, data + g, source + i, t);
      }
   if (grid->d == 3) for(g = 0; g < par->g; g++) for(i = 0; i < par->f; i++) {
      addSource3D(par, grid + g, data + g, source + i, t);
      }
}




addSource2D(par, grid, data, source, t)
Parameter       *par;
FdGrid         *grid;
VelocityStress *data;
FdSource     *source;
int                t;
{
   float M, F, f, ft, dx, dy, dz, ds, dv, dt, irho, dy2d;
   int l, m, n;

   /* determine source location */
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;
   l  = 0;
   n  = (int) floor((source->x - grid->x0)/dx);
   m  = (int) floor((source->z - grid->z0)/dz);
   if (n < -1 || n > grid->n) return(0);
   if (m < -1 || m > grid->m) return(0);

   if ((f = source->f[t]) == 0.0) return(0);
   M  = f;
   F  = f;
   dt = par->dt;
   ds = dx   *dz;
   dv = dx*dy*dz;
   dy2d = dy/(KMTOM); /* force-density in y direction because of line source */
   irho = grid->irho[l][m][n];
   switch(source->type) {
      case PSOURCE:
           data->Vx[l  ][m  ][n  ] += dt*(((-M            /dx)/dv)*dy2d)*irho;
         data->Vx[l  ][m  ][n+1] += dt*((( M            /dx)/dv)*dy2d)*irho;
           data->Vz[l  ][m  ][n  ] += dt*((( M            /dz)/dv)*dy2d)*irho;
         data->Vz[l  ][m-1][n  ] += dt*(((-M            /dz)/dv)*dy2d)*irho;
         break;
      case MSOURCE:
         data->Vx[l  ][m  ][n  ] += dt*(((-M*source->Mxx/dx)/dv)*dy2d)*irho;
         data->Vx[l  ][m  ][n+1] += dt*((( M*source->Mxx/dx)/dv)*dy2d)*irho;
         data->Vz[l  ][m-1][n  ] += dt*(((-M*source->Mzz/dz)/dv)*dy2d)*irho;
         data->Vz[l  ][m  ][n  ] += dt*((( M*source->Mzz/dz)/dv)*dy2d)*irho;
         data->Vx[l  ][m-1][n  ] += dt*(((-M*source->Mxz/dx)/dv)*dy2d)*irho/4.;
         data->Vx[l  ][m-1][n+1] += dt*(((-M*source->Mxz/dx)/dv)*dy2d)*irho/4.; 
         data->Vx[l  ][m+1][n  ] += dt*((( M*source->Mxz/dx)/dv)*dy2d)*irho/4.;
         data->Vx[l  ][m+1][n+1] += dt*((( M*source->Mxz/dx)/dv)*dy2d)*irho/4.;
         data->Vz[l  ][m-1][n-1] += dt*(((-M*source->Mxz/dz)/dv)*dy2d)*irho/4.;
         data->Vz[l  ][m  ][n-1] += dt*(((-M*source->Mxz/dz)/dv)*dy2d)*irho/4.;
         data->Vz[l  ][m-1][n+1] += dt*((( M*source->Mxz/dz)/dv)*dy2d)*irho/4.;
         data->Vz[l  ][m  ][n+1] += dt*((( M*source->Mxz/dz)/dv)*dy2d)*irho/4.;
         break;
      case FSOURCE:
         data->Vx[l  ][m  ][n  ] += dt*((( F*source->Fx    )/dv)*dy2d)*irho/2.;
         data->Vx[l  ][m  ][n+1] += dt*((( F*source->Fx    )/dv)*dy2d)*irho/2.;
         data->Vz[l  ][m-1][n  ] += dt*((( F*source->Fz    )/dv)*dy2d)*irho/2.;
         data->Vz[l  ][m  ][n  ] += dt*((( F*source->Fz    )/dv)*dy2d)*irho/2.;
         break;
      case SSOURCE:
           data->Vx[l  ][m  ][n  ] += dt*(((-M            /dx)/dv)*dy2d)*irho;
         data->Vx[l  ][m+1][n  ] += dt*((( M            /dx)/dv)*dy2d)*irho;
           data->Vz[l  ][m  ][n  ] += dt*(((-M            /dz)/dv)*dy2d)*irho;
         data->Vz[l  ][m  ][n-1] += dt*((( M            /dz)/dv)*dy2d)*irho;
         break;
      case TSOURCE:
         data->Txx[l ][m  ][n  ] += f;
         data->Tzz[l ][m  ][n  ] += f;
         fprintf(stderr, "Error: TSOURCE non-functional\n");
         exit(-1);
         break;
      default:
         fprintf(stderr, "ERROR: unknown source type %d\n", source->type);
         exit(-1);
         break;
      }

}




addSource3D(par, grid, data, source, t)
Parameter       *par;
FdGrid         *grid;
VelocityStress *data;
FdSource     *source;
int                t;
{
   float M, F, f, ft, dx, dy, dz, dv, dt, irho;
   int l, m, n;

   /* determine source location */
   dx = grid->dx;
   dy = grid->dy;
   dz = grid->dz;
   n  = (int) floor((source->x - grid->x0)/dx);
   m  = (int) floor((source->z - grid->z0)/dz);
   l  = (int) floor((source->y - grid->y0)/dy);
   if (n < -1 || n > grid->n) return(0);
   if (m < -1 || m > grid->m) return(0);
   if (l < -1 || l > grid->l) return(0);

   if ((f = source->f[t]) == 0.0) return(0);
   M  = f;
   F  = f;
   dt = par->dt;
   dv = dx*dy*dz;
   irho = grid->irho[l][m][n];
   switch(source->type) {
      case PSOURCE:
           data->Vx[l  ][m  ][n  ] += dt*((-M            /dx)/dv)*irho;
         data->Vx[l  ][m  ][n+1] += dt*(( M            /dx)/dv)*irho;
           data->Vy[l-1][m  ][n  ] += dt*((-M            /dy)/dv)*irho;
         data->Vy[l  ][m  ][n  ] += dt*(( M            /dy)/dv)*irho;
           data->Vz[l  ][m  ][n  ] += dt*(( M            /dz)/dv)*irho;
         data->Vz[l  ][m-1][n  ] += dt*((-M            /dz)/dv)*irho;
         break;
      case MSOURCE:
         data->Vx[l  ][m  ][n  ] += dt*((-M*source->Mxx/dx)/dv)*irho;
         data->Vx[l  ][m  ][n+1] += dt*(( M*source->Mxx/dx)/dv)*irho;
         data->Vy[l-1][m  ][n  ] += dt*((-M*source->Myy/dy)/dv)*irho;
         data->Vy[l  ][m  ][n  ] += dt*(( M*source->Myy/dy)/dv)*irho;
         data->Vz[l  ][m-1][n  ] += dt*((-M*source->Mzz/dz)/dv)*irho;
         data->Vz[l  ][m  ][n  ] += dt*(( M*source->Mzz/dz)/dv)*irho;
         data->Vx[l+1][m  ][n  ] += dt*(( M*source->Mxy/dx)/dv)*irho/4.;
         data->Vx[l+1][m  ][n+1] += dt*(( M*source->Mxy/dx)/dv)*irho/4.; 
         data->Vx[l-1][m  ][n  ] += dt*((-M*source->Mxy/dx)/dv)*irho/4.;
         data->Vx[l-1][m  ][n+1] += dt*((-M*source->Mxy/dx)/dv)*irho/4.;
         data->Vy[l  ][m  ][n-1] += dt*((-M*source->Mxy/dy)/dv)*irho/4.;
         data->Vy[l-1][m  ][n-1] += dt*((-M*source->Mxy/dy)/dv)*irho/4.;
         data->Vy[l  ][m  ][n+1] += dt*(( M*source->Mxy/dy)/dv)*irho/4.;
         data->Vy[l-1][m  ][n+1] += dt*(( M*source->Mxy/dy)/dv)*irho/4.;
         data->Vx[l  ][m-1][n  ] += dt*((-M*source->Mxz/dx)/dv)*irho/4.;
         data->Vx[l  ][m-1][n+1] += dt*((-M*source->Mxz/dx)/dv)*irho/4.; 
         data->Vx[l  ][m+1][n  ] += dt*(( M*source->Mxz/dx)/dv)*irho/4.;
         data->Vx[l  ][m+1][n+1] += dt*(( M*source->Mxz/dx)/dv)*irho/4.;
         data->Vz[l  ][m-1][n-1] += dt*((-M*source->Mxz/dz)/dv)*irho/4.;
         data->Vz[l  ][m  ][n-1] += dt*((-M*source->Mxz/dz)/dv)*irho/4.;
         data->Vz[l  ][m-1][n+1] += dt*(( M*source->Mxz/dz)/dv)*irho/4.;
         data->Vz[l  ][m  ][n+1] += dt*(( M*source->Mxz/dz)/dv)*irho/4.;
         data->Vy[l-1][m-1][n  ] += dt*((-M*source->Myz/dy)/dv)*irho/4.;
         data->Vy[l  ][m-1][n  ] += dt*((-M*source->Myz/dy)/dv)*irho/4.; 
         data->Vy[l-1][m+1][n  ] += dt*(( M*source->Myz/dy)/dv)*irho/4.;
         data->Vy[l  ][m+1][n  ] += dt*(( M*source->Myz/dy)/dv)*irho/4.;
         data->Vz[l-1][m-1][n  ] += dt*((-M*source->Myz/dz)/dv)*irho/4.;
         data->Vz[l-1][m  ][n  ] += dt*((-M*source->Myz/dz)/dv)*irho/4.;
         data->Vz[l+1][m-1][n  ] += dt*(( M*source->Myz/dz)/dv)*irho/4.;
         data->Vz[l+1][m  ][n  ] += dt*(( M*source->Myz/dz)/dv)*irho/4.;
         break;
      case FSOURCE:
         data->Vx[l  ][m  ][n  ] += dt*(( F*source->Fx    )/dv)*irho/2.;
         data->Vx[l  ][m  ][n+1] += dt*(( F*source->Fx    )/dv)*irho/2.;
         data->Vy[l-1][m  ][n  ] += dt*(( F*source->Fy    )/dv)*irho/2.;
         data->Vy[l  ][m  ][n  ] += dt*(( F*source->Fy    )/dv)*irho/2.;
         data->Vz[l  ][m-1][n  ] += dt*(( F*source->Fz    )/dv)*irho/2.;
         data->Vz[l  ][m  ][n  ] += dt*(( F*source->Fz    )/dv)*irho/2.;
         break;
      case SSOURCE:
           data->Vx[l  ][m  ][n  ] += dt*((-M            /dx)/dv)*irho;
         data->Vx[l  ][m+1][n  ] += dt*(( M            /dx)/dv)*irho;
           data->Vz[l  ][m  ][n  ] += dt*((-M            /dz)/dv)*irho;
         data->Vz[l  ][m  ][n-1] += dt*(( M            /dz)/dv)*irho;
         fprintf(stderr, "Error: SSOURCE non-functional in 3D\n");
         exit(-1);
         break;
      case TSOURCE:
         data->Txx[l ][m  ][n  ] += f;
         data->Tyy[l ][m  ][n  ] += f;
         data->Tzz[l ][m  ][n  ] += f;
         fprintf(stderr, "Error: TSOURCE non-functional\n");
         exit(-1);
         break;
      default:
         fprintf(stderr, "ERROR: unknown source type %d\n", source->type);
         exit(-1);
         break;
      }

}




getSourceF(par, source)
Parameter *par;
FdSource  *source;
{
   int i;
   for(i = 0; i < par->f; i++) getSource(par, source + i);
}




getSource(par, source)
Parameter *par;
FdSource  *source;
{
   float f, ft, dt, t0, factor;
   int file, t;
   char header[SACSIZE];

   /* compute source time function by reading sac file */ 
   if (source->file[0]) {
      if ((file = open(source->file, O_RDONLY)) == -1) {
         fprintf(stderr, "ERROR: Can't open source file \"%s\"\n",source->file);
         exit(-1);
         };
      for(t = 0; t < source->t; t++) source->f[t] = 0.;
      read(file, header, SACSIZE);
      dt = *((float *) &header[WORDSIZE*DDELTA]);
      t  = *((int   *) &header[WORDSIZE*DNPTS ]);
      if (dt != par->dt) {
         fprintf(stderr, "ERROR: source file dt not compatible\n");
         exit(-1);
         }
      /*
      if (t != par->t) {
         fprintf(stderr, "ERROR: source file npts not compatible\n");
         exit(-1);
         }
      */
      read(file, source->f, ((source->t < t) ? (source->t):(t))*sizeof(float));
      for(t = 0; t < source->t; t++) source->f[t] *= source->amp;
      close(file);
      }
   else {
      t0 = 1./source->freq + source->t0;
      for(t = 0; t < source->t; t++) {
         ft = par->dt*((float) t);
         factor = ((PI)*source->freq*(ft - t0))*((PI)*source->freq*(ft - t0));
         f = source->amp*(2.*factor - 1.)*exp(-factor);
         source->f[t] = f;
         }
      }
   
}



/******************************************************************************
   getMoment3D: get moment tensor components from fault orientation definitions

   usage: void = getMoment(source)

   note: follows aki and richards convention (P117) except that X and Y
         axes are interchanged. In this convention x is east, y is north,
         and z is down. Strike is clockwise from north (0 <= strike < 360); 
         dip is measured down from the right of the strike (0 <= dip <= 90); 
         rake is the angle between the slip and strike directions, where
         the slip is the direction of the hanging wall relative to the 
         foot wall (a positive rake is a reverse fault, a negative rake
         is a normal fault). A stike of 0., a dip of 90., and a rake of 0
         corresponds to a north-south left-lateral fault or an east-west
         right-lateral fault. 

   note: input units are km, s, km/s; g/cm3; dyne-cm;
   note: used  units are mks (m, s, m/s, kg/m3; N-m)
*****************************************************************************/
getMoment3D(source)
FdSource  *source;
{
   double cos(), sin(), Mxx, Myy, Mzz, Mxy, Mxz, Myz, M0, S, D, R;
   double sin1S, sin2S, sin1D, sin2D, sin1R, cos1S, cos2S, cos1D, cos2D, cos1R;

   /* initialize */
   S  = (double) source->strike/(RAD);
   D  = (double) source->dip   /(RAD);
   R  = (double) source->rake  /(RAD);
   M0 = (double) source->M0;

   /* calculate terms */
   sin1S = sin(   S); sin1D = sin(   D); sin1R = sin(   R);
   cos1S = cos(   S); cos1D = cos(   D); cos1R = cos(   R);
   sin2S = sin(2.*S); sin2D = sin(2.*D);
   cos2S = cos(2.*S); cos2D = cos(2.*D);

   /* calculate moments */
   Mxx =  M0*(sin1D*cos1R*sin2S -     sin2D*sin1R*cos1S*cos1S);
   Mxy =  M0*(sin1D*cos1R*cos2S + 0.5*sin2D*sin1R*sin2S      );
   Mxz = -M0*(cos1D*cos1R*sin1S -     cos2D*sin1R*cos1S      );
   Myy = -M0*(sin1D*cos1R*sin2S +     sin2D*sin1R*sin1S*sin1S);
   Myz = -M0*(cos1D*cos1R*cos1S +     cos2D*sin1R*sin1S      );
   Mzz =  M0*(sin2D*sin1R);

   /* put back into source structure */
   source->Mxx = (float) Mxx; 
   source->Mxy = (float) Mxy; 
   source->Mxz = (float) Mxz; 
   source->Myy = (float) Myy; 
   source->Myz = (float) Myz; 
   source->Mzz = (float) Mzz; 
   
}
