/******************************************************************************
   stencil2d: routines for elastic 2-D finite-difference stencils (Levandar)

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   update4Vx2d     : 4th order (spatial) velocity update in x direction
   update4Vz2d     : 4th order (spatial) velocity update in z direction
   update4Txx2d    : 4th order (spatial) stress   update (normal x)
   update4Tzz2d    : 4th order (spatial) stress   update (normal z)
   update4Txz2d    : 4th order (spatial) stress   update (shear   )
   update2Vx2d     : 2th order (spatial) velocity update in x direction
   update2Vz2d     : 2th order (spatial) velocity update in z direction
   update2Txx2d    : 2th order (spatial) stress   update (normal x)
   update2Tzz2d    : 2th order (spatial) stress   update (normal z)
   update2Txz2d    : 2th order (spatial) stress   update (shear   )
   getP2D          : get P potential from Vx and Vz (return float)
   getS2D          : get S potential from Vx and Vz (return float)

   par  - input parameters
   grid - input parameters for grid
   Vx   - velocity in x direction
   Vz   - velocity in z direction
   Txx  - stress (normal x)
   Tzz  - stress (normal z)
   Txz  - stress (shear   )
   l    - grid point in y direction (plane ) (positive to back)  
   m    - grid point in z direction (row   ) (positive down)
   n    - grid point in x direction (column)
   getP - P potential at point
   getS - S potential at point

   compile: cc -g -c stencil2d.c

   note: used units are mks (m, s, m/s, kg/m3)
   note: assumes 3d interface
   note: there are 3d stencils commented out at end of file

   author: shawn larsen
   date  : march 1989
   where : caltech
*****************************************************************************/

#include "elas.h"



update4Vx2d(par, grid, Vx, Txx, Txz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Txx, **Txz;
int m, n;
{
   float irho;
   float rho;
   irho = grid->irho[0][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[0][m][n] + grid->rho[0][m][n-1]);
      irho = 1./rho;
      }
   Vx[m][n]  = par->dt*((C1*(Txx[m  ][n  ] - Txx[m  ][n-1])                + 
                         C2*(Txx[m  ][n+1] - Txx[m  ][n-2]))/grid->dx      +
                        (C1*(Txz[m  ][n  ] - Txz[m-1][n  ])                +
                         C2*(Txz[m+1][n  ] - Txz[m-2][n  ]))/grid->dz)*irho +
                         Vx[m][n];
}



update4Vz2d(par, grid, Vz, Txz, Tzz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vz, **Txz, **Tzz;
int m, n;
{
   float irho;
   float rho;
   irho = grid->irho[0][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[0][m][n] + grid->rho[0][m+1][n]);
      irho = 1./rho;
      }
   Vz[m][n]  = par->dt*((C1*(Txz[m  ][n+1] - Txz[m  ][n  ])                + 
                         C2*(Txz[m  ][n+2] - Txz[m  ][n-1]))/grid->dx      +
                        (C1*(Tzz[m+1][n  ] - Tzz[m  ][n  ])                +
                         C2*(Tzz[m+2][n  ] - Tzz[m-1][n  ]))/grid->dz)*irho +
                         Vz[m][n];
}



update4Txx2d(par, grid, Txx, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Txx;
int m, n;
{
   float lamda, mu;
   lamda = grid->lamda[0][m][n];
   mu    = grid->mu   [0][m][n];
   Txx[m][n] = par->dt*
               ((C1*(Vx[m  ][n+1] - Vx[m  ][n  ])                            +
                 C2*(Vx[m  ][n+2] - Vx[m  ][n-1]))*(lamda + 2.*mu)/grid->dx  +
                (C1*(Vz[m  ][n  ] - Vz[m-1][n  ])                            +
                 C2*(Vz[m+1][n  ] - Vz[m-2][n  ]))*(lamda        )/grid->dz) +
                 Txx[m][n];
}



update4Tzz2d(par, grid, Tzz, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Tzz;
int m, n;
{
   float lamda, mu;
   lamda = grid->lamda[0][m][n];
   mu    = grid->mu   [0][m][n];
   Tzz[m][n] = par->dt*
               ((C1*(Vx[m  ][n+1] - Vx[m  ][n  ])                            +
                 C2*(Vx[m  ][n+2] - Vx[m  ][n-1]))*(lamda        )/grid->dx  +
                (C1*(Vz[m  ][n  ] - Vz[m-1][n  ])                            +
                 C2*(Vz[m+1][n  ] - Vz[m-2][n  ]))*(lamda + 2.*mu)/grid->dz) +
                 Tzz[m][n];
}



update4Txz2d(par, grid, Txz, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Txz;
int m, n;
{
   float mu;
   mu    = grid->mu[0][m][n];
   Txz[m][n] = par->dt*
               ((C1*(Vx[m+1][n  ] - Vx[m  ][n  ])               +
                 C2*(Vx[m+2][n  ] - Vx[m-1][n  ]))/grid->dz     +
                (C1*(Vz[m  ][n  ] - Vz[m  ][n-1])               +
                 C2*(Vz[m  ][n+1] - Vz[m  ][n-2]))/grid->dx)*mu +
                 Txz[m][n];
}



update2Vx2d(par, grid, Vx, Txx, Txz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Txx, **Txz;
int m, n;
{
   float irho;
   float rho;
   irho = grid->irho[0][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[0][m][n] + grid->rho[0][m][n-1]);
      irho = 1./rho;
      }
   Vx[m][n]  = par->dt*((Txx[m  ][n  ] - Txx[m  ][n-1])/grid->dx      + 
                        (Txz[m  ][n  ] - Txz[m-1][n  ])/grid->dz)*irho +
                         Vx[m][n];
}



update2Vz2d(par, grid, Vz, Txz, Tzz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vz, **Txz, **Tzz;
int m, n;
{
   float irho;
   float rho;
   irho = grid->irho[0][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[0][m][n] + grid->rho[0][m+1][n]);
      irho = 1./rho;
      }
   Vz[m][n]  = par->dt*((Txz[m  ][n+1] - Txz[m  ][n  ])/grid->dx      + 
                        (Tzz[m+1][n  ] - Tzz[m  ][n  ])/grid->dz)*irho +
                         Vz[m][n];
}



update2Txx2d(par, grid, Txx, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Txx;
int m, n;
{
   float lamda, mu;
   lamda = grid->lamda[0][m][n];
   mu    = grid->mu   [0][m][n];
   Txx[m][n] = par->dt*
               ((Vx[m  ][n+1] - Vx[m  ][n  ])*(lamda + 2.* mu)/grid->dx  +
                (Vz[m  ][n  ] - Vz[m-1][n  ])*(lamda         )/grid->dz) +
                 Txx[m][n];
}



update2Tzz2d(par, grid, Tzz, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Tzz;
int m, n;
{
   float lamda, mu;
   lamda = grid->lamda[0][m][n];
   mu    = grid->mu   [0][m][n];
   Tzz[m][n] = par->dt*
               ((Vx[m  ][n+1] - Vx[m  ][n  ])*(lamda        )/grid->dx  +
                (Vz[m  ][n  ] - Vz[m-1][n  ])*(lamda + 2.*mu)/grid->dz) +
                 Tzz[m][n];
}



update2Txz2d(par, grid, Txz, Vx, Vz, m, n)
Parameter *par;
FdGrid    *grid;
float **Vx, **Vz, **Txz;
int m, n;
{
   float mu;
   mu    = grid->mu[0][m][n];
   Txz[m][n] = par->dt*
               ((Vx[m+1][n  ] - Vx[m  ][n  ])/grid->dz     +
                (Vz[m  ][n  ] - Vz[m  ][n-1])/grid->dx)*mu +
                 Txz[m][n];
}




float
getP2D(Vx, Vz, dx, dz, m, n)
float **Vx, **Vz, dx, dz;
int m, n;
{
   float dVxdx, dVzdz;
   /*
   dVxdx = (Vx[m][n+1] - Vx[m  ][n-1])/(2.*dx);
   dVzdz = (Vz[m][n  ] - Vz[m-1][n  ] +
            Vz[m][n-1] - Vz[m-1][n-1])/(2.*dz);
   */
   dVxdx = (Vx[m][n+1] - Vx[m  ][n])/dx;
   dVzdz = (Vz[m][n  ] - Vz[m-1][n])/dz;
   return(dVxdx + dVzdz);
}
 
 
 
float
getS2D(Vx, Vz, dx, dz, m, n)
float **Vx, **Vz, dx, dz;
int m, n;
{
   float dVxdz, dVzdx;
   /*
   dVxdz = (Vx[m+1][n] - Vx[m-1][n  ])/(2.*dz);
   dVzdx = (Vz[m  ][n] - Vz[m  ][n-1] +
            Vz[m-1][n] - Vz[m-1][n-1])/(2.*dx);
   */
   dVxdz = ((Vx[m+1][n] - Vx[m-1][n]) + (Vx[m+1][n+1] - Vx[m-1][n+1]))/(4.*dz);
   dVzdx = ((Vz[m][n+1] - Vz[m][n-1]) + (Vz[m-1][n+1] - Vz[m-1][n-1]))/(4.*dx);
   return(0.5*(dVxdz - dVzdx));
}

