/****************************************************************
   stencil3d:  routines for elastic 3-D finite difference stencils

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   update2Vx3d    : 2nd order (spatial) velocity update in x direction
   update2Vy3d    : 2nd order (spatial) velocity update in y direction
   update2Vz3d    : 2nd order (spatial) velocity update in z direction
   update4Vx3d    : 4th order (spatial) velocity update in x direction
   update4Vy3d    : 2nd order (spatial) velocity update in y direction
   update4Vz3d    : 4th order (spatial) velocity update in z direction
   update2Txx3d   : 2nd order (spatial) stress update (normal x)
   update2Tyy3d   : 2nd order (spatial) stress update (normal y)
   update2Tzz3d   : 2nd order (spatial) stress update (normal z)
   update2Txy3d   : 2nd order (spatial) stress update (shear   )
   update2Txz3d   : 2nd order (spatial) stress update (shear   )
   update2Tyz3d   : 2nd order (spatial) stress update (shear   )
   update4Txx3d   : 4th order (spatial) stress update (normal x)
   update4Tyy3d   : 4th order (spatial) stress update (normal y)
   update4Tzz3d   : 4th order (spatial) stress update (normal z)
   getP3D         : get P potential from Vx, Vy, and Vz (return float)
   getS3D         : get S potential from Vx, Vy, and Vz (return float)


   par - input parameters
   grid - input parameters for grid
   Vx   - velocity in x direction
   Vy   - velocity in y direction
   Vz   - velocity in z direction
   Txx  - stress (normal x)
   Txy  - stress (shear)
   Txz  - stress (shear)
   Tyy  - stress (normal)
   Tyz  - stress (shear)
   Tzz  - stress (normal)
   l    - grid point in y direction 
   m    - grid point in z direction (row) (positive down)
   n    - grid point in x direction (column)
   
   compile:  cc -g -c stencil3d.c
   
   note: used units are mks (m, s, m/s, kg/m3)

   author: shawn larsen
   date:   june 1989
   where:  caltech

   author: denise maddix (modified)
   date:   11 January 1995
   where:  llnl

   modified: shawn larsen
***************************************************************************/

#include "elas.h"


update2Vx3d(par, grid, Vx, Txx, Txy, Txz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vx, ***Txx, ***Txy, ***Txz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l][m][n-1]);
      irho = 1./rho;
      }
   /* Update Velocity in x direction using 2nd order (spatial) diff scheme */
   Vx[l][m][n] = par->dt*((Txx[l][m][n] - Txx[l  ][m  ][n-1])/grid->dx +
                          (Txy[l][m][n] - Txy[l-1][m  ][n  ])/grid->dy +
                          (Txz[l][m][n] - Txz[l  ][m-1][n  ])/grid->dz)*irho +
                            Vx[l][m][n];
}


update2Vy3d(par,grid, Vy, Txy, Tyy, Tyz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vy, ***Txy, ***Tyy, ***Tyz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l+1][m][n]);
      irho = 1./rho;
      }
   /* Update Velocity in y direction using 2nd order difference scheme */
   Vy[l][m][n] = par->dt*((Txy[l  ][m][n+1] - Txy[l][m  ][n])/grid->dx +
                          (Tyy[l+1][m][n  ] - Tyy[l][m  ][n])/grid->dy +
                          (Tyz[l  ][m][n  ] - Tyz[l][m-1][n])/grid->dz)*irho +
                            Vy[l  ][m][n  ];
}


update2Vz3d(par,grid, Vz, Txz, Tyz, Tzz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vz, ***Txz, ***Tyz, ***Tzz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l][m+1][n]);
      irho = 1./rho;
      }
   /* Update Velocity in z direction using 2nd order difference scheme */
   Vz[l][m][n] = par->dt*((Txz[l][m  ][n+1] - Txz[l  ][m][n])/grid->dx +
                          (Tyz[l][m  ][n  ] - Tyz[l-1][m][n])/grid->dy +
                          (Tzz[l][m+1][n  ] - Tzz[l  ][m][n])/grid->dz)*irho +
                            Vz[l][m  ][n  ];
}

 
/* Update normal stress direction using 2nd order difference scheme */
update2T3d(par, grid, T, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***T, ***Vx, ***Vy, ***Vz;
int l, m, n;
{
   T[l][m][n] += par->dt*grid->lamda[l][m][n]/grid->dx*(
                 Vx[l  ][m  ][n+1] - Vx[l  ][m  ][n  ] + 
                 Vy[l  ][m  ][n  ] - Vy[l-1][m  ][n  ] + 
                 Vz[l  ][m  ][n  ] - Vz[l  ][m-1][n  ]);
}

update2Txx3d(par,grid, Txx, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Txx, ***Vx, ***Vy, ***Vz;
int l, m, n;

{
  float lamda, mu;
  lamda = grid-> lamda[l][m][n];
  mu    = grid-> mu[l][m][n];
/* Update normal stress in x direction using 2nd order difference scheme */
  Txx[l][m][n] = par->dt*((Vx[l][m][n+1] - Vx[l][m][n])*(lamda + 2.*mu)/
                                                             grid->dx
                   +lamda*((Vy[l][m][n]- Vy[l-1][m][n])/grid->dy 
                           + (Vz[l][m][n]-Vz[l][m-1][n])/grid->dz)) 
                   + Txx[l][m][n];
}

update2Tyy3d(par,grid, Tyy, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Tyy, ***Vx, ***Vy, ***Vz;
int l, m, n;

{
  float lamda, mu;  
  lamda = grid-> lamda[l][m][n];
  mu    = grid-> mu[l][m][n];
 /* Update normal stress in y direction using 2nd order difference scheme */
  Tyy[l][m][n] = par->dt*(((Vx[l][m][n+1]- Vx[l][m][n])/grid->dx + 
                            (Vz[l][m][n] - Vz[l][m-1][n])/grid->dz)*lamda
                      + (Vy[l][m][n] - Vy[l-1][m][n])*(lamda+2.*mu)/grid->dy)
                      + Tyy[l][m][n];

}

update2Tzz3d(par,grid, Tzz, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Tzz, ***Vx, ***Vy, ***Vz;
int l, m,n;

{
  float lamda, mu;
  lamda = grid-> lamda[l][m][n];
  mu    = grid->   mu[l][m][n];
 /* Update normal stress in z direction using 2nd order difference scheme */
  Tzz[l][m][n] = par->dt*(((Vx[l][m][n+1] - Vx[l][m][n])/grid->dx
                           + (Vy[l][m][n] - Vy[l-1][m][n])/grid->dy)*lamda
                           + (Vz[l][m][n] - Vz[l][m-1][n])*(lamda+2.*mu)/grid->dz)
                           + Tzz[l][m][n];
}

update2Txy3d(par,grid, Txy, Vx, Vy, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Txy, ***Vx, ***Vy;
int l, m, n;

{
  float lamda, mu;
  lamda = grid-> lamda[l][m][n];
  mu =    grid->  mu[l][m][n];
 /* Update shear stress using 2nd order difference scheme */
  Txy[l][m][n] = par->dt*((Vx[l+1][m][n] - Vx[l][m][n])/grid->dy
                          + (Vy[l][m][n] - Vy[l][m][n-1])/grid->dx)*mu
                          + Txy[l][m][n];
}

update2Txz3d(par,grid, Txz, Vx, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Txz, ***Vx, ***Vz;
int l, m, n;

{
  float lamda, mu;
  lamda = grid -> lamda[l][m][n];
  mu    = grid ->mu[l][m][n];
 /* Update shear stress using 2nd order difference scheme */
  Txz[l][m][n] = par->dt*((Vx[l][m+1][n] - Vx[l][m][n])/grid->dz
                          + (Vz[l][m][n] - Vz[l][m][n-1])/grid->dx)*mu
                          + Txz[l][m][n];

}

update2Tyz3d(par,grid, Tyz, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Tyz, ***Vy, ***Vz;
int l, m, n;

{
  float lamda, mu;
  lamda = grid-> lamda[l][m][n];
  mu    = grid-> mu[l][m][n];
 /* Update shear stress using 2nd order difference scheme */
  Tyz[l][m][n] = par->dt*((Vy[l][m+1][n] - Vy[l][m][n])/grid->dz
                          +(Vz[l+1][m][n] - Vz[l][m][n])/grid->dy)*mu
                          + Tyz[l][m][n];
}


update4Vx3d(par,grid, Vx, Txx, Txy, Txz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vx, ***Txx, ***Txy, ***Txz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l][m][n-1]);
      irho = 1./rho;
      }
   /* Update Velocity in x direction using 4th order differencing scheme */
     Vx[l][m][n] = par->dt*((C1*(Txx[l][m][n] - Txx[l][m][n-1])
                      +  C2*(Txx[l][m][n+1] - Txx[l][m][n-2]))/grid->dx
                      +  (C1*(Txy[l][m][n] - Txy[l-1][m][n])
                      +  C2*(Txy[l+1][m][n] - Txy[l-2][m][n]))/grid->dy
                      +  (C1*(Txz[l][m][n] - Txz[l][m-1][n])
                      +  C2*(Txz[l][m+1][n] - Txz[l][m-2][n]))/grid->dz)*irho
                      + Vx[l][m][n];
}

  
update4Vz3d(par,grid, Vz, Txz, Tyz, Tzz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vz, ***Txz, ***Tyz, ***Tzz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l][m+1][n]);
      irho = 1./rho;
      }
   /* Update Velocity in z direction using 4th order differencing scheme */
      Vz[l][m][n] = par->dt*((C1*(Txz[l][m][n+1] - Txz[l][m][n])
                       +   C2*(Txz[l][m][n+2] - Txz[l][m][n-1]))/grid->dx
                       +   (C1*(Tyz[l][m][n] - Tyz[l-1][m][n])
                       +   C2*(Tyz[l+1][m][n] - Tyz[l-2][m][n]))/grid->dy
                       +   (C1*(Tzz[l][m+1][n] - Tzz[l][m][n])
                       +   C2*(Tzz[l][m+2][n] - Tzz[l][m-1][n]))/grid->dz)*irho
                       +   Vz[l][m][n]; 
}


update4Vy3d(par,grid, Vy, Txy, Tyy, Tyz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Vy, ***Txy, ***Tyy, ***Tyz;
int l,m,n;
{
   float irho;
   float  rho;
   irho = grid->irho[l][m][n];
   if (grid->ave) {
      rho = 0.5*(grid->rho[l][m][n] + grid->rho[l+1][m][n]);
      irho = 1./rho;
      }
   /* Update Velocity in y direction using 4th order differencing scheme */
      Vy[l][m][n] = par->dt*((C1*(Txy[l][m][n+1] - Txy[l][m][n])
                       +    C2*(Txy[l][m][n+2] - Txy[l][m][n-1]))/grid->dx
                       +    (C1*(Tyy[l+1][m][n]- Tyy[l][m][n])
                       +    C2*(Tyy[l+2][m][n] - Tyy[l-1][m][n]))/grid->dy
                       +    (C1*(Tyz[l][m][n] - Tyz[l][m-1][n])
                       +     C2*(Tyz[l][m+1][n] - Tyz[l][m-2][n]))/grid->dz)*irho
                       +     Vy[l][m][n];  
                       
       
}


update4Txx3d(par,grid, Txx, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Txx, ***Vx, ***Vy, ***Vz;
int l, m, n;

{

     float lamda, mu;
     lamda = grid-> lamda[l][m][n];
     mu    = grid-> mu[l][m][n];
/*  Update stress in normal x direction using 4th order differencing scheme */
     Txx[l][m][n] = par->dt*((C1*(Vx[l][m][n+1] - Vx[l][m][n]) 
                      + C2*(Vx[l][m][n+2] - Vx[l][m][n-1]))*(lamda +2.*mu)/grid->dx                      + lamda*((C1*(Vy[l][m][n] - Vy[l-1][m][n]) 
                      +      C2*(Vy[l+1][m][n] - Vy[l-2][m][n]))/grid->dy
                      +      (C1*(Vz[l][m][n] - Vz[l][m-1][n])
                      +      C2*(Vz[l][m+1][n] - Vz[l][m-2][n])))/grid->dz)
                      + Txx[l][m][n];
}

update4Tyy3d(par,grid, Tyy, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Tyy, ***Vx, ***Vy, ***Vz;
int l, m, n;

{

     float lamda, mu;
     lamda = grid-> lamda[l][m][n];
     mu    = grid-> mu[l][m][n];
/* Update stress in normal y direction using 4th order differencing scheme */
     Tyy[l][m][n] = par->dt*(((C1*(Vx[l][m][n+1] - Vx[l][m][n])
                          + C2*(Vx[l][m][n+2] - Vx[l][m][n-1]))/grid->dx
                          + (C1*(Vz[l][m][n] - Vz[l][m-1][n])
                          + C2*(Vz[l][m+1][n] - Vz[l][m-2][n]))/grid->dz)*lamda
                          + (lamda +2.*mu)*(C1*(Vy[l][m][n] - Vy[l-1][m][n])
                          +      C2*(Vy[l+1][m][n] - Vy[l-2][m][n]))/grid->dy)
                          + Tyy[l][m][n];
}

update4Tzz3d(par,grid, Tzz, Vx, Vy, Vz, l, m, n)
Parameter *par;
FdGrid *grid;
float ***Tzz, ***Vx, ***Vy, ***Vz;
int l, m, n;


{

     float lamda, mu;
     lamda = grid-> lamda[l][m][n];
     mu    = grid-> mu[l][m][n];
/* Update stress in normal z direction using 4th order differencing scheme */
     Tzz[l][m][n] = par->dt*(((C1*(Vx[l][m][n+1] - Vx[l][m][n])
                          + C2*(Vx[l][m][n+2] - Vx[l][m][n-1]))/grid->dx
                          + (C1*(Vy[l][m][n] - Vy[l-1][m][n])
                          + C2*(Vy[l+1][m][n] - Vy[l-2][m][n]))/grid->dy)*lamda
                          + (lamda +2.*mu)*(C1*(Vz[l][m][n] - Vz[l][m-1][n])
                          +    C2*(Vz[l][m+1][n] - Vz[l][m-2][n]))/grid->dz)
                          + Tzz[l][m][n];
}


float
getP3D(Vx, Vy, Vz, dx, dy, dz, l, m, n)
float ***Vx, ***Vy, ***Vz, dx, dy, dz;
int l, m, n;
{
   float dVxdx, dVydy, dVzdz;
   dVxdx = (Vx[l][m][n+1] - Vx[l  ][m  ][n])/dx;
   dVydy = (Vy[l][m][n  ] - Vy[l-1][m  ][n])/dy;
   dVzdz = (Vz[l][m][n  ] - Vz[l  ][m-1][n])/dz;
   return(dVxdx + dVydy + dVzdz);
}


float
getS3D(Vx, Vy, Vz, dx, dy, dz, l, m, n)
float ***Vx, ***Vy, ***Vz, dx, dy, dz;
int l, m, n;
{
   double sqrt();
   float dVxdy, dVxdz, dVydx, dVydz, dVzdx, dVzdy;
   float f, s, s1, s2, s3;
   f = (1./(4.*dx));
   dVxdy = ((Vx[l+1][m][n]-Vx[l-1][m][n])+(Vx[l+1][m][n+1]-Vx[l-1][m][n+1]))*f;
   dVxdz = ((Vx[l][m+1][n]-Vx[l][m-1][n])+(Vx[l][m+1][n+1]-Vx[l][m-1][n+1]))*f;
   dVydx = ((Vy[l][m][n+1]-Vy[l][m][n-1])+(Vy[l-1][m][n+1]-Vy[l-1][m][n-1]))*f;
   dVydz = ((Vy[l][m+1][n]-Vy[l][m-1][n])+(Vy[l-1][m+1][n]-Vy[l-1][m-1][n]))*f;
   dVzdy = ((Vz[l+1][m][n]-Vz[l-1][m][n])+(Vz[l+1][m-1][n]-Vz[l-1][m-1][n]))*f;
   dVzdx = ((Vz[l][m][n+1]-Vz[l][m][n-1])+(Vz[l][m-1][n+1]-Vz[l][m-1][n-1]))*f;
   s1 = dVzdy - dVydz;
   s2 = dVxdz - dVzdx;
   s3 = dVydx - dVxdy;
   /*s  = sqrt(s1*s1 + s2*s2 + s3*s3);*/
   s  = s1 + s2 + s3;
/*
s = -dVzdy + dVydz + dVxdz - dVzdx;
s = sqrt(s1*s1 + s2*s2 + s3*s3);
*/
   return(0.5*s);
}

