/******************************************************************************
   surface: surface conditions for 2D/3D elastic finite-difference

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   surface2VT : 2D free-surface velocity boundary condition (top)
   surface2TT : 2D free-surface   stress boundary condition (top)
   surface3VT : 3D free-surface velocity boundary condition (top)
   surface3TT : 3D free-surface   stress boundary condition (top) 
   surface3VTv: 3D free-surface velocity boundary condition (top) (vectorized)
   surface3TTv: 3D free-surface   stress boundary condition (top) (vectorized)
   surface3VTa: 3D free-surface velocity boundary condition (top) (acoustic  )
   surface3TTa: 3D free-surface   stress boundary condition (top) (acoustic  )

   aboveSurfaceVx2d: 2D velocity update above free-surface (x direction)
   aboveSurfaceVz2d: 2D velocity update above free-surface (z direction)
   aboveSurfaceVx3d: 3D velocity update above free-surface (x direction)
   aboveSurfaceVy3d: 3D velocity update above free-surface (y direction)
   aboveSurfaceVz3d: 3D velocity update above free-surface (z direction)

   compile: cc -g -c surface.c 

   note: input units are km, s, km/s; g/cm3

   author: shawn larsen
   date  : march 1991
   where : caltech
*****************************************************************************/

#include "elas.h"


surface2VT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = -1;
   /**/
   for(n = data->nmin; n < data->nmax; n++) {
      update2Vx2d(par, grid, data->Vx[0], data->Txx[0], data->Txz[0], 0, n);
      update2Vz2d(par, grid, data->Vz[0], data->Txz[0], data->Tzz[0], 0, n);
      }
   /* order is important */
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 1);
   for(n = data->nmin; n < data->nmax; n++) 
      aboveSurfaceVz2d(par, grid, data->Vx[0], data->Vz[0], m, n);
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 3);
   for(n = data->nmin; n < data->nmax; n++) 
      aboveSurfaceVx2d(par, grid, data->Vx[0], data->Vz[0], m, n);
   /**/
   /*** old stuff
   for(n = data->nmin; n < data->nmax; n++) {
      update2Vxd(par, grid, data->Vx, data->Txx, data->Txz, m, n);
      update2Vzd(par, grid, data->Vz, data->Txz, data->Tzz, m, n);
      update2Vzd(par, grid, data->Vz, data->Txz, data->Tzz, m - 1, n);
      ***/
      /*
      aboveSurfaceVz2d(par, grid, data->Vx[0], data->Vz[0], m, n);
      aboveSurfaceVx2d(par, grid, data->Vx[0], data->Vz[0], m, n);
      }
      */
}



surface2TT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int n, m;
   m = -1;
   /**/
   for(n = data->nmin; n < data->nmax; n++) {
      update2Tzz2d(par, grid, data->Tzz[0], data->Vx[0], data->Vz[0], 0, n);
      update2Txz2d(par, grid, data->Txz[0], data->Vx[0], data->Vz[0], 0, n);
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], 0, n);
      }
   /**/
   for(n = data->nmin; n < data->nmax; n++) {
      data->Tzz[0][ 0][n] =               0.;
      data->Tzz[0][-1][n] = -data->Tzz[0][1][n];
      data->Tzz[0][-2][n] = -data->Tzz[0][2][n];
      data->Txz[0][-1][n] = -data->Txz[0][0][n];
      data->Txz[0][-2][n] = -data->Txz[0][1][n];
      update2Txx2d(par, grid, data->Txx[0], data->Vx[0], data->Vz[0], m, n);
      }
}



surface3VT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int l, n, m;
   m = -1;
   /**/
{
        float c1, c2;
        int d;
        c1 = C1*par->dt/grid->dx;
        c2 = C2*par->dt/grid->dx;
        d = data->nmax - data->nmin;
        n = data->nmin;
        m = 0;
        for(l = data->lmin; l < data->lmax; l++) {
      getDT(data->Vz[l][m]+n,data->Txz[l][m]+n-1,data->Txz[l][m]+n,data->Txz[l][m]+n+1,data->Txz[l][m]+n+2,
            c1,c2,d,grid->irho[l][m]+n);
      getDT(data->Vz[l][m]+n,data->Tyz[l-2][m]+n,data->Tyz[l-1][m]+n,data->Tyz[l][m]+n,data->Tyz[l+1][m]+n,
            c1,c2,d,grid->irho[l][m]+n);
      getDT(data->Vz[l][m]+n,data->Tzz[l][m-1]+n,data->Tzz[l][m]+n,data->Tzz[l][m+1]+n,data->Tzz[l][m+2]+n,
            c1,c2,d,grid->irho[l][m]+n);
                }
        m = -1;
}
/*
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         update2Vz3d(par, grid, data->Vz, data->Txz, data->Tyz,data->Tzz,l,0,n);
         }
      }
*/
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         update2Vx3d(par, grid, data->Vx, data->Txx, data->Txy,data->Txz,l,0,n);
         update2Vy3d(par, grid, data->Vy, data->Txy, data->Tyy,data->Tyz,l,0,n);
         }
      }
   /**/
   /* order is important */
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 1);
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 2);
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         aboveSurfaceVz3d(par, grid, data->Vx, data->Vy, data->Vz, l, m, n);
         }
      }
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 3);
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         aboveSurfaceVy3d(par, grid, data->Vx, data->Vy, data->Vz, l, m, n);
         }
      }
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         aboveSurfaceVx3d(par, grid, data->Vx, data->Vy, data->Vz, l, m, n);
         }
      }
}



surface3TT(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   int l, n, m;
   m = -1;
   /**/
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy,          l,0,n);
         update2Txz3d(par, grid, data->Txz, data->Vx,           data->Vz,l,0,n);
         update2Tyz3d(par, grid, data->Tyz,           data->Vy, data->Vz,l,0,n);
         update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz,l,0,n);
         update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz,l,0,n);
         update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz,l,0,n);
         }
      }
   /**/
   for(l = data->lmin; l < data->lmax; l++) {
      for(n = data->nmin; n < data->nmax; n++) {
         data->Tzz[l][ 0][n] =                  0.;
         data->Tzz[l][-1][n] = -data->Tzz[l][1][n];
         data->Txz[l][-1][n] = -data->Txz[l][0][n];
         data->Tyz[l][-1][n] = -data->Tyz[l][0][n];
         }
      }
}



aboveSurfaceVx2d(par, grid, Vx, Vz, m, n)
Parameter       *par;
FdGrid         *grid;
float **Vx, **Vz;
int m, n;
{
   Vx[m][n] = Vx[m+2][n] + (grid->dz/grid->dx)*
              ((Vz[m+1][n] - Vz[m+1][n-1]) + (Vz[m][n] - Vz[m][n-1])); 
}



aboveSurfaceVz2d(par, grid, Vx, Vz, m, n)
Parameter       *par;
FdGrid         *grid;
float **Vx, **Vz;
int m, n;
{
   float lamda, mu;
   lamda = grid->lamda[0][m][n];
   mu    = grid->mu   [0][m][n];
   Vz[m][n] = Vz[m+1][n] + (lamda)/(lamda + 2.*mu)*(grid->dz/grid->dx)*
              (Vx[m+1][n+1] - Vx[m+1][n]);
}



aboveSurfaceVx3d(par, grid, Vx, Vy, Vz, l, m, n)
Parameter       *par;
FdGrid         *grid;
float ***Vx, ***Vy, ***Vz;
int l, m, n;
{
   Vx[l][m][n] = Vx[l][m+2][n] + (grid->dz/grid->dx)*
               ((Vz[l][m+1][n]-Vz[l][m+1][n-1]) + (Vz[l][m][n]-Vz[l][m][n-1]));
}



aboveSurfaceVy3d(par, grid, Vx, Vy, Vz, l, m, n)
Parameter       *par;
FdGrid         *grid;
float ***Vx, ***Vy, ***Vz;
int l, m, n;
{
   Vy[l][m][n] = Vy[l][m+2][n] + (grid->dz/grid->dy)*
               ((Vz[l+1][m+1][n]-Vz[l][m+1][n]) + (Vz[l+1][m][n]-Vz[l][m][n]));
}
 


aboveSurfaceVz3d(par, grid, Vx, Vy, Vz, l, m, n)
Parameter       *par;
FdGrid         *grid;
float ***Vx, ***Vy, ***Vz;
int l, m, n;
{
   float lamda, mu;
   lamda = grid->lamda[l][m][n];
   mu    = grid->mu[l][m][n];
   Vz[l][m][n] = Vz[l][m+1][n] + (lamda)/(lamda + 2.*mu)*(grid->dz/grid->dx)*
                (Vx[l][m+1][n+1] - Vx[l][m+1][n]) +
                                 (lamda)/(lamda + 2.*mu)*(grid->dz/grid->dy)*
                (Vy[l][m+1][n] - Vy[l-1][m+1][n]);
}



surface3TTv(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Point at top boundary */
   m = 0;

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      */
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

   /* use symmetry condition to get stresses at and above free surface */
   for(l = data->lmin; l < data->lmax; l++) {
      getNullV(Tzz[l][ 0]+n,              d);
      getNegV (Tzz[l][-1]+n, Tzz[l][1]+n, d);
      getNegV (Txz[l][-1]+n, Txz[l][0]+n, d);
      getNegV (Tyz[l][-1]+n, Tyz[l][0]+n, d);
      }

}



surface3VTv(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p, c, c1, c2;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   g   = grid->lamda;
   u   = grid->mu;
   p   = grid->irho;

   /* get array differencing limits and coefficient */
   d  = data->nmax - data->nmin;
   n  = data->nmin;
   c  = 1.0*par->dt/grid->dx;
   c1 =  C1*par->dt/grid->dx;
   c2 =  C2*par->dt/grid->dx;

   /* Points along surface boundary */
   m = 0;
   
   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDT(Vz[l][m]+n,   Txz[l][m]+n-1, Txz[l][m]+n, 
                          Txz[l][m]+n+1, Txz[l][m]+n+2, c1, c2, d, p[l][m]+n);
      getDT(Vz[l][m]+n,   Tyz[l-2][m]+n, Tyz[l-1][m]+n, 
                          Tyz[l][m]+n,   Tyz[l+1][m]+n, c1, c2, d, p[l][m]+n);
      getDT(Vz[l][m]+n,   Tzz[l][m-1]+n, Tzz[l][m]+n,
                          Tzz[l][m+1]+n, Tzz[l][m+2]+n, c1, c2, d, p[l][m]+n);
      getDT12(Vx[l][m]+n, Txx[l][m]+n-1, Txx[l][m]+n,
                          Txy[l-1][m]+n, Txy[l][m]+n,
                          Txz[l][m-1]+n, Txz[l][m]+n,   c, d, p[l][m]+n);
      getDT12(Vy[l][m]+n, Txy[l][m]+n,   Txy[l][m]+n+1,
                          Tyy[l][m]+n,   Tyy[l+1][m]+n,
                          Tyz[l][m-1]+n, Tyz[l][m]+n,   c, d, p[l][m]+n);
      */
      dtf_(Vz[l][m]+n,   Txz[l][m]+n-1, Txz[l][m]+n, 
                         Txz[l][m]+n+1, Txz[l][m]+n+2, &c1, &c2, &d, p[l][m]+n);
      dtf_(Vz[l][m]+n,   Tyz[l-2][m]+n, Tyz[l-1][m]+n, 
                         Tyz[l][m]+n,   Tyz[l+1][m]+n, &c1, &c2, &d, p[l][m]+n);
      dtf_(Vz[l][m]+n,   Tzz[l][m-1]+n, Tzz[l][m]+n,
                         Tzz[l][m+1]+n, Tzz[l][m+2]+n, &c1, &c2, &d, p[l][m]+n);
      dt12f_(Vx[l][m]+n, Txx[l][m]+n-1, Txx[l][m]+n,
                         Txy[l-1][m]+n, Txy[l][m]+n,
                         Txz[l][m-1]+n, Txz[l][m]+n,   &c, &d, p[l][m]+n);
      dt12f_(Vy[l][m]+n, Txy[l][m]+n,   Txy[l][m]+n+1,
                         Tyy[l][m]+n,   Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n, Tyz[l][m]+n,   &c, &d, p[l][m]+n);
      }

   /* get points above surface */
   m = -1;

   /* update velocities for grid planes (order is important) */
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 1);
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 2);
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getSV12(Vz[l][m]+n,     Vz[l][m+1]+n, Vx[l][m+1]+n, Vx[l][m+1]+n+1,
              Vy[l-1][m+1]+n, Vy[l][m+1]+n, d, g[l][m]+n, u[l][m]+n);
      */
      sv12f_(Vz[l][m]+n,     Vz[l][m+1]+n, Vx[l][m+1]+n, Vx[l][m+1]+n+1,
             Vy[l-1][m+1]+n, Vy[l][m+1]+n, &d, g[l][m]+n, u[l][m]+n);
      }
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 3);
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getSV02(Vy[l][m]+n,   Vy[l][m+2]+n, Vz[l][m+1]+n,   Vz[l+1][m+1]+n,
              Vz[l][m]+n,   Vz[l+1][m]+n, d);
      getSV02(Vx[l][m]+n,   Vx[l][m+2]+n, Vz[l][m+1]+n-1, Vz[l][m+1]+n,
              Vz[l][m]+n-1, Vz[l][m]+n, d);
      */
      sv02f_(Vy[l][m]+n,   Vy[l][m+2]+n, Vz[l][m+1]+n,   Vz[l+1][m+1]+n,
             Vz[l][m]+n,   Vz[l+1][m]+n, &d);
      sv02f_(Vx[l][m]+n,   Vx[l][m+2]+n, Vz[l][m+1]+n-1, Vz[l][m+1]+n,
             Vz[l][m]+n-1, Vz[l][m]+n,   &d);
      }

}

/* acoustic cases below */

surface3TTa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       **dX, **dY, **dZ, c;
   int l, m, n, d;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Point at top boundary */
   m = 0;

   /* update stresses for grid planes */
   /* not needed since nothing below (stress 0 at surface) */
/*
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
*/
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      */
      /* use fortran this isn't needed (stress 0 at surface) */
      /*
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n          );
      */
      }

   /* free memory */
/* not used above (stress 0 at surface) */
/*
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);
*/

   /* use symmetry condition to get stresses at and above free surface */
   for(l = data->lmin; l < data->lmax; l++) {
      getNullV(T  [l][ 0]+n,              d);
      getNegV (T  [l][-1]+n, T  [l][1]+n, d);
      }

}



surface3VTa(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       ***p, c, c1, c2;
   int l, m, n, d;

   /* initialize */
   T   = data->T;
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   g   = grid->lamda;
   p   = grid->irho;

   /* get array differencing limits and coefficient */
   d  = data->nmax - data->nmin;
   n  = data->nmin;
   c  = 1.0*par->dt/grid->dx;
   c1 =  C1*par->dt/grid->dx;
   c2 =  C2*par->dt/grid->dx;

   /* Points along surface boundary */
   m = 0;
   
   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDT(Vz[l][m]+n,   T  [l][m-1]+n, T  [l][m]+n,
                          T  [l][m+1]+n, T  [l][m+2]+n, c1, c2, d, p[l][m]+n);
      getDA12(Vx[l][m]+n, T  [l][m]+n-1, T  [l][m]+n,
                                                     
                                                        c, d, p[l][m]+n);
      getDA12(Vy[l][m]+n,                              
                          T  [l][m]+n,   T  [l+1][m]+n,
                                                        c, d, p[l][m]+n);
      */
      dtf_(Vz[l][m]+n,   T  [l][m-1]+n, T  [l][m]+n,
                         T  [l][m+1]+n, T  [l][m+2]+n, &c1, &c2, &d, p[l][m]+n);
      da12f_(Vx[l][m]+n, T  [l][m]+n-1, T  [l][m]+n,
                                                    
                                                       &c, &d, p[l][m]+n);
      da12f_(Vy[l][m]+n,                              
                         T  [l][m]+n,   T  [l+1][m]+n,
                                                       &c, &d, p[l][m]+n);
      }

   /* get points above surface */
   m = -1;

   /* update velocities for grid planes (order is important) */
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 1);
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 2);
   for(l = data->lmin; l < data->lmax; l++) {
      /* acoustic changes getsv12 to getsv02
      getSV02(Vz[l][m]+n,     Vz[l][m+1]+n, Vx[l][m+1]+n, Vx[l][m+1]+n+1,
              Vy[l-1][m+1]+n, Vy[l][m+1]+n, d                      );
      */
      sv02f_(Vz[l][m]+n,     Vz[l][m+1]+n, Vx[l][m+1]+n, Vx[l][m+1]+n+1,
             Vy[l-1][m+1]+n, Vy[l][m+1]+n, &d                      );
      }
   if(par->nnode>1)mpSurfaceUpdate(par, grid, data->Vx, data->Vy, data->Vz, 3);
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getSV02(Vy[l][m]+n,   Vy[l][m+2]+n, Vz[l][m+1]+n,   Vz[l+1][m+1]+n,
              Vz[l][m]+n,   Vz[l+1][m]+n, d);
      getSV02(Vx[l][m]+n,   Vx[l][m+2]+n, Vz[l][m+1]+n-1, Vz[l][m+1]+n,
              Vz[l][m]+n-1, Vz[l][m]+n, d);
      */
      sv02f_(Vy[l][m]+n,   Vy[l][m+2]+n, Vz[l][m+1]+n,   Vz[l+1][m+1]+n,
             Vz[l][m]+n,   Vz[l+1][m]+n, &d);
      sv02f_(Vx[l][m]+n,   Vx[l][m+2]+n, Vz[l][m+1]+n-1, Vz[l][m+1]+n,
             Vz[l][m]+n-1, Vz[l][m]+n,   &d);
      }

}

