/******************************************************************************
   update: routines for elastic 3-D finite-difference updates (Levandar)

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   update4V   : 4th order (spatial) velocity update
   update4T   : 4th order (spatial) stress   update
   update4V2D : 2D 4th order (spatial) velocity update
   update4T2D : 2D 4th order (spatial) stress   update
   update4A2D : 2D 4th order (spatial) stress   update (attenuation)
   update4V3D : 3D 4th order (spatial) velocity update
   update4T3D : 3D 4th order (spatial) stress   update
   update4A3D : 3D 4th order (spatial) stress   update (attenuation)
   update4V3DA: 3D 4th order (spatial) velocity update (acoustic only)
   update4T3DA: 3D 4th order (spatial) stress   update (acoustic only)

   par  - input parameters
   grid - input parameters for grid
   data - input data
   Vx   - velocity in x direction
   Vy   - velocity in y direction
   Vz   - velocity in z direction
   Txx  - stress (normal x)
   Tyy  - stress (normal y)
   Tzz  - stress (normal z)
   Txy  - stress (shear   )
   Txz  - stress (shear   )
   Tyz  - stress (shear   )
   l    - grid points in y direction (plane )  
   m    - grid points in z direction (row   ) (positive down)
   n    - grid points in x direction (column)

   compile: cc -g -c update.c

   note: used  units are mks (m, s, m/s, kg/m3)
   note: regular mode (2D, 3D; attenuation; averaging/topography; C, FORTRAN)
         acoustic mode (3D only, no attenuation, ave/topo; C, FORTRAN)
         topography (2D, 3D; attenuation, C, FORTRAN) 
         attenuation (2D, 3D; no ave/topo; FORTRAN only in 3D)
         C pragma's in zsave
         
   author: shawn larsen
   date  : march 1990
   where : caltech
*****************************************************************************/

#include "elas.h"

#define C_SOURCE_CODE 0

update4V(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   if (par->model == 1) {
      if (grid->d == 2) update4V2D (par, grid, data);
      if (grid->d == 3) update4V3D (par, grid, data);
      }
   if (par->model == 2) {
      if (grid->d == 2) update4V2D (par, grid, data);
      if (grid->d == 3) update4V3DA(par, grid, data);
      }
}




update4T(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   if (par->model == 1) {
      if (grid->d == 2 && par->q == 0) update4T2D (par, grid, data);
      if (grid->d == 2 && par->q == 1) update4A2D (par, grid, data);
      if (grid->d == 3 && par->q == 0) update4T3D (par, grid, data);
      if (grid->d == 3 && par->q == 1) update4A3D (par, grid, data);
      }
   if (par->model == 2) {
      if (grid->d == 2 && par->q == 0) update4T2D (par, grid, data);
      if (grid->d == 2 && par->q == 1) update4A2D (par, grid, data);
      if (grid->d == 3 && par->q == 0) update4T3DA(par, grid, data);
      if (grid->d == 3 && par->q == 1) update4A3D (par, grid, data);
      }
}



/* equations being solved
   next 2 equations update velocities
   dVx/dt  = (1./rho)*(dTxx/dx + dTxz/dz)
   dVz/dt  = (1./rho)*(dTxz/dx + dTzz/dz)
*/
update4V2D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***p, *rx, *rz, c1, c2;
   float *v1, *v2, *v3, *v4;
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08; 
   float *f09, *f10, *f11, *f12, *f13, *f14, *f15, *f16; 
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08;
   float  g09,  g10,  g11,  g12,  g13,  g14,  g15,  g16;
   int l, m, n, mmin, nmin, mmax, nmax, d;

   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   p   = grid->irho;
   mmin = data->mmin;
   nmin = data->nmin;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx;
   c2 = C2*par->dt/grid->dx;
   if (grid->ave) {
      rx = (float *) malloc(grid->n*sizeof(float));
      rz = (float *) malloc(grid->n*sizeof(float));
      p = grid->rho;
      }
 
   /* update velocities along surface (now done in surface routine) */
 
   /* update velocities for rest of grid */
   l = 0;
   for(m = mmin; m < mmax; m++) {
      v1  = Vx [l  ][m  ]+nmin;
      v2  = Vz [l  ][m  ]+nmin;
      v3  = p  [l  ][m  ]+nmin;
      v4  = p  [l  ][m  ]+nmin;
      f01 = Txx[l  ][m  ]+nmin-2;
      f02 = Txx[l  ][m  ]+nmin-1;
      f03 = Txx[l  ][m  ]+nmin  ;
      f04 = Txx[l  ][m  ]+nmin+1;
      f05 = Txz[l  ][m-2]+nmin  ;
      f06 = Txz[l  ][m-1]+nmin  ;
      f07 = Txz[l  ][m  ]+nmin  ;
      f08 = Txz[l  ][m+1]+nmin  ;
      f09 = Txz[l  ][m  ]+nmin-1;
      f10 = Txz[l  ][m  ]+nmin  ;
      f11 = Txz[l  ][m  ]+nmin+1;
      f12 = Txz[l  ][m  ]+nmin+2;
      f13 = Tzz[l  ][m-1]+nmin  ;
      f14 = Tzz[l  ][m  ]+nmin  ;
      f15 = Tzz[l  ][m+1]+nmin  ;
      f16 = Tzz[l  ][m+2]+nmin  ;
      /* c formulation */
      #if C_SOURCE_CODE
      if (grid->ave) {
         getR(rx+nmin, p[l][m  ]+nmin-1, p[l][m  ]+nmin, d);
         getR(rz+nmin, p[l][m  ]+nmin  , p[l][m+1]+nmin, d);
         v3 = rx+nmin;
         v4 = rz+nmin;
         }
      for(n = nmin; n < nmax; n++) {
         g01 = *f01++;
         g02 = *f02++;
         g03 = *f03++;
         g04 = *f04++;
         g05 = *f05++;
         g06 = *f06++;
         g07 = *f07++;
         g08 = *f08++;
         g09 = *f09++;
         g10 = *f10++;
         g11 = *f11++;
         g12 = *f12++;
         g13 = *f13++;
         g14 = *f14++;
         g15 = *f15++;
         g16 = *f16++;
         *v1 += (c1*(g03 - g02 + g07 - g06) +
                 c2*(g04 - g01 + g08 - g05))*(*v3);
         *v2 += (c1*(g11 - g10 + g15 - g14) + 
                 c2*(g12 - g09 + g16 - g13))*(*v4);
         v1++;
         v2++;
         v3++;
         v4++;
         }
      /* fortran formulation */
      #else
      if (grid->ave) {
         rf_(rx+nmin, p[l][m  ]+nmin-1, p[l][m  ]+nmin, &d);
         rf_(rz+nmin, p[l][m  ]+nmin  , p[l][m+1]+nmin, &d);
         v3 = rx+nmin;
         v4 = rz+nmin;
         }
      dv2d2_(v1, v2, v3, v4, f01, f02, f03, f04, f05, f06, f07, f08, f09, f10, 
                             f11, f12, f13, f14, f15, f16, &c1, &c2, &d);
      #endif
      }
 
   /* free averaging memory */
   if (grid->ave) {free(rx); free(rz);}

}




/* equations being solved
   next 6 equations update stresses
   h01     = (dVx/dx)
   h02     = (dVz/dz)
   h03     = (lamda + 2.*mu)
   dTxx/dt = (lamda + 2.*mu)*(dVx/dx) + lamda*(dVz/dz)
   dTzz/dt = (lamda + 2.*mu)*(dVz/dz) + lamda*(dVx/dx)
   dTxz/dt = (mu)*(dVx/dz + dVz/dx)
*/
update4T2D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, *r, c1, c2;
   float *v1, *v2, *v3, *v4, *v5, *v6; 
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08; 
   float *f09, *f10, *f11, *f12, *f13, *f14, *f15, *f16; 
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08;
   float  g09,  g10,  g11,  g12,  g13,  g14,  g15,  g16;
   float  h01, h02, h03;
   int l, m, n, mmin, nmin, mmax, nmax, d;
 
   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   g   = grid->lamda;
   u   = grid->mu;
   mmin = data->mmin;
   nmin = data->nmin;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx;
   c2 = C2*par->dt/grid->dx;
   if (grid->ave) r = (float *) malloc(grid->n*sizeof(float));
 
   /* update stresses along surface (now done in surface routine) */
 
   /* update stresses for rest of grid */
   l = 0;
   for(m = mmin; m < mmax; m++) {
      v1  = Txx[l  ][m  ]+nmin;
      v2  = Tzz[l  ][m  ]+nmin;
      v3  = Txz[l  ][m  ]+nmin;
      v4  = g  [l  ][m  ]+nmin;
      v5  = u  [l  ][m  ]+nmin;
      v6  = u  [l  ][m  ]+nmin;
      f01 = Vx [l  ][m  ]+nmin-1;
      f02 = Vx [l  ][m  ]+nmin  ;
      f03 = Vx [l  ][m  ]+nmin+1;
      f04 = Vx [l  ][m  ]+nmin+2;
      f05 = Vz [l  ][m-2]+nmin  ;
      f06 = Vz [l  ][m-1]+nmin  ;
      f07 = Vz [l  ][m  ]+nmin  ;
      f08 = Vz [l  ][m+1]+nmin  ;
      f09 = Vx [l  ][m-1]+nmin  ;
      f10 = Vx [l  ][m  ]+nmin  ;
      f11 = Vx [l  ][m+1]+nmin  ;
      f12 = Vx [l  ][m+2]+nmin  ;
      f13 = Vz [l  ][m  ]+nmin-2;
      f14 = Vz [l  ][m  ]+nmin-1;
      f15 = Vz [l  ][m  ]+nmin  ;
      f16 = Vz [l  ][m  ]+nmin+1;
      /* c formulation */
      #if C_SOURCE_CODE
      if (grid->ave) {
         getU(r+nmin, u[l  ][m  ]+nmin-1, u[l  ][m  ]+nmin  ,
                      u[l  ][m+1]+nmin-1, u[l  ][m+1]+nmin  , d);
         v6 = r+nmin;
         }
      for(n = nmin; n < nmax; n++) {
         g01 = *f01++;
         g02 = *f02++;
         g03 = *f03++;
         g04 = *f04++;
         g05 = *f05++;
         g06 = *f06++;
         g07 = *f07++;
         g08 = *f08++;
         g09 = *f09++;
         g10 = *f10++;
         g11 = *f11++;
         g12 = *f12++;
         g13 = *f13++;
         g14 = *f14++;
         g15 = *f15++;
         g16 = *f16++;
         h01 = (c1*(g03 - g02) + c2*(g04 - g01));   
         h02 = (c1*(g07 - g06) + c2*(g08 - g05));   
         h03 = ((*v5) + (*v5) + (*v4));              
         *v1 += h03*h01 + h02*(*v4);                 
         *v2 += h03*h02 + h01*(*v4);
         *v3 += (c1*(g11 - g10 + g15 - g14) + 
                 c2*(g12 - g09 + g16 - g13))*(*v6);
         v1++;
         v2++;
         v3++;
         v4++;
         v5++;
         v6++;
         }
      /* fortran formulation */
      #else
      if (grid->ave) {
         uf_(r+nmin, u[l][m  ]+nmin-1, u[l][m  ]+nmin  ,
                     u[l][m+1]+nmin-1, u[l][m+1]+nmin  , &d);
         v6 = r+nmin;
         }
      dt2d2_(v1 , v2 , v3 , v4 , v5 , v6, f01, f02, f03, f04, f05, f06,
             f07, f08, f09, f10, f11, f12, f13, f14, f15, f16,
             &c1, &c2, &d);
      #endif
      }

   /* free averaging memory */
   if (grid->ave) free(r);
 
}


update4A2D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float dt, *dv1, *dv2, c1, c2, ***Vx, ***Vz; 
   float *Txx, *Tzz, *Txz, *lamda, *mu, *r;
   float *F0, *F1, *F2, *F3, *q1, *q2, *q3;
   float s1, s2, s3, D1, D2; 
   int d, l, m, n, mmin, nmin;

   /* initialize and allocate memory for derivative arrays */
   Vx = data->Vx; 
   Vz = data->Vz; 
   dt = par->dt;
   c1 = C1/grid->dx;
   c2 = C2/grid->dx;
   dv1 = (float *) malloc(grid->n*sizeof(float));
   dv2 = (float *) malloc(grid->n*sizeof(float));
   d = data->nmax - data->nmin;
   l = 0;
   mmin = data->mmin;
   nmin = data->nmin;
   /*if (mmin < 1) mmin = 1;*/
   if (grid->bz1 == SURFACE && mmin < 1) mmin = 1;
   if (grid->ave) r = (float *) malloc(grid->n*sizeof(float));

   /* calculate dVx/dx and dVz/dz derivatives; update q1,q2,q3; stress */
   for(m = mmin; m < data->mmax; m++) {
      n = data->nmin;
      Txx = data->Txx[l][m]; Tzz   = data->Tzz  [l][m];
      F0  = grid->F0 [l][m]; F1    = grid->F1   [l][m]; 
      F2  = grid->F2 [l][m]; F3    = grid->F3   [l][m];
      q1  = data->q1 [l][m]; q2    = data->q2   [l][m];
      mu  = grid->mu [l][m]; lamda = grid->lamda[l][m];
      /* use c derivatives */
      #if C_SOURCE_CODE
      getDV(dv1+n,Vx[l][m]+n-1,Vx[l][m]+n,Vx[l][m]+n+1,Vx[l][m]+n+2,c1,c2,d);
      getDV(dv2+n,Vz[l][m-2]+n,Vz[l][m-1]+n,Vz[l][m]+n,Vz[l][m+1]+n,c1,c2,d);
      for(n = data->nmin; n < data->nmax; n++) {
         s1 = q1[n];
         s2 = q2[n];
         q1[n]=F0[n]*s1 + F1[n]*(dv1[n] + dv2[n]) + F2[n]*dv2[n];
         q2[n]=F0[n]*s2 + F1[n]*(dv1[n] + dv2[n]) + F2[n]*dv1[n];
         s1 = (s1 + q1[n])*0.5;
         s2 = (s2 + q2[n])*0.5;
         D1 = (lamda[n] + 2.*mu[n]);
         D2 = (lamda[n]);
         Txx[n] += dt*(D1*dv1[n] + D2*dv2[n] + s1);
         Tzz[n] += dt*(D2*dv1[n] + D1*dv2[n] + s2);
         }
      #else
      dvf_(dv1+n,Vx[l][m]+n-1,Vx[l][m]+n,Vx[l][m]+n+1,Vx[l][m]+n+2,&c1,&c2,&d);
      dvf_(dv2+n,Vz[l][m-2]+n,Vz[l][m-1]+n,Vz[l][m]+n,Vz[l][m+1]+n,&c1,&c2,&d);
      da1f_(q1+n,q2+n,Txx+n,Tzz+n,lamda+n,mu+n,F0+n,F1+n,F2+n,dv1+n,dv2+n,
            &dt,&d);
      #endif
      }

   /* calculate dVx/dz and dVz/dx derivatives; update e3; stress */
   for(m = mmin; m < data->mmax; m++) {
      n = data->nmin;
      F0  = grid->F0 [l][m]; F3 = grid->F3[l][m];
      Txz = data->Txz[l][m]; 
      q3  = data->q3 [l][m]; 
      mu  = grid->mu [l][m];
      /* use c derivatives and formulation */
      #if C_SOURCE_CODE
      getDV(dv2+n,Vx[l][m-1]+n,Vx[l][m]+n,Vx[l][m+1]+n,Vx[l][m+2]+n,c1,c2,d);
      getDV(dv1+n,Vz[l][m]+n-2,Vz[l][m]+n-1,Vz[l][m]+n,Vz[l][m]+n+1,c1,c2,d);
      if (grid->ave) {
         getU(r+nmin, grid->mu[l][m  ]+nmin-1, grid->mu[l][m  ]+nmin  ,
                      grid->mu[l][m+1]+nmin-1, grid->mu[l][m+1]+nmin  , d);
         mu = r;
         }
      for(n = data->nmin; n < data->nmax; n++) {
         s3 = q3[n];
         q3[n] = F0[n]*s3 + F3[n]*(dv1[n] + dv2[n]);
         s3 = (s3 + q3[n])*0.5;
         Txz[n] += dt*mu[n]*(dv1[n] + dv2[n]) +dt*s3;
         }
      #else
      dvf_(dv2+n,Vx[l][m-1]+n,Vx[l][m]+n,Vx[l][m+1]+n,Vx[l][m+2]+n,&c1,&c2,&d);
      dvf_(dv1+n,Vz[l][m]+n-2,Vz[l][m]+n-1,Vz[l][m]+n,Vz[l][m]+n+1,&c1,&c2,&d);
      if (grid->ave) {
         uf_(r+nmin, grid->mu[l][m  ]+nmin-1, grid->mu[l][m  ]+nmin  ,
                     grid->mu[l][m+1]+nmin-1, grid->mu[l][m+1]+nmin  , &d);
         mu = r;
         }
      da2f_(q3+n, Txz+n, mu+n, F0+n, F3+n, dv1+n, dv2+n, &dt, &d);
      #endif
      }

   /* free memory */
   if (grid->ave) free(r);
   free(dv1);
   free(dv2);

}


/* equations being solved
   dVx/dt  = (1./rho)*(dTxx/dx + dTxy/dy + dTxz/dz) 
   dVy/dt  = (1./rho)*(dTxy/dx + dTyy/dy + dTyz/dz) 
   dVz/dt  = (1./rho)*(dTxz/dx + dTyz/dy + dTzz/dz)
*/
update4V3D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***p, *r, c1, c2;
   float *v1, *v2, *v3, *v4;
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08, *f09, *f10, *f11, *f12;
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08,  g09,  g10,  g11,  g12;
   int l, m, n, lmin, mmin, nmin, lmax, mmax, nmax, d;

   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz; 
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   p = grid->irho;
   lmin = data->lmin;
   mmin = data->mmin;
   nmin = data->nmin;
   lmax = data->lmax;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx; 
   c2 = C2*par->dt/grid->dx;
   if (grid->ave) { 
      r = (float *) malloc(grid->n*sizeof(float));
      p = grid->rho;
      }

   /* update velocities along surface (now done in surface routine) */

   /* update velocities for rest of grid */
   for(l = lmin; l < lmax; l++) {
      for(m = mmin; m < mmax; m++) {
         v1  = Vx [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f01 = Txx[l  ][m  ]+nmin-2;
         f02 = Txx[l  ][m  ]+nmin-1;
         f03 = Txx[l  ][m  ]+nmin  ;
         f04 = Txx[l  ][m  ]+nmin+1;
         f05 = Txy[l-2][m  ]+nmin  ;
         f06 = Txy[l-1][m  ]+nmin  ;
         f07 = Txy[l  ][m  ]+nmin  ;
         f08 = Txy[l+1][m  ]+nmin  ;
         f09 = Txz[l  ][m-2]+nmin  ;
         f10 = Txz[l  ][m-1]+nmin  ;
         f11 = Txz[l  ][m  ]+nmin  ;
         f12 = Txz[l  ][m+1]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getR(r+nmin, p[l][m]+nmin-1, p[l][m]+nmin, d); 
            v4 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            *v1 += (c1*(g03 - g02 + g07 - g06 + g11 - g10) +
                    c2*(g04 - g01 + g08 - g05 + g12 - g09))*(*v4);
            v1++;
            v4++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            rf_(r+nmin, p[l][m]+nmin-1, p[l][m]+nmin, &d); 
            v4 = r+nmin;
            }
         dv3d1_(v1, v4, f01, f02, f03, f04, f05, f06, 
                        f07, f08, f09, f10, f11, f12, &c1, &c2, &d);
         #endif
         v2  = Vy [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f01 = Txy[l  ][m  ]+nmin-1;
         f02 = Txy[l  ][m  ]+nmin  ;
         f03 = Txy[l  ][m  ]+nmin+1;
         f04 = Txy[l  ][m  ]+nmin+2;
         f05 = Tyy[l-1][m  ]+nmin  ;
         f06 = Tyy[l  ][m  ]+nmin  ;
         f07 = Tyy[l+1][m  ]+nmin  ;
         f08 = Tyy[l+2][m  ]+nmin  ;
         f09 = Tyz[l  ][m-2]+nmin  ;
         f10 = Tyz[l  ][m-1]+nmin  ;
         f11 = Tyz[l  ][m  ]+nmin  ;
         f12 = Tyz[l  ][m+1]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getR(r+nmin, p[l][m]+nmin, p[l+1][m]+nmin, d);
            v4 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            *v2 += (c1*(g03 - g02 + g07 - g06 + g11 - g10) +
                    c2*(g04 - g01 + g08 - g05 + g12 - g09))*(*v4);
            v2++;
            v4++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            rf_(r+nmin, p[l][m]+nmin, p[l+1][m]+nmin, &d);
            v4 = r+nmin;
            }
         dv3d1_(v2, v4, f01, f02, f03, f04, f05, f06, 
                        f07, f08, f09, f10, f11, f12, &c1, &c2, &d);
         #endif
         v3  = Vz [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f01 = Txz[l  ][m  ]+nmin-1;
         f02 = Txz[l  ][m  ]+nmin  ;
         f03 = Txz[l  ][m  ]+nmin+1;
         f04 = Txz[l  ][m  ]+nmin+2;
         f05 = Tyz[l-2][m  ]+nmin  ;
         f06 = Tyz[l-1][m  ]+nmin  ;
         f07 = Tyz[l  ][m  ]+nmin  ;
         f08 = Tyz[l+1][m  ]+nmin  ;
         f09 = Tzz[l  ][m-1]+nmin  ;
         f10 = Tzz[l  ][m  ]+nmin  ;
         f11 = Tzz[l  ][m+1]+nmin  ;
         f12 = Tzz[l  ][m+2]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getR(r+nmin, p[l][m]+nmin, p[l][m+1]+nmin, d);
            v4 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            *v3 += (c1*(g03 - g02 + g07 - g06 + g11 - g10) +
                    c2*(g04 - g01 + g08 - g05 + g12 - g09))*(*v4);
            v3++;
            v4++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            rf_(r+nmin, p[l][m]+nmin, p[l][m+1]+nmin, &d);
            v4 = r+nmin;
            }
         dv3d1_(v3, v4, f01, f02, f03, f04, f05, f06, 
                        f07, f08, f09, f10, f11, f12, &c1, &c2, &d);
         #endif
         }
      }

   /* free averaging memory */
   if (grid->ave) free(r);
              
}




/* next 8 equations update normal stresses
   h01     = (dVx/dx)
   h02     = (dVy/dy)
   h03     = (dVz/dz)
   h04     = (lamda)*(dVx/dx + dVy/dy + dVz/dz)
   h05     = (2.*mu)
   dTxx/dt = (2.*mu)*(dVx/dx) + lamda*(dVx/dx + dVy/dy + dVz/dz) 
   dTyy/dt = (2.*mu)*(dVy/dy) + lamda*(dVx/dx + dVy/dy + dVz/dz) 
   dTzz/dt = (2.*mu)*(dVz/dz) + lamda*(dVx/dx + dVy/dy + dVz/dz) 
   next equation updates shear stress component
   dTxy/dt = (mu)*(dVx/dy + dVy/dx)
   [(g03 - g02 + g07 - g06); (g04 - g01 + g08 - g05)]
   comment? next equation updates shear stress component
   dTxz/dt = (mu)*(dVx/dz + dVz/dx)
   [(g03 - g02 + g07 - g06); (g04 - g01 + g08 - g05)]
   next equation updates shear stress component
   dTyz/dt = (mu)*(dVy/dz + dVz/dy)
   [(g03 - g02 + g07 - g06); (g04 - g01 + g08 - g05)]
*/
update4T3D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, *r, c1, c2; 
   float *v1, *v2, *v3, *v4, *v5, *v6, *v7, *v8;
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08, *f09, *f10, *f11, *f12;
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08,  g09,  g10,  g11,  g12;
   float  h01,  h02,  h03,  h04,  h05;
   int l, m, n, lmin, mmin, nmin, lmax, mmax, nmax, d;

   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz; 
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   lmin = data->lmin;
   mmin = data->mmin;
   nmin = data->nmin;
   lmax = data->lmax;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx; 
   c2 = C2*par->dt/grid->dx;
   if (grid->ave) r = (float *) malloc(grid->n*sizeof(float));

   /* update stresses along surface (now done in surface routine) */

   /* update stresses for rest of grid */
   for(l = lmin; l < lmax; l++) {
      for(m = mmin; m < mmax; m++) {
         v1  = Txx[l  ][m  ]+nmin;
         v2  = Tyy[l  ][m  ]+nmin;
         v3  = Tzz[l  ][m  ]+nmin;
         v7  = g  [l  ][m  ]+nmin;
         v8  = u  [l  ][m  ]+nmin;
         f01 = Vx [l  ][m  ]+nmin-1;
         f02 = Vx [l  ][m  ]+nmin  ;
         f03 = Vx [l  ][m  ]+nmin+1;
         f04 = Vx [l  ][m  ]+nmin+2;
         f05 = Vy [l-2][m  ]+nmin  ;
         f06 = Vy [l-1][m  ]+nmin  ;
         f07 = Vy [l  ][m  ]+nmin  ;
         f08 = Vy [l+1][m  ]+nmin  ;
         f09 = Vz [l  ][m-2]+nmin  ;
         f10 = Vz [l  ][m-1]+nmin  ;
         f11 = Vz [l  ][m  ]+nmin  ;
         f12 = Vz [l  ][m+1]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            h01 = (c1*(g03 - g02) + c2*(g04 - g01));
            h02 = (c1*(g07 - g06) + c2*(g08 - g05));
            h03 = (c1*(g11 - g10) + c2*(g12 - g09));
            h04 = (h01 + h02 + h03)*(*v7);
            h05 = ((*v8) + (*v8));
            *v1 += h05*h01 + h04;
            *v2 += h05*h02 + h04; 
            *v3 += h05*h03 + h04;
            v1++;
            v2++;
            v3++;
            v7++;
            v8++;
            }
         /* fortran formulation */
         #else
         dt3d1_(v1 , v2 , v3 , v7 , v8 , f01, f02, f03, f04, f05, f06,
                f07, f08, f09, f10, f11, f12, &c1, &c2, &d);
         #endif
         v4  = Txy[l  ][m  ]+nmin;
         v8  = u  [l  ][m  ]+nmin;
         f01 = Vx [l-1][m  ]+nmin  ;
         f02 = Vx [l  ][m  ]+nmin  ;
         f03 = Vx [l+1][m  ]+nmin  ;
         f04 = Vx [l+2][m  ]+nmin  ;
         f05 = Vy [l  ][m  ]+nmin-2;
         f06 = Vy [l  ][m  ]+nmin-1;
         f07 = Vy [l  ][m  ]+nmin  ;
         f08 = Vy [l  ][m  ]+nmin+1;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getU(r+nmin, u[l  ][m  ]+nmin-1, u[l  ][m  ]+nmin  , 
                         u[l+1][m  ]+nmin-1, u[l+1][m  ]+nmin  , d);
            v8 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            *v4 += (c1*(g07 - g06 - g02 + g03) +
                    c2*(g08 - g05 - g01 + g04))*(*v8);
            v4++;
            v8++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            uf_(r+nmin, u[l  ][m  ]+nmin-1, u[l  ][m  ]+nmin  , 
                        u[l+1][m  ]+nmin-1, u[l+1][m  ]+nmin  , &d);
            v8 = r+nmin;
            }
         dt3d2_(v4 , v8 , f01, f02, f03, f04, f05, f06, f07, f08,
                &c1, &c2, &d);
         #endif
         v5  = Txz[l  ][m  ]+nmin;
         v8  = u  [l  ][m  ]+nmin;
         f01 = Vx [l  ][m-1]+nmin  ;
         f02 = Vx [l  ][m  ]+nmin  ;
         f03 = Vx [l  ][m+1]+nmin  ;
         f04 = Vx [l  ][m+2]+nmin  ;
         f05 = Vz [l  ][m  ]+nmin-2;
         f06 = Vz [l  ][m  ]+nmin-1;
         f07 = Vz [l  ][m  ]+nmin  ;
         f08 = Vz [l  ][m  ]+nmin+1;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getU(r+nmin, u[l  ][m  ]+nmin-1, u[l  ][m  ]+nmin  , 
                         u[l  ][m+1]+nmin-1, u[l  ][m+1]+nmin  , d);
            v8 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            *v5 += (c1*(g07 - g06 - g02 + g03) +
                    c2*(g08 - g05 - g01 + g04))*(*v8);
            v5++;
            v8++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            uf_(r+nmin, u[l  ][m  ]+nmin-1, u[l  ][m  ]+nmin  , 
                        u[l  ][m+1]+nmin-1, u[l  ][m+1]+nmin  , &d);
            v8 = r+nmin;
            }
         dt3d2_(v5 , v8 , f01, f02, f03, f04, f05, f06, f07, f08,
                &c1, &c2, &d);
         #endif
         v6  = Tyz[l  ][m  ]+nmin;
         v8  = u  [l  ][m  ]+nmin;
         f01 = Vy [l  ][m-1]+nmin  ;
         f02 = Vy [l  ][m  ]+nmin  ;
         f03 = Vy [l  ][m+1]+nmin  ;
         f04 = Vy [l  ][m+2]+nmin  ;
         f05 = Vz [l-1][m  ]+nmin  ;
         f06 = Vz [l  ][m  ]+nmin  ;
         f07 = Vz [l+1][m  ]+nmin  ;
         f08 = Vz [l+2][m  ]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         if (grid->ave) {
            getU(r+nmin, u[l  ][m  ]+nmin  , u[l+1][m  ]+nmin  , 
                         u[l  ][m+1]+nmin  , u[l+1][m+1]+nmin  , d);
            v8 = r+nmin;
            }
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            *v6 += (c1*(g07 - g06 - g02 + g03) +
                    c2*(g08 - g05 - g01 + g04))*(*v8);
            v6++;
            v8++;
            }
         /* fortran formulation */
         #else
         if (grid->ave) {
            uf_(r+nmin, u[l  ][m  ]+nmin  , u[l+1][m  ]+nmin  , 
                        u[l  ][m+1]+nmin  , u[l+1][m+1]+nmin  , &d);
            v8 = r+nmin;
            }
         dt3d2_(v6 , v8 , f01, f02, f03, f04, f05, f06, f07, f08,
                &c1, &c2, &d);
         #endif
         }
      }

   /* free averaging memory */
   if (grid->ave) free(r);

}

/****************************************************************************
   attenuate3D: three dimensional attenuation. 

   NOTE: The global memory variables (q1, ..., q6) have the following 
   mapping within this funciton:
   
            rxx = q1
            rzz = q2
            rxz = q3
            ryy = q4
            ryx = q5
            ryz = q6
   
   NOTE: the memory variable now have the following format:
   
            q1[L][l][m][n]

   where-
            L - solids dim
            l - y dim
            m - z dim
            n - x dim
****************************************************************************/

update4A3D(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float dt, *dv1, *dv2, *dv3, *dv4, *dv5, *dv6;
   float c1, c2, ***Vx, ***Vy, ***Vz;
   float *Txx, *Tyy, *Tzz, *Txy, *Tyz, *Txz;
   float *lamda, *mu, *F0, *F1, *F2, *F3;
   float *muxy, *muxz, *muyz, *r1, *r2, *r3;
   /* right now these are 1D, but later they'll have an L dim. */
   float *rxx, *ryy, *rzz, *rxz, *rxy, *ryz;
   float s1, s2, s3, s4, s5, D1, D2, D3;
   int d, l, m, n, mmin, L;

   /* initialize and allocate memory for derivative arrays */
   Vx  = data->Vx;  Vy = data->Vy;  Vz = data->Vz;
   dt  = par->dt;
   c1  = C1/grid->dx;  c2 = C2/grid->dx;            /* dx == dy == dz */
   dv1 = (float *) malloc(grid->n*sizeof(float));
   dv2 = (float *) malloc(grid->n*sizeof(float));
   dv3 = (float *) malloc(grid->n*sizeof(float));
   dv4 = (float *) malloc(grid->n*sizeof(float));
   dv5 = (float *) malloc(grid->n*sizeof(float));
   dv6 = (float *) malloc(grid->n*sizeof(float));
   d   = data->nmax - data->nmin;  
   mmin = data->mmin;
   L = 0;
   if (grid->ave) {
      r1 = (float *) malloc(grid->n*sizeof(float));
      r2 = (float *) malloc(grid->n*sizeof(float));
      r3 = (float *) malloc(grid->n*sizeof(float));
      }

   /* if (mmin < 1) mmin = 1; */
   if (grid->bz1 == SURFACE && mmin < 1) mmin = 1;
 
   /* l in the outer loop */
   for (l = data->lmin; l < data->lmax; l++) {

      /* calculate dVx/dx, dVy/dy, dVz/dz; update diag. memory & stress */
      for (m = mmin; m < data->mmax; m++) {
         n = data->nmin;

         /* get the data and pointers we need */
         Txx = data->Txx   [l][m]; Tyy = data->Tyy    [l][m]; 
         Tzz = data->Tzz   [l][m]; F0  = grid->F0     [l][m]; 
         F1  = grid->F1    [l][m]; F2  = grid->F2     [l][m]; 
         F3  = grid->F3    [l][m]; rxx = data->q1     [l][m];
         rzz = data->q2    [l][m]; ryy = data->q4     [l][m];
         mu  = grid->mu    [l][m]; lamda = grid->lamda[l][m];
       
         /* get derivatives, fortran function to compute vector of derivs */
         dvf_(dv1+n,Vx[l][m]+n-1,Vx[l][m]+n,Vx[l][m]+n+1,Vx[l][m]+n+2,
              &c1,&c2,&d);
         /* old from leeann?
         dvf_(dv2+n,Vy[l-1][m]+n,Vy[l][m]+n,Vy[l+1][m]+n,Vy[l+2][m]+n,
              &c1,&c2,&d);
         */
         dvf_(dv2+n,Vy[l-2][m]+n,Vy[l-1][m]+n,Vy[l][m]+n,Vy[l+1][m]+n,
              &c1,&c2,&d);
         dvf_(dv3+n,Vz[l][m-2]+n,Vz[l][m-1]+n,Vz[l][m]+n,Vz[l][m+1]+n,
              &c1,&c2,&d);
     
         /* In C: This does the same thing as the fortran defined below. */
         /* See the file vectorf.f for the actual function. */ 
         /* I don't believe these c routines are tested (why f2, not F2)*/
         /*
         for(n = data->nmin; n < data->nmax; n++) {
            s1 = rxx[n]; s2 = ryy[n]; s3 = rzz[n];
            rxx[n] = F0[n]*s1 + F1[n]*(dv1[n]+dv2[n]+dv3[n]) + 
                                f2[n]*(dv2[n]+dv3[n]);
            ryy[n] = F0[n]*s2 + F1[n]*(dv1[n]+dv2[n]+dv3[n]) + 
                                f2[n]*(dv1[n]+dv3[n]);
            rzz[n] = F0[n]*s3 + F1[n]*(dv1[n]+dv2[n]+dv3[n]) + 
                                f2[n]*(dv1[n]+dv2[n]);
            s1 = (s1 + rxx[n])*0.5;
            s2 = (s2 + ryy[n])*0.5;
            s3 = (s3 + rzz[n])*0.5;
            D1 = (lamda[n] + 2.*mu[n]);
            D2 = (lamda[n]);
            Txx[n] += dt*(D1*dv1[n] + D2*(dv2[n] + dv3[n]) + s1);
            Tyy[n] += dt*(D1*dv2[n] + D2*(dv1[n] + dv3[n]) + s2);
            Tzz[n] += dt*(D1*dv3[n] + D2*(dv1[n] + dv2[n]) + s3);
            }
            */
       
         da3f_(rxx+n, ryy+n, rzz+n, Txx+n, Tyy+n, Tzz+n, lamda+n, mu+n, 
               F0+n, F1+n, F2+n, dv1+n, dv2+n, dv3+n, &dt, &d);
         }
     
      /* now do the same as above with the off-diagonal terms */
      for (m = mmin; m < data->mmax; m++) {
         n = data->nmin;
         F0  = grid->F0   [l][m]; F3  = grid->F3   [l][m];
         Txz = data->Txz  [l][m]; Txy = data->Txy  [l][m];
         Tyz = data->Tyz  [l][m]; rxz = data->q3   [l][m];
         rxy = data->q5   [l][m]; ryz = data->q6   [l][m];
         mu  = grid->mu   [l][m];
         muxy = mu;
         muxz = mu;
         muyz = mu;
       
         /* fortran function to compute vector of derivatives */
         dvf_(dv1+n,Vx[l][m-1]+n,Vx[l][m]+n,Vx[l][m+1]+n,Vx[l][m+2]+n,
              &c1,&c2,&d);  /* dVx/dz */
         dvf_(dv2+n,Vy[l][m-1]+n,Vy[l][m]+n,Vy[l][m+1]+n,Vy[l][m+2]+n,
              &c1,&c2,&d);  /* dVy/dz */
         dvf_(dv3+n,Vx[l-1][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,Vx[l+2][m]+n,
              &c1,&c2,&d);  /* dVx/dy */
         dvf_(dv4+n,Vz[l][m]+n-2,Vz[l][m]+n-1,Vz[l][m]+n,Vz[l][m]+n+1,
              &c1,&c2,&d);  /* dVz/dx */
         dvf_(dv5+n,Vz[l-1][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,Vz[l+2][m]+n,
              &c1,&c2,&d);  /* dVz/dy */
         dvf_(dv6+n,Vy[l][m]+n-2,Vy[l][m]+n-1,Vy[l][m]+n,Vy[l][m]+n+1,
              &c1,&c2,&d);  /* dVy/dx */

         /* In C: Same as routines in fortran */
         /*
         if (grid->ave) {
            getU(r1+n, grid->mu[l  ][m  ]+n-1, grid->mu[l  ][m  ]+n  ,
                       grid->mu[l+1][m  ]+n-1, grid->mu[l+1][m  ]+n  , d);
            getU(r2+n, grid->mu[l  ][m  ]+n-1, grid->mu[l  ][m  ]+n  ,
                       grid->mu[l  ][m+1]+n-1, grid->mu[l  ][m+1]+n  , d);
            getU(r3+n, grid->mu[l  ][m  ]+n  , grid->mu[l+1][m  ]+n  ,
                       grid->mu[l  ][m+1]+n  , grid->mu[l+1][m+1]+n  , d);
            muxy = r1;
            muxz = r2;
            muyz = r3;
            }
         for(n = data->nmin; n < data->nmax; n++) { 
            s3 = rxy[n]; s4 = rxz[n]; s5 = rxy[n];
            rxy[n] = F0[n]*s3 + F3[n]*(dv3[n] + dv6[n]);
            rxz[n] = F0[n]*s4 + F3[n]*(dv1[n] + dv4[n]);
            ryz[n] = F0[n]*s5 + F3[n]*(dv2[n] + dv5[n]);
            s3 = (s3 + rxy[n])*0.5;
            s4 = (s4 + rxz[n])*0.5;
            s5 = (s5 + ryz[n])*0.5;
            Txy[n] += dt*muxy[n]*(dv4[n] + dv1[n]) +dt*s3;
            Txz[n] += dt*muxz[n]*(dv5[n] + dv2[n]) +dt*s4;
            Tyz[n] += dt*muyz[n]*(dv6[n] + dv3[n]) +dt*s5;
            }
            */

         /* fortran version */
         if (grid->ave) {
            uf_(r1+n, grid->mu[l  ][m  ]+n-1, grid->mu[l  ][m  ]+n  ,
                      grid->mu[l+1][m  ]+n-1, grid->mu[l+1][m  ]+n  , &d);
            uf_(r2+n, grid->mu[l  ][m  ]+n-1, grid->mu[l  ][m  ]+n  ,
                      grid->mu[l  ][m+1]+n-1, grid->mu[l  ][m+1]+n  , &d);
            uf_(r3+n, grid->mu[l  ][m  ]+n  , grid->mu[l+1][m  ]+n  ,
                      grid->mu[l  ][m+1]+n  , grid->mu[l+1][m+1]+n  , &d);
            muxy = r1;
            muxz = r2;
            muyz = r3;
            }
         da4f_(rxy+n, rxz+n, ryz+n, Txy+n, Txz+n, Tyz+n, muxy+n, muxz+n, muyz+n,
               F0+n, F3+n, dv1+n, dv2+n, dv3+n, dv4+n, dv5+n, dv6+n, &dt, &d);
         }
      }

   /* free memory */
   if (grid->ave) {
      free(r1);
      free(r2);
      free(r3);
      }
   free(dv1);
   free(dv2);
   free(dv3);
   free(dv4);
   free(dv5);
   free(dv6);

}



/* acoustic (with density) equations below */
/* equations being solved
   dVx/dt  = (1./rho)*(dT/dx) 
   dVy/dt  = (1./rho)*(dT/dy) 
   dVz/dt  = (1./rho)*(dT/dz)
   note: use Txx value for T (coding)
*/
update4V3DA(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T; 
   float ***p, c1, c2;
   float *v1, *v2, *v3, *v4;
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08, *f09, *f10, *f11, *f12;
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08,  g09,  g10,  g11,  g12;
   int l, m, n, lmin, mmin, nmin, lmax, mmax, nmax, d;

   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   T   = data->T; 
   Vx  = data->Vx;  Vy  = data->Vy;  Vz  = data->Vz;
   p = grid->irho;
   lmin = data->lmin;
   mmin = data->mmin;
   nmin = data->nmin;
   lmax = data->lmax;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx; 
   c2 = C2*par->dt/grid->dx;

   /* update velocities along surface (done in surface routines) */

   /* update velocities for rest of grid */
   for(l = lmin; l < lmax; l++) {
      for(m = mmin; m < mmax; m++) {
         v1  = Vx [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f01 = T  [l  ][m  ]+nmin-2;
         f02 = T  [l  ][m  ]+nmin-1;
         f03 = T  [l  ][m  ]+nmin  ;
         f04 = T  [l  ][m  ]+nmin+1;
         /* c formulation */
         #if C_SOURCE_CODE
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            *v1 += (c1*(g03 - g02                        ) +
                    c2*(g04 - g01                        ))*(*v4);
            v1++;
            v4++;
            }
         /* fortran formulation */
         #else
         dtf_(v1, f01, f02, f03, f04, &c1, &c2, &d, v4);
         #endif
         v2  = Vy [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f05 = T  [l-1][m  ]+nmin  ;
         f06 = T  [l  ][m  ]+nmin  ;
         f07 = T  [l+1][m  ]+nmin  ;
         f08 = T  [l+2][m  ]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         for(n = nmin; n < nmax; n++) {
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            *v2 += (c1*(            g07 - g06            ) +
                    c2*(            g08 - g05            ))*(*v4);
            v2++;
            v4++;
            }
         /* fortran formulation */
         #else
         dtf_(v2, f05, f06, f07, f08, &c1, &c2, &d, v4);
         #endif
         v3  = Vz [l  ][m  ]+nmin;
         v4  = p  [l  ][m  ]+nmin;
         f09 = T  [l  ][m-1]+nmin  ;
         f10 = T  [l  ][m  ]+nmin  ;
         f11 = T  [l  ][m+1]+nmin  ;
         f12 = T  [l  ][m+2]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         for(n = nmin; n < nmax; n++) {
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            *v3 += (c1*(                        g11 - g10) +
                    c2*(                        g12 - g09))*(*v4);
            v3++;
            v4++;
            }
         /* fortran formulation */
         #else
         dtf_(v3, f09, f10, f11, f12, &c1, &c2, &d, v4);
         #endif
         }
      }
              
}




/* next 8 equations update normal stresses
   h01     = (dVx/dx)
   h02     = (dVy/dy)
   h03     = (dVz/dz)
   h04     = (lamda)*(dVx/dx + dVy/dy + dVz/dz)
   dT  /dt =                    lamda*(dVx/dx + dVy/dy + dVz/dz) 
*/
update4T3DA(par, grid, data)
Parameter      *par;
FdGrid         *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       c1, c2; 
   float *v1,                          *v7     ;
   float *f01, *f02, *f03, *f04, *f05, *f06, *f07, *f08, *f09, *f10, *f11, *f12;
   float  g01,  g02,  g03,  g04,  g05,  g06,  g07,  g08,  g09,  g10,  g11,  g12;
   float  h01,  h02,  h03,  h04;
   int l, m, n, lmin, mmin, nmin, lmax, mmax, nmax, d;

   /* initialize */
   if (grid->dx != grid->dy) {fprintf(stderr, "ERROR: dx != dy\n"); exit(-1);}
   if (grid->dx != grid->dz) {fprintf(stderr, "ERROR: dx != dz\n"); exit(-1);}
   T   = data->T; 
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   lmin = data->lmin;
   mmin = data->mmin;
   nmin = data->nmin;
   lmax = data->lmax;
   mmax = data->mmax;
   nmax = data->nmax;
   d    = nmax - nmin;
   c1 = C1*par->dt/grid->dx; 
   c2 = C2*par->dt/grid->dx;

   /* update stresses along surface (done in surface routines) */

   /* update stresses for rest of grid */
   for(l = lmin; l < lmax; l++) {
      for(m = mmin; m < mmax; m++) {
         v1  = T  [l  ][m  ]+nmin;
         v7  = g  [l  ][m  ]+nmin;
         f01 = Vx [l  ][m  ]+nmin-1;
         f02 = Vx [l  ][m  ]+nmin  ;
         f03 = Vx [l  ][m  ]+nmin+1;
         f04 = Vx [l  ][m  ]+nmin+2;
         f05 = Vy [l-2][m  ]+nmin  ;
         f06 = Vy [l-1][m  ]+nmin  ;
         f07 = Vy [l  ][m  ]+nmin  ;
         f08 = Vy [l+1][m  ]+nmin  ;
         f09 = Vz [l  ][m-2]+nmin  ;
         f10 = Vz [l  ][m-1]+nmin  ;
         f11 = Vz [l  ][m  ]+nmin  ;
         f12 = Vz [l  ][m+1]+nmin  ;
         /* c formulation */
         #if C_SOURCE_CODE
         for(n = nmin; n < nmax; n++) {
            g01 = *f01++;
            g02 = *f02++;
            g03 = *f03++;
            g04 = *f04++;
            g05 = *f05++;
            g06 = *f06++;
            g07 = *f07++;
            g08 = *f08++;
            g09 = *f09++;
            g10 = *f10++;
            g11 = *f11++;
            g12 = *f12++;
            h01 = (c1*(g03 - g02) + c2*(g04 - g01));
            h02 = (c1*(g07 - g06) + c2*(g08 - g05));
            h03 = (c1*(g11 - g10) + c2*(g12 - g09));
            h04 = (h01 + h02 + h03)*(*v7);
            *v1 +=           h04;
            v1++;
            v7++;
            }
         /* fortran formulation */
         #else
         dt3da_(v1, f01, f02, f03, f04, f05, f06,
                    f07, f08, f09, f10, f11, f12, &c1, &c2, &d, v7);
         #endif
         }
      }

}

