/******************************************************************************
   vectorc: vector derivative routines (C) for elastic 3-D finite-difference

    ================================================= 
   |     Copyright (c) Shawn C. Larsen               |
   |     Seismological Laboratory                    |
   |     California Institute of Technology          |
   |     Pasadena, CA 91125                          |
   |     All Rights Reserved.                        |
    ================================================= 

   getDT     : 4th order (spatial) derivative for stress
   getDT12   : 2nd order (spatial) derivative for stress
   getDV     : 4th order (spatial) derivative for velocity
   getDV0    : 4th order (spatial) derivative for velocity
   getDV02   : 2nd order (spatial) derivative for velocity
   getDV1    : 4th order (spatial) derivative for velocity
   getDV2    : 4th order (spatial) derivative for velocity
   getDV3    : 4th order (spatial) derivative for velocity
   getDV32   : 4th order (spatial) derivative for velocity
   addVector2: add velocity derivatives
   addVector3: add velocity derivatives
   getR      : averaging for density
   getU      : averaging for mu's
   getNullV  : null a vector
   getNegV   : get negative of a vector
   getA1b    : get P velocity squared
   getB1b    : get S velocity squared
   getS1b    : square root of vector
   getZ1b    : paraxial coefficients
   getP1b    : paraxial extrapolators
   getSV02   : 2nd order updates for surface (stress)
   getSV12   : 2nd order updates for surface (stress)

   compile: cc -g -c vectorc.c

   note: input units are km, s, km/s, g/cm3
   note: used  units are mks (m, s, m/s, kg/m3, N-m)

   author: shawn larsen
   date  : july 1989
   where : caltech
*****************************************************************************/

#include "elas.h"


getDT(d, r0, r1, r2, r3, c1, c2, l, p)
float *d, *r0, *r1, *r2, *r3, c1, c2, *p;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) += (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)))*(*(p++));
}



getDV(d, r0, r1, r2, r3, c1, c2, l)
float *d, *r0, *r1, *r2, *r3, c1, c2;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)));
}



getDV0(d, r0, r1, r2, r3, c1, c2, l, g, u)
float *d, *r0, *r1, *r2, *r3, c1, c2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) = (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)));
}



/* not used */
getDV1(d, r0, r1, r2, r3, c1, c2, l, g, u)   
float *d, *r0, *r1, *r2, *r3, c1, c2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) += (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)))* 
      (*(g++)            );
}



/* not used */
getDV2(d, r0, r1, r2, r3, c1, c2, l, g, u)   
float *d, *r0, *r1, *r2, *r3, c1, c2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) += (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)))*
      (*(g++) + *(u++)*2.);
}



getDV3(d, r0, r1, r2, r3, c1, c2, l, g, u)
float *d, *r0, *r1, *r2, *r3, c1, c2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) += (c1*(*(r2++) - *(r1++)) + c2*(*(r3++) - *(r0++)))*
      (         *(u++)   );
}



addVector2(d, v1, v2, l, g, u)
float *d, *v1, *v2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) { 
      *(d++) += (*(g) + *(u++)*2.)*(*(v1++)) + *(g)*(*(v2++));
      g++;
      }
}



addVector3(d, v1, v2, v3, l, g, u)
float *d, *v1, *v2, *v3, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) { 
      *(d++) += (*(g) + *(u++)*2.)*(*(v1++)) + *(g)*((*(v2++)) + (*(v3++)));
      g++;
      }
}



addVectorA(d, v1, v2, v3, l, g   )
float *d, *v1, *v2, *v3, *g    ;
int l;
{
   int n;
   for(n = 0; n < l; n++) { 
      *(d++) += *(g)*((*(v1++)) + (*(v2++)) + (*(v3++)));
      g++;
      }
}



getR(d, r0, r1, l)
float *d, *r0, *r1;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) = 2.0/(*(r0++) + *(r1++));
}



getU(d, u0, u1, u2, u3, l)
float *d, *u0, *u1, *u2, *u3;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      /* inverse averaging the efficient but precision unacceptable way */
      /*
      *(d++) = 4.0*((*u0)*(*u1)*(*u2)*(*u3))/
                   ((*u2)*(*u3)*((*u0) + (*u1)) +
                    (*u0)*(*u1)*((*u2) + (*u3)));
      u0++; u1++; u2++; u3++;
      */

      /* inverse averaging the precision acceptable way */
      *(d++) = 4.0/(1./(*u0++) + 1./(*u1++) + 1./(*u2++) + 1./(*u3++));

      /* straight averaging */
      /*
      *(d++) = 0.25*((*u0++) + (*u1++) + (*u2++) + (*u3++));
      */
      }
}


/* following is vectorized paraxial updates */

getDV02(d, r1, r2, c, l, g, u)
float *d, *r1, *r2, c, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = c*(*(r2++) - *(r1++));
}


getDA02(d, r1, r2, c, l      )
float *d, *r1, *r2, c        ;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = c*(*(r2++) - *(r1++));
}


getDV32(d, r1, r2, c, l, g, u)
float *d, *r1, *r2, *g, *u, c;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) += c*((*(r2++) - *(r1++)))*(*(u++));
}


getNullV(d, l)
float *d; 
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) = 0.;
}


getNegV(d, v, l)
float *d, *v; 
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) = -(*(v++));
}


getDT12(d, r1, r2, s1, s2, t1, t2, c, l, p)
float *d, *r1, *r2, *s1, *s2, *t1, *t2, *p, c;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) += c*((*(r2++) - *(r1++)) + 
                   (*(s2++) - *(s1++)) + 
                   (*(t2++) - *(t1++)))*(*(p++));
}


getDA12(d, r1, r2,                 c, l, p) /* acoustic */
float *d, *r1, *r2,                     *p, c;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) += c*((*(r2++) - *(r1++)))*(*(p++));
}


getA1b(d, l, g, u, p)
float *d, *g, *u, *p;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = ((*(g++)) + 2.*(*(u++)))*(*(p++));
}


getA1a(d, l, g,    p)
float *d, *g,     *p;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = (*(g++))*(*(p++));
}


getB1b(d, l, g, u, p)
float *d, *g, *u, *p;
int l;
{
   int n;
   for(n = 0; n < l; n++)
      *(d++) = (*(u++))*(*(p++));
}


getS1b(d, l)
float *d;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      *(d) = sqrt(*(d));
      d++;
      }
}


getZ1b(d, v, dh, dt, l)
float *d, *v, dh, dt;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      *(d++) = (dh - (*v)*dt)/(dh + (*v)*dt);
      v++;
      }
}



getP1b(d, v, r, t, l)
float *d, *v, *r, *t;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      (*d) = *(r++) + (*(t++))*((*d) - *(v++));
      d++;
      }
}


getP1a(d, v, r,    l)
float *d, *v, *r    ;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      (*d) = *(r++) +          ((*d) - *(v++));
      d++;
      }
}


getSV02(d, v, r1, r2, s1, s2, l)
float *d, *v, *r1, *r2, *s1, *s2;
int l;
{
   int n;
   for(n = 0; n < l; n++) 
      *(d++) = *(v++) + (*(r2++) - *(r1++)) + (*(s2++) - *(s1++));
}


getSV12(d, v, r1, r2, s1, s2, l, g, u)
float *d, *v, *r1, *r2, *s1, *s2, *g, *u;
int l;
{
   int n;
   for(n = 0; n < l; n++) {
      *(d++) = *(v++) + ((*(r2++) - *(r1++)) + (*(s2++) - *(s1++)))*
                        (*g)/((*g) + 2.*(*(u++)));
      g++;
      }
}
