/************************************************************************
   paraxial3d: paraxial boundary condition routines for 3D elastic code

   vabsorbLVR  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVL  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVT  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVB  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVF  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVK  :  Clayton-Engquist A1 absorbing boundary condition - vector
   vabsorbLVLa :  account for acoustic instability
   vabsorbLVBa :  account for acoustic instability
   vabsorbLVKa :  account for acoustic instability
   tabsorbLVR  :  topographic (averaging) boundary conditions              
   tabsorbLVL  :  topographic (averaging) boundary conditions              
   tabsorbLVT  :  topographic (averaging) boundary conditions              
   tabsorbLVB  :  topographic (averaging) boundary conditions              
   tabsorbLVF  :  topographic (averaging) boundary conditions              
   tabsorbLVK  :  topographic (averaging) boundary conditions              
   tabsorbLVLa :  topographic (averaging) boundary conditions (instability)
   tabsorbLVBa :  topographic (averaging) boundary conditions (instability)
   tabsorbLVKa :  topographic (averaging) boundary conditions (instability)
   aabsorbLVF  :  acoustic absorbing side boundary condition
   aabsorbLVR  :  acoustic absorbing side boundary condition
   aabsorbLVKa :  acoustic absorbing side boundary condition (instability)
   aabosrbLVBa :  acoustic absorbing side boundary condition (instability)
   aabsorbLVLa :  acoustic absorbing side boundary condition (instability)
   dedgeVBL    :  edge velocity boundary condition (bottom left)
   dedgeVTL    :  edge velocity boundary condition (top left)
   dedgeVBR    :  edge velocity boundary condition (bottom right)
   dedgeVTR    :  edge velocity boundary condition (top right)
   dedgeVFL    :  edge velocity boundary condition (front left)
   dedgeVFR    :  edge velocity boundary condition (front right)
   dedgeVKL    :  edge velocity boundary condition (back left)
   dedgeVKR    :  edge velocity boundary condition (back right)
   dedgeVBF    :  edge velocity boundary condition (bottom front)
   dedgeVTF    :  edge velocity boundary condition (top front)
   dedgeVBK    :  edge velocity boundary condition (bottom back)
   dedgeVTK    :  edge velocity boundary condition (top back)
   aedgeVBL    :  edge velocity boundary condition (bottom left)
   aedgeVTL    :  edge velocity boundary condition (top left)
   aedgeVBR    :  edge velocity boundary condition (bottom right)
   aedgeVTR    :  edge velocity boundary condition (top right)
   aedgeVFL    :  edge velocity boundary condition (front left)
   aedgeVFR    :  edge velocity boundary condition (front right)
   aedgeVKL    :  edge velocity boundary condition (back left)
   aedgeVKR    :  edge velocity boundary condition (back right)
   aedgeVBF    :  edge velocity boundary condition (bottom front)
   aedgeVTF    :  edge velocity boundary condition (top front)
   aedgeVBK    :  edge velocity boundary condition (bottom back)
   aedgeVTK    :  edge velocity boundary condition (top back)
   dcornerVBLF :  corner velocity boundary condition (front bottom left)
   dcornerVBRF :  corner velocity boundary condition (front bottom right)
   dcornerVTLF :  corner velocity boundary condition (front top left)
   dcornerVTRF :  corner velocity boundary condition (front top right)
   dcornerVBLK :  corner velocity boundary condition (back bottom left)
   dcornerVBRK :  corner velocity boundary condition (back bottom right)
   dcornerVTLK :  corner velocity boundary condition (back top left)
   dcornerVTRK :  corner velocity boundary condition (back top right)
   acornerVBLF :  corner velocity boundary condition (front bottom left)
   acornerVBRF :  corner velocity boundary condition (front bottom right)
   acornerVTLF :  corner velocity boundary condition (front top left)
   acornerVTRF :  corner velocity boundary condition (front top right)
   acornerVBLK :  corner velocity boundary condition (back bottom left)
   acornerVBRK :  corner velocity boundary condition (back bottom right)
   acornerVTLK :  corner velocity boundary condition (back top left)
   acornerVTRK :  corner velocity boundary condition (back top right)
   vboundTR    :  stress boundary condition (vectorized) (right side) 
   vboundTL    :  stress boundary condition (vectorized) (left side) 
   vboundTT    :  stress boundary condition (vectorized) (top)   
   vboundTB    :  stress boundary condition (vectorized) (bottom) 
   vboundTF    :  stress boundary condition (vectorized) (front) 
   vboundTK    :  stress boundary condition (vectorized) (back) 
   vboundTKa   :  account for acoustic instability
   vboundTBa   :  account for acoustic instability
   vboundTLa   :  account for acoustic instability
   aboundTF    :  acoustic absorbing side boundary condition
   aboundTR    :  acoustic absorbing side boundary condition
   aboundTBa   :  acoustic absorbing side boundary condition (instability)
   aboundTLa   :  acoustic absorbing side boundary condition (instability)
   aboundTKa   :  acoustic absorbing side boundary condition (instability)
   dedgeTBL    :  edge stress boundary condition (bottom left)
   dedgeTTL    :  edge stress boundary condition (top left)
   dedgeTBR    :  edge stress boundary condition (bottom right)
   dedgeTTR    :  edge stress boundary condition (top right)
   dedgeTFL    :  edge stress boundary condition (front left)
   dedgeTFR    :  edge stress boundary condition (front right)
   dedgeTKL    :  edge stress boundary condition (back left)
   dedgeTKR    :  edge stress boundary condition (back right)
   dedgeTBF    :  edge stress boundary condition (bottom front)
   dedgeTTF    :  edge stress boundary condition (top front)
   dedgeTBK    :  edge stress boundary condition (bottom back )
   dedgeTTK    :  edge stress boundary condition (top back )
   aedgeTBL    :  edge stress boundary condition (bottom left)
   aedgeTTL    :  edge stress boundary condition (top left)
   aedgeTBR    :  edge stress boundary condition (bottom right)
   aedgeTTR    :  edge stress boundary condition (top right)
   aedgeTFL    :  edge stress boundary condition (front left)
   aedgeTFR    :  edge stress boundary condition (front right)
   aedgeTKL    :  edge stress boundary condition (back left)
   aedgeTKR    :  edge stress boundary condition (back right)
   aedgeTBF    :  edge stress boundary condition (bottom front)
   aedgeTTF    :  edge stress boundary condition (top front)
   aedgeTBK    :  edge stress boundary condition (bottom back )
   aedgeTTK    :  edge stress boundary condition (top back )
   dcornerTBLF :  corner stress boundary condition (front bottom left)
   dcornerTBRF :  corner stress boundary condition (front bottom right)
   dcornerTTLF :  corner stress boundary condition (front top left)
   dcornerTTRF :  corner stress boundary condition (front top right)
   dcornerTBLK :  corner stress boundary condition (back bottom left)
   dcornerTBRK :  corner stress boundary condition (back bottom right)
   dcornerTTLK :  corner stress boundary condition (back top left)
   dcornerTTRK :  corner stress boundary condition (back top right)
   acornerTBLF :  corner stress boundary condition (front bottom left)
   acornerTBRF :  corner stress boundary condition (front bottom right)
   acornerTTLF :  corner stress boundary condition (front top left)
   acornerTTRF :  corner stress boundary condition (front top right)
   acornerTBLK :  corner stress boundary condition (back bottom left)
   acornerTBRK :  corner stress boundary condition (back bottom right)
   acornerTTLK :  corner stress boundary condition (back top left)
   acornerTTRK :  corner stress boundary condition (back top right)

   in testc.c
   ----------

   * means not made acoustic
   
   compile: cc -g -c paraxial.c

   author:  shawn larsen
   date:    april 1989
   where:   caltech

   author:  denise maddix (modified)
   date  :  10 January 1995
   where :  llnl

   author:  shawn larsen  (modified acoustic routines)
   date:    january 1997, august 2001                       
   where:   llnl                               
**************************************************************************/

#include "elas.h"


/*************************************************************************
VELOCITY SIDE BOUNDARY CONDITIONS
*************************************************************************/

vabsorbLVR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVR(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from right boundary */
   n = grid->n;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n];
         YLast = Vy[m][n];
         ZLast = Vz[m][n];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                      (T1             - Txy[l-1][m][n]) +
                      (T2             - Txz[l][m-1][n]))*P1;
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n+1] = XLast + xP*(Vx[m][n+1] - Vx[m][n]);
         Vy[m][n+1] = YLast + xS*(Vy[m][n+1] - Vy[m][n]);
         Vz[m][n+1] = ZLast + xS*(Vz[m][n+1] - Vz[m][n]);
         }
      }

}


vabsorbLVL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVL(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n];
         YLast = Vy[m][n];
         ZLast = Vz[m][n];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                      (T1             - Txy[l-1][m][n]) +
                      (T2             - Txz[l][m-1][n]))*P1;
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n-1] = XLast + xP*(Vx[m][n-1] - Vx[m][n]);
         Vy[m][n-1] = YLast + xS*(Vy[m][n-1] - Vy[m][n]);
         Vz[m][n-1] = ZLast + xS*(Vz[m][n-1] - Vz[m][n]);
         }
      }

}


vabsorbLVT(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVT(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one up from top boundary */
   m = -1;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m-1]+n, Vx[l][m]+n, XLast+n, zS+n, d);
      getP1b(Vy[l][m-1]+n, Vy[l][m]+n, YLast+n, zS+n, d);
      getP1b(Vz[l][m-1]+n, Vz[l][m]+n, ZLast+n, zP+n, d);
      */
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m-1]+n, Vx[l][m]+n, XLast+n, zS+n, &d);
      dp1bf_(Vy[l][m-1]+n, Vy[l][m]+n, YLast+n, zS+n, &d);
      dp1bf_(Vz[l][m-1]+n, Vz[l][m]+n, ZLast+n, zP+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );

}


vabsorbLVB(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVB(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m+1]+n, Vx[l][m]+n, XLast+n, zS+n, d);
      getP1b(Vy[l][m+1]+n, Vy[l][m]+n, YLast+n, zS+n, d);
      getP1b(Vz[l][m+1]+n, Vz[l][m]+n, ZLast+n, zP+n, d);
      */
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m+1]+n, Vx[l][m]+n, XLast+n, zS+n, &d);
      dp1bf_(Vy[l][m+1]+n, Vy[l][m]+n, YLast+n, zS+n, &d);
      dp1bf_(Vz[l][m+1]+n, Vz[l][m]+n, ZLast+n, zP+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );

}


vabsorbLVF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVF(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = -1;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n, yS+n, d);
      getP1b(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, d);
      getP1b(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, d);
      */
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n, yS+n, &d);
      dp1bf_(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, &d);
      dp1bf_(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );

}


vabsorbLVK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVK(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = grid->l;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l+1][m]+n, Vx[l][m]+n, XLast+n, yS+n, d);
      getP1b(Vy[l+1][m]+n, Vy[l][m]+n, YLast+n, yP+n, d);
      getP1b(Vz[l+1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, d);
      */
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l+1][m]+n, Vx[l][m]+n, XLast+n, yS+n, &d);
      dp1bf_(Vy[l+1][m]+n, Vy[l][m]+n, YLast+n, yP+n, &d);
      dp1bf_(Vz[l+1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );

}


vabsorbLVLa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVLa(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n  ];
         YLast = Vy[m][n-1];
         ZLast = Vz[m][n-1];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                      (T1             - Txy[l-1][m][n]) +
                      (T2             - Txz[l][m-1][n]))*P1;
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         n--;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         n++;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n-1] = XLast + xP*(Vx[m][n-1] - Vx[m][n  ]);
         Vy[m][n-2] = YLast + xS*(Vy[m][n-2] - Vy[m][n-1]);
         Vz[m][n-2] = ZLast + xS*(Vz[m][n-2] - Vz[m][n-1]);
         }
      }

}


vabsorbLVBa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVBa(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m  ]+n), d*sizeof(float));
      /*
      */
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      m++;
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      m--;
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n, zS+n, d);
      getP1b(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n, zS+n, d);
      getP1b(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, d);
      /*
      */
      /*
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      m++;
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      m--;
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n, zS+n, &d);
      dp1bf_(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n, zS+n, &d);
      dp1bf_(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );

}


vabsorbLVKa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* do paraxial averaging, if required */
   if (grid->ave) {tabsorbLVKa(par, grid, data); return;}

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = grid->l;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l+1][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l  ][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l+1][m]+n), d*sizeof(float));
      /*
      */
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      l++;
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,p[l][m]+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,p[l][m]+n);
      l--;
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      getS1b(A+n, d);
      getS1b(B+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n, yS+n, d);
      getP1b(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, d);
      getP1b(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n, yS+n, d);
      /*
      */
      /*
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      l++;
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,p[l][m]+n);
      l--;
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, p[l][m]+n);
      ds1bf_(A+n, &d);
      ds1bf_(B+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n, yS+n, &d);
      dp1bf_(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, &d);
      dp1bf_(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n, yS+n, &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );

}


tabsorbLVR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from right boundary */
   n = grid->n;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n];
         YLast = Vy[m][n];
         ZLast = Vz[m][n];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m  ][n-1]));
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                       (T1             - Txy[l-1][m][n]) +
                       (T2             - Txz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ]));
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                       (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                       (T3             - Tyz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ]));
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                       (T3             - Tyz[l-1][m][n]) +
                       (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n+1] = XLast + xP*(Vx[m][n+1] - Vx[m][n]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vy[m][n+1] = YLast + xS*(Vy[m][n+1] - Vy[m][n]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vz[m][n+1] = ZLast + xS*(Vz[m][n+1] - Vz[m][n]);
         }
      }

}


tabsorbLVL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n];
         YLast = Vy[m][n];
         ZLast = Vz[m][n];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m  ][n-1]));
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                      (T1             - Txy[l-1][m][n]) +
                      (T2             - Txz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ]));
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ]));
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n-1] = XLast + xP*(Vx[m][n-1] - Vx[m][n]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vy[m][n-1] = YLast + xS*(Vy[m][n-1] - Vy[m][n]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vz[m][n-1] = ZLast + xS*(Vz[m][n-1] - Vz[m][n]);
         }
      }

}


tabsorbLVT(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one up from top boundary */
   m = -1;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m-1]+n, Vx[l][m]+n, XLast+n, zS+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vy[l][m-1]+n, Vy[l][m]+n, YLast+n, zS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(A+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getP1b(Vz[l][m-1]+n, Vz[l][m]+n, ZLast+n, zP+n, d);
      */
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m-1]+n, Vx[l][m]+n, XLast+n, zS+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vy[l][m-1]+n, Vy[l][m]+n, YLast+n, zS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(A+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vz[l][m-1]+n, Vz[l][m]+n, ZLast+n, zP+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


tabsorbLVB(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m+1]+n, Vx[l][m]+n, XLast+n, zS+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vy[l][m+1]+n, Vy[l][m]+n, YLast+n, zS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(A+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getP1b(Vz[l][m+1]+n, Vz[l][m]+n, ZLast+n, zP+n, d);
      */
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m+1]+n, Vx[l][m]+n, XLast+n, zS+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vy[l][m+1]+n, Vy[l][m]+n, YLast+n, zS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(A+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vz[l][m+1]+n, Vz[l][m]+n, ZLast+n, zP+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


tabsorbLVF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = -1;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n, yS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(A+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getP1b(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, d);
      */
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n, yS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(A+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


tabsorbLVK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = grid->l;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l+1][m]+n, Vx[l][m]+n, XLast+n, yS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(A+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getP1b(Vy[l+1][m]+n, Vy[l][m]+n, YLast+n, yP+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vz[l+1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, d);
      */
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l+1][m]+n, Vx[l][m]+n, XLast+n, yS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(A+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vy[l+1][m]+n, Vy[l][m]+n, YLast+n, yP+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vz[l+1][m]+n, Vz[l][m]+n, ZLast+n, yS+n, &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


tabsorbLVLa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float **Vx, **Vy, **Vz, **G, **U, **P;
   float G0, U0, P0, P1, T1, T2, T3, c;
   float XLast, YLast, ZLast, A, B, xP, xS;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l]; U   = grid->mu [l]; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n  ];
         YLast = Vy[m][n-1];
         ZLast = Vz[m][n-1];
         G0    = G[m][n];
         U0    = U[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m  ][n-1]));
         Vx[m][n] += ((Txx[l][m][n]   - Txx[l][m][n-1]) +
                      (T1             - Txy[l-1][m][n]) +
                      (T2             - Txz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ]));
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ]));
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         n--;
         T1    = Txy[l][m][n];
         T2    = Txz[l][m][n];
         T3    = Tyz[l][m][n];
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ]));
         Vy[m][n] += ((Txy[l][m][n+1] - T1            ) +
                      (Tyy[l+1][m][n] - Tyy[l][m][n]  ) +
                      (T3             - Tyz[l][m-1][n]))*P1;
         P1 = c*(2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ]));
         Vz[m][n] += ((Txz[l][m][n+1] - T2            ) +
                      (T3             - Tyz[l-1][m][n]) +
                      (Tzz[l][m+1][n] - Tzz[l][m][n]  ))*P1;
         n++;
         A = sqrt((G0 + 2.*U0)*P0);
         B = sqrt((U0*P0));
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vx[m][n-1] = XLast + xP*(Vx[m][n-1] - Vx[m][n  ]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l+1][m  ][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vy[m][n-2] = YLast + xS*(Vy[m][n-2] - Vy[m][n-1]);
         B = sqrt((U0*2./(grid->rho[l][m][n] + grid->rho[l  ][m+1][n  ])));
         xS = (grid->dx - B*par->dt)/(grid->dx + B*par->dt);
         Vz[m][n-2] = ZLast + xS*(Vz[m][n-2] - Vz[m][n-1]);
         }
      }

}


tabsorbLVBa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *zP, *zS, c;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   zS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m  ]+n), d*sizeof(float));
      /*
      */
      m++;
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      m--;
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n, zS+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(B+n, d);
      getZ1b(zS+n, B+n, grid->dz, par->dt, d);
      getP1b(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n, zS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(A+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getP1b(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, d);
      /*
      */
      /*
      m++;
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      m--;
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);

      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n, zS+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(B+n, &d);
      dz1bf_(zS+n, B+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n, zS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(A+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dp1bf_(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(zP   );
   free(zS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


tabsorbLVKa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, ***p; 
   float *rx, *ry, *rz;
   float *XLast, *YLast, *ZLast, *A, *B, *yP, *yS, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;
   p = grid->rho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = grid->l;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   B     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   yS    = (float *) malloc(grid->n*sizeof(float));
   rx    = (float *) malloc(grid->n*sizeof(float));
   ry    = (float *) malloc(grid->n*sizeof(float));
   rz    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l+1][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l  ][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l+1][m]+n), d*sizeof(float));
      /*
      */
      l++;
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      l--;
      getR(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, d);
      getR(ry+n, p[l][m]+n, p[l+1][m  ]+n  , d);
      getR(rz+n, p[l][m]+n, p[l  ][m+1]+n  , d);
      getDT12(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                         Txy[l-1][m]+n,Txy[l][m]+n,
                         Txz[l][m-1]+n,Txz[l][m]+n,c,d,rx+n);
      getDT12(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                         Tyy[l][m]+n,Tyy[l+1][m]+n,
                         Tyz[l][m-1]+n,Tyz[l][m]+n,c,d,ry+n);
      getDT12(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                         Tyz[l-1][m]+n,Tyz[l][m]+n,
                         Tzz[l][m]+n,Tzz[l][m+1]+n,c,d,rz+n);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rx+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n, yS+n, d);
      getA1b(A+n, d, g[l][m]+n, u[l][m]+n, ry+n);
      getS1b(A+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getP1b(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, d);
      getB1b(B+n, d, g[l][m]+n, u[l][m]+n, rz+n);
      getS1b(B+n, d);
      getZ1b(yS+n, B+n, grid->dy, par->dt, d);
      getP1b(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n, yS+n, d);
      /*
      */
      /*
      l++;
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      l--;
      rf_(rx+n, p[l][m]+n, p[l  ][m  ]+n-1, &d);
      rf_(ry+n, p[l][m]+n, p[l+1][m  ]+n  , &d);
      rf_(rz+n, p[l][m]+n, p[l  ][m+1]+n  , &d);
      dt12f_(Vx[l][m]+n,Txx[l][m]+n-1,Txx[l][m]+n,
                        Txy[l-1][m]+n,Txy[l][m]+n,
                        Txz[l][m-1]+n,Txz[l][m]+n,&c,&d,rx+n);
      dt12f_(Vy[l][m]+n,Txy[l][m]+n,Txy[l][m]+n+1,
                        Tyy[l][m]+n,Tyy[l+1][m]+n,
                        Tyz[l][m-1]+n,Tyz[l][m]+n,&c,&d,ry+n);
      dt12f_(Vz[l][m]+n,Txz[l][m]+n,Txz[l][m]+n+1,
                        Tyz[l-1][m]+n,Tyz[l][m]+n,
                        Tzz[l][m]+n,Tzz[l][m+1]+n,&c,&d,rz+n);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rx+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n, yS+n, &d);
      da1bf_(A+n, &d, g[l][m]+n, u[l][m]+n, ry+n);
      ds1bf_(A+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, &d);
      db1bf_(B+n, &d, g[l][m]+n, u[l][m]+n, rz+n);
      ds1bf_(B+n, &d);
      dz1bf_(yS+n, B+n, &(grid->dy), &(par->dt), &d);
      dp1bf_(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n, yS+n, &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(B    );
   free(yP   );
   free(yS   );
   free(rx   );
   free(ry   );
   free(rz   );

}


aabsorbLVF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       ***p; 
   float *XLast, *YLast, *ZLast, *A,     *yP,      c;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = -1;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m]+n), d*sizeof(float));
      /*                    
      getDA12(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                   
                                                   c,d,p[l][m]+n);
      getDA12(Vy[l][m]+n,                          
                         T  [l][m]+n,T  [l+1][m]+n,
                                                   c,d,p[l][m]+n);
      getDA12(Vz[l][m]+n,                          
                                                   
                         T  [l][m]+n,T  [l][m+1]+n,c,d,p[l][m]+n);
      getA1a(A+n, d, g[l][m]+n,            p[l][m]+n);
      getS1b(A+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getP1a(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n,       d);
      getP1b(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, d);
      getP1a(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n,       d);
      */
      da12f_(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                  
                                                  &c,&d,p[l][m]+n);
      da12f_(Vy[l][m]+n,                          
                        T  [l][m]+n,T  [l+1][m]+n,
                                                  &c,&d,p[l][m]+n);
      da12f_(Vz[l][m]+n,                          
                                                  
                        T  [l][m]+n,T  [l][m+1]+n,&c,&d,p[l][m]+n);
      da1af_(A+n, &d, g[l][m]+n,            p[l][m]+n);
      ds1bf_(A+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dp1af_(Vx[l-1][m]+n, Vx[l][m]+n, XLast+n,       &d);
      dp1bf_(Vy[l-1][m]+n, Vy[l][m]+n, YLast+n, yP+n, &d);
      dp1af_(Vz[l-1][m]+n, Vz[l][m]+n, ZLast+n,       &d);
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(yP   );

}


aabsorbLVR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***T;
   float **Vx, **Vy, **Vz, **G,      **P;
   float G0,     P0, P1,             c;
   float XLast, YLast, ZLast, A,    xP    ;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from right boundary */
   n = grid->n;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l];                   ; P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n];
         YLast = Vy[m][n];
         ZLast = Vz[m][n];
         G0    = G[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         Vx[m][n] += ((T  [l][m][n]   - T  [l][m][n-1])  
                                                         
                                                       )*P1;
         Vy[m][n] += (                                   
                      (T  [l+1][m][n] - T  [l][m][n]  )  
                                                       )*P1;
         Vz[m][n] += (                                   
                                                         
                      (T  [l][m+1][n] - T  [l][m][n]  ))*P1;
         A = sqrt((G0        )*P0);
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         Vx[m][n+1] = XLast + xP*(Vx[m][n+1] - Vx[m][n]);
         Vy[m][n+1] = YLast +    (Vy[m][n+1] - Vy[m][n]);
         Vz[m][n+1] = ZLast +    (Vz[m][n+1] - Vz[m][n]);
         }
      }

}


aabsorbLVKa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       ***p; 
   float *XLast, *YLast, *ZLast, *A,     *yP,      c;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = grid->l;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   yP    = (float *) malloc(grid->n*sizeof(float));
   for(m = mmin; m < data->mmax; m++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l+1][m]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l  ][m]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l+1][m]+n), d*sizeof(float));
      /*
      */
      getDA12(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                    
                                                   c,d,p[l][m]+n);
      getDA12(Vy[l][m]+n,                          
                         T  [l][m]+n,T  [l+1][m]+n,
                                                   c,d,p[l][m]+n);
      getDA12(Vz[l][m]+n,                          
                                                   
                         T  [l][m]+n,T  [l][m+1]+n,c,d,p[l][m]+n);
      l++;
      getDA12(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                   
                                                   c,d,p[l][m]+n);
      getDA12(Vz[l][m]+n,                          
                                                   
                         T  [l][m]+n,T  [l][m+1]+n,c,d,p[l][m]+n);
      l--;
      getA1a(A+n, d, g[l][m]+n,            p[l][m]+n);
      getS1b(A+n, d);
      getZ1b(yP+n, A+n, grid->dy, par->dt, d);
      getP1a(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n,       d);
      getP1b(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, d);
      getP1a(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n,       d);
      /*
      */
      /* not complete?: check?
      da12f_(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                  
                                                  &c,&d,p[l][m]+n);
      da12f_(Vy[l][m]+n,                          
                        T  [l][m]+n,T  [l+1][m]+n,
                                                  &c,&d,p[l][m]+n);
      da12f_(Vz[l][m]+n,                          
                                                  
                        T  [l][m]+n,T  [l][m+1]+n,&c,&d,p[l][m]+n);
      l++;
      da12f_(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                  
                                                  &c,&d,p[l][m]+n);
      da12f_(Vz[l][m]+n,                          
                                                  
                        T  [l][m]+n,T  [l][m+1]+n,&c,&d,p[l][m]+n);
      l--;
      da1af_(A+n, &d, g[l][m]+n,            p[l][m]+n);
      ds1bf_(A+n, &d);
      dz1bf_(yP+n, A+n, &(grid->dy), &(par->dt), &d);
      dp1af_(Vx[l+2][m]+n, Vx[l+1][m]+n, XLast+n,       &d);
      dp1bf_(Vy[l+1][m]+n, Vy[l  ][m]+n, YLast+n, yP+n, &d);
      dp1af_(Vz[l+2][m]+n, Vz[l+1][m]+n, ZLast+n,       &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(yP   );

}


aabsorbLVBa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       ***p; 
   float *XLast, *YLast, *ZLast, *A,     *zP,      c;
   int l, m, n, d;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   p = grid->irho;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;

   /* update velocities for grid planes */
   XLast = (float *) malloc(grid->n*sizeof(float));
   YLast = (float *) malloc(grid->n*sizeof(float));
   ZLast = (float *) malloc(grid->n*sizeof(float));
   A     = (float *) malloc(grid->n*sizeof(float));
   zP    = (float *) malloc(grid->n*sizeof(float));
   for(l = data->lmin; l < data->lmax; l++) {
      memcpy((char *) (XLast+n), (char *) (Vx[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (YLast+n), (char *) (Vy[l][m+1]+n), d*sizeof(float));
      memcpy((char *) (ZLast+n), (char *) (Vz[l][m  ]+n), d*sizeof(float));
      /*
      */
      getDA12(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                   
                                                   c,d,p[l][m]+n);
      getDA12(Vy[l][m]+n,                          
                         T  [l][m]+n,T  [l+1][m]+n,
                                                   c,d,p[l][m]+n);
      getDA12(Vz[l][m]+n,                          
                                                   
                         T  [l][m]+n,T  [l][m+1]+n,c,d,p[l][m]+n);
      m++;
      getDA12(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                   
                                                   c,d,p[l][m]+n);
      getDA12(Vy[l][m]+n,                          
                         T  [l][m]+n,T  [l+1][m]+n,
                                                   c,d,p[l][m]+n);
      m--;
      getA1a(A+n, d, g[l][m]+n,            p[l][m]+n);
      getS1b(A+n, d);
      getZ1b(zP+n, A+n, grid->dz, par->dt, d);
      getP1a(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n,       d);
      getP1a(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n,       d);
      getP1b(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, d);
      /*
      */
      /* not complete?: check?
      dt12f_(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                  
                                                  &c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,                          
                        T  [l][m]+n,T  [l+1][m]+n,
                                                  &c,&d,p[l][m]+n);
      dt12f_(Vz[l][m]+n,                          
                                                  
                        T  [l][m]+n,T  [l][m+1]+n,&c,&d,p[l][m]+n);
      m++;
      dt12f_(Vx[l][m]+n,T  [l][m]+n-1,T  [l][m]+n,
                                                  
                                                  &c,&d,p[l][m]+n);
      dt12f_(Vy[l][m]+n,                          
                        T  [l][m]+n,T  [l+1][m]+n,
                                                  &c,&d,p[l][m]+n);
      m--;
      da1af_(A+n, &d, g[l][m]+n,            p[l][m]+n);
      ds1bf_(A+n, &d);
      dz1bf_(zP+n, A+n, &(grid->dz), &(par->dt), &d);
      dp1af_(Vx[l][m+2]+n, Vx[l][m+1]+n, XLast+n,       &d);
      dp1af_(Vy[l][m+2]+n, Vy[l][m+1]+n, YLast+n,       &d);
      dp1bf_(Vz[l][m+1]+n, Vz[l][m  ]+n, ZLast+n, zP+n, &d);
      */
      }

   /* free memory */
   free(XLast);
   free(YLast);
   free(ZLast);
   free(A    );
   free(zP   );

}


aabsorbLVLa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***T;
   float **Vx, **Vy, **Vz, **G,      **P;
   float G0,     P0, P1,             c;
   float XLast, YLast, ZLast, A,    xP    ;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T; 

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;

   /* update velocities for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Vx  = data->Vx   [l]; Vy  = data->Vy [l]; Vz  = data->Vz  [l];
      G   = grid->lamda[l];                     P   = grid->irho[l];
      for(m = mmin; m < data->mmax; m++) {
         XLast = Vx[m][n  ];
         YLast = Vy[m][n-1];
         ZLast = Vz[m][n-1];
         G0    = G[m][n];
         P0    = P[m][n];
         P1    = c*P0;
         Vx[m][n] += ((T  [l][m][n]   - T  [l][m][n-1])  
                                                         
                                                       )*P1;
         Vy[m][n] += (                                   
                      (T  [l+1][m][n] - T  [l][m][n]  )  
                                                       )*P1;
         Vz[m][n] += (                                   
                                                         
                      (T  [l][m+1][n] - T  [l][m][n]  ))*P1;
         n--;
         Vy[m][n] += (                                   
                      (T  [l+1][m][n] - T  [l][m][n]  )  
                                                       )*P1;
         Vz[m][n] += (                                   
                                                         
                      (T  [l][m+1][n] - T  [l][m][n]  ))*P1;
         n++;
         A = sqrt((G0        )*P0);
         xP = (grid->dx - A*par->dt)/(grid->dx + A*par->dt);
         Vx[m][n-1] = XLast + xP*(Vx[m][n-1] - Vx[m][n  ]);
         Vy[m][n-2] = YLast +    (Vy[m][n-2] - Vy[m][n-1]);
         Vz[m][n-2] = ZLast +    (Vz[m][n-2] - Vz[m][n-1]);
         }
      }

}



/*************************************************************************
VELOCITY EDGE BOUNDARY CONDITIONS
*************************************************************************/

dedgeVBL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 m = grid-> m;      /* Points one down from bottom */
 n = -1;            /* Points one out from left side */
 for (l=0; l< grid->l; l++); {   /* bottom left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m-1][n+1]);
   data->Vy[l][m][n] = VyLast + lS*(data->Vy[l][m][n] - data->Vy[l][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m-1][n+1];
    VyLast = data->Vy[l][m-1][n+1]; 
    VzLast = data->Vz[l][m-1][n+1];
  }

}


dedgeVTL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 m = -1;      /* Points one out from top */
 n = -1;            /* Points one out from left side */
 for (l=0; l< grid->l; l++); {   /* top left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m+1][n+1]);
   data->Vy[l][m][n] = VyLast + lS*(data->Vy[l][m][n] - data->Vy[l][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m+1][n+1];
    VyLast = data->Vy[l][m+1][n+1]; 
    VzLast = data->Vz[l][m+1][n+1];
  }

}


dedgeVBR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 m = grid-> m;      /* Points one down from bottom */
 n = grid-> n;      /* Points one out from right side */
 for (l=0; l< grid->l; l++); {   /* bottom right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom right edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m-1][n-1]);
   data->Vy[l][m][n] = VyLast + lS*(data->Vy[l][m][n] - data->Vy[l][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m-1][n-1];
    VyLast = data->Vy[l][m-1][n-1]; 
    VzLast = data->Vz[l][m-1][n-1];
  }

}


dedgeVTR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 m = -1;      /* Points one out from top */
 n = grid-> n;      /* Points one out from right side */
 for (l=0; l< grid->l; l++); {   /* top right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top right edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m+1][n-1]);
   data->Vy[l][m][n] = VyLast + lS*(data->Vy[l][m][n] - data->Vy[l][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m+1][n-1];
    VyLast = data->Vy[l][m+1][n-1]; 
    VzLast = data->Vz[l][m+1][n-1];
  }

}


dedgeVFL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Points one out from front */
 n = -1;      /* Points one out from left side */
 for (m=0; m< grid->m; m++); {   /* front left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   front left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m][n+1]);
   data->Vz[l][m][n] = VzLast + lS*(data->Vz[l][m][n] - data->Vz[l+1][m][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m][n+1];
    VyLast = data->Vy[l+1][m][n+1]; 
    VzLast = data->Vz[l+1][m][n+1];
  }

}


dedgeVFR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Points one out from front */
 n = grid->n;      /* Points one out from right side */
 for (m=0; m< grid->m; m++); {   /* front right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])/grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   front left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m][n-1]);
   data->Vz[l][m][n] = VzLast + lS*(data->Vz[l][m][n] - data->Vz[l+1][m][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m][n-1];
    VyLast = data->Vy[l+1][m][n-1]; 
    VzLast = data->Vz[l+1][m][n-1];
  }

}


dedgeVKL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 n = -1;      /* Points one out from left side */
 for (m=0; m< grid->m; m++); {   /* back left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   back left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m][n+1]);
   data->Vz[l][m][n] = VzLast + lS*(data->Vz[l][m][n] - data->Vz[l-1][m][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m][n+1];
    VyLast = data->Vy[l-1][m][n+1]; 
    VzLast = data->Vz[l-1][m][n+1];
  }

}


dedgeVKR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 n = grid->n;      /* Points one out from right side */
 for (m=0; m< grid->m; m++); {   /* back right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);
    lS = (SQRT2*grid->dx - beta*par->dt) /(SQRT2*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   back left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m][n-1]);
   data->Vz[l][m][n] = VzLast + lS*(data->Vz[l][m][n] - data->Vz[l-1][m][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m][n-1];
    VyLast = data->Vy[l-1][m][n-1]; 
    VzLast = data->Vz[l-1][m][n-1];
  }

}


dedgeVBF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Points one out from front */
 m = grid->m;      /* Points one down from bottom */
 for (n=0; n< grid->n; n++); {   /* bottom front edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);
    lS = (SQRT2*grid->dy - beta*par->dt) /(SQRT2*grid->dy + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom front edge */
   data->Vx[l][m][n] = VxLast + lS*(data->Vx[l][m][n] - data->Vx[l+1][m-1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n];
    VyLast = data->Vy[l+1][m-1][n]; 
    VzLast = data->Vz[l+1][m-1][n];
  }

}


dedgeVTF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Points one out from front */
 m = -1;      /* Points one up from top */
 for (n=0; n< grid->n; n++); {   /* top front edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);
    lS = (SQRT2*grid->dy - beta*par->dt) /(SQRT2*grid->dy + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom front edge */
   data->Vx[l][m][n] = VxLast + lS*(data->Vx[l][m][n] - data->Vx[l+1][m+1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n];
    VyLast = data->Vy[l+1][m+1][n]; 
    VzLast = data->Vz[l+1][m+1][n];
  }

}


dedgeVBK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 m = grid->m;      /* Points one down from bottom */
 for (n=0; n< grid->n; n++); {   /* bottom back edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);
    lS = (SQRT2*grid->dy - beta*par->dt) /(SQRT2*grid->dy + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom back edge */
   data->Vx[l][m][n] = VxLast + lS*(data->Vx[l][m][n] - data->Vx[l-1][m-1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n];
    VyLast = data->Vy[l-1][m-1][n]; 
    VzLast = data->Vz[l-1][m-1][n];
  }

}


dedgeVTK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 m = -1;      /* Points one out from top */
 for (n=0; n< grid->n; n++); {   /* back top edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);
    lS = (SQRT2*grid->dy - beta*par->dt) /(SQRT2*grid->dy + beta*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top back edge */
   data->Vx[l][m][n] = VxLast + lS*(data->Vx[l][m][n] - data->Vx[l-1][m+1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n];
    VyLast = data->Vy[l-1][m+1][n]; 
    VzLast = data->Vz[l-1][m+1][n];
  }

}


aedgeVBL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 m = grid-> m;      /* Points one down from bottom */
 n = -1;            /* Points one out from left side */
 for (l=0; l< grid->l; l++); {   /* bottom left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m-1][n+1]);
   data->Vy[l][m][n] = VyLast +    (data->Vy[l][m][n] - data->Vy[l][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m-1][n+1];
    VyLast = data->Vy[l][m-1][n+1]; 
    VzLast = data->Vz[l][m-1][n+1];
  }

}

 
aedgeVTL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 m = -1;      /* Points one out from top */
 n = -1;            /* Points one out from left side */
 for (l=0; l< grid->l; l++); {   /* top left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m+1][n+1]);
   data->Vy[l][m][n] = VyLast +    (data->Vy[l][m][n] - data->Vy[l][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m+1][n+1];
    VyLast = data->Vy[l][m+1][n+1]; 
    VzLast = data->Vz[l][m+1][n+1];
  }

}


aedgeVBR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, lP;
 int l,m,n;
 m = grid-> m;      /* Points one down from bottom */
 n = grid-> n;      /* Points one out from right side */
 for (l=0; l< grid->l; l++); {   /* bottom right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom right edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m-1][n-1]);
   data->Vy[l][m][n] = VyLast +    (data->Vy[l][m][n] - data->Vy[l][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m-1][n-1];
    VyLast = data->Vy[l][m-1][n-1]; 
    VzLast = data->Vz[l][m-1][n-1];
  }

}


aedgeVTR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, lP;
 int l,m,n;
 m = -1;      /* Points one out from top */
 n = grid-> n;      /* Points one out from right side */
 for (l=0; l< grid->l; l++); {   /* top right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top right edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l][m+1][n-1]);
   data->Vy[l][m][n] = VyLast +    (data->Vy[l][m][n] - data->Vy[l][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l][m+1][n-1];
    VyLast = data->Vy[l][m+1][n-1]; 
    VzLast = data->Vz[l][m+1][n-1];
  }

}


aedgeVFL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Points one out from front */
 n = -1;      /* Points one out from left side */
 for (m=0; m< grid->m; m++); {   /* front left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   front left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m][n+1]);
   data->Vz[l][m][n] = VzLast +    (data->Vz[l][m][n] - data->Vz[l+1][m][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m][n+1];
    VyLast = data->Vy[l+1][m][n+1]; 
    VzLast = data->Vz[l+1][m][n+1];
  }

}


aedgeVFR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Points one out from front */
 n = grid->n;      /* Points one out from right side */
 for (m=0; m< grid->m; m++); {   /* front right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )/grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   front left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m][n-1]);
   data->Vz[l][m][n] = VzLast +    (data->Vz[l][m][n] - data->Vz[l+1][m][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m][n-1];
    VyLast = data->Vy[l+1][m][n-1]; 
    VzLast = data->Vz[l+1][m][n-1];
  }

}


aedgeVKL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 n = -1;      /* Points one out from left side */
 for (m=0; m< grid->m; m++); {   /* back left edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   back left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m][n+1]);
   data->Vz[l][m][n] = VzLast +    (data->Vz[l][m][n] - data->Vz[l-1][m][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m][n+1];
    VyLast = data->Vy[l-1][m][n+1]; 
    VzLast = data->Vz[l-1][m][n+1];
  }

}


aedgeVKR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 n = grid->n;      /* Points one out from right side */
 for (m=0; m< grid->m; m++); {   /* back right edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dx - alpha*par->dt)/(SQRT2*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   back left edge */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m][n-1]);
   data->Vz[l][m][n] = VzLast +    (data->Vz[l][m][n] - data->Vz[l-1][m][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m][n-1];
    VyLast = data->Vy[l-1][m][n-1]; 
    VzLast = data->Vz[l-1][m][n-1];
  }

}


aedgeVBF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Points one out from front */
 m = grid->m;      /* Points one down from bottom */
 for (n=0; n< grid->n; n++); {   /* bottom front edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom front edge */
   data->Vx[l][m][n] = VxLast +    (data->Vx[l][m][n] - data->Vx[l+1][m-1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n];
    VyLast = data->Vy[l+1][m-1][n]; 
    VzLast = data->Vz[l+1][m-1][n];
  }

}


aedgeVTF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Points one out from front */
 m = -1;      /* Points one up from top */
 for (n=0; n< grid->n; n++); {   /* top front edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom front edge */
   data->Vx[l][m][n] = VxLast +    (data->Vx[l][m][n] - data->Vx[l+1][m+1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n];
    VyLast = data->Vy[l+1][m+1][n]; 
    VzLast = data->Vz[l+1][m+1][n];
  }

}


aedgeVBK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 m = grid->m;      /* Points one down from bottom */
 for (n=0; n< grid->n; n++); {   /* bottom back edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   bottom back edge */
   data->Vx[l][m][n] = VxLast +    (data->Vx[l][m][n] - data->Vx[l-1][m-1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n];
    VyLast = data->Vy[l-1][m-1][n]; 
    VzLast = data->Vz[l-1][m-1][n];
  }

}


aedgeVTK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Points one out from back */
 m = -1;      /* Points one out from top */
 for (n=0; n< grid->n; n++); {   /* back top edge */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT2*grid->dy - alpha*par->dt)/(SQRT2*grid->dy + alpha*par->dt);

/* Update velocities using paraxial extrapolations for points one out along
   top back edge */
   data->Vx[l][m][n] = VxLast +    (data->Vx[l][m][n] - data->Vx[l-1][m+1][n]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n];
    VyLast = data->Vy[l-1][m+1][n]; 
    VzLast = data->Vz[l-1][m+1][n];
  }

}



/*************************************************************************
VELOCITY CORNER BOUNDARY CONDITIONS
*************************************************************************/

dcornerVBLF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = grid-> m;      /* Point one down from bottom */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for front bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m-1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n+1];
    VyLast = data->Vy[l+1][m-1][n+1]; 
    VzLast = data->Vz[l+1][m-1][n+1];

}


dcornerVBRF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = grid-> m;      /* Point one down from bottom */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for front bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m-1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n-1];
    VyLast = data->Vy[l+1][m-1][n-1]; 
    VzLast = data->Vz[l+1][m-1][n-1];

}


dcornerVTLF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = -1;      /* Point one out from top */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for front top left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m+1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n+1];
    VyLast = data->Vy[l+1][m+1][n+1]; 
    VzLast = data->Vz[l+1][m+1][n+1];

}


dcornerVTRF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = -1;      /* Point one out from top */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for front top right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m+1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n-1];
    VyLast = data->Vy[l+1][m+1][n-1]; 
    VzLast = data->Vz[l+1][m+1][n-1];

}


dcornerVBLK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = grid-> m;      /* Point one down from bottom */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for back bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m-1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n+1];
    VyLast = data->Vy[l-1][m-1][n+1]; 
    VzLast = data->Vz[l-1][m-1][n+1];

}


dcornerVBRK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = grid-> m;      /* Point one down from bottom */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for back bottom right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m-1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n-1];
    VyLast = data->Vy[l-1][m-1][n-1]; 
    VzLast = data->Vz[l-1][m-1][n-1];

}


dcornerVTLK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = -1;      /* Point one out from top */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for back top left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m+1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n+1];
    VyLast = data->Vy[l-1][m+1][n+1]; 
    VzLast = data->Vz[l-1][m+1][n+1];

}


dcornerVTRK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha, beta, lP,lS;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = -1;      /* Point one out from top */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]+2.*grid->mu[l][m][n])*grid->irho[l][m][n]);
    beta =sqrt((grid->mu[l][m][n])*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);
    lS = (SQRT3*grid->dx - beta*par->dt) /(SQRT3*grid->dx + beta*par->dt);

/* Update velocities using paraxial extrapolations for back top right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m+1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n-1];
    VyLast = data->Vy[l-1][m+1][n-1]; 
    VzLast = data->Vz[l-1][m+1][n-1];

}


acornerVBLF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = grid-> m;      /* Point one down from bottom */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for front bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m-1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n+1];
    VyLast = data->Vy[l+1][m-1][n+1]; 
    VzLast = data->Vz[l+1][m-1][n+1];

}


acornerVBRF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = grid-> m;      /* Point one down from bottom */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for front bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m-1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m-1][n-1];
    VyLast = data->Vy[l+1][m-1][n-1]; 
    VzLast = data->Vz[l+1][m-1][n-1];

}


acornerVTLF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = -1;      /* Point one out from top */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for front top left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m+1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n+1];
    VyLast = data->Vy[l+1][m+1][n+1]; 
    VzLast = data->Vz[l+1][m+1][n+1];

}


acornerVTRF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = -1;      /* Point one out from front */
 m = -1;      /* Point one out from top */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for front top right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l+1][m+1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l+1][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l+1][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l+1][m+1][n-1];
    VyLast = data->Vy[l+1][m+1][n-1]; 
    VzLast = data->Vz[l+1][m+1][n-1];

}


acornerVBLK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = grid-> m;      /* Point one down from bottom */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for back bottom left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m-1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n+1];
    VyLast = data->Vy[l-1][m-1][n+1]; 
    VzLast = data->Vz[l-1][m-1][n+1];

}


acornerVBRK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = grid-> m;      /* Point one down from bottom */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for back bottom right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m-1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m-1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m-1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m-1][n-1];
    VyLast = data->Vy[l-1][m-1][n-1]; 
    VzLast = data->Vz[l-1][m-1][n-1];

}


acornerVTLK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = -1;      /* Point one out from top */
 n = -1;      /* Point one out from left */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for back top left corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m+1][n+1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n+1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n+1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n+1];
    VyLast = data->Vy[l-1][m+1][n+1]; 
    VzLast = data->Vz[l-1][m+1][n+1];

}


acornerVTRK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 static float VxLast = 0., VyLast = 0., VzLast = 0.; 
 float alpha,       lP   ;
 int l,m,n;
 l = grid->l;      /* Point one out from back */
 m = -1;      /* Point one out from top */
 n = grid->n;      /* Point one out from right */
   
/* Calculate wave speed parameters */
    alpha=sqrt((grid->lamda[l][m][n]                     )*grid->irho[l][m][n]);

/* Determine the coefficients for paraxial extrapolation */
    lP = (SQRT3*grid->dx - alpha*par->dt)/(SQRT3*grid->dx + alpha*par->dt);

/* Update velocities using paraxial extrapolations for back top right corner point */
   data->Vx[l][m][n] = VxLast + lP*(data->Vx[l][m][n] - data->Vx[l-1][m+1][n-1]);
   data->Vy[l][m][n] = VyLast + lP*(data->Vy[l][m][n] - data->Vy[l-1][m+1][n-1]);
   data->Vz[l][m][n] = VzLast + lP*(data->Vz[l][m][n] - data->Vz[l-1][m+1][n-1]);

 /* Save the current values of velocities */
    VxLast = data->Vx[l-1][m+1][n-1];
    VyLast = data->Vy[l-1][m+1][n-1]; 
    VzLast = data->Vz[l-1][m+1][n-1];

}



/****************************************************************************
STRESS SIDE BOUNDARY CONDITIONS
*****************************************************************************/

vboundTT(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one out from top boundary */
   m = -1;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      m--;
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      m++;
      */
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      m--;
/*
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
*/
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      m++;
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


vboundTB(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      m++;
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      m--;
      */
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      m++;
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      m--;
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


vboundTF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = -1;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(m = mmin; m < data->mmax; m++) {
      /*
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      l--;
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      l++;
      */
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      l--;
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      l++;
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


vboundTK(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from back boundary */
   l = grid->l;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(m = mmin; m < data->mmax; m++) {
      /*
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      l++;
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      l--;
      */
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      l++;
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      l--;
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


/* NOTE: at free surface mmin = 1, not 0 so this might not be correct */
vboundTR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz; 
   float **Txx, **Tyy, **Tzz, **Txy, **Txz, **Tyz, **G, **U;
   float dX, dY, dZ, G0, U0, U1, U2, V1, V2, V3, c;
   int l, m, n, d, mmin;

   /* initialize */
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from right boundary */
   n = grid->n;   

   /* update stresses for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Txx = data->Txx  [l]; Tyy = data->Tyy[l]; Tzz = data->Tzz[l];
      Txy = data->Txy  [l]; Txz = data->Txz[l]; Tyz = data->Tyz[l];
      G   = grid->lamda[l]; U   = grid->mu [l]; 
      for(m = mmin; m < data->mmax; m++) {
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         G0 = c*G[m][n];
         U0 = c*U[m][n];
         U2 = U0 + U0;
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G0 + U2) + (dY + dZ)*(G0);
         Tyy[m][n] += dY*(G0 + U2) + (dX + dZ)*(G0);
         Tzz[m][n] += dZ*(G0 + U2) + (dX + dY)*(G0);
         Txy[m][n] += U0*((Vx[l+1][m][n] - V1) + (V2 - Vy[l][m][n-1]));
         Txz[m][n] += U0*((Vx[l][m+1][n] - V1) + (V3 - Vz[l][m][n-1]));
         Tyz[m][n] += U0*((Vy[l][m+1][n] - V2) + (Vz[l+1][m][n] - V3));
         }
      n++;
      for(m = mmin; m < data->mmax; m++) {
         U1 = c*U[m][n];
         Txy[m][n] += U1*((Vx[l+1][m  ][n  ] - Vx[l  ][m  ][n  ]) + 
                          (Vy[l  ][m  ][n  ] - Vy[l  ][m  ][n-1]));
         Txz[m][n] += U1*((Vx[l  ][m+1][n  ] - Vx[l  ][m  ][n  ]) + 
                          (Vz[l  ][m  ][n  ] - Vz[l  ][m  ][n-1]));
         }
      n--;
      }

   /* need to worry about free surface condition case (?) */

   /* free memory */

}


vboundTL(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz; 
   float **Txx, **Tyy, **Tzz, **Txy, **Txz, **Tyz, **G, **U;
   float dX, dY, dZ, G0, G1, U0, U1, U2, V1, V2, V3, c;
   int l, m, n, d, mmin;

   /* initialize */
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;   

   /* update stresses for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Txx = data->Txx  [l]; Tyy = data->Tyy[l]; Tzz = data->Tzz[l];
      Txy = data->Txy  [l]; Txz = data->Txz[l]; Tyz = data->Tyz[l];
      G   = grid->lamda[l]; U   = grid->mu [l]; 
      for(m = mmin; m < data->mmax; m++) {
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         G0 = c*G[m][n];
         U0 = c*U[m][n];
         U2 = U0 + U0;
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G0 + U2) + (dY + dZ)*(G0);
         Tyy[m][n] += dY*(G0 + U2) + (dX + dZ)*(G0);
         Tzz[m][n] += dZ*(G0 + U2) + (dX + dY)*(G0);
         Txy[m][n] += U0*((Vx[l+1][m][n] - V1) + (V2 - Vy[l][m][n-1]));
         Txz[m][n] += U0*((Vx[l][m+1][n] - V1) + (V3 - Vz[l][m][n-1]));
         Tyz[m][n] += U0*((Vy[l][m+1][n] - V2) + (Vz[l+1][m][n] - V3));
         }
      n--;
      for(m = mmin; m < data->mmax; m++) {
         U1 = c*U[m][n];
         G1 = c*G[m][n];
         dX = (Vx[l][m][n+1] - Vx[l][m][n]  );
         dY = (Vy[l][m][n]   - Vy[l-1][m][n]);
         dZ = (Vz[l][m][n]   - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G1 + 2.0*U1) + (dY + dZ)*(G1);
         }
      n++;
      if (grid->bz1 == SURFACE) Tzz[0][n] = 0.;
      }

   /* need to worry about free surface condition case (?) */

   /* free memory */

}


vboundTKa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from back boundary */
   l = grid->l;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(m = mmin; m < data->mmax; m++) {
      /*
      */
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      l++;
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      l--;
      /*
      */
      /*
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      l++;
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      l--;
      */
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


vboundTBa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***Txx, ***Tyy, ***Tzz, ***Txy, ***Txz, ***Tyz;
   float ***g, ***u, **dX, **dY, **dZ, c;
   int l, m, n, d;

   /* initialize */
   Txx = data->Txx; Tyy = data->Tyy; Tzz = data->Tzz;
   Txy = data->Txy; Txz = data->Txz; Tyz = data->Tyz;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;
   u = grid->mu;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      */
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      m++;
      getDV02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      addVector3(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,d,g[l][m]+n,u[l][m]+n);
      addVector3(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,c,d,g[l][m]+n,u[l][m]+n);
      getDV32(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,c,d,g[l][m]+n,u[l][m]+n);
      m--;
      /*
      */
      /*
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      m++;
      dv02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Txx[l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tyy[l][m]+n,dY[0]+n,dX[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      addv3f_(Tzz[l][m]+n,dZ[0]+n,dX[0]+n,dY[0]+n,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vx[l][m]+n,Vx[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txy[l][m]+n,Vy[l][m]+n-1,Vy[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vx[l][m]+n,Vx[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Txz[l][m]+n,Vz[l][m]+n-1,Vz[l][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vy[l][m]+n,Vy[l][m+1]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      dv32f_(Tyz[l][m]+n,Vz[l][m]+n,Vz[l+1][m]+n,&c,&d,g[l][m]+n,u[l][m]+n);
      m--;
      */
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


vboundTLa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz; 
   float **Txx, **Tyy, **Tzz, **Txy, **Txz, **Tyz, **G, **U;
   float dX, dY, dZ, G0, G1, U0, U1, U2, V1, V2, V3, c;
   int l, m, n, d, mmin;

   /* initialize */
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;   

   /* update stresses for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      Txx = data->Txx  [l]; Tyy = data->Tyy[l]; Tzz = data->Tzz[l];
      Txy = data->Txy  [l]; Txz = data->Txz[l]; Tyz = data->Tyz[l];
      G   = grid->lamda[l]; U   = grid->mu [l]; 
      for(m = mmin; m < data->mmax; m++) {
         G0 = c*G[m][n];
         U0 = c*U[m][n];
         U2 = U0 + U0;
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G0 + U2) + (dY + dZ)*(G0);
         Tyy[m][n] += dY*(G0 + U2) + (dX + dZ)*(G0);
         Tzz[m][n] += dZ*(G0 + U2) + (dX + dY)*(G0);
         Txy[m][n] += U0*((Vx[l+1][m][n] - V1) + (V2 - Vy[l][m][n-1]));
         Txz[m][n] += U0*((Vx[l][m+1][n] - V1) + (V3 - Vz[l][m][n-1]));
         Tyz[m][n] += U0*((Vy[l][m+1][n] - V2) + (Vz[l+1][m][n] - V3));
         n--;
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G0 + U2) + (dY + dZ)*(G0);
         Tyy[m][n] += dY*(G0 + U2) + (dX + dZ)*(G0);
         Tzz[m][n] += dZ*(G0 + U2) + (dX + dY)*(G0);
         Txy[m][n] += U0*((Vx[l+1][m][n] - V1) + (V2 - Vy[l][m][n-1]));
         Txz[m][n] += U0*((Vx[l][m+1][n] - V1) + (V3 - Vz[l][m][n-1]));
         Tyz[m][n] += U0*((Vy[l][m+1][n] - V2) + (Vz[l+1][m][n] - V3));
         n++;
         }
      /*
      n--;
      for(m = mmin; m < data->mmax; m++) {
         U1 = c*U[m][n];
         G1 = c*G[m][n];
         dX = (Vx[l][m][n+1] - Vx[l][m][n]  );
         dY = (Vy[l][m][n]   - Vy[l-1][m][n]);
         dZ = (Vz[l][m][n]   - Vz[l][m-1][n]);
         Txx[m][n] += dX*(G1 + 2.0*U1) + (dY + dZ)*(G1);
         }
      n++;
      */
      if (grid->bz1 == SURFACE) {Tzz[0][n] = 0.; Tzz[0][n-1] = 0.;}
      }

   /* need to worry about free surface condition case (?) */

   /* free memory */

}


aboundTF(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from front boundary */
   l = -1;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(m = mmin; m < data->mmax; m++) {
      /*
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      l--;
      l++;
      */
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n          );
      l--;
      l++;
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


/* NOTE: at free surface mmin = 1, not 0 so this might not be correct */
aboundTR(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz; 
   float **T,                                      **G     ;
   float dX, dY, dZ, G0,             V1, V2, V3, c;
   int l, m, n, d, mmin;

   /* initialize */
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from right boundary */
   n = grid->n;   

   /* update stresses for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      T   = data->T    [l]; 
      G   = grid->lamda[l]; 
      for(m = mmin; m < data->mmax; m++) {
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         G0 = c*G[m][n];
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         T  [m][n] +=                (dX + dY + dZ)*(G0);
         }
      n++;
      n--;
      }

   /* need to worry about free surface condition case (?) */

   /* free memory */

}


aboundTKa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       **dX, **dY, **dZ, c;
   int l, m, n, d, mmin;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from back boundary */
   l = grid->l;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(m = mmin; m < data->mmax; m++) {
      /*
      */
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      l++;
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      l--;
      /*
      */
      /*
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n,u[l][m]+n);
      l++;
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n          );
      l--;
      */
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


aboundTBa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz, ***T;
   float ***g,       **dX, **dY, **dZ, c;
   int l, m, n, d;

   /* initialize */
   T   = data->T;
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;
   g = grid->lamda;

   /* get array differencing limits and coefficient */
   d = data->nmax - data->nmin;
   n = data->nmin;
   c = 1.0*par->dt/grid->dx;

   /* Points one down from bottom boundary */
   m = grid->m;   

   /* update stresses for grid planes */
   dX    = (float **) malloc(      1*sizeof(float *));
   dX[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dY    = (float **) malloc(      1*sizeof(float *));
   dY[0] = (float  *) malloc(grid->n*sizeof(float  ));
   dZ    = (float **) malloc(      1*sizeof(float *));
   dZ[0] = (float  *) malloc(grid->n*sizeof(float  ));
   for(l = data->lmin; l < data->lmax; l++) {
      /*
      */
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      m++;
      getDA02(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,c,d                    );
      getDA02(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,c,d                    );
      getDA02(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,c,d                    );
      addVectorA(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,d,g[l][m]+n          );
      m--;
      /*
      */
      /*
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n          );
      m++;
      da02f_(dX[0]+n,Vx[l][m]+n,Vx[l][m]+n+1,&c,&d                    );
      da02f_(dY[0]+n,Vy[l-1][m]+n,Vy[l][m]+n,&c,&d                    );
      da02f_(dZ[0]+n,Vz[l][m-1]+n,Vz[l][m]+n,&c,&d                    );
      addvaf_(T  [l][m]+n,dX[0]+n,dY[0]+n,dZ[0]+n,&d,g[l][m]+n          );
      m--;
      */
      }

   /* free memory */
   free(dX[0]); free(dX);
   free(dY[0]); free(dY);
   free(dZ[0]); free(dZ);

}


aboundTLa(par, grid, data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
   float ***Vx, ***Vy, ***Vz; 
   float **T,               **G;
   float dX, dY, dZ, G0, G1,             V1, V2, V3, c;
   int l, m, n, d, mmin;

   /* initialize */
   Vx = data->Vx; Vy = data->Vy; Vz = data->Vz;

   /* get array differencing limits and coefficient */
   mmin = (data->mmin == 1 && grid->bz1 == SURFACE) ? (0) : (data->mmin);
   c = 1.0*par->dt/grid->dx;

   /* Points one out from left boundary */
   n = -1;   

   /* update stresses for grid planes */
   for(l = data->lmin; l < data->lmax; l++) {
      T   = data->T[l];
      G   = grid->lamda[l]; 
      for(m = mmin; m < data->mmax; m++) {
         G0 = c*G[m][n];
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         T  [m][n] +=                (dX + dY + dZ)*(G0);
         n--;
         V1 = Vx[l][m][n]; 
         V2 = Vy[l][m][n]; 
         V3 = Vz[l][m][n];
         dX = (Vx[l][m][n+1] - V1           );
         dY = (V2            - Vy[l-1][m][n]);
         dZ = (V3            - Vz[l][m-1][n]);
         T  [m][n] +=                (dX + dY + dZ)*(G0);
         n++;
         }
      /*
      n--;
      for(m = mmin; m < data->mmax; m++) {
         G1 = c*G[m][n];
         dX = (Vx[l][m][n+1] - Vx[l][m][n]  );
         dY = (Vy[l][m][n]   - Vy[l-1][m][n]);
         dZ = (Vz[l][m][n]   - Vz[l][m-1][n]);
         T  [m][n] += (dX + dY + dZ)*(G1);
         }
      n++;
      */
      if (grid->bz1 == SURFACE) {T  [0][n] = 0.; T  [0][n-1] = 0.;}
      }

   /* need to worry about free surface condition case (?) */

   /* free memory */

}



/****************************************************************************
STRESS EDGE BOUNDARY CONDITIONS
*****************************************************************************/

dedgeTBL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from bottom left edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTTL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 m = -1;    /* Points one out from top */
 n = -1;         /* Points one out from left */
 for (l=0; l< grid->l; l++) {    /* Points along top left edge */
   
 /* Update stresses for points one out from top left edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTBR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 m = grid->m;    /* Points one down from bottom */
 n = grid->n;         /* Points one out from right */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from bottom left edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTTR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 m = -1;    /* Points one out from top */
 n = grid->n;         /* Points one out from right */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from top right edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTFL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;    /* Points one out from front */
 n = -1;         /* Points one out from left */
 for (m=0; m< grid->m; m++) {    /* Points along front left edge */
   
 /* Update stresses for points one out from front left edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTFR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;    /* Points one out from front */
 n = grid->n;         /* Points one out from right */
 for (m=0; m< grid->m; m++) {    /* Points along front right edge */
   
 /* Update stresses for points one out from front right edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTKL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 n = -1;         /* Points one out from left */
 for (m=0; m< grid->m; m++) {    /* Points along back left edge */
   
 /* Update stresses for points one out from back left edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTKR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 n = grid->n;         /* Points one out from right */
 for (m=0; m< grid->m; m++) {    /* Points along back right edge */
   
 /* Update stresses for points one out from back right edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTBF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;    /* Points one out from front */
 m = grid->m;         /* Points one down from bottom */
 for (n=0; n< grid->n; n++) {    /* Points along bottom front edge */
   
 /* Update stresses for points one out from bottom front edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTTF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;    /* Points one out from front */
 m = -1;         /* Points one out from top */
 for (n=0; n< grid->n; n++) {    /* Points along top front edge */
   
 /* Update stresses for points one out from top front edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTBK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 m = grid->m;         /* Points one down from bottom */
 for (n=0; n< grid->n; n++) {    /* Points along bottom back edge */
   
 /* Update stresses for points one out from bottom back edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


dedgeTTK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 m = -1;         /* Points one out from top */
 for (n=0; n< grid->n; n++) {    /* Points along top back edge */
   
 /* Update stresses for points one out from top back edge */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTBL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from bottom left edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTTL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 m = -1;    /* Points one out from top */
 n = -1;         /* Points one out from left */
 for (l=0; l< grid->l; l++) {    /* Points along top left edge */
   
 /* Update stresses for points one out from top left edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTBR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 m = grid->m;    /* Points one down from bottom */
 n = grid->n;         /* Points one out from right */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from bottom left edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTTR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 m = -1;    /* Points one out from top */
 n = grid->n;         /* Points one out from right */
 for (l=0; l< grid->l; l++) {    /* Points along bottom edge */
   
 /* Update stresses for points one out from top right edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTFL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;    /* Points one out from front */
 n = -1;         /* Points one out from left */
 for (m=0; m< grid->m; m++) {    /* Points along front left edge */
   
 /* Update stresses for points one out from front left edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTFR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;    /* Points one out from front */
 n = grid->n;         /* Points one out from right */
 for (m=0; m< grid->m; m++) {    /* Points along front right edge */
   
 /* Update stresses for points one out from front right edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTKL(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 n = -1;         /* Points one out from left */
 for (m=0; m< grid->m; m++) {    /* Points along back left edge */
   
 /* Update stresses for points one out from back left edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTKR(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 n = grid->n;         /* Points one out from right */
 for (m=0; m< grid->m; m++) {    /* Points along back right edge */
   
 /* Update stresses for points one out from back right edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTBF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;    /* Points one out from front */
 m = grid->m;         /* Points one down from bottom */
 for (n=0; n< grid->n; n++) {    /* Points along bottom front edge */
   
 /* Update stresses for points one out from bottom front edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTTF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;    /* Points one out from front */
 m = -1;         /* Points one out from top */
 for (n=0; n< grid->n; n++) {    /* Points along top front edge */
   
 /* Update stresses for points one out from top front edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTBK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 m = grid->m;         /* Points one down from bottom */
 for (n=0; n< grid->n; n++) {    /* Points along bottom back edge */
   
 /* Update stresses for points one out from bottom back edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}


/* not made acoustic */
aedgeTTK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;    /* Points one out from back */
 m = -1;         /* Points one out from top */
 for (n=0; n< grid->n; n++) {    /* Points along top back edge */
   
 /* Update stresses for points one out from top back edge */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);
  }
}



/****************************************************************************
STRESS CORNER BOUNDARY CONDITIONS
*****************************************************************************/

dcornerTBLF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for bottom left corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTBRF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = grid->m;    /* Points one down from bottom */
 n = grid->n;         /* Points one out from right */
   
 /* Update stresses for front bottom right corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTTLF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = -1;    /* Point one out from top */
 n = -1;         /* Point one out from left */
   
 /* Update stresses for top left corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTTRF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = -1;    /* Point one out from top */
 n = grid->n;         /* Point one out from right */
   
 /* Update stresses for top right corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTBLK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for back bottom left corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTBRK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = grid->m;    /* Point one down from bottom */
 n = grid->n;         /* Point one out from right */
   
 /* Update stresses for back bottom right corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTTLK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = -1;    /* Points one out from top */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for back top left corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


dcornerTTRK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;

{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = -1;    /* Points one out from top */
 n = grid->n;         /* Points one out from right */
   
 /* Update stresses for back top right corner point */
    update2Txx3d(par, grid, data->Txx, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tyy3d(par, grid, data->Tyy, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Tzz3d(par, grid, data->Tzz, data->Vx, data->Vy, data->Vz, l, m, n);
    update2Txy3d(par, grid, data->Txy, data->Vx, data->Vy, l, m, n);
    update2Txz3d(par, grid, data->Txz, data->Vx, data->Vz, l, m, n);
    update2Tyz3d(par, grid, data->Tyz, data->Vy, data->Vz, l, m, n);
  }


/* not made acoustic */
acornerTBLF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for bottom left corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTBRF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = grid->m;    /* Points one down from bottom */
 n = grid->n;         /* Points one out from right */
   
 /* Update stresses for front bottom right corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTTLF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = -1;    /* Point one out from top */
 n = -1;         /* Point one out from left */
   
 /* Update stresses for top left corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTTRF(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = -1;         /* Point one out from front */
 m = -1;    /* Point one out from top */
 n = grid->n;         /* Point one out from right */
   
 /* Update stresses for top right corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTBLK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = grid->m;    /* Points one down from bottom */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for back bottom left corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTBRK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = grid->m;    /* Point one down from bottom */
 n = grid->n;         /* Point one out from right */
   
 /* Update stresses for back bottom right corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTTLK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = -1;    /* Points one out from top */
 n = -1;         /* Points one out from left */
   
 /* Update stresses for back top left corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}


/* not made acoustic */
acornerTTRK(par,grid,data)
Parameter *par;
FdGrid *grid;
VelocityStress *data;
{
 int l,m,n;
 l = grid->l;         /* Point one out from back */
 m = -1;    /* Points one out from top */
 n = grid->n;         /* Points one out from right */
   
 /* Update stresses for back top right corner point */
    update2T3d(par, grid, data->T  , data->Vx, data->Vy, data->Vz, l, m, n);

}

